/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File name:       mdl_auxiliary_Config.cpp                                 */
/* Description:     static class for command line arguments                  */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_AUXILIARY_CONFIG_CPP_
#define MDL_AUXILIARY_CONFIG_CPP_

#include "mdl/auxiliary/mdl_auxiliary_Config.hpp"

namespace mdl {
namespace auxiliary {

	/****************************
	 *		Public members
	 ****************************/

	inline Config :: RunMode_
	Config :: getRunMode() {
		return runMode_;
	}

	// general flags
	inline bool
	Config :: check() {
		return check_;
	}
	inline bool
	Config :: prove() {
		return prove_;
	}
	inline bool
	Config :: compress() {
		return compress_;
	}
	inline bool
	Config :: useProver() {
		return (prove_ || compress_);
	}
	inline bool
	Config :: deepWrite() {
		return deepWrite_;
	}

	// proof search options
	inline Config :: SearchMode_
	Config :: getSearchMode() {
		return searchMode_;
	}
	inline Size_t
	Config :: getStackVolume() {
		return stackVolume_;
	}
	inline Size_t
	Config :: getProverVolume()
	{
		if (useProver()) {
			return proverVolume_;
		} else {
			return 0;
		}
	}
	inline Size_t
	Config :: getReserveVolume()
	{
		if (useProver()) {
			return reserveVolume_;
		} else {
			return 0;
		}
	}
	inline Time_t
	Config :: getMaxProveTime() {
		return maxProveTime_;
	}
	inline Time_t
	Config :: getMaxCompressTime() {
		return maxCompressTime_;
	}
	inline Clock_t
	Config :: getMinProveTime() {
		return minProveTime_;
	}
	inline Clock_t
	Config :: getMinCompressTime() {
		return minCompressTime_;
	}
	inline Clock_t
	Config :: getRefreshClock() {
		return refreshClock_;
	}
	inline Clock_t
	Config :: getQuantumClock() {
		return quantumClock_;
	}

	// all factors for prover
	inline value :: Factor
	Config :: getFitnessFactor() {
		return fitnessFactor_;
	}
	inline value :: Factor
	Config :: getProvedFactor() {
		return provedFactor_;
	}
	inline value :: Factor
	Config :: getProofFactor() {
		return proofFactor_;
	}
	inline value :: Factor
	Config :: getExperienceFactor() {
		return experienceFactor_;
	}
	inline value :: Factor
	Config :: getPositiveFactor() {
		return positiveFactor_;
	}
	inline value :: Factor
	Config :: getNegativeFactor() {
		return negativeFactor_;
	}
	inline value :: Factor
	Config :: getFormFactor() {
		return formFactor_;
	}
	inline value :: Factor
	Config :: getContentsFactor() {
		return contentsFactor_;
	}
	inline value :: Integer
	Config :: getFormDepth() {
		return formDepth_;
	}
	inline value :: Integer
	Config :: getContentsDepth() {
		return contentsDepth_;
	}
	inline value :: Integer
	Config :: getMaxSubtermCount() {
		return maxSubtermCount_;
	}
	inline value :: Integer
	Config :: getMaxProverDepth() {
		return maxProverDepth_;
	}
	inline value :: Integer
	Config :: getMaxSubstitutionComplexity() {
		return maxSubstitutionComplexity_;
	}
	inline value :: Integer
	Config :: getMaxExpressionComplexity() {
		return maxExpressionComplexity_;
	}
	inline value :: Integer
	Config :: getCompressionFactor() {
		return compressionFactor_;
	}

	inline bool
	Config :: checkUnambiguity() {
		return checkUnambiguity_;
	}
	inline bool
	Config :: keepLabels() {
		return keepLabels_;
	}

	// statistics output options
	inline bool
	Config :: showStatistics() {
		return showStatistics_;
	}

	// info output options
	inline bool
	Config :: showTimings() {
		return showTimings_;
	}
	inline bool
	Config :: showMath() {
		return showMath_;
	}
	inline bool
	Config :: showVolume() {
		return showVolume_;
	}

using std :: cout;
using std :: cerr;
using std :: endl;

	bool
	Config :: proceed (const int argCount, const char* argValues[])
	{
		init();
		if (argCount == 1) {
			cerr << getProgramName() << ": no input files" << endl;
			return false;
		}

		for (int i = 1; i < argCount; i ++) {
			const char* opt = argValues [i];
			if (*opt == '-') {
				opt ++;
			} else {
				if (i != argCount - 1) {
					std :: cerr << "option \"" << opt;
					std :: cerr << "\" does not start with \"-\" or \"--\"" << std :: endl;
					return false;
				}
				if (!proceedStringOption (opt, argCount, argValues, i)) {
					std :: cerr << "option \"" << opt;
					std :: cerr << "\" does not start with \"-\" or \"--\"" << std :: endl;
					return false;
				}
				continue;
			}
			if (*opt == '-') {
				opt ++;
				if (!proceedStringOption (opt, argCount, argValues, i)) {
					cerr << "unknown option: \"" << opt << "\"" << endl;
					return false;
				}
			} else if (!proceedOneCharOption (opt, argCount, argValues, i)) {
				cerr << "unknown option: \"" << opt << "\"" << endl;
				return false;
			}
		}
		return analyseArguments();
	}
	inline void
	Config :: release() {
		nstd :: Config :: release();
	}
	inline Size_t
	Config :: getVolume() {
		return nstd :: Config :: getVolume();
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Config :: showUsageMessage()
	{
		Ancestor_ :: showProgramUsage();
		cout << "Options:" << endl;
		Ancestor_ :: showOptions();
		cout << "  -c, --check                   check the specified file" << endl;
		cout << "  -p, --prove                   prove to the specified file" << endl;
		cout << "      --compress                compress all proofs " << endl;
		cout << "      --compress-one-of <int>   compress the proof for one of <int> theorems. By default = " << DEFAULT_COMPRESSION_FACTOR << endl;
		cout << "      --deep-write              write all included source files" << endl;
		cout << endl;
		cout << "      --max-prove-time <time>    max time for proving (minutes). By default = " << DEFAULT_MAX_PROVE_TIME << endl;
		cout << "      --max-compress-time <time> max time for compression (minutes). By default = " << DEFAULT_MAX_COMPRESS_TIME << endl;
		cout << "      --min-prove-time <time>    minimal time for proving (milliseconds). By default = " << DEFAULT_MIN_PROVE_TIME << endl;
		cout << "      --min-compress-time <time> minimal time for compression (milliseconds). By default = " << DEFAULT_MIN_COMPRESS_TIME << endl;
		cout << "      --refresh-time <time>      prover status refresh period (milliseconds). By default = " << DEFAULT_REFRESH_CLOCK << endl;
		cout << "      --quantum-time <time>      prover time quantum (microseconds). By default = " << DEFAULT_QUANTUM_CLOCK << endl;
		cout << endl;
		cout << "      --stack-volume <size>     use <size> volume for stack (megabytes). By default = " << DEFAULT_STACK_VOLUME << endl;
		cout << "      --prover-volume <size>    use <size> volume for proving (megabytes). By default = " << DEFAULT_PROVER_VOLUME << endl;
		cout << "      --reserve-volume <size>   use <size> volume for reserve (megabytes). By default = " << DEFAULT_RESERVE_VOLUME << endl;
		cout << endl;
		cout << "      --minimal-search          use greedy search for the proof:" << endl;
		cout << "                                stop search immediately after the first proof is found." << endl;
		cout << "      --optimal-search          search for the reasonably shorter proof:" << endl;
		cout << "                                search some time for a shorter proof after the first proof is found." << endl;
		cout << "      --maximal-search          search for the shortest proof of a minimal depth:" << endl;
		cout << "                                take the shortest proof of all of the current depth." << endl;
		cout << endl;
		cout << "      --factor-fitness <factor>    fitness prover factor. By default = " << fitnessFactor_ << endl;
		cout << "      --factor-proved <factor>     proved prover factor. By default = " << provedFactor_ << endl;
		cout << "      --factor-proof <factor>      proof prover factor. By default = " << proofFactor_ << endl;
		cout << "      --factor-experience <factor> experience prover factor. By default = " << experienceFactor_ << endl;
		cout << "      --factor-positive <factor>   positive prover factor. By default = " << positiveFactor_ << endl;
		cout << "      --factor-negative <factor>   negative prover factor. By default = " << negativeFactor_ << endl;
		cout << "      --factor-form <factor>       form prover factor. By default = " << formFactor_ << endl;
		cout << "      --factor-contents <factor>   contents prover factor. By default = " << contentsFactor_ << endl;
		cout << "      --depth-form <integer>       form depth (in prover). By default = " << formDepth_ << endl;
		cout << "      --depth-contents <integer>   contents depth (in prover). By default = " << contentsDepth_ << endl;
		cout << "      --max-subterm-count <int>    maximum exponent base for expression factorization. By default = " << maxSubtermCount_ << endl;
		cout << "      --max-prover-depth <int>     maximum prover depth. By default = " << maxProverDepth_ << endl;
		cout << "      --max-s-complexity <int>     maximum complexity of substitutions. By default = " << maxSubstitutionComplexity_ << endl;
		cout << "      --max-e-complexity <int>     maximum complexity of expressions. By default = " << maxExpressionComplexity_ << endl;
		cout << endl;
		cout << "      --stat-proof              show proof statistics" << endl;
		cout << "      --stat-all                show all statistics" << endl;
		cout << endl;
		cout << "      --info-timing             show thorough timing info" << endl;
		cout << "      --info-math               show thorough math info" << endl;
		cout << "      --info-volume             show thorough volume info" << endl;
		cout << "      --info-all                show all thorough info" << endl;
		cout << endl;
		cout << "      --check-unambiguity       check unambiguity of expressions in grammar (time consuming option)" << endl;
		cout << "      --keep-labels             keep assertion and hypothesis labels in output" << endl;
		cout << endl;
	}
	inline void
	Config :: init ()
	{
		Ancestor_ :: init ();
		programName() << "mdl";
		programDescription() << "mdl prover & translator to smm (Simplified MetaMath)";
	}

	bool
	Config :: proceedStringOption (const char* opt, const int argCount, const char* argValues[], int& i)
	{
		if (Ancestor_ :: proceedStringOption (opt, argCount, argValues, i)) {
			return true;
		}
		if (equalStrings (opt, "check")) {
			check_ = true;
			return true;
		}
		if (equalStrings (opt, "prove")) {
			prove_ = true;
			return true;
		}
		if (equalStrings (opt, "compress")) {
			compress_ = true;
			return true;
		}
		if (equalStrings (opt, "deep-write")) {
			deepWrite_ = true;
			return true;
		}
		if (equalStrings (opt, "max-prove-time")) {
			sscanf (argValues [++ i], "%ld", &maxProveTime_);
			return true;
		}
		if (equalStrings (opt, "max-compress-time")) {
			sscanf (argValues [++ i], "%ld", &maxCompressTime_);
			return true;
		}
		if (equalStrings (opt, "min-prove-time")) {
			sscanf (argValues [++ i], "%ld", &minProveTime_);
			return true;
		}
		if (equalStrings (opt, "min-compress-time")) {
			sscanf (argValues [++ i], "%ld", &minCompressTime_);
			return true;
		}
		if (equalStrings (opt, "refresh-time")) {
			sscanf (argValues [++ i], "%ld", &refreshClock_);
			return true;
		}
		if (equalStrings (opt, "quantum-time")) {
			sscanf (argValues [++ i], "%ld", &quantumClock_);
			return true;
		}
		if (equalStrings (opt, "minimal-search")) {
			searchMode_ = MINIMAL_SEARCH;
			return true;
		}
		if (equalStrings (opt, "optimal-search")) {
			searchMode_ = OPTIMAL_SEARCH;
			return true;
		}
		if (equalStrings (opt, "maximal-search")) {
			searchMode_ = MAXIMAL_SEARCH;
			return true;
		}
		if (equalStrings (opt, "stack-volume")) {
			sscanf (argValues [++ i], "%d", &stackVolume_);
			return true;
		}
		if (equalStrings (opt, "prover-volume")) {
			sscanf (argValues [++ i], "%d", &proverVolume_);
			return true;
		}
		if (equalStrings (opt, "reserve-volume")) {
			sscanf (argValues [++ i], "%d", &reserveVolume_);
			return true;
		}
		if (equalStrings (opt, "factor-fitness")) {
			sscanf (argValues [++ i], "%lf", &fitnessFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-proved")) {
			sscanf (argValues [++ i], "%lf", &provedFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-proof")) {
			sscanf (argValues [++ i], "%lf", &proofFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-experience")) {
			sscanf (argValues [++ i], "%lf", &experienceFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-positive")) {
			sscanf (argValues [++ i], "%lf", &positiveFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-negative")) {
			sscanf (argValues [++ i], "%lf", &negativeFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-form")) {
			sscanf (argValues [++ i], "%lf", &formFactor_);
			return true;
		}
		if (equalStrings (opt, "factor-contents")) {
			sscanf (argValues [++ i], "%lf", &contentsFactor_);
			return true;
		}
		if (equalStrings (opt, "depth-form")) {
			sscanf (argValues [++ i], "%d", &formDepth_);
			return true;
		}
		if (equalStrings (opt, "depth-contents")) {
			sscanf (argValues [++ i], "%d", &contentsDepth_);
			return true;
		}
		if (equalStrings (opt, "max-subterm-count")) {
			sscanf (argValues [++ i], "%d", &maxSubtermCount_);
			return true;
		}
		if (equalStrings (opt, "max-prover-depth")) {
			sscanf (argValues [++ i], "%d", &maxProverDepth_);
			return true;
		}
		if (equalStrings (opt, "max-s-complexity")) {
			sscanf (argValues [++ i], "%d", &maxSubstitutionComplexity_);
			return true;
		}
		if (equalStrings (opt, "max-e-complexity")) {
			sscanf (argValues [++ i], "%d", &maxExpressionComplexity_);
			return true;
		}
		if (equalStrings (opt, "compress-one-of")) {
			sscanf (argValues [++ i], "%d", &compressionFactor_);
			compress_ = true;
			return true;
		}
		if (equalStrings (opt, "check-unambiguity")) {
			checkUnambiguity_ = true;
			return true;
		}
		if (equalStrings (opt, "keep-labels")) {
			keepLabels_ = true;
			return true;
		}
		if (equalStrings (opt, "stat-all")) {
			showStatistics_ = true;
			return true;
		}
		if (equalStrings (opt, "info-timing")) {
			showTimings_ = true;
			return true;
		}
		if (equalStrings (opt, "info-math")) {
			showMath_ = true;
			return true;
		}
		if (equalStrings (opt, "info-volume")) {
			showVolume_ = true;
			return true;
		}
		if (equalStrings (opt, "info-all")) {
			showTimings_ = true;
			showMath_ = true;
			showVolume_ = true;
			return true;
		}
		return false;
	}
	bool
	Config :: proceedOneCharOption (const char* opt, const int argCount, const char* argValues[], int& i)
	{
		if (Ancestor_ :: proceedOneCharOption (opt, argCount, argValues, i)) {
			return true;
		}
		switch (*opt) {
		case 'p' :
			prove_ = true;
			return true;
		case 'c' :
			check_ = true;
			return true;
		default :
			return false;
		}
	}

	inline bool
	Config :: initRunMode()
	{
		if (check_) {
			runMode_ = CHECK_WORK;
		}
		if (prove_) {
			runMode_ = PROVE_WORK;
		}
		if (Ancestor_ :: translate()) {
			runMode_ = TRANSLATE_WORK;
		}
		if (compress_) {
			runMode_ = COMPRESS_WORK;
		}
		return true;
	}
	inline bool
	Config :: initMemorySize()
	{
		stackVolume_ *= MEMORY_UNIT_SIZE;
		proverVolume_ *= MEMORY_UNIT_SIZE;
		reserveVolume_ *= MEMORY_UNIT_SIZE;
		if (reserveVolume_ * 2 > proverVolume_) {
			cerr << "too large reserve in comparence with main memory volume." << std :: endl;
			return false;
		}
		return true;
	}
	inline bool
	Config :: initTimes()
	{
		maxProveTime_    *= 60; // seconds in one minute
		maxCompressTime_ *= 60; // seconds in one minute
		minProveTime_    *= 1000;  // microseconds in one millisecond
		minCompressTime_ *= 1000;  // microseconds in one millisecond
		refreshClock_ *= 1000; // microseconds in one millisecond
		quantumClock_ *= 1;
		return true;
	}
	bool
	Config :: isConsistent()
	{
		if (!translate() && !prove() && !compress() && !check()) {
			cerr << "no action is chosen." << std :: endl;
			return false;
		}
		if (translate() && (check() || prove() || compress())) {
			cerr << "translation together with other activity is chosen." << std :: endl;
			return false;
		}
		if (compress() && (check() || prove() || translate())) {
			cerr << "compression together with other activity is chosen." << std :: endl;
			return false;
		}
		if (prove() && (check() || compress() || translate())) {
			cerr << "proving together with other activity is chosen." << std :: endl;
			return false;
		}
		if (check() && (prove() || compress() || translate())) {
			cerr << "checking together with other activity is chosen." << std :: endl;
			return false;
		}
		return true;
	}
	bool
	Config :: checkPaths()
	{
		if (source()->extension() != "mdl") {
			cerr << "wrong source extension. Should be *.mdl" << endl;
			return false;
		}
		if (translate()) {
			if (getTarget() == NULL) {
				target() = new Path_ (*getSource());
				target()->extension() = "smm";
			}
			if (target()->extension() != "smm") {
				cerr << "wrong target extension. Should be *.smm" << endl;
				return false;
			}
		}
		if (prove() || compress()) {
			if (getTarget() == NULL) {
				target() = new Path_ (*getSource());
				target()->extension() = "mdl";
				target()->name() << "_out";
			}
			if (target()->extension() != "mdl") {
				cerr << "wrong target extension. Should be *.mdl" << endl;
				return false;
			}
		}
		return Paths :: check();
	}
	bool
	Config :: analyseArguments()
	{
		if (showHelp()) {
			showUsageMessage();
		}
		if (abortParsing()) {
			return false;
		}
		if (!initRunMode()) {
			return false;
		}
		if (!isConsistent()) {
			return false;
		}
		if (!checkPaths()) {
			return false;
		}
		if (!initMemorySize()) {
			return false;
		}
		if (!initTimes()) {
			return false;
		}
		return true;
	}

	/*************************************
	 *		Private static attributes
	 *************************************/

	// general flags
	bool Config :: check_= false;
	bool Config :: prove_= false;
	bool Config :: compress_= false;
	bool Config :: deepWrite_ = false;

	Config :: RunMode_
		Config :: runMode_ = Config :: DEFAULT_WORK;
	Config :: SearchMode_
		Config :: searchMode_ = Config :: DEFAULT_SEARCH;

	// volume options
	Size_t
		Config :: stackVolume_ = DEFAULT_STACK_VOLUME;
	Size_t
		Config :: proverVolume_ = DEFAULT_PROVER_VOLUME;
	Size_t
		Config :: reserveVolume_ = DEFAULT_RESERVE_VOLUME;

	// timing options
	Time_t
		Config :: maxProveTime_ = DEFAULT_MAX_PROVE_TIME;
	Time_t
		Config :: maxCompressTime_ = DEFAULT_MAX_COMPRESS_TIME;
	Clock_t
		Config :: minProveTime_ = DEFAULT_MIN_PROVE_TIME;
	Clock_t
		Config :: minCompressTime_ = DEFAULT_MIN_COMPRESS_TIME;
	Clock_t
		Config :: refreshClock_ = DEFAULT_REFRESH_CLOCK;
	Clock_t
		Config :: quantumClock_ = DEFAULT_QUANTUM_CLOCK;

	// all prover factors
	value :: Factor
		Config :: fitnessFactor_ = 1;
	value :: Factor
		Config :: provedFactor_ = 1;
	value :: Factor
		Config :: proofFactor_ = 1;
	value :: Factor
		Config :: experienceFactor_ = 1;
	value :: Factor
		Config :: positiveFactor_ = 1;
	value :: Factor
		Config :: negativeFactor_ = 1;
	value :: Factor
		Config :: formFactor_ = 1;
	value :: Factor
		Config :: contentsFactor_ = 1;
	value :: Integer
		Config :: formDepth_ = 3;
	value :: Integer
		Config :: contentsDepth_ = 3;
	value :: Integer
		Config :: maxSubtermCount_ = 10;
	value :: Integer
		Config :: maxProverDepth_ = 1024; //value :: undefined :: INTEGER;
	value :: Integer
		Config :: maxSubstitutionComplexity_ = 32; //value :: undefined :: INTEGER;
	value :: Integer
		Config :: maxExpressionComplexity_ = 256; //value :: undefined :: INTEGER;
	value :: Integer
		Config :: compressionFactor_ = DEFAULT_COMPRESSION_FACTOR;

	bool Config :: checkUnambiguity_ = false;
	bool Config :: keepLabels_ = false;

	// statistics output options
	bool Config :: showStatistics_ = false;

	// info output options
	bool Config :: showTimings_ = false;
	bool Config :: showMath_ = false;
	bool Config :: showVolume_ = false;
}
}

#endif /*MDL_AUXILIARY_CONFIG_CPP_*/
