/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File name:       mdl_auxiliary_Config.hpp                                 */
/* Description:     static class for command line arguments                  */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_AUXILIARY_CONFIG_HPP_
#define MDL_AUXILIARY_CONFIG_HPP_

#include "mdl/types/mdl_types.hpp"

namespace mdl {
namespace auxiliary {

// Class which contain boolean flags and string parameters for mdl.
class Config : public nstd :: Config {
	typedef nstd :: Config Ancestor_;
public :
	typedef int Percent_;

	// mdl general mode
	enum RunMode_ {
		NO_WORK        = 0,
		PROVE_WORK     = 1,
		CHECK_WORK     = 2,
		TRANSLATE_WORK = 3,
		COMPRESS_WORK  = 4,
		DEFAULT_WORK   = NO_WORK
	};
	static RunMode_ getRunMode();

	// check proofs
	static bool check();
	// prove problems
	static bool prove();
	// compress proofs
	static bool compress();
	// prover is not needed for translation
	static bool useProver();
	// write all included source files
	static bool deepWrite();

	// proof search options
	enum SearchMode_ {
		MINIMAL_SEARCH = 0,
		OPTIMAL_SEARCH = 1,
		MAXIMAL_SEARCH = 2,
		DEFAULT_SEARCH = OPTIMAL_SEARCH
	};
	static SearchMode_ getSearchMode();

	// volume for stack memory space
	static Size_t getStackVolume();
	// volume of memory needed by prover
	static Size_t getProverVolume();
	// volume of reserve memory needed by prover
	static Size_t getReserveVolume();

	// maximum time for proof, seconds
	static Time_t getMaxProveTime();
	// maximum time for compression, seconds
	static Time_t getMaxCompressTime();
	// minimum time for proof, microseconds
	static Clock_t getMinProveTime();
	// minimum time for compression, microseconds
	static Clock_t getMinCompressTime();
	// clocks for the refresh of the prover status, microseconds
	static Clock_t getRefreshClock();
	// clocks quantum for prover, microseconds
	static Clock_t getQuantumClock();

	// all factors for prover
	static value :: Factor getFitnessFactor();
	static value :: Factor getProvedFactor();
	static value :: Factor getProofFactor();
	static value :: Factor getExperienceFactor();
	static value :: Factor getPositiveFactor();
	static value :: Factor getNegativeFactor();
	static value :: Factor getFormFactor();
	static value :: Factor getContentsFactor();
	static value :: Integer getFormDepth();
	static value :: Integer getContentsDepth();
	static value :: Integer getMaxSubtermCount();
	static value :: Integer getMaxProverDepth();
	static value :: Integer getMaxSubstitutionComplexity();
	static value :: Integer getMaxExpressionComplexity();
	// compress one of each getCompressionFactor() theorems
	static value :: Integer getCompressionFactor();

	// check unambiguity of expressions in grammar
	static bool checkUnambiguity();
	// keep labels in smm output (should be removed)
	static bool keepLabels();

	// statistics output options
	static bool showStatistics();

	// info output options
	static bool showTimings();
	static bool showMath();
	static bool showVolume();

	static bool proceed (const int, const char* []);
	static void release();
	static Size_t getVolume();

private :
	enum {
		MEMORY_UNIT_SIZE = nstd :: MEGABYTE,
		DEFAULT_STACK_VOLUME = 64,                // in memory units
		DEFAULT_PROVER_VOLUME = 64,               // in memory units
		DEFAULT_RESERVE_VOLUME = 4,               // in memory units
		DEFAULT_MEMORY_USAGE_PERCENT = 30,        // %

		DEFAULT_MAX_PROVE_TIME = 5,               // minutes
		DEFAULT_MAX_COMPRESS_TIME = 5,            // minutes
		DEFAULT_MIN_PROVE_TIME = 10,              // milliseconds
		DEFAULT_MIN_COMPRESS_TIME = 10,           // milliseconds
		DEFAULT_REFRESH_CLOCK = 1000,             // milliseconds
		DEFAULT_QUANTUM_CLOCK = 10000,            // microseconds

		DEFAULT_COMPRESSION_FACTOR = 1            // compress one of each DEFAULT_COMPRESSION_FACTOR theorems
	};

	static void showUsageMessage();
	static void init();

	static bool proceedStringOption (const char*, const int, const char* [], int&);
	static bool proceedOneCharOption (const char*, const int, const char* [], int&);

	static bool initRunMode();
	static bool initMemorySize();
	static bool initTimes();
	static bool isConsistent();
	static bool checkPaths();
	static bool analyseArguments();

	// general flags
	static bool check_;
	static bool prove_;
	static bool compress_;
	static bool deepWrite_;

	// general mode option
	static RunMode_ runMode_;
	// search options
	static SearchMode_ searchMode_;

	// volume options
	static Size_t stackVolume_;
	static Size_t proverVolume_;
	static Size_t reserveVolume_;

	// timing options
	static Time_t maxProveTime_;
	static Time_t maxCompressTime_;
	static Clock_t minProveTime_;
	static Clock_t minCompressTime_;
	static Clock_t refreshClock_;
	static Clock_t quantumClock_;

	// all prover factors
	static value :: Factor fitnessFactor_;
	static value :: Factor provedFactor_;
	static value :: Factor proofFactor_;
	static value :: Factor experienceFactor_;
	static value :: Factor positiveFactor_;
	static value :: Factor negativeFactor_;
	static value :: Factor formFactor_;
	static value :: Factor contentsFactor_;
	static value :: Integer formDepth_;
	static value :: Integer contentsDepth_;
	static value :: Integer maxSubtermCount_;
	static value :: Integer maxProverDepth_;
	static value :: Integer maxSubstitutionComplexity_;
	static value :: Integer maxExpressionComplexity_;
	static value :: Integer compressionFactor_;

	static bool checkUnambiguity_;
	static bool keepLabels_;

	// statistics output options
	static bool showStatistics_;

	// info output options
	static bool showTimings_;
	static bool showMath_;
	static bool showVolume_;
};

}
}

#endif /*MDL_AUXILIARY_CONFIG_HPP_*/
