/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File name:       mdl_expression_Branching.ipp                             */
/* Description:     general base for expression trees (set of expressions)   */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_EXPRESSION_BRANCHING_IPP_
#define MDL_EXPRESSION_BRANCHING_IPP_

#include "mdl/auxiliary/mdl_auxiliary.hpp"

namespace mdl {
namespace expression {

	/****************************
	 *		Public members
	 ****************************/

	template<class M, class N, class V, class A>
	inline
	Branching<M, N, V, A> :: Branching() :
	root_ (NULL),
	symbolVolume_ (0) {
	}
	template<class M, class N, class V, class A>
	inline
	Branching<M, N, V, A> :: ~ Branching()
	{
		if (root_ != NULL) {
			root_->deleteNodes();
			delete root_;
			root_ = NULL;
		}
	}

	template<class M, class N, class V, class A>
	inline bool
	Branching<M, N, V, A> :: isNull() const {
		return (root_ == NULL);
	}
	template<class M, class N, class V, class A>
	inline bool
	Branching<M, N, V, A> :: isEmpty() const {
		return (root_ == NULL);
	}

	template<class M, class N, class V, class A>
	inline typename Branching<M, N, V, A> :: Iterator_
	Branching<M, N, V, A> :: getBegin() {
		return Iterator_ (root_);
	}
	template<class M, class N, class V, class A>
	inline typename Branching<M, N, V, A> :: const_Iterator_
	Branching<M, N, V, A> :: getBegin() const {
		return const_Iterator_ (root_);
	}
	template<class M, class N, class V, class A>
	inline value :: Integer
	Branching<M, N, V, A> :: getSymbolVolume() const {
		return symbolVolume_;
	}

	template<class M, class N, class V, class A>
	template<class T>
	inline const typename Branching<M, N, V, A> :: Term_*
	Branching<M, N, V, A> :: createTerm (const T* const term)
	{
		typedef T Term;
		return Builder<Implementation_, Term> :: build (term);
	}

	template<class M, class N, class V, class A>
	template<class E>
	inline typename Branching<M, N, V, A> :: Iterator_
	Branching<M, N, V, A> :: addBranch
	(
		const E* const expression,
		const Value_ value,
		const bool addData
	)
	{
		typedef E Expression;
		typedef
			typename Expression :: Term_
			Term;
		const bool addRoot = isEmpty();
		Term_* term = NULL;
		if (addData) {
			term = Builder<Implementation_, Term, Value_> :: build
				(expression->getTerm(), dynamic_cast<Implementation_*>(this), value);
		} else {
			term = Builder<Implementation_, Term> :: build
				(expression->getTerm(), dynamic_cast<Implementation_*>(this));
			term->getEnd()->data().addValue (value);
		}
		if (addRoot) {
			root_ = term->getBegin().getNode();
		}
		symbolVolume_ += expression->getSymbolVolume();
		return term->getEnd();
	}
	template<class M, class N, class V, class A>
	template<class E>
	inline void
	Branching<M, N, V, A> :: addSubBranches (const E* const expression, const Value_ value)
	{
		typedef E Expression;
		typedef
			typename Expression :: const_Iterator_
			const_Iterator;
		typedef
			typename Expression :: Term_
			Term;

		const const_Iterator begin = expression->getBegin();
		const const_Iterator end = expression->getEnd();
		const_Iterator iterator = begin;

		while (true) {
			const Term* const
				term = iterator->getTerms().getInitial().getTerm();
			if (term->getDepth() <= Config :: getContentsDepth()) {
				if (iterator == term->getBegin()) {
					addBranch (term, value);
				}
			}
			if (iterator == end) {
				break;
			}
			++ iterator;
		}
	}
	template<class M, class N, class V, class A>
	template<class E>
	inline void
	Branching<M, N, V, A> :: addFactorBranches (const E* const expression, const Value_ value)
	{
		for (value :: Integer depth = 1; depth < Config :: getFormDepth(); ++ depth) {
			const value :: Integer
				cardinality = getLevelCardinality (expression, depth);
			insertFactorBranches (expression, value, depth, cardinality);
		}
	}

	template<class M, class N, class V, class A>
	template<class E>
	typename Branching<M, N, V, A> ::
	template Unify_
	<
		E,
		unify :: policy :: DIRECT,
		unify :: kind :: PLAIN,
		allocator :: Stack
	> :: Result_
	Branching<M, N, V, A> :: equifyDirect (const E* expression) const
	{
		typedef Branching Expression_1;
		typedef E Expression_2;
		typedef
			Unify_
			<
				Expression_2,
				unify :: policy :: DIRECT,
				unify :: kind :: PLAIN,
				allocator :: Stack
			>
			Unifier;
		typedef
			typename Unifier :: Result_
			Result;
		typedef
			typename Unifier :: Unit_
			Unit;

		if (root_ == NULL) {
			return Result();
		}
		const Expression_1* expression_1 = this;
		const Expression_2* expression_2 = expression;

		const Unit unit
		(
			expression_1->getBegin(),
			expression_2->getBegin()
		);
		return Unifier :: proceed (unit);
	}
	template<class M, class N, class V, class A>
	template<class E>
	typename Branching<M, N, V, A> ::
	template Unify_
	<
		E,
		unify :: policy :: INVERSE,
		unify :: kind :: PLAIN,
		allocator :: Stack
	> :: Result_
	Branching<M, N, V, A> :: equifyInverse (const E* expression) const
	{
		typedef Branching Expression_1;
		typedef E Expression_2;
		typedef
			Unify_
			<
				Expression_2,
				unify :: policy :: INVERSE,
				unify :: kind :: PLAIN,
				allocator :: Stack
			>
			Unifier;
		typedef
			typename Unifier :: Result_
			Result;
		typedef
			typename Unifier :: Unit_
			Unit;

		if (root_ == NULL) {
			return Result();
		}
		const Expression_1* expression_1 = this;
		const Expression_2* expression_2 = expression;

		const Unit unit
		(
			expression_1->getBegin(),
			expression_2->getBegin()
		);
		return Unifier :: proceed (unit);
	}
	template<class M, class N, class V, class A>
	template<class E>
	typename Branching<M, N, V, A> ::
	template Unify_<E, unify :: policy :: DIRECT> :: Result_
	Branching<M, N, V, A> :: unifyDirect
	(
		const E* expression,
		const Substitution_* substitution_1,
		const typename E :: Substitution_* substitution_2,
		const index :: Step level_1,
		const index :: Step level_2
	) const
	{
		typedef Branching Expression_1;
		typedef E Expression_2;
		typedef
			Unify_<Expression_2, unify :: policy :: DIRECT>
			Unifier;
		typedef
			typename Unifier :: Result_
			Result;
		typedef
			typename Unifier :: Unit_
			Unit;

		if (root_ == NULL) {
			return Result();
		}
		const Expression_1* expression_1 = this;
		const Expression_2* expression_2 = expression;

		const Unit unit
		(
			expression_1->getBegin(),
			expression_2->getBegin(),
			substitution_2,
			substitution_1,
			level_1,
			level_2
		);
		return Unifier :: proceed (unit);
	}
	template<class M, class N, class V, class A>
	template<class E>
	typename Branching<M, N, V, A> ::
	template Unify_<E, unify :: policy :: INVERSE> :: Result_
	Branching<M, N, V, A> :: unifyInverse
	(
		const E* expression,
		const Substitution_* substitution_1,
		const typename E :: Substitution_* substitution_2,
		const index :: Step level_1,
		const index :: Step level_2
	) const
	{
		typedef Branching Expression_1;
		typedef E Expression_2;
		typedef
			Unify_<Expression_2, unify :: policy :: INVERSE>
			Unifier;
		typedef
			typename Unifier :: Result_
			Result;
		typedef
			typename Unifier :: Unit_
			Unit;

		if (root_ == NULL) {
			return Result();
		}
		const Expression_1* expression_1 = this;
		const Expression_2* expression_2 = expression;

		const Unit unit
		(
			expression_1->getBegin(),
			expression_2->getBegin(),
			substitution_2,
			substitution_1,
			level_1,
			level_2
		);
		return Unifier :: proceed (unit);
	}
	template<class M, class N, class V, class A>
	template<class E>
	typename Branching<M, N, V, A> ::
	template Unify_<E, unify :: policy :: SYMMETRIC> :: Result_
	Branching<M, N, V, A> :: unifySymmetric
	(
		const E* expression,
		const index :: Step level_1,
		const index :: Step level_2
	) const
	{
		typedef Branching Expression_1;
		typedef E Expression_2;
		typedef
			Unify_<Expression_2, unify :: policy :: SYMMETRIC>
			Unifier;
		typedef
			typename Unifier :: Result_
			Result;
		typedef
			typename Unifier :: Unit_
			Unit;

		if (root_ == NULL) {
			return Result();
		}
		const Expression_1* expression_1 = this;
		const Expression_2* expression_2 = expression;

		const Unit unit
		(
			expression_1->getBegin(),
			expression_2->getBegin(),
			level_1,
			level_2
		);

		return Unifier :: proceed (unit);
	}
	template<class M, class N, class V, class A>
	template<class E>
	typename Branching<M, N, V, A> ::
	template Simify_<E> :: Similarity_
	Branching<M, N, V, A> :: simify
	(
		const E* expression,
		const index :: Step level_1,
		const index :: Step level_2
	) const
	{
		typedef Branching Expression_1;
		typedef E Expression_2;
		typedef
			Simify_<Expression_2>
			Simifier;
		typedef
			typename Simifier :: Similarity_
			Similarity;
		if (root_ == NULL) {
			return Similarity();
		}
		typedef
			typename Simifier :: Unit_
			Unit;

		const Expression_1* expression_1 = this;
		const Expression_2* expression_2 = expression;

		const Unit unit
		(
			expression_1->getBegin(),
			expression_2->getBegin(),
			level_1,
			level_2
		);
		return Simifier :: proceed (unit);
	}
	template<class M, class N, class V, class A>
	inline void
	Branching<M, N, V, A> :: setKind (const value :: Kind kind) const {
		setKind (kind, root_);
	}

using manipulator :: endLine;
using manipulator :: tab;

	// object :: Object interface
	template<class M, class N, class V, class A>
	Size_t
	Branching<M, N, V, A> :: getVolume() const
	{
		Size_t volume = 0;
		if (root_ != NULL) {
			volume += root_->getVolume();
			volume += root_->getSizeOf();
		}
		return volume;
	}
	template<class M, class N, class V, class A>
	Size_t
	Branching<M, N, V, A> :: getSizeOf() const {
		return sizeof (Branching);
	}
	template<class M, class N, class V, class A>
	void
	Branching<M, N, V, A> :: show (String& string) const
	{
		string << "tree:" << endLine;
		string << "symbol volume: " << symbolVolume_ << endLine;
		if (root_ != NULL) {
			root_->showTree (string);
		} else {
			string << tab << "NULL";
		}
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class M, class N, class V, class A>
	template<class I>
	void
	Branching<M, N, V, A> :: copyNode (Iterator_& target, const I& source) const {
		*target = static_cast<const Symbol&>(*source);
	}
	template<class M, class N, class V, class A>
	template<class I>
	bool
	Branching<M, N, V, A> :: findNode (Iterator_& target, const I& source) const
	{
		while (true) {
			if (*target == *source) {
				return true;
			}
			+ target;
		}
		return false;
	}
	template<class M, class N, class V, class A>
	template<class I>
	void
	Branching<M, N, V, A> :: addFindNode (Iterator_& target, const I& source) const
	{
		while (true) {
			if (*target == *source) {
				break;
			}
			if (target->isUndefined()) {
				*target = static_cast<const Symbol&>(*source);
				break;
			}
			+ target;
		}
	}
	template<class M, class N, class V, class A>
	template<class E>
	void
	Branching<M, N, V, A> :: insertFactorBranches
	(
		const E* const expression,
		const Value_ value,
		const value :: Integer depth,
		const value :: Integer cardinality
	)
	{
		typedef
			List<allocator :: Stack>
			FactorExpression;
		value :: Integer
			congruence = 0;
		while (congruence >> cardinality == 0) {
			Memory :: stack().push();
			FactorExpression factorExpression (expression);
			factorExpression.factorize (depth, congruence);
			addBranch (&factorExpression, value);
			Memory :: stack().pop();
			++ congruence;
		}
	}
	template<class M, class N, class V, class A>
	template<class E>
	value :: Integer
	Branching<M, N, V, A> :: getLevelCardinality
	(
		const E* const expression,
		const value :: Integer depth
	) const
	{
		typedef E Expression;
		typedef
			typename Expression :: const_Iterator_
			const_Iterator;
		typedef
			typename Expression :: Term_
			Term;

		const const_Iterator begin = expression->getBegin();
		const const_Iterator end = expression->getEnd();
		const_Iterator iterator = begin;

		value :: Integer
			termCounter = 0;
		while (true) {
			const Term* const
				term = iterator->getTerms().getInitial().getTerm();
			if (term != NULL &&
				term->getDepth() == depth &&
				term->getBegin() == iterator &&
				term->isComplex()) {
					++ termCounter;
			}
			if ((iterator == end) ||
				(termCounter >= Config :: getMaxSubtermCount())) {
				break;
			}
			++ iterator;
		}
		return termCounter;
	}

	template<class M, class N, class V, class A>
	void
	Branching<M, N, V, A> :: setKind (const value :: Kind kind, const Node_* node) const
	{
		if (node = NULL) {
			return;
		}
		const_cast<Node_*>(node)->setKind (kind);
		const Node_* left = node->goLeft();
		const Node_* up = node->goUp();
		setKind (kind, left);
		setKind (kind, up);
	}
}
}

#endif /*MDL_EXPRESSION_BRANCHING_IPP_*/
