/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File name:       mdl_expression_parse_Parser.ipp                          */
/* Description:     static class: expression parser                          */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_EXPRESSION_PARSE_PARSER_IPP_
#define MDL_EXPRESSION_PARSE_PARSER_IPP_

#include "mdl/expression/parse/mdl_expression_parse_Unit.ipp"

namespace mdl {
namespace expression {
namespace parse {

	/****************************
	 *		Public members
	 ****************************/

	template<bool d, bool f>
	inline bool
	Parser<d, f> :: proceed (Expression_* const expression, const mdl :: Type* const type)
	{
		if (expression->isEmpty()) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << "expression fails parsing:" << endLine;
			error->message() << tab << "empty expression." << endLine;
			throw error;
		}
		const expr_Iterator_ begin = expression->getBegin();
		const tree :: Rule* ruleTree = type->getRuleTree();
		const tree_Iterator_ root = ruleTree->getBegin();
		expression_ = expression;

		const Result_ result =
			(expression->isVariable()) ?
			parseVariable (begin, type) :
			parseTerm (begin, type);

		bool success = false;
		try {
			success = result.confirm();
		} catch (Error* error) {
			error->message() << "expression: " << *expression << endLine;
			throw error;
		}
		return success;
	}

	/****************************
	 *		Private members
	 ****************************/

	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: parseVariable (expr_Iterator_ begin, const mdl :: Type* type)
	{
		if (begin->getType() != type) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << "variable expression fails parsing:" << endLine;
			error->message() << tab << "type of a variable differs a the type of the expression." << endLine;
			throw error;
		}
		const Term_*
			term = Term_ :: create<allocator :: Stack> (begin, type);
		const Tree_* tree = new Tree_ (term, true, true);
		return Result_ (tree);
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: parseTerm (expr_Iterator_ begin, const mdl :: Type* type, const Tree_* tree)
	{
		const tree :: Rule* rules = type->getRuleTree();
		if (rules == NULL) {
			return Result_();
		}
		Unit_ unit (begin, rules->getBegin(), tree);
		return parseNode (unit, type);
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: parseTerm (Unit_ unit, const mdl :: Type* type)
	{
		const expr_Iterator_ begin = unit.getEnd();
		const Tree_* tree = unit.getTree();
		return parseTerm (begin, type, tree);
	}
	template<bool d, bool f>
	typename Parser<d, f> :: Result_
	Parser<d, f> :: parseNode (Unit_ unit, const mdl :: Type* type)
	{
		Result_ result;
		while (unit != NULL) {
			result.join (tryConstant (unit, type));
			if (!deep_ && result.isObtained()) {
				return result;
			}
			+ unit;
		}
		unit.recall();
		while (unit != NULL) {
			result.join (tryVariable (unit, type));
			if (!deep_ && result.isObtained()) {
				return result;
			}
			+ unit;
		}
		return result;
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: tryConstant (Unit_ unit, const mdl :: Type* type)
	{
		if (!unit.isConstant()) {
			return Result_();
		}
		if (!unit.isEqual()) {
			return Result_();
		}
		return parseNext (unit, type);
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: tryVariable (Unit_ unit, const mdl :: Type* type)
	{
		if (!unit.isVariable()) {
			return Result_();
		}
		const mdl :: Type* localType = unit.getTreeType();
		Result_ result = tryType (unit, type, localType);
		if (!deep_ && result.isObtained()) {
			return result;
		}
		for (index :: Type i = 0; i < localType->getSubtypeCount(); ++ i) {
			const mdl :: Type* localSubtype = localType->getSubtype (i);
			result.join (tryType (unit, type, localSubtype));
			if (!deep_ && result.isObtained()) {
				return result;
			}
		}
		return result;
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: tryType (Unit_ unit, const mdl :: Type* type, const mdl :: Type* localType)
	{
		Result_ result;
		result.join (tryRule (unit, type, localType));
		if (!deep_ && result.isObtained()) {
			return result;
		}
		result.join (tryVar (unit, type, localType));
		if (!deep_ && result.isObtained()) {
			return result;
		}
		return result;
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: tryRule (Unit_ unit, const mdl :: Type* type, const mdl :: Type* localType)
	{
		Result_ result;
		const Result_
			subterms = parseTerm (unit, localType);
		const typename Result_ :: Node_*
			node = subterms.getList().getFirst();
		while (node != NULL) {
			const Tree_* tree = node->getValue();
			result.join (tryTerm (unit, type, tree));
			if (!deep_ && result.isObtained()) {
				return result;
			}
			node = node->getNext();
		}
		return result;
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: tryTerm (Unit_ unit, const mdl :: Type* type, const Tree_* tree)
	{
		const Term_* term = tree->getTerm();
		unit.newSubstitution (term);
		unit.setEnd (const_cast<Term_*>(term)->getEnd());
		unit.setTree (tree);
		return parseNext (unit, type);
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: tryVar (Unit_ unit, const mdl :: Type* type, const mdl :: Type* localType)
	{
		if (unit.getExprType() != localType) {
			return Result_();
		}
		unit.newVarTerm (localType);
		return parseNext (unit, type);
	}
	template<bool d, bool f>
	inline typename Parser<d, f> :: Result_
	Parser<d, f> :: parseNext (Unit_ unit, const mdl :: Type* type)
	{
		switch (unit.getState()) {
		case Unit_ :: TRANSIT :
			++ unit;
			return parseNode (unit, type);
		case Unit_ :: FINAL :
			return Result_ (unit.newRuleTerm (type, true));
		case Unit_ :: TERMINAL :
			return Result_ (unit.newRuleTerm (type, false));
		case Unit_ :: DEADLOCK :
			return Result_();
		}
		return Result_();
	}

	/***********************************
	 *		Static private attribute
	 ***********************************/

	template<bool d, bool f>
	typename Parser<d, f> :: Expression_*
		Parser<d, f> :: expression_ = NULL;
}
}
}

#endif /*MDL_EXPRESSION_PARSE_PARSER_IPP_*/
