/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File name:       mdl_expression_parse_Unit.ipp                            */
/* Description:     expression parser unit: pair of iterators& substitutiuon */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_EXPRESSION_PARSE_UNIT_IPP_
#define MDL_EXPRESSION_PARSE_UNIT_IPP_

namespace mdl {
namespace expression {
namespace parse {

	/****************************
	 *		Public members
	 ****************************/

	template<class P>
	inline
	Unit<P> :: Unit
	(
		const expr_Iterator_& begin,
		const tree_Iterator_& root,
		const Tree_* tree
	) :
	Pair_ (begin,  root),
	begin_ (begin),
	root_ (root),
	substitution_ (NULL),
	tree_ (tree) {
	}
	template<class P>
	inline
	Unit<P> :: Unit (const Unit& unit) :
	Pair_ (unit),
	begin_ (unit.begin_),
	root_ (unit.root_),
	substitution_ (unit.substitution_),
	tree_ (unit.tree_) {
	}

	template<class P>
	inline bool
	Unit<P> :: isConstant() const {
		return
			(Pair_ :: getSecond() != NULL) &&
			(Pair_ :: getSecond()->isConstant());
	}
	template<class P>
	inline bool
	Unit<P> :: isVariable() const {
		return
			(Pair_ :: getSecond() != NULL) &&
			(Pair_ :: getSecond()->isVariable());
	}
	template<class P>
	inline typename Unit<P> :: State
	Unit<P> :: getState() const
	{
		const bool
			expressionEnd = Pair_ :: getFirst()->template
				isTerminal<direction :: RIGHT>(Parser_ :: expression_);
		const bool
			treeEnd = Pair_ :: getSecond()->template
				isTerminal<direction :: RIGHT>();

		if (expressionEnd) {
			return treeEnd ? FINAL : DEADLOCK;
		} else {
			return treeEnd ? TERMINAL : TRANSIT;
		}
	}

	template<class P>
	inline void
	Unit<P> :: operator ++ () {
		Pair_ :: operator ++();
	}
	template<class P>
	inline void
	Unit<P> :: operator + ()
	{
		Pair_ :: operator +();
		root_ = Pair_ :: getSecond();
	}
	template<class P>
	inline void
	Unit<P> :: operator - ()
	{
		Pair_ :: operator -();
		root_ = Pair_ :: getSecond();
	}

	template<class P>
	inline const Symbol&
	Unit<P> :: getTreeSymbol() const {
		return *Pair_ :: getSecond();
	}
	template<class P>
	inline const mdl :: Type*
	Unit<P> :: getTreeType() const {
		return Pair_ :: getSecond()->getType();
	}
	template<class P>
	inline const mdl :: Type*
	Unit<P> :: getExprType() const {
		return Pair_ :: getFirst()->getType();
	}
	template<class P>
	inline const mdl :: Rule*
	Unit<P> :: getTreeRule() const
	{
		return
			Config :: prove() ?
			NULL :
			Pair_ :: getSecond()->getData().getValue();
	}

	template<class P>
	inline typename Unit<P> :: expr_Iterator_
	Unit<P> :: getBegin() const {
		return begin_;
	}
	template<class P>
	inline typename Unit<P> :: expr_Iterator_
	Unit<P> :: getEnd() const {
		return Pair_ :: getFirst();
	}
	template<class P>
	inline const typename Unit<P> :: Tree_*
	Unit<P> :: getTree() const {
		return tree_;
	}
	template<class P>
	inline void
	Unit<P> :: setEnd (const expr_Iterator_ end) {
		Pair_ :: first() = end;
	}
	template<class P>
	inline void
	Unit<P> :: setTree (const Tree_* tree) {
		tree_ = tree;
	}

	template<class P>
	inline const typename Unit<P> :: Substitution_*
	Unit<P> :: getSubstitution() const {
		return substitution_;
	}
	template<class P>
	inline void
	Unit<P> :: newSubstitution (const Term_* term)
	{
		if (isVariable()) {
			substitution_ =
			Config :: prove() ? NULL :
			const_cast<Substitution_*>
			(
				Substitution_ :: template create <allocator :: Stack>
				(
					getTreeSymbol(),
					term,
					substitution_
				)->
				template castAway<allocator :: Heap>()
			);
		}
	}

	template<class P>
	inline const typename Unit<P> :: Tree_*
	Unit<P> :: newRuleTerm (const mdl :: Type* type, const bool final)
	{
		const Term_* term = Term_ :: template create<allocator :: Stack>
		(
			getBegin(), getEnd(),
			type,
			getTreeRule(),
			substitution_
		);
		tree_ = new Tree_
		(
			term,
			getBegin()->template isTerminal<direction :: LEFT>(Parser_ :: expression_),
			getEnd()->template isTerminal<direction :: RIGHT>(Parser_ :: expression_),
			tree_
		);
		return tree_;
	}
	template<class P>
	inline const typename Unit<P> :: Tree_*
	Unit<P> :: newVarTerm (const mdl :: Type* type)
	{
		const Term_*
			term = Term_ :: template create<allocator :: Stack> (getEnd(), type);
		newSubstitution (term);
		tree_ = new Tree_ (term, false, false, tree_);
		return tree_;
	}

	template<class P>
	inline value :: Integer
	Unit<P> :: getExpressionLength() const
	{
		value :: Integer length = 1;
		for (expr_Iterator_ iterator = getEnd(); iterator != getBegin(); -- iterator) {
			++ length;
		}
		return length;
	}
	template<class P>
	inline value :: Integer
	Unit<P> :: getTreeLength() const
	{
		value :: Integer length = 1;
		const tree_Iterator_ end (root_);
		if (end.isNull()) {
			return 0;
		}
		for (tree_Iterator_ iterator = Pair_ :: getSecond(); iterator != end; -- iterator) {
			if (iterator.isNull()) {
				-- length;
				break;
			}
			++ length;
		}
		return length;
	}

using manipulator :: endLine;
using manipulator :: tab;
using manipulator :: space;

	template<class P>
	inline void
	Unit<P> :: show (String& string) const
	{
		string << "unit." << endLine;
		string << "-------" << endLine;
		string << "expression: " << endLine;
		string << tab << "length: " << getExpressionLength() << endLine;
		string << tab;
		showExpression (string); string << endLine;
		string << endLine;
		string << "tree: " << endLine;
		string << tab <<"length: " << getTreeLength() << endLine;
		string << tab << "root: " << endLine;
		if (root_ == NULL) {
			string << tab << "NULL" << endLine;
		} else {
			root_->showTree (string); string << endLine;
			string << tab << "branch:" << endLine;
			string << tab;
			showTreeBranch (string); string << endLine;
		}

		string << tab << "end tree:" << endLine;
		if (Pair_ :: getSecond().isNull()) {
			string << tab << "NULL" << endLine;
		} else {
			string << *Pair_ :: getSecond();
		}
		string << endLine;
		string << tab << "substitution:" << endLine;
		if (substitution_ != NULL) {
			string << tab;
			substitution_->show (string);
		} else {
			string << tab << "NULL";
		}
	}
	template<class P>
	inline void
	Unit<P> :: show () const
	{
		String string (1000);
		show (string);
		std :: cerr << string;
	}

	template<class P>
	inline String&
	operator << (String& string, const Unit<P> unit)
	{
		unit.show (string);
		return string;
	}
	template<class P>
	inline std :: ostream&
	operator << (std :: ostream& os, const Unit<P> unit)
	{
		static String string (1024);
		string.clear();
		unit.show (string);
		os << string;
		return os;
	}

	/****************************
	 *		Private members
	 ****************************/

using manipulator :: space;

	template<class P>
	void
	Unit<P> :: showExpression (String& string) const
	{
		const value :: Integer length = getExpressionLength();
		Symbol buffer [length];

		expr_Iterator_ iterator = getEnd();
		for (value :: Integer i = 0; i < length; ++ i) {
			buffer [i] = *iterator;
			-- iterator;
		}
		for (value :: Integer i = 0; i < length; ++ i) {
			string << buffer [length - i - 1] << space;
		}
	}
	template<class P>
	void
	Unit<P> :: showTreeBranch (String& string) const
	{
		const value :: Integer length = getTreeLength();
		Symbol buffer [length];

		tree_Iterator_ iterator = Pair_ :: getSecond();
		for (value :: Integer i = 0; i < length; ++ i) {
			buffer [i] = *iterator;
			-- iterator;
		}
		for (value :: Integer i = 0; i < length; ++ i) {
			string << buffer [length - i - 1] << space;
		}
	}
}
}
}

#endif /*MDL_EXPRESSION_PARSE_UNIT_IPP_*/
