/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_expression_unify_sub_prod_Multyindex.ipp             */
/* Description:     vector of proof indexes                                  */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_EXPRESSION_UNIFY_SUB_PROD_MULTYINDEX_IPP_
#define MDL_EXPRESSION_UNIFY_SUB_PROD_MULTYINDEX_IPP_

namespace mdl {
namespace expression {
namespace unify {
namespace sub {
namespace prod {

	/****************************
	 *		Public members
	 ****************************/

	template<class S>
	inline
	Multyindex<S> :: Multyindex
	(
		const index :: Arity dimension,
		const prover :: proof :: Matrix& matrix
	) :
	dimension_ (dimension),
	matrix_ (matrix),
	indexVector_ (dimension)
	{
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			indexVector_.push() = 0;
		}
	}

	template<class S>
	inline const prover :: Evidence*
	Multyindex<S> :: getValue (const index :: Arity d) const
	{
		const prover :: Confirmation* const
			proofs = matrix_.getValue (d);
		const value :: Integer
			index = indexVector_.getValue (d);
		return proofs->get (index);
	}
	template<class S>
	inline bool
	Multyindex<S> :: operator ++() {
		return increment ();
	}
	template<class S>
	inline bool
	Multyindex<S> :: isOld() const {
		return !isNew();
	}
	template<class S>
	inline bool
	Multyindex<S> :: isNew() const
	{
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Confirmation* const
				proofs = matrix_.getValue (d);
			const value :: Integer
				index = indexVector_.getValue (d);
			if (!(index < proofs->getBoundary())) {
				return true;
			}
		}
		return false;
	}
	template<class S>
	inline void
	Multyindex<S> :: assign (Substivector_& substivector) const
	{
		substivector.clear();
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Evidence*
				evidence = getValue (d);
			substivector.push() = evidence->getSubstitution();
		}
	}
	template<class S>
	inline bool
	Multyindex<S> :: fitsDepth () const
	{
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Evidence*
				evidence = getValue (d);
			if (!Prover :: fitsDepth (evidence->getDepth())) {
				return false;
			}
		}
		return true;
	}
	template<class S>
	inline bool
	Multyindex<S> :: fitsComplexity () const
	{
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Evidence*
				evidence = getValue (d);
			if (!Prover :: fitsSubstitutionComplexity (evidence->getSubstitution()->getComplexity())) {
				return false;
			}
		}
		return true;
	}

	template<class S>
	inline value :: Integer
	Multyindex<S> :: getSearchVolume() const
	{
		const value :: Integer
			wholeVolume = getWholeVolume();
		const value :: Integer
			observedVolume = getObservedVolume();
		return (wholeVolume - observedVolume);
	}
	template<class S>
	inline value :: Integer
	Multyindex<S> :: getWholeVolume() const
	{
		value :: Integer wholeVolume = 1;
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Confirmation* const
				confirmation = matrix_.getValue (d);
			wholeVolume *= confirmation->getSize();
		}
		return wholeVolume;
	}
	template<class S>
	inline value :: Integer
	Multyindex<S> :: getObservedVolume() const
	{
		value :: Integer observedVolume = 1;
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Confirmation* const
				confirmation = matrix_.getValue (d);
			observedVolume *= confirmation->getBoundary();
		}
		return observedVolume;
	}

using manipulator :: endLine;
using manipulator :: tab;
using manipulator :: space;

	template<class S>
	template<class S1>
	inline void
	Multyindex<S> :: show (S1& string) const
	{
		string << "multyindex:" << endLine;
		string << "-----------" << endLine;
		string << tab << "dimension: " << dimension_ << endLine;
		string << tab << "whole volume: " << getWholeVolume() << endLine;
		string << tab << "observed volume: " << getObservedVolume() << endLine;
		string << tab << "search volume: " << getSearchVolume() << endLine;
		string << endLine;
		string << tab << "proofs:" << endLine;
		string << tab << "----------" << endLine;
		for (index :: Arity d = 0; d < dimension_; ++ d) {
			const prover :: Confirmation* const
				confirmation = matrix_.getValue (d);
			string << tab << tab << d << " size: " << confirmation->getSize() << endLine;
			string << tab << tab << d << " boundary: " << confirmation->getBoundary() << endLine;
			string << endLine;
		}
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class S>
	bool
	Multyindex<S> :: increment (const index :: Arity d)
	{
		if (d == dimension_) {
			return true;
		}
		const value :: Integer
			index = indexVector_.getValue (d);
		const value :: Integer
			boundary = matrix_.getValue (d)->getSize();

		if (index < boundary - 1) {
			++ indexVector_.getReference (d);
			return false;
		} else {
			indexVector_.getReference (d) = 0;
			return increment (d + 1);
		}
	}

	/********************************
	 *		Overloaded operators
	 ********************************/

	template<class S1, class S2>
	inline S1&
	operator << (S1& string, const Multyindex<S2>& multyindex)
	{
		multyindex.show (string);
		return string;
	}
	template<class S>
	inline std :: ostream&
	operator << (std :: ostream& os, const Multyindex<S>& multyindex)
	{
		nstd :: String<> string (1024);
		return os << (string << multyindex);
	}
}
}
}
}
}

#endif /*MDL_EXPRESSION_UNIFY_SUB_PROD_MULTYINDEX_IPP_*/
