/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File name:       mdl_expression_sub_tree_Unifier.ipp                      */
/* Description:     main unification algorithm. Static class                 */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_EXPRESSION_UNIFY_SUB_TREE_UNIFIER_IPP_
#define MDL_EXPRESSION_UNIFY_SUB_TREE_UNIFIER_IPP_

namespace mdl {
namespace expression {
namespace unify {
namespace sub {
namespace tree {

	/****************************
	 *		Public members
	 ****************************/

	template<class U>
	inline
	Unifier<U> :: Unifier
	(
		const index :: Arity dimension,
		const Substitution_* substitutionInt,
		prover :: evidence :: Vector& proofs,
		const value :: Integer boundary
	) :
	dimension_ (dimension),
	substitutionInt_ (substitutionInt),
	proofs_ (proofs),
	boundary_ (boundary)
	/*flags_ {true, true, true, true, true},
	stop_ {2, -1, -1, -1, -1},
	count_ {0, 0, 0, 0, 0}*/
	{ }

	template<class U>
	inline void
	Unifier<U> :: operator () (const Matrix_& matrix)
	{
		Unit_ unit (dimension_, matrix);
		unifyNode (unit);
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class U>
	void
	Unifier<U> :: unifyNode (Unit_ unit)
	{
		while (!unit.isNull()) {
			if (unit.isGood()) {
				if (unit.isLambda()) {
					unifyLast (unit);
				} else {
					Memory :: stack().push();
					tryUnit (unit);
					Memory :: stack().pop();
				}
			}
			+ unit;
		}
	}
	template<class U>
	inline void
	Unifier<U> :: tryUnit (Unit_ unit)
	{
		applyUnit (unit);
		if (!unit.fixedAreEqual()) {
			return;
		}
		if (unit.allAreEqual()) {
			unifyNext (unit); return;
		}
		const index :: Arity i = unit.getVariable();
		if (i == index :: undefined :: ARITY) {
			unifyNext (unit); return;
		}
		const index :: Arity j = unit.getTerm (i);
		if (j == index :: undefined :: ARITY) {
			unifyNext (unit); return;
		}
		tryReplace (unit, i, j);
	}
	template<class U>
	inline void
	Unifier<U> :: applyUnit (Unit_& unit)
	{
		for (index :: Arity i = 0; i < unit.getDimension(); ++ i) {
			if (unit.skip (i)) {
				continue;
			}
			const Substitution_* const
				substitution = unit.findSubstitution (i);
			if (substitution != NULL) {
				const Term_* const
					term = substitution->getTerm();
				applyTerm (unit, i, term);
			}
		}
	}
	template<class U>
	inline void
	Unifier<U> :: applyTerm
	(
		Unit_& unit,
		const index :: Arity i,
		const Term_* term
	)
	{
		typedef
			Builder
			<
				typename Unit_ :: Term_ :: template Alternative<allocator :: Stack> :: Result_,
				Term_
			>
			Builder;

		typename Unit_ :: Term_* const
			insertion = Builder :: build (term)->template
					castAway<typename Term_ :: Allocator_>();

		insertion->getEnd()->template
			link<expression :: direction :: RIGHT> (unit.iterator (i).getNode()->goRight());
		insertion->getBegin()->template
			link<expression :: direction :: LEFT> (unit.iterator (i).getNode()->goLeft());
		unit.insert (i, insertion);
	}
	template<class U>
	inline void
	Unifier<U> :: tryReplace
	(

		Unit_ unit,
		const index :: Arity i,
		const index :: Arity j
	)
	{
		const typename Terms_ :: Initial_ :: Node_*
			node = unit.getIterator (j)->getTerms().getInitial().getFirst();
		while (node != NULL)
		{
			const typename Unit_ :: Term_*
				term = node->getValue();
			Memory :: stack().push();
			tryTerm (unit, i, j, term);
			Memory :: stack().pop();
			node = node->getNext();
		}
	}
	template<class U>
	inline void
	Unifier<U> :: tryTerm
	(
		Unit_ unit,
		const index :: Arity i,
		const index :: Arity j,
		const typename Unit_ :: Term_* term
	)
	{
		if (!unit.checkTermType (i, j, term)) {
			return;
		}
		if (!unit.checkTermAccessible (j, term)) {
			return;
		}
		if (!unit.checkInsertionLength (i, term)) {
			return;
		}
		if (!unit.checkTermVariables (i, term)) {
			return;
		}
		typedef
			Builder
			<
				Term_,
				typename Unit_ :: Term_ :: template Alternative<allocator :: Stack> :: Result_
			>
			Builder;
		Term_* const
			insertion = Builder :: build (term);

		const Symbol&
			variable = *unit.getIterator (i);
		Substitution_*
			delta = Substitution_ :: template
				create<allocator :: Stack>(variable, insertion, NULL, true);
		unit.newSubstitution (delta);
		unit.intersectData (j, term);
		applyTerm (unit, i, insertion);

		tryUnit (unit);
	}
	template<class U>
	inline void
	Unifier<U> :: unifyNext (Unit_ unit)
	{
		Prover :: refresh ("unifying tree matrix");
		if (unit.isTerminal()) {
			if (unit.isFinal()) {
				unifyLast(unit);
			}
			return;
		}
		++ unit;
		/*if (!Prover :: fitsSubstitutionComplexity (unit.getComplexity())) {
			return;
		}*/
		Memory :: stack().push();
		unifyNode (unit);
		Memory :: stack().pop();
	}
	template<class U>
	inline void
	Unifier<U> :: unifyLast (Unit_ unit)
	{
		if (!unit.isLastColumn()) {
			//showBranch (unit);
			{
				Unit_ next (unit);
				next.shiftColumn (&unit);
				unifyNode (next);
			}
			//hideBranch (unit);
		} else {
			//showBranch (unit);
			Finalize<Unit_> finalize
			(
				substitutionInt_,
				proofs_,
				boundary_
			);
			finalize (unit);
			//hideBranch (unit);
		}
	}

	/****************************
	 *		Debug stuff
	 ****************************/

	/*template<class U>
	void
	Unifier<U> :: showBranch (Unit_ unit, const value :: Integer i)
	{
		if (!Counter :: stop (1)) {
			return;
		}
		if (i == DEBUG_DIM) {
			return;
		}
		if (unit.getIndex() == i) {
			if ((i == 0) && flags_ [0]) {
				++ count_ [0];
			} else if ((count_ [i - 1] == stop_ [i - 1] && flags_ [i])) {
				++ count_ [i];
			}
			if ((count_ [i] == stop_ [i]) && flags_ [i]) {
				std :: cerr << unit;
				if (i == 1) {
					Counter :: set (0, 2);
				}
			}
			if ((stop_ [i] == -1) &&
				((i == 0) || ((count_ [i - 1] == stop_ [i - 1]) && flags_ [i - 1]))) {
				std :: cerr << "count " << i << " = " << count_[i] << std :: endl;
				std :: cerr << unit;
				Counter :: set (0, 2);
			}
		}
		showBranch (unit, i + 1);
	}
	template<class U>
	void
	Unifier<U> :: hideBranch (Unit_ unit, const value :: Integer i)
	{
		if (!Counter :: stop (1)) {
			return;
		}
		if (i == DEBUG_DIM) {
			return;
		}
		if (unit.getIndex() == i) {
			if ((count_ [i] == stop_ [i]) && flags_ [i]) {
				flags_ [i] = false;
				if (i == 4) {
					Counter :: set (1000, 2);
				}
			}
		}
		hideBranch (unit, i + 1);
	}*/
}
}
}
}
}

#endif /*MDL_EXPRESSION_UNIFY_SUB_TREE_UNIFIER_IPP_*/

