/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_block_Assertion.cpp                             */
/* Description:     assertion class                                          */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_BLOCK_ASSERTION_CPP_
#define MDL_FORM_BLOCK_ASSERTION_CPP_

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	template<class H, class P>
	inline
	Assertion<H, P> :: Assertion
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const lexer :: Indent barIndent,
		const value :: Name name,
		const mdl :: Variables* const variables,
		const mdl :: Disjointeds* const disjointeds,
		const ParserHypothesisVector_& hypothesisVector,
		const ParserPropositionVector_& propositionVector
	) :
	Ancestor_ (location, comments, name),
	variables_ (variables),
	disjointeds_ (disjointeds),
	hypothesisVector_ (hypothesisVector),
	propositionVector_ (propositionVector),
	index_ (index :: undefined :: ASSERTION),
	barIndent_(barIndent) {
	}
	template<class H, class P>
	Assertion<H, P> :: ~ Assertion()
	{
		deleteAggregateObject (variables_);
		deleteAggregateObject (disjointeds_);
	}

	// mdl :: Assertion implementation
	template<class H, class P>
	inline index :: Assertion
	Assertion<H, P> :: getIndex() const {
		return index_;
	}

	// general mdl :: Assertion implementation
	template<class H, class P>
	const mdl :: Variables*
	Assertion<H, P> :: getVariables() const {
		return variables_;
	}
	template<class H, class P>
	const mdl :: Disjointeds*
	Assertion<H, P> :: getDisjointeds() const {
		return disjointeds_;
	}
	template<class H, class P>
	index :: Arity
	Assertion<H, P> :: getHypArity() const {
		return hypothesisVector_.getSize();
	}
	template<class H, class P>
	index :: Arity
	Assertion<H, P> :: getPropArity() const {
		return propositionVector_.getSize();
	}
	template<class H, class P>
	const mdl :: Hypothesis*
	Assertion<H, P> :: getHypothesis (const index :: Arity i) const {
		return hypothesisVector_.getValue (i);
	}
	template<class H, class P>
	const mdl :: Proposition*
	Assertion<H, P> :: getProposition (const index :: Arity i) const {
		return propositionVector_.getValue (i);
	}
	template<class H, class P>
	void
	Assertion<H, P> :: completeWithHypotheses (prover :: tree :: Premise* premises) const
	{
		for (index :: Arity i = 0; i < hypothesisVector_.getSize(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis = hypothesisVector_.getValue (i);
			const mdl :: proof :: Node* const
				proofNode = hypothesis->getProofNode();
			premises->addBranch
			(
				proofNode->getExpression(),
				proofNode
			);
		}
	}

using manipulator :: space;
using manipulator :: endLine;
using manipulator :: underline;

	// object :: Identifiable implementation
	template<class H, class P>
	const mdl :: Theory*
	Assertion<H, P> :: getTheory() const {
		return Identifiable :: getTheory();
	}
	template<class H, class P>
	value :: Name
	Assertion<H, P> :: getName() const {
		return Identifiable :: getName();
	}
	template<class H, class P>
	void
	Assertion<H, P> :: translateName (String& string, const index :: Arity index) const
	{
		if (Config :: keepLabels()) {
			Identifiable :: translateName (string);
			if (index > 0) {
				string << underline << index + 1;
			}
		} else {
			if (getKind() == Token :: AXIOM || getKind() == Token :: DEFINITION) {
				string << :: smm :: Token :: prefix_a_;
			} else {
				string << :: smm :: Token :: prefix_p_;
			}
			string << propositionVector_.getValue (index)->getTargetIndex();
		}
	}
	template<class H, class P>
	void
	Assertion<H, P> :: replicateName (String& string) const {
		Identifiable :: replicateName (string);
	}

	// object :: Buildable implementation
	template<class H, class P>
	void
	Assertion<H, P> :: buildTheory (mdl :: Theory* const theory)
	{
		//std :: cerr << Table :: names()->get(name_) << std :: endl;
		buildHeader (theory);
		buildHypothesis (theory);
		buildPropositions (theory);
	}

	// object :: Writable implementation
	template<class H, class P>
	void
	Assertion<H, P> :: translate (String& string) const
	{
		try {
			if (!Config :: removeComments() && (comments_ != NULL)) {
				string << location_.indent();
				comments_->translate (string);
			}
			string << :: smm :: Token :: blockBegin_ << endLine;
			variables_->translateVariables (string);
			disjointeds_->translate (string);
			for (index :: Arity i = 0; i < hypothesisVector_.getSize(); ++ i) {
				hypothesisVector_.getValue (i)->translate (string);
			}
			variables_->translateFloatings (string);
			for (index :: Arity i = 0; i < propositionVector_.getSize(); ++ i) {
				propositionVector_.getValue (i)->translate (string);
			}
			string << :: smm :: Token :: blockEnd_ << endLine;
			string << endLine;
		} catch (Error* error) {
			error->add (location_,"translation failed");
			throw error;
		}
	}
	template<class H, class P>
	void
	Assertion<H, P> :: replicate (String& string) const
	{
		replicateHeader (string);
		string << location_.indent() << Token :: openBrace_<< endLine;
		replicateHypothesis (string);
		if (getHypArity() > 0) {
			replicateBar (string);
		}
		replicatePropositions (string);
		string << location_.indent() << Token :: closeBrace_<< endLine;
	}

	// object :: Object implementation
	template<class H, class P>
	Size_t
	Assertion<H, P> :: getSizeOf() const {
		return sizeof (Assertion);
	}
	template<class H, class P>
	Size_t
	Assertion<H, P> :: getVolume() const
	{
		Size_t result = 0;
		result += Identifiable :: getVolume();
		result += getAggregateVolume (variables_);
		result += getAggregateVolume (disjointeds_);
		result += hypothesisVector_.getVolume();
		result += propositionVector_.getVolume();
		return result;
	}

	/****************************
	 *		Protected members
	 ****************************/

	template<class H, class P>
	inline void
	Assertion<H, P> :: replicateHeader (String& string) const
	{
		replicateName (string);
		string << space;
		variables_->replicate (string);
		if (disjointeds_ != NULL) {
			disjointeds_->replicate (string);
		}
		string << endLine;
	}
	template<class H, class P>
	inline void
	Assertion<H, P> :: replicateHypothesis (String& string) const
	{
		for (index :: Arity i = 0; i < hypothesisVector_.getSize(); i++) {
			hypothesisVector_.getValue (i)->replicate (string);
		}
	}
	template<class H, class P>
	inline void
	Assertion<H, P> :: replicateBar (String& string) const {
		string << barIndent_ << Token :: horisontalBar_ << endLine;
	}
	template<class H, class P>
	inline void
	Assertion<H, P> :: replicatePropositions (String& string) const
	{
		for (index :: Arity i = 0; i < propositionVector_.getSize(); ++ i) {
			propositionVector_.getValue (i)->replicate (string);
		}
	}

	template<class H, class P>
	inline void
	Assertion<H, P> :: buildHeader (mdl :: Theory* const theory)
	{
		form :: Identifiable :: buildTheory (theory);
		const_cast<mdl :: Variables*> (variables_)->buildTheory (theory);
	}
	template<class H, class P>
	void
	Assertion<H, P> :: buildHypothesis (mdl :: Theory* const theory)
	{
		for (index :: Arity i = 0; i < hypothesisVector_.getSize(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis = hypothesisVector_.getValue (i);
			hypothesis->setAssertion (this);
			const_cast<mdl :: Hypothesis* const>(hypothesis)->buildTheory (theory);
		}
	}
	template<class H, class P>
	void
	Assertion<H, P> :: buildPropositions (mdl :: Theory* const theory)
	{
		for (index :: Arity i = 0; i < propositionVector_.getSize(); ++ i) {
			const mdl :: Statement* const
				proposition = propositionVector_.getValue (i);
			proposition->setAssertion (this);
			const_cast<mdl :: Statement* const>(proposition)->buildTheory (theory);
		}
	}
	template<class H, class P>
	template<class T>
	void
	Assertion<H, P> :: add (const T* assertion, mdl :: Theory* const theory)
	{
		const_cast<index :: Assertion&>(index_) = Math :: assertions()->add (assertion);
		theory->add (index_, name_);
	}
}
}
}

#endif /*MDL_FORM_BLOCK_ASSERTION_CPP_*/
