/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_block_Definition.cpp                            */
/* Description:     definition class                                         */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_BLOCK_DEFINITION_CPP_
#define MDL_FORM_BLOCK_DEFINITION_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form.hpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Definition :: Definition
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const lexer :: Indent barIndent,
		const value :: Name name,
		const mdl :: Variables* const variables,
		const mdl :: Disjointeds* const disjointeds,
		const mdl :: term :: Definition* const defiendum,
		const mdl :: term :: Definition* const definiens,
		const vector :: parser :: Hypothesis& hypothesisVector,
		const vector :: parser :: Definition& propositionVector
	) :
	Ancestor_
	(
		location,
		comments,
		barIndent,
		name,
		variables,
		disjointeds,
		hypothesisVector,
		propositionVector
	),
	defiendum_ (defiendum),
	definiens_ (definiens) {
	}
	Definition :: ~ Definition()
	{
		deleteAggregateObject (defiendum_);
		deleteAggregateObject (definiens_);
	}

	// mdl :: Definition implementation
	const mdl :: term :: Definition*
	Definition :: getDefiendum() const {
		return defiendum_;
	}
	const mdl :: term :: Definition*
	Definition :: getDefiniens() const {
		return definiens_;
	}

	// mdl :: Assertion implementation
	value :: Kind
	Definition :: getKind() const {
		return Token :: DEFINITION;
	}

	const mdl :: Variables*
	Definition :: getVariables() const {
		return Ancestor_ :: getVariables();
	}
	const mdl :: Disjointeds*
	Definition :: getDisjointeds() const {
		return Ancestor_ :: getDisjointeds();
	}

	index :: Arity
	Definition :: getHypArity() const {
		return Ancestor_ :: getHypArity();
	}
	index :: Arity
	Definition :: getPropArity() const {
		return Ancestor_ :: getPropArity();
	}

	const mdl :: Hypothesis*
	Definition :: getHypothesis (const index :: Arity index) const {
		return Ancestor_ :: getHypothesis (index);
	}
	const mdl :: Proposition*
	Definition :: getProposition (const index :: Arity index) const {
		return Ancestor_ :: getProposition (index);
	}

	// object :: Identifiable implementation
	const mdl :: Theory*
	Definition :: getTheory() const {
		return Ancestor_ :: getTheory();
	}
	value :: Name
	Definition :: getName() const {
		return Ancestor_ :: getName();
	}
	void
	Definition :: translateName (String& string, const index :: Arity) const {
		Ancestor_ :: translateName (string);
	}
	void
	Definition :: replicateName (String& string) const {
		Ancestor_ :: replicateName (string);
	}

	// object :: Verifiable implementation
	void
	Definition :: check() const {
	}
	void
	Definition :: compress(){
	}
	void
	Definition :: prove() {
	}
	bool
	Definition :: isVerified() const {
		return true;
	}

	// object :: Buildable implementation
	void
	Definition :: buildTheory (mdl :: Theory* const theory)
	{
		Ancestor_ :: buildHeader (theory);
		Ancestor_ :: buildHypothesis (theory);
		build (theory, defiendum_);
		build (theory, definiens_);
		const array :: Expression*
			defiendum = defiendum_->getExpression();
		if (!defiendum->isPrimitive()) {
			if (Config :: warnings()) {
				std :: cerr << "Warning: defiendum term " << *defiendum << " ";
				std :: cerr << "is not primitive." << std :: endl;
			}
		}
		Ancestor_ :: buildPropositions (theory);
		Ancestor_ :: add (this, theory);
	}

using nstd :: manipulator :: space;
using nstd :: manipulator :: endLine;

	// object :: Writable implementation
	void
	Definition :: translate (String& string) const {
		Ancestor_ :: translate (string);
	}
	void
	Definition :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent() << Token :: definition_ << space;
		replicateHeader (string);
		string << location_.indent() << Token :: openBrace_<< endLine;
		defiendum_->replicate (string);
		definiens_->replicate (string);
		replicateHypothesis (string);
		replicateBar (string);
		replicatePropositions (string);
		string << location_.indent() << Token :: closeBrace_<< endLine;
	}

	// object :: Object implementation
	void
	Definition :: commitSuicide() {
		delete this;
	}
	Size_t
	Definition :: getVolume() const
	{
		Size_t result = 0;
		result += Ancestor_ :: getVolume();
		result += getAggregateVolume (defiendum_);
		result += getAggregateVolume (definiens_);
		return result;
	}
	Size_t
	Definition :: getSizeOf() const {
		return sizeof (Definition);
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Definition :: build
	(
		mdl :: Theory* const theory,
		const mdl :: term :: Definition* const term
	)
	{
		term->setDefinition (this);
		const_cast<mdl :: term :: Definition* const >(term)->buildTheory (theory);
	}
}
}
}

#endif /*MDL_FORM_BLOCK_DEFINITION_HPP_*/
