/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_block_Proof.cpp                                 */
/* Description:     proof class                                              */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_BLOCK_PROOF_CPP_
#define MDL_FORM_BLOCK_PROOF_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/block/mdl_form_block_Proof.hpp"
#include "mdl/form/proof/scope/mdl_form_proof_scope_Proof.hpp"

namespace mdl {
namespace form {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Proof :: Proof
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const mdl :: proof :: Line* const firstLine,
		const vector :: parser :: Step& stepVector
	) :
	location_ (location),
	comments_ (comments),
	firstLine_ (firstLine),
	lastLine_ (NULL),
	stepVector_ (stepVector),
	scope_ (NULL),
	isVerified_ (false),
	isCompressed_ (false) {
	}
	inline
	Proof :: Proof
	(
		const mdl :: proof :: Step* step,
		const index :: Step length
	) :
	location_ (step->getLocation()),
	comments_ (NULL),
	firstLine_ (NULL),
	lastLine_ (NULL),
	stepVector_ (length),
	scope_ (NULL),
	isVerified_ (false),
	isCompressed_ (false) {
	}
	Proof :: ~ Proof()
	{
		deleteAggregateObject (comments_);
		deleteAggregateObject (scope_);

		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			const mdl :: proof :: Line*
				next = line->getNext();
			deleteAggregateObject (line);
			line = next;
		}
	}

	inline const Log&
	Proof :: getLog() {
		return log_;
	}
	inline void
	Proof :: init()
	{
		Refiner :: init();
		Timer :: init();
	}
	inline void
	Proof :: release()
	{
		Refiner :: release();
		Timer :: release();
		log_.destroy();
	}

	// mdl :: Proof implementation
	bool
	Proof :: isCompressed() const {
		return isCompressed_;
	}
	index :: Step
	Proof :: getLength() const {
		return stepVector_.getSize();
	}
	const mdl :: proof :: Step*
	Proof :: getStep (const index :: Step index) const
	{
		if (index >= stepVector_.getSize()) {
			return NULL;
		}
		return stepVector_.getValue (index);
	}
	const mdl :: proof :: Line*
	Proof :: getFirst() const {
		return firstLine_;
	}
	const mdl :: proof :: Line*
	Proof :: getLast() const {
		return lastLine_;
	}
	void
	Proof :: addStep (const mdl :: proof :: Step* step) {
		addLine (step);
		const_cast<vector :: proof :: Step&>(stepVector_).add (step);
	}
	void
	Proof :: addVariable (const mdl :: proof :: Variable* variable) {
		addLine (variable);
	}
	void
	Proof :: addQed (const mdl :: proof :: Qed* qed) {
		addLine (qed);
	}

	// object :: Scoping implementation
	const mdl :: proof :: Scope*
	Proof :: buildScope (const mdl :: proof :: Scope* const scope)
	{
		scope_ = new proof :: scope :: Proof (scope, this);
		const mdl :: proof :: Scope* sc = scope_;

		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			sc = const_cast<mdl :: proof :: Line* const>(line)->buildScope (sc);
			if (line->getNext() == NULL) {
				lastLine_ = line;
			}
			line = line->getNext();
		}
		return sc;
	}
	const mdl :: proof :: Scope*
	Proof :: getScope() const {
		return scope_;
	}
	void
	Proof :: setScope (const mdl :: proof :: Scope* scope) const {
		scope_ = scope;
	}

	// object :: Verifiable implementation
	void
	Proof :: check() const
	{
		isVerified_ = false;
		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			line->check();
			line = line->getNext();
		}
		isVerified_ = true;
	}
	void
	Proof :: compress()
	{
		const mdl :: Assertion* const
			assertion = scope_->getAssertion();
		Refiner :: beforeCompression (assertion);
		if (!Refiner :: isCompressable (assertion)) {
			return;
		}
		questionSelf();
		Timer :: calculateTimeLimit (this);
		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			const mdl :: proof :: Line*
				next = line->getNext();
			if (line->getSort() == mdl :: proof :: Line :: STEP) {
				compressStep (line);
			}
			line = next;
		}
		Timer :: incrementProveActs (assertion);
		Refiner :: afterCompression (assertion);
		writeCompressionLog (assertion);
		isCompressed_ = true;
	}
	void
	Proof :: prove()
	{
		Timer :: calculateTimeLimit (this);
		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			const mdl :: proof :: Line*
				next = line->getNext();
			const_cast<mdl :: proof :: Line*>(line)->prove();
			if (line->getSort() == mdl :: proof :: Line :: STEP) {
				const mdl :: proof :: Step* const
					step = dynamic_cast<const mdl :: proof :: Step*>(line);
				if (step->getClass() == mdl :: proof :: Step :: QUESTION) {
					const mdl :: proof :: Question* const
						question = dynamic_cast<const mdl :: proof :: Question*>(step);
					if (question->isProved() && (question->getAnswer() == NULL)) {
						removeLine (line);
						const_cast<mdl :: proof :: Line*>(line)->commitSuicide();
					}
				}
			}
			line = next;
		}
		Timer :: incrementProveActs (scope_->getAssertion());
	}
	bool
	Proof :: isVerified() const {
		return isVerified_;
	}

	// object :: Buildable implementation
	void
	Proof :: buildTheory (mdl :: Theory* const theory)
	{
		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			const_cast<mdl :: proof :: Line* const>(line)->buildTheory (theory);
			line = line->getNext();
		}
	}

using nstd :: manipulator :: endLine;

	// object :: Writable implementation
	void
	Proof :: translate (String&) const {
	}
	void
	Proof :: replicate (String& string) const
	{
		string << location_.indent() << mdl :: Token :: proof_ << space;
		if (!Config :: removeComments() && (comments_ != NULL)) {
			comments_->replicate (string);
		}
		string << Token :: openBrace_ << endLine;
		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			line->replicate (string);
			line = line->getNext();
		}
		string << location_.indent() << Token :: closeBrace_ << endLine;
	}

	// object :: Object implementation
	void
	Proof :: commitSuicide() {
		delete this;
	}
	Size_t
	Proof :: getSizeOf() const  {
		return sizeof (Proof);
	}
	Size_t
	Proof :: getVolume() const
	{
		Size_t volume = 0;
		volume += Refiner :: getVolume();
		volume += Timer :: getVolume();
		volume += location_.getVolume();
		volume += getAggregateVolume (comments_);
		volume += stepVector_.getVolume();
		volume += getAggregateVolume (scope_);

		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			volume += getAggregateVolume (line);
			line = line->getNext();
		}
		return volume;
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Proof :: addLine (const mdl :: proof :: Line* line)
	{
		if (firstLine_ == NULL) {
			firstLine_ = line;
			lastLine_ = line;
			const_cast<mdl :: proof :: Line*>(line)->buildScope (scope_);
		} else {
			const mdl :: proof :: Scope* const
				scope = lastLine_->getScope();
			lastLine_->setNext (line);
			line->setPrev (lastLine_);
			lastLine_ = line;
			const_cast<mdl :: proof :: Line*>(line)->buildScope (scope);
		}
	}
	inline void
	Proof :: replaceLine
	(
		const mdl :: proof :: Line* oldLine,
		const mdl :: proof :: Line* newLine
	)
	{
		if (firstLine_ == oldLine) {
			firstLine_ = newLine;
		}
		if (lastLine_ == oldLine) {
			lastLine_ = newLine;
		}
		if (oldLine->getPrev() != NULL) {
			newLine->setPrev (oldLine->getPrev());
			oldLine->getPrev()->setNext (newLine);
			oldLine->setPrev (NULL);
		}
		if (oldLine->getNext() != NULL) {
			newLine->setNext (oldLine->getNext());
			oldLine->getNext()->setPrev (newLine);
			oldLine->setNext (NULL);
		}
		if (oldLine->getSort() == mdl :: proof :: Line :: STEP) {
			const mdl :: proof :: Step*
				oldStep = dynamic_cast<const mdl :: proof :: Step*>(oldLine);
			const mdl :: proof :: Step*
				newStep = dynamic_cast<const mdl :: proof :: Step*>(newLine);
			const_cast<vector :: proof :: Step&>(stepVector_).getReference (oldStep->getIndex()) = newStep;
		}
	}
	inline void
	Proof :: insertLine (const mdl :: proof :: Line* prev, const mdl :: proof :: Line* line)
	{
		const mdl :: proof :: Line*
			next = (prev == NULL) ? firstLine_: prev->getNext();
		if (prev == NULL) {
			firstLine_ = line;
		}
		if (lastLine_ == prev) {
			lastLine_ = line;
		}
		if (prev != NULL) {
			prev->setNext (line);
			line->setPrev (prev);
		}
		if (next != NULL) {
			next->setPrev (line);
			line->setNext (next);
		}
		if (line->getSort() == mdl :: proof :: Line :: STEP) {
			const mdl :: proof :: Step*
				step = dynamic_cast<const mdl :: proof :: Step*>(line);
			const_cast<vector :: proof :: Step&>(stepVector_).insert (step, step->getIndex());
			while (next != NULL) {
				next->incIndex (step->getIndex());
				next = next->getNext();
			}
		}
	}
	inline const mdl :: proof :: Line*
	Proof :: removeLine (const mdl :: proof :: Line* line)
	{
		const mdl :: proof :: Line*
			next = line->getNext();
		const mdl :: proof :: Line*
			prev = line->getPrev();
		if (firstLine_ == line) {
			firstLine_ = next;
		}
		if (lastLine_ == line) {
			lastLine_ = prev;
		}
		if (prev != NULL) {
			prev->setNext (next);
			line->setPrev (NULL);
		}
		if (next != NULL) {
			next->setPrev (prev);
			line->setNext (NULL);
		}
		if (line->getSort() == mdl :: proof :: Line :: STEP) {
			const mdl :: proof :: Step*
				step = dynamic_cast<const mdl :: proof :: Step*>(line);
			const_cast<vector :: proof :: Step&>(stepVector_).remove (step->getIndex());
			while (next != NULL) {
				next->decIndex (step->getIndex());
				next = next->getNext();
			}
		}
		return prev;
	}

	inline void
	Proof :: questionSelf()
	{
		const mdl :: proof :: Line*
			line = firstLine_;
		while (line != NULL) {
			const mdl :: proof :: Line*
				next = line->getNext();
			if (line->getSort() == mdl :: proof :: Line :: STEP) {
				const mdl :: proof :: Step*
					step = dynamic_cast<const mdl :: proof :: Step*>(line);
				const mdl :: proof :: Question*
					question = step->questionSelf();
				replaceLine (step, question);
			}
			line = next;
		}
	}
	inline bool
	Proof :: compressRest (const mdl :: proof :: Line* line)
	{
		const mdl :: proof :: Line*
			rest = line;
		while (rest != NULL) {
			const mdl :: proof :: Line*
				next = rest->getNext();
			if (rest->getSort() == mdl :: proof :: Line :: STEP) {
				const mdl :: proof :: Question* const
					question = dynamic_cast<const mdl :: proof :: Question*>(rest);
				const_cast<mdl :: proof :: Question*>(question)->compress();
				if (!question->isProved()) {
					return false;
				} /*else if (question->getGuess() == NULL) {
					removeLine (rest);
					const_cast<mdl :: proof :: Line*>(rest)->commitSuicide();
				}*/
			}
			rest = next;
		}
		return true;
	}
	inline void
	Proof :: confirmCompression (const mdl :: proof :: Line* line)
	{
		const mdl :: proof :: Line*
			rest = line;
		while (rest != NULL) {
			const mdl :: proof :: Line*
				next = rest->getNext();
			if (rest->getSort() == mdl :: proof :: Line :: STEP) {
				const mdl :: proof :: Question* const
					question = dynamic_cast<const mdl :: proof :: Question*>(rest);
				const_cast<mdl :: proof :: Question*>(question)->confirmCompression();
			}
			rest = next;
		}
	}
	inline void
	Proof :: rejectCompression (const mdl :: proof :: Line* line)
	{
		const mdl :: proof :: Line*
			rest = line;
		while (rest != NULL) {
			if (rest->getSort() == mdl :: proof :: Line :: STEP) {
				const mdl :: proof :: Question* const
					question = dynamic_cast<const mdl :: proof :: Question*>(rest);
				const_cast<mdl :: proof :: Question*>(question)->rejectCompression();
			}
			rest = rest->getNext();
		}
	}
	inline void
	Proof :: compressStep (const mdl :: proof :: Line* line)
	{
		const mdl :: proof :: Line*
			next = line->getNext();
		if (next->getSort() == mdl :: proof :: Line :: QED) {
			return;
		}
		const mdl :: proof :: Question*
			question = dynamic_cast<const mdl :: proof :: Question*>(line);

		const mdl :: proof :: Line*
			prev = removeLine (question);

		if (!compressRest (next)) {
			rejectCompression (next);
			insertLine (prev, line);
		} else {
			confirmCompression (next);
			const_cast<mdl :: proof :: Question*>(question)->commitSuicide();
		}
	}

using manipulator :: tab;
using manipulator :: endLine;
using manipulator :: space;

	void
	Proof :: writeCompressionLog (const mdl :: Assertion* assertion)
	{
		if (Config :: verbosity() == Config :: VERBOSE_STANDARD) {
			if (log_.init ("compression.log")) {
				log_.message() << "count:" << tab;
				log_.message() << "index:" << tab;
				log_.message() << "original:" << tab;
				log_.message() << "compressed:" << tab;
				log_.message() << "factor:" << tab;
				log_.message() << endLine;
			}
			log_.message() << Refiner :: getCompressedCount() << tab;
			log_.message() << assertion->getIndex() << tab;
			log_.message() << Refiner :: getOriginalProofLength() << tab;
			log_.message() << Refiner :: getCompressedProofLength() << tab;
			log_.message() << Refiner :: getCompressionFactor() << tab;
			log_.message() << endLine;
		}
	}

	/************************************
	 *		Private static attributes
	 ************************************/

	Log
		Proof :: log_;
}
}
}

#endif /*MDL_FORM_BLOCK_PROOF_CPP_*/
