/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_block_Rule.cpp                                  */
/* Description:     syntax grammar rule class                                */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_BLOCK_RULE_CPP_
#define MDL_FORM_BLOCK_RULE_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/block/mdl_form_block_Rule.hpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Rule :: Rule
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const value :: Name name,
		const mdl :: Variables* const variables,
		const mdl :: term :: Rule* const term
	) :
	form :: Identifiable (location, comments, name),
	type_ (NULL),
	variables_ (variables),
	term_ (term) {
	}
	Rule :: ~ Rule()
	{
		deleteAggregateObject (variables_);
		deleteAggregateObject (term_);
	}

	// mdl :: Rule implementation
	const mdl :: Type*
	Rule :: getType() const {
		return type_;
	}
	const mdl :: Variables*
	Rule :: getVariables() const {
		return variables_;
	}
	const mdl :: term :: Rule*
	Rule :: getTerm() const {
		return term_;
	}
	void
	Rule :: setType (const mdl :: Type* type) const {
		type_ = type;
	}

	// object :: Identifiable implementation
	const mdl :: Theory*
	Rule :: getTheory() const {
		return Identifiable :: getTheory();
	}
	value :: Name
	Rule :: getName() const {
		return Identifiable :: getName();
	}
	void
	Rule :: translateName (String& string, const index :: Arity) const
	{
		if (Config :: keepLabels()) {
			Identifiable :: translateName (string);
		} else {
			string << :: smm :: Token :: prefix_a_;
			string << term_->getTargetIndex();
		}
	}
	void
	Rule :: replicateName (String& string) const {
		Identifiable :: replicateName (string);
	}

	// object :: Verifiable implementation
	void
	Rule :: check() const {
	}
	void
	Rule :: compress() {
	}
	void
	Rule :: prove() {
	}
	bool
	Rule :: isVerified() const {
		return true;
	}

using manipulator :: space;
using manipulator :: endLine;

	// object :: Buildable implementation
	void
	Rule :: buildTheory (mdl :: Theory* const theory)
	{
		try {
			form :: Identifiable :: buildTheory (theory);
			term_->setRule (this);
			const_cast<mdl :: Variables* const>(variables_)->buildTheory (theory);
			const_cast<mdl :: term :: Rule* const>(term_)->buildTheory (theory);
			Math :: rules()->add (this);
		} catch (Error* error) {
			error->add (location_);
			error->message() << "rule ";
			replicateName (error->message());
			error->message() << endLine;
			throw error;
		}
	}

	// object :: Writable implementation
	void
	Rule :: translate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->translate (string);
			string << endLine;
		}
		string << :: smm :: Token :: blockBegin_ << endLine;
		variables_->translate (string);
		term_->translate (string);
		string << :: smm :: Token :: blockEnd_ << endLine;
		string << endLine;
	}
	void
	Rule :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent() << Token :: rule_ << space;
		replicateName (string);
		string << space;
		variables_->replicate (string); string << endLine;
		string << location_.indent() << Token :: openBrace_ << endLine;
		term_->replicate (string);
		string << location_.indent() << Token :: closeBrace_ << endLine;
		string << endLine;
	}

	// object :: Object implementation
	void
	Rule :: commitSuicide() {
		delete this;
	}
	Size_t
	Rule :: getVolume() const
	{
		Size_t result = 0;
		result += Identifiable :: getVolume();
		result += getAggregateVolume (variables_);
		result += getAggregateVolume (term_);
		return result;
	}
	Size_t
	Rule :: getSizeOf() const {
		return sizeof (Rule);
	}
}
}
}

#endif /*MDL_FORM_BLOCK_RULE_CPP_*/
