/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_block_Theorem.cpp                               */
/* Description:     theorem class                                            */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_BLOCK_THEOREM_CPP_
#define MDL_FORM_BLOCK_THEOREM_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form.hpp"
#include "mdl/form/proof/scope/mdl_form_proof_scope_Assertion.hpp"

namespace mdl {
namespace form {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Theorem :: Theorem
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const lexer :: Indent barIndent,
		const value :: Name name,
		const mdl :: Variables* const variables,
		const mdl :: Disjointeds* const disjointeds,
		const vector :: parser :: Hypothesis& hypothesisVector,
		const vector :: parser :: Provable& propositionVector,
		const mdl :: Proof* const proof
	) :
	Ancestor_
	(
		location,
		comments,
		barIndent,
		name,
		variables,
		disjointeds,
		hypothesisVector,
		propositionVector
	),
	proof_ (proof),
	scope_ (NULL) {
	}
	Theorem :: ~ Theorem()
	{
		deleteAggregateObject (proof_);
		deleteAggregateObject (scope_);
	}

	const mdl :: Proof*
	Theorem :: getProof() const {
		return proof_;
	}

	// mdl :: Assertion interface
	value :: Kind
	Theorem :: getKind() const {
		return Token :: THEOREM;
	}

	// object :: Verifiable implementation
	void
	Theorem :: check() const {
		//std :: cerr << Table :: names()->get(name_) << std :: endl;
		proof_->check();
	}
	void
	Theorem :: compress()
	{
		//std :: cerr << Table :: names()->get(name_) << std :: endl;
		const_cast<mdl :: Proof*>(proof_)->compress();
		try {
			proof_->check();
		} catch (Error* error) {
			error->add (location_, "in theorem:");
			error->message() << *this << endLine;
			throw error;
		}
	}
	void
	Theorem :: prove() {
		proof_->check();
		//std :: cerr << Table :: names()->get(name_) << std :: endl;
	}
	bool
	Theorem :: isVerified() const {
		return proof_->isVerified();
	}

using manipulator :: space;
using manipulator :: endLine;

	// object :: Buildable implementation
	void
	Theorem :: buildTheory (mdl :: Theory* const theory)
	{
		scope_ = new proof :: scope :: Assertion (this);
		buildComponents (theory);
		Ancestor_ :: add (this, theory);
	}

	// object :: Writable implementation
	void
	Theorem :: translate (String& string) const {
		Ancestor_ :: translate (string);
	}
	void
	Theorem :: replicate (String& string) const
	{
		replicateHeader (string);
		replicateBody (string);
	}

	// object :: Object implementation
	void
	Theorem :: commitSuicide() {
		delete this;
	}
	Size_t
	Theorem :: getSizeOf() const {
		return sizeof (Theorem);
	}
	Size_t
	Theorem :: getVolume() const
	{
		Size_t result = 0;
		result += Ancestor_ :: getVolume ();
		result += getAggregateVolume (proof_);
		result += getAggregateVolume (scope_);
		return result;
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Theorem :: buildComponents (mdl :: Theory* const theory)
	{
		Ancestor_ :: buildTheory (theory);
		try {
			const_cast<mdl :: Proof* const>(proof_)->buildTheory (theory);
			const_cast<mdl :: Proof* const>(proof_)->buildScope (scope_);
			buildHypothesis();
			buildPropositions();
		} catch (Error* error) {
			error->add (location_);
			error->message() << "in theorem ";
			replicateName (error->message());
			error->message() << endLine;
			throw error;
		}
	}
	void
	Theorem :: buildHypothesis()
	{
		for (index :: Arity i = 0; i < hypothesisVector_.getSize(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis = hypothesisVector_.getValue (i);
			mdl :: proof :: Node* const
				node = const_cast<mdl :: proof :: Node*>(hypothesis->getProofNode());
			node->setScope (scope_);
		}
	}
	void
	Theorem :: buildPropositions()
	{
		for (index :: Arity i = 0; i < propositionVector_.getSize(); ++ i) {
			const mdl :: Proposition* const
				proposition = propositionVector_.getValue (i);
			mdl :: proof :: Node* const
				node = const_cast<mdl :: proof :: Node*>(proposition->getProofNode());
			node->setScope (scope_);
		}
	}
	void
	Theorem :: replicateHeader (String& string, const bool isTheorem) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent();
		if (Config :: prove()) {
			if (isTheorem || proof_->isVerified()) {
				string << Token :: theorem_ << space;
			} else {
				string << Token :: problem_ << space;
			}
		} else {
			if (isTheorem && !proof_->isCompressed()) {
				string << Token :: theorem_ << space;
			} else {
				string << Token :: problem_ << space;
			}
		}
	}
	void
	Theorem :: replicateBody (String& string) const
	{
		Ancestor_ :: replicate (string);
		proof_->replicate (string);
		string << endLine;
	}
}
}
}

#endif /*MDL_FORM_BLOCK_THEOREM_CPP_*/
