/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_block_Type.cpp                                  */
/* Description:     type class                                               */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_BLOCK_TYPE_CPP_
#define MDL_FORM_BLOCK_TYPE_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/block/mdl_form_block_Type.hpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Type :: Type
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const value :: Name name,
		const vector :: parser :: Identificator& superNames
	) :
	form :: Identifiable (location, comments, name),
	superNames_ (superNames),
	supers_ (superNames.getSize()),
	subTypes_ (),
	ruleTree_ (new tree :: Rule())
	{
		for (index :: Type i = 0; i < superNames_.getSize(); ++ i) {
			const mdl :: Identificator* const
				name = superNames_.getValue (i);
			mdl :: Super* const
				super = new form :: block :: Super (this, name);
			supers_.add (super);
		}
	}
	Type :: ~ Type() {
		deleteAggregateObject (ruleTree_);
	}

using nstd :: manipulator :: underline;

	// mdl :: Type implementation
	const tree :: Rule*
	Type :: getRuleTree() const {
		return ruleTree_;
	}
	void
	Type :: addRule (const mdl :: Rule* const rule) {
		ruleTree_->addBranch (rule->getTerm()->getExpression(), rule);
	}

	bool
	Type :: isSuper (const mdl :: Type* const type) const
	{
		if (type == this) {
			return true;
		}
		for (index :: Type i = 0; i < supers_.getSize(); ++ i) {
			if (supers_.getValue (i)->getType() == type) {
				return true;
			}
		}
		return false;
	}
	bool
	Type :: isInfer (const mdl :: Type* const type) const
	{
		if (type == this) {
			return true;
		}
		for (index :: Type i = 0; i < subTypes_.getSize(); ++ i) {
			if (subTypes_.getValue (i) == type) {
				return true;
			}
		}
		return false;
	}

	index :: Type
	Type :: getSuperCount() const {
		return supers_.getSize();
	}
	const mdl :: Super*
	Type :: getSuper (const index :: Type index) const {
		return supers_.getValue (index);
	}
	const  mdl :: Super*
	Type :: getSuper (const mdl :: Type* const type) const
	{
		const mdl :: Super* super = NULL;
		for (index :: Type i = 0; i < supers_.getSize(); ++ i) {
			super = supers_.getValue(i);
			if (super->getType() == type) {
				return super;
			}
		}
		return NULL;
	}
	index :: Type
	Type :: getSubtypeCount() const {
		return subTypes_.getSize();
	}
	const  mdl :: Type*
	Type :: getSubtype (const index :: Type index) const {
		return subTypes_.getValue (index);
	}
	void
	Type :: addSubtype (const mdl :: Type* type)
	{
		if (subTypes_.contains (type)) {
			return;
		}
		subTypes_.add (type);
		for (index :: Type i = 0; i < supers_.getSize(); ++ i) {
			const mdl :: Type* const
				superType = supers_.getValue (i)->getType();
			const_cast<mdl :: Type*>(superType)->addSubtype (type);
		}
	}

	// object :: Identifiable implementation
	const mdl :: Theory*
	Type :: getTheory() const {
		return Identifiable :: getTheory();
	}
	value :: Name
	Type :: getName() const {
		return Identifiable :: getName();
	}
	const mdl :: Identificator*
	Type :: getIdentificator() const {
		return Identifiable :: getIdentificator();
	}
	void
	Type :: translateName (String& string, const index :: Arity) const {
		Identifiable :: translateName (string);
	}
	void
	Type :: replicateName (String& string) const {
		Identifiable :: replicateName (string);
	}

	// object :: Verifiable implementation
	void
	Type :: check() const {
	}
	void
	Type :: compress() {
	}
	void
	Type :: prove() {
	}
	bool
	Type :: isVerified() const {
		return true;
	}

using nstd :: manipulator :: space;
using nstd :: manipulator :: tab;
using nstd :: manipulator :: endLine;

	// object :: Buildable implementation
	void
	Type :: buildTheory (mdl :: Theory* const theory)
	{
		try {
			form :: Identifiable :: buildTheory (theory);
			for (index :: Type i = 0; i < supers_.getSize(); ++ i) {
				mdl :: Super* const
					super = const_cast<mdl :: Super*>(supers_.getValue (i));
				super->buildTheory (theory);
				super->addSuperSuper (supers_);
			}
			const index :: Type
				thisIndex = Math :: types()->add (this);
			theory->add (thisIndex, name_);
		} catch (Error* error) {
			error->add (location_);
			throw error;
		}
	}

	// object :: Writable implementation
	void
	Type :: translate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->translate (string);
			string << endLine;
		}
		string << location_.indent() << :: smm :: Token :: constant_ << space;
		translateName (string);
		string << space << :: smm :: Token :: end_ << endLine;
		string << endLine;
		for (index :: Type i = 0; i < supers_.getSize(); ++ i) {
			const mdl :: Super* const
				super = supers_.getValue (i);
			super->translate (string);
		}
		string << endLine;
	}
	void
	Type :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent() << Token :: type_ << space;
		replicateName (string);
		string << space;
		if (superNames_.isEmpty()) {
			string << Token :: semicolon_ << endLine;
		} else {
			string << Token :: colon_ <<  space;
			for (index :: Type i = 0; i < superNames_.getSize(); ++ i) {
				const mdl :: Identificator* const
					superName = superNames_.getValue (i);
				superName->replicate (string);
				if (i < superNames_.getSize() - 1) {
					string << Token :: comma_ << space;
				}
			}
			string << Token :: semicolon_ << endLine;
		}
		string << endLine;
	}

	// object :: Object implementation
	void
	Type :: commitSuicide() {
		delete this;
	}
	Size_t
	Type :: getVolume() const {
		Size_t volume = 0;
		volume += Identifiable :: getVolume();
		volume += superNames_.getVolume();
		volume += supers_.getVolume();
		volume += subTypes_.getVolume();
		volume += getAggregateVolume (ruleTree_);
		return volume;
	}
	Size_t
	Type :: getSizeOf() const {
		return sizeof (Type);
	}
}
}
}

#endif /*MDL_FORM_BLOCK_TYPE_CPP_*/
