/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_header_Variable.cpp                             */
/* Description:     variable declaration                                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_HEADER_VARIABLE_CPP_
#define MDL_FORM_HEADER_VARIABLE_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/header/mdl_form_header_Variable.hpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace header {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Variable :: Variable
	(
		const Location& location,
		const value :: Literal literal,
		const mdl :: Identificator* const identificator
	) :
	location_ (location),
	index_ (index :: undefined :: ARITY),
	symbol_ (literal),
	identificator_ (identificator) {
		const_cast<Symbol&>(symbol_).setReplaceable();
	}
	inline
	Variable :: Variable
	(
		const Location& location,
		const Symbol& symbol
	) :
	location_ (location),
	index_ (index :: undefined :: ARITY),
	symbol_ (symbol),
	identificator_ (new Identificator (symbol.getType()->getIdentificator())) {
		const_cast<Symbol&>(symbol_).setReplaceable();
	}
	Variable :: ~ Variable() {
		deleteAggregateObject (identificator_);
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;
using manipulator :: underline;

	// mdl :: header :: Variable implementation
	const Symbol&
	Variable :: getSymbol() const {
		return symbol_;
	}
	value :: Literal
	Variable :: getLiteral() const {
		return symbol_.getLiteral();
	}
	const mdl :: Type*
	Variable :: getType () const {
		return symbol_.getType();
	}
	void
	Variable :: setIndex (const index :: Arity index) const {
		index_ = index;
	}
	void
	Variable :: translateReference (String& string) const
	{
		if (Config :: keepLabels()) {
			getType()->translateName (string);
			string << underline << Table :: literals()->get (getLiteral());
		} else {
			string << smm :: Token :: prefix_f_ << index_;
		}
	}

	// object :: Buildable implementation
	void
	Variable :: buildTheory (mdl :: Theory* const theory)
	{
		const index :: Type
			index = theory->getIndex (identificator_);
		const_cast<Symbol&>(symbol_).setType (Math :: types()->get (index));
		if (getType() == NULL) {
			Error* error = new Error (location_, Error :: SEMANTIC);
			error->message() << "type " << *identificator_ << " is not declared." << endLine;
			throw error;
		}
	}

using manipulator :: underline;

	// object :: Writable implementation
	void
	Variable :: translate (String& string) const
	{
		translateReference (string);
		string << space;
		string << :: smm :: Token :: floating_ << space;
		getType()->translateName (string);
		string << space;
		string << Table :: literals()->get (getLiteral()) << space;
		string << :: smm :: Token :: end_ << endLine;
	}
	void
	Variable :: replicate (String& string) const
	{
		string << Token :: variable_ << space;
		string << Table :: literals()->get (getLiteral()) << space;
		string << Token :: colon_ << space;
		if (identificator_ != NULL) {
			identificator_->replicate (string);
		} else {
			string << "<undefined>";
		}
	}

	// object :: Object implementation
	void
	Variable :: commitSuicide() {
		delete this;
	}
	Size_t
	Variable :: getVolume() const {
		return getAggregateVolume (identificator_);
	}
	Size_t
	Variable :: getSizeOf() const {
		return sizeof (Variable);
	}
}
}
}

#endif /*MDL_FORM_HEADER_VARIABLE_CPP_*/
