/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_header_Variables.cpp                            */
/* Description:     variables declarations                                   */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_HEADER_VARIABLES_CPP_
#define MDL_FORM_HEADER_VARIABLES_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/header/mdl_form_header_Variables.hpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace header {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Variables :: Variables
	(
		const Location& location,
		const vector :: parser :: Variable& variableVector
	) :
	location_ (location),
	variableVector_ (variableVector),
	next_ (NULL) {
	}
	Variables :: ~ Variables() {
	}

	const mdl :: Variables*
	Variables :: getCurrent() {
		return current_;
	}

	// mdl :: Variables implementation
	index :: Arity
	Variables :: getSize() const {
		return variableVector_.getSize();
	}
	const mdl :: header :: Variable*
	Variables :: getVariable (const index :: Arity index) const {
		return variableVector_.getValue (index);
	}
	bool
	Variables :: contain (const Symbol& symbol) const
	{
		for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
			const mdl :: header :: Variable* const
				variable = variableVector_.getValue (i);
			if (variable->getLiteral() == symbol.getLiteral()) {
				return true;
			}
		}
		return false;
	}

	// object :: Typing implementation
	const mdl :: Type*
	Variables :: getType (const value :: Literal literal) const
	{
		for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
			const mdl :: header :: Variable* const
				variable = variableVector_.getValue (i);
			if (variable->getLiteral() == literal) {
				return variable->getType();
			}
		}
		return NULL;
	}
	const mdl :: header :: Variable*
	Variables :: getDeclaration (const value :: Literal literal) const
	{
		for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
			const mdl :: header :: Variable* const
				variable = variableVector_.getValue (i);
			if (variable->getLiteral() == literal) {
				return variable;
			}
		}
		if (next_ != NULL) {
			return next_->getDeclaration (literal);
		}
		return NULL;
	}
	const object :: Typing*
	Variables :: getPrevious() const {
		return NULL;
	}
	void
	Variables :: setNext (const object :: Typing*  next) const {
		next_ = next;
	}
	void
	Variables :: translateVariables (String& string) const
	{
		current_ = this;
		if (!variableVector_.isEmpty()) {
			string << tab << :: smm :: Token :: variable_ << space;
			for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
				const value :: Literal
					literal = variableVector_.getValue (i)->getLiteral();
				string << Table :: literals()->get (literal) << space;
			}
			string << :: smm :: Token :: end_ << endLine;
		}
		if (next_ != NULL) {
			next_->translateVariables (string);
		}
	}
	void
	Variables :: translateFloatings (String& string, const index :: Arity index) const
	{
		for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
			string << tab;
			variableVector_.getValue (i)->setIndex (index + i);
			variableVector_.getValue (i)->translate (string);
		}
		if (next_ != NULL) {
			next_->translateFloatings (string, index);
		}
	}

	// object :: Buildable implementation
	void
	Variables :: buildTheory (mdl :: Theory* const theory)
	{
		for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
			const_cast<mdl :: header :: Variable* const>(variableVector_.getValue (i))->buildTheory (theory);
		}
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// object :: Writable implementation
	void
	Variables :: translate (String& string) const
	{
		translateVariables (string);
		translateFloatings (string);
	}
	void
	Variables :: replicate (String& string) const
	{
		string << Token :: openBracket_;
		for (index :: Arity i = 0; i < variableVector_.getSize(); ++ i) {
			variableVector_.getValue (i)->replicate (string);
			if (i < variableVector_.getSize() - 1) {
				string << Token :: comma_ << space;
			}
		}
		string << Token :: closeBracket_ << space;
	}

	// object :: Object implementation
	void
	Variables :: commitSuicide() {
		delete this;
	}
	Size_t
	Variables :: getVolume() const {
		return variableVector_.getVolume();
	}
	Size_t
	Variables :: getSizeOf() const {
		return sizeof (Variables);
	}

	/***********************************
	 *		Static private attributes
	 ***********************************/

	const mdl :: Variables*
		Variables :: current_ = NULL;
}
}
}

#endif /*MDL_FORM_HEADER_VARIABLES_CPP_*/
