/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_header_statement_Definition.cpp                 */
/* Description:     definition statement class                               */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_HEADER_STATEMENT_DEFINITION_CPP_
#define MDL_FORM_HEADER_STATEMENT_DEFINITION_CPP_

#include "smm/lexer/smm_lexer_Token.hpp"
#include "mdl/form/mdl_form.hpp"

namespace mdl {
namespace form {
namespace header {
namespace statement {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Definition :: Definition
	(
		const Location& location,
		const index :: Arity index,
		const mdl :: Identificator* const identificator,
		const array :: Expression* const expression
	) :
	header :: statement :: Proposition (location, index, identificator, expression),
	expandedExpression_ (NULL),
	originalExpression_ (NULL),
	defiendum_ (NULL),
	definiens_ (NULL),
	metaVariables_ (NULL) {
	}
	Definition :: ~ Definition()
	{
		if (expression_ != originalExpression_) {
			deleteAggregateObject (originalExpression_);
		}
		if (expression_ != expandedExpression_) {
			deleteAggregateObject (expandedExpression_);
		}
		deleteAggregateObject (metaVariables_);
	}

	// statement :: Definition implementation
	void
	Definition :: setDefiendum (const mdl :: term :: Definition* defiendum) {
		defiendum_ = defiendum;
	}
	void
	Definition :: setDefiniens (const mdl :: term :: Definition* definiens) {
		definiens_ = definiens;
	}

	// mdl :: Proposition interface
	index :: Assertion
	Definition :: getSourceIndex() const {
		return Proposition :: getSourceIndex();
	}
	index :: Assertion
	Definition :: getTargetIndex() const {
		return Proposition :: getTargetIndex();
	}
	prover :: Experience&
	Definition :: getExperience() const {
		return Proposition :: getExperience();
	}

	// mdl :: Statement implementation
	void
	Definition :: setAssertion (const mdl :: Assertion* const assertion) const
	{
		Proposition :: setAssertion (assertion);
		defiendum_ = dynamic_cast<const mdl :: Definition*> (assertion_)->getDefiendum();
		definiens_ = dynamic_cast<const mdl :: Definition*> (assertion_)->getDefiniens();
	}
	index :: Arity
	Definition :: getIndex() const {
		return Proposition :: getIndex();
	}
	const mdl :: proof :: Node*
	Definition :: getProofNode() const {
		return Proposition :: proofNode_;
	}

	// object :: Expressive implementation
	const array :: Expression*
	Definition :: getExpression() const {
		return Proposition :: expression_;
	}
	const mdl :: Type*
	Definition :: getType() const {
		return header :: Line :: type_;
	}

	// object :: Buildable implementation
	void
	Definition :: buildTheory (mdl :: Theory* const theory)
	{
		substituteTerms();
		originalExpression_ = expression_;
		expression_ = expandedExpression_;
		Proposition :: buildTheory (theory);
		buildOriginal ();
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// object :: Writable implementation
	void
	Definition :: translate (String& string) const
	{
		string << tab;
		assertion_->translateName (string);
		string << space;
		string << :: smm :: Token :: axiomatic_ << space;
		Proposition :: translate (string);
		string << :: smm :: Token :: end_ << endLine;
	}
	void
	Definition :: replicate (String& string) const
	{
		if (originalExpression_ != NULL) {
			expression_ = originalExpression_;
		}
		Proposition :: replicate (string);
		if (originalExpression_ != NULL) {
			expression_ = expandedExpression_;
		}
	}

	// object :: Object implementation
	void
	Definition :: commitSuicide() {
		delete this;
	}
	Size_t
	Definition :: getVolume() const
	{
		Size_t volume = 0;
		volume += Proposition :: getVolume();
		if (expression_ != originalExpression_) {
			volume += getAggregateVolume (originalExpression_);
		}
		if (expression_ != expandedExpression_) {
			volume += getAggregateVolume (expandedExpression_);
		}
		volume += getAggregateVolume (metaVariables_);
		return volume;
	}
	Size_t
	Definition :: getSizeOf() const {
		return sizeof (Definition);
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Definition :: insert
	(
		value :: Literal* literals,
		index :: Literal& index,
		const mdl :: term :: Definition* term
	) const
	{
		const array :: Expression*
			expression = term->getExpression();
		for (index :: Literal i = 0 ; i < expression->getSize(); ++ i) {
			literals [index ++] = expression->getSymbol (i).getLiteral();
		}
	}
	inline index :: Literal
	Definition :: calculateLength() const
	{
		index :: Literal length = 0;
		length += expression_->getSize();
		length += defiendum_->getExpression()->getSize();
		length += definiens_->getExpression()->getSize();
		return length - 2;
	}
	inline void
	Definition :: substituteTerms ()
	{
		const index :: Literal
			length = calculateLength();

		value :: Literal literals [length];
		index :: Literal index = 0;

		if (defiendumKeyword_ == value :: undefined :: LITERAL) {
			defiendumKeyword_= Table :: literals()->find (Token :: defiendum_.getString());
			definiensKeyword_= Table :: literals()->find (Token :: definiens_.getString());
		}
		bool defiendumMet = false;
		bool definiensMet = false;
		for (index :: Literal i = 0; i < Proposition :: expression_->getSize(); ++ i) {
			substituteSymbol (i, literals, index, defiendumMet, definiensMet);
		}
		const_cast<const array :: Expression*&>(expandedExpression_) =
			new array :: Expression (literals, length);
	}
	inline void
	Definition :: substituteSymbol
	(
		const index :: Literal i,
		value :: Literal* literals,
		index :: Literal& index,
		bool& defiendumMet,
		bool& definiensMet
	) const
	{
		const Symbol& symbol = Proposition :: expression_->getSymbol (i);
		if (symbol.getLiteral() == defiendumKeyword_) {
			if (defiendumMet) {
				Error* error = new Error (location_,Error :: SEMANTIC);
				error->message() << "defiendum already met in definition.";
				throw error;
			}
			insert (literals, index, defiendum_);
			defiendumMet = true;
		} else if (symbol.getLiteral() == definiensKeyword_) {
			if (definiensMet) {
				Error* error = new Error (location_,Error :: SEMANTIC);
				error->message() << "definiens already met in definition.";
				throw error;
			}
			insert (literals, index, definiens_);
			definiensMet = true;
		} else {
			literals [index ++] = symbol.getLiteral();
		}
	}
	inline void
	Definition :: buildOriginal ()
	{
		try {
			metaVariables_ = new MetaVariables_
			(
				assertion_->getVariables(),
				defiendum_,
				definiens_
			);
			array :: Expression* const
				expression = const_cast<array :: Expression* const>(originalExpression_);
			expression->build (metaVariables_);
			expression->parse (type_);
		} catch (Error* error) {
			error->add (location_);
			throw error;
		}
		if (!originalExpression_->isPrimitive()) {
			Error* error = new Error (location_, Error :: SEMANTIC);
			error->message() << "definition proposition: " << *originalExpression_ << space;
			error->message() << "is not primitive." << endLine;
			throw error;
		}
	}

	/*************************************
	 *		Static private attributes
	 *************************************/

	value :: Literal
		Definition :: defiendumKeyword_ = value :: undefined :: LITERAL;
	value :: Literal
		Definition :: definiensKeyword_ = value :: undefined :: LITERAL;
}
}
}
}

#include "form/header/statement/mdl_form_header_statement_Definition_MetaVariables.cpp"

#endif /*MDL_FORM_HEADER_STATEMENT_DEFINITION_CPP_*/

