/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_Contents.cpp                                    */
/* Description:     contents of a theory                                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_CONTENTS_CPP_
#define MDL_FORM_CONTENTS_CPP_

#include "mdl/auxiliary/mdl_auxiliary.hpp"
#include "mdl/lexer/mdl_lexer.hpp"
#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form.hpp"

namespace mdl {
namespace form {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Contents :: Contents
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const value :: Name name
	) :
	form :: Identifiable (location, comments, name),
	componentList_ (),
	isVerified_ (false) {
	}
	Contents :: ~ Contents() {
	}

	// mdl :: Contents implementation
	void
	Contents :: addComponent (const object :: Verifiable* const component) {
		componentList_.addLast (component);
	}

	// object :: Identifiable implementation
	const mdl :: Theory*
	Contents :: getTheory() const {
		return form :: Identifiable :: getTheory();
	}
	value :: Name
	Contents :: getName() const {
		return form :: Identifiable :: getName();
	}
	void
	Contents :: translateName (String& string) const {
		form :: Identifiable :: translateName (string);
	}
	void
	Contents :: replicateName (String& string) const {
		form :: Identifiable :: replicateName (string);
	}

	// object :: Verifiable implementation
	void
	Contents :: check() const
	{
		isVerified_ = false;
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			node->getValue()->check();
			node = node->getNext();
		}
		isVerified_ = true;
	}
	void
	Contents :: compress()
	{
		isVerified_ = false;
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->compress();
			node = node->getNext();
		}
		isVerified_ = true;
	}
	void
	Contents :: prove()
	{
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->prove();
			node = node->getNext();
		}
	}
	bool
	Contents :: isVerified() const {
		return isVerified_;
	}

	// object :: Buildable implementation
	void
	Contents :: buildTheory (mdl :: Theory* const theory)
	{
		theory_ = theory;

		const index :: Form
			selfIndex = theory->getIndex (identificator_);
		if  (selfIndex == index :: undefined :: FORM) {
			throw new Error (Error :: SEMANTIC, "undeclared theory");
		}

		mdl :: Theory* const
			selfTheory = const_cast<mdl :: Theory* const> (Math :: theories()->get (selfIndex));
		if (selfTheory == NULL) {
			throw new Error (Error :: SEMANTIC, "undeclared theory");
		}

		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->buildTheory (selfTheory);
			node = node->getNext();
		}
	}

using manipulator :: endLine;
using manipulator :: space;

	// object :: Writable implementation
	void
	Contents :: translate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->translate (string);
			string << endLine;
		}
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			node->getValue()->translate (string);
			node = node->getNext();
		}
	}
	void
	Contents :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent();
		string << mdl :: Token :: contents_ << space;
		string << mdl :: Token :: of_ << space;
		form :: Identifiable :: replicateName (string);
		string << space << Token :: openBrace_ << endLine;
		string << endLine;

		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			node->getValue()->replicate (string);
			node = node->getNext();
		}
		string << location_.indent();
		string << Token :: closeBrace_ << endLine;
		string << endLine;
	}
	void
	Contents :: write (Output& output) const
	{
		if (!Config :: translate()) {
			if (!Config :: removeComments() && (comments_ != NULL)) {
				output.buffer() << location_.indent();
				comments_->replicate (output.buffer());
				output.buffer() << endLine;
			}
			output.buffer() << location_.indent();
			output.buffer() << mdl :: Token :: contents_ << space;
			output.buffer() << mdl :: Token :: of_ << space;
			form :: Identifiable :: replicateName (output.buffer());
			output.buffer() << space << Token :: openBrace_ << endLine;
			output.buffer() << endLine;
		}

		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			node->getValue()->write (output);
			node = node->getNext();
		}
		if (!Config :: translate()) {
			output.buffer() << location_.indent();
			output.buffer() << Token :: closeBrace_ << endLine;
			output.buffer() << endLine;
		}
	}

	// object :: Object implementation
	void
	Contents :: commitSuicide() {
		delete this;
	}
	Size_t
	Contents :: getVolume() const {
		Size_t volume = 0;
		volume += form :: Identifiable :: getVolume();
		volume += componentList_.getVolume();
		return volume;
	}
	Size_t
	Contents :: getSizeOf() const {
		return sizeof (Contents);
	}
}
}

#endif /*MDL_FORM_CONTENTS_CPP_*/
