/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_Source.cpp                                      */
/* Description:     source class                                             */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_SOURCE_CPP_
#define MDL_FORM_SOURCE_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form_Source.hpp"

namespace mdl {
namespace form {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Source :: Source
	(
		const bool isTop,
		const Lexer* const lexer,
		const mdl :: Comments* comments,
		const Path& sourcePath,
		const Path& targetPath
	) :
	isTop_ (isTop),
	lexer_ (lexer),
	sourcePath_ (sourcePath),
	targetPath_ (targetPath),
	comments_ (comments),
	componentList_ (),
	output_ (NULL),
	isVerified_ (false)
	{
		const Config :: Path_& path = *Config :: getTarget();
		if (path.getDepth() > 0) {
			const_cast<Path&> (targetPath_).directory(0) = path.getDirectory (0);
		}
		if (Config :: translate()) {
			targetPath_.extension() = "smm";
		} else {
			targetPath_.extension() = "mdl";
		}
	}
	Source :: ~ Source()
	{
		deleteAggregateObject (comments_);
		deleteAggregateObject (output_);
	}

using manipulator :: endLine;

	// mdl :: Source implementation
	void
	Source :: build ()
	{
		mdl :: Theory* const
			theory = Math :: theories()->getRoot();
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->buildTheory (theory);
			node = node->getNext();
		}
	}
	void
	Source :: addComponent (const object :: Verifiable* const component) {
		componentList_.addLast (component);
	}

	// object :: Verifiable implementation
	void
	Source :: check() const
	{
		isVerified_ = false;
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			node->getValue()->check();
			node = node->getNext();
		}
		isVerified_ = true;
	}
	void
	Source :: compress()
	{
		isVerified_ = false;
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->compress();
			node = node->getNext();
		}
		isVerified_ = true;
	}
	void
	Source :: prove()
	{
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->prove();
			node = node->getNext();
		}
	}
	bool
	Source :: isVerified() const {
		return isVerified_;
	}

	// object :: Buildable implementation
	void
	Source :: buildTheory (mdl :: Theory* const theory)
	{
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->buildTheory (theory);
			node = node->getNext();
		}
	}

	// object :: Writable implementation
	void
	Source :: translate (String& string) const {
	}
	void
	Source :: replicate (String& string) const {
	}
	void
	Source :: write() const
	{
		if (!isTop_ && !Config :: deepWrite()) {
			return;
		}
		const String& file = targetPath_.getString();
		targetPath_.provideDirectory();
		output_ = new Output (file);
		output_->open();
		if (comments_ != NULL) {
			if (Config :: translate()) {
				comments_->translate (output_->buffer());
			} else {
				comments_->replicate (output_->buffer());
			}
			output_->buffer() << endLine;
		}
		const Node_* node = componentList_.getFirst();
		while (node != NULL) {
			object :: Verifiable* const
				component = const_cast<object :: Verifiable* const>(node->getValue ());
			component->write();
			component->write (*output_);
			node = node->getNext();
		}
		output_->close();
	}

	// object :: Object implementation
	void
	Source :: commitSuicide() {
		delete this;
	}
	Size_t
	Source :: getSizeOf() const {
		return sizeof (Source);
	}
	Size_t
	Source :: getVolume() const
	{
		Size_t volume = 0;
		volume += componentList_.getVolume();
		volume += sourcePath_.getVolume();
		volume += targetPath_.getVolume();
		volume += getAggregateVolume (output_);
		volume += getAggregateVolume (comments_);
		return volume;
	}
}
}

#endif /*MDL_FORM_SOURCE_CPP_*/
