/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_Theory.cpp                                      */
/* Description:     theory class                                             */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_THEORY_CPP_
#define MDL_FORM_THEORY_CPP_

#include "mdl/auxiliary/mdl_auxiliary.hpp"
#include "mdl/lexer/mdl_lexer.hpp"
#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form_Identificator.hpp"
#include "mdl/form/mdl_form_Theory.hpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Theory :: Theory () :
	form :: Identifiable (),
	path_ (new form :: Identificator()),
	formMap_ () {
	}
	inline
	Theory :: Theory
	(
		const Location& location,
		const mdl :: Comments* const comments,
		const value :: Name name
	) :
	form :: Identifiable (location, comments, name),
	path_ (NULL),
	formMap_ () {
	}
	Theory :: ~ Theory() {
		deleteAggregateObject (path_);
	}

using nstd :: manipulator :: endLine;

	// mdl :: Theory implementation
	void
	Theory :: add (const index :: Form index, const value :: Name name) {
		formMap_.add (name, index);
	}
	void
	Theory :: checkUnused (const value :: Name name) const
	{
		if (formMap_.find (name) != index :: undefined :: FORM) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << "name " << Table :: names()->get (name) << " is already used." << endLine;
			throw error;
		}
		if (theory_ != NULL) {
			theory_->checkUnused (name);
		}
	}
	index :: Form
	Theory :: checkUsed (const mdl :: Identificator* const identificator) const
	{
		const index :: Form
			index = getIndexUp (identificator);
		if (index == index :: undefined :: FORM) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << "unknown identificator " << *identificator << endLine;
			throw error;
		}
		return index;
	}
	const mdl :: Identificator*
	Theory :: getPath() const {
		return path_;
	}
	index :: Form
	Theory :: getIndex (const mdl :: Identificator* const identificator) const {
		return getIndexUp (identificator);
	}

using nstd :: manipulator :: space;

	// object :: Identifiable implementation
	const mdl :: Theory*
	Theory :: getTheory() const {
		return form :: Identifiable :: getTheory();
	}
	value :: Name
	Theory :: getName() const {
		return form :: Identifiable :: getName();
	}
	void
	Theory :: translateName (String& string) const {
		form :: Identifiable :: translateName (string);
	}
	void
	Theory :: replicateName (String& string) const {
		form :: Identifiable :: replicateName (string);
	}

	// object :: Verifiable implementation
	void
	Theory :: check() const {
	}
	void
	Theory :: compress() {
	}
	void
	Theory :: prove() {
	}
	bool
	Theory :: isVerified() const {
		return true;
	}

	// object :: Buildable implementation
	void
	Theory :: buildTheory (mdl :: Theory* const theory)
	{
		try {
			form :: Identifiable :: buildTheory (theory);
			const index :: Theory
				thisIndex = Math :: theories()->add (this);
			theory->add (thisIndex, name_);

			const mdl :: Identificator* const
				theoryPath = theory_->getPath();
			path_ = new form :: Identificator (theoryPath);
			const mdl :: Identificator* path = path_;
			while (path->getNext() != NULL) {
				path = path->getNext();
			}
		path->setNext (new form :: Identificator (name_));
		} catch (Error* error) {
			error->add (location_);
			throw error;
		}
	}

using manipulator :: endLine;

	// object :: Writable implementation
	void
	Theory :: translate (String& string) const
	{
		if (theory_ != NULL) {
			if (!Config :: removeComments() && (comments_ != NULL)) {
				string << location_.indent();
				comments_->translate (string);
				string << endLine;
			}
			string << location_.indent() << :: smm :: Token :: commentBegin_ << space;
			string << Token :: theory_ << space;
			form :: Identifiable :: translateName (string);
			string << space << :: smm :: Token :: commentEnd_ << endLine << endLine;
		}
	}
	void
	Theory :: replicate (String& string) const
	{
		if (theory_ != NULL) {
			if (!Config :: removeComments() && (comments_ != NULL)) {
				string << location_.indent();
				comments_->replicate (string);
				string << endLine;
			}
			string << location_.indent() << Token :: theory_ << space;
			form :: Identifiable :: replicateName (string);
			string << Token :: semicolon_ << endLine;
			string << endLine;
		}
	}

	// object :: Object implementation
	void
	Theory :: commitSuicide() {
		delete this;
	}
	Size_t
	Theory :: getVolume() const
	{
		Size_t result = 0;
		result += Identifiable :: getVolume();
		result += formMap_.getVolume();
		result += getAggregateVolume (path_);
		return result;
	}
	Size_t
	Theory :: getSizeOf() const {
		return sizeof (Theory);
	}

	/****************************
	 *		Private members
	 ****************************/

	// mdl :: Theory implementation
	index :: Form
	Theory :: getIndexUp (const mdl :: Identificator* const identificator) const
	{
		const value :: Name
			name = identificator->getName();
		const index :: Form
			nameIndex = formMap_.find (name);

		if (nameIndex == index :: undefined :: FORM) {
			if (theory_ != NULL) {
				return theory_->getIndexUp (identificator);
			} else {
				return index :: undefined :: FORM;
			}
		} else {
			if (identificator->getNext() == NULL) {
				return nameIndex;
			} else {
				const mdl :: Theory* const
					theory = Math :: theories()->get (nameIndex);
				if (theory != NULL) {
					return theory->getIndexDown (identificator->getNext());
				} else {
					return index :: undefined :: FORM;
				}
			}
		}
	}
	index :: Form
	Theory :: getIndexDown (const mdl :: Identificator* const identificator) const
	{
		const value :: Name
			name = identificator->getName();
		const index :: Form
			nameIndex = formMap_.find (name);

		if (nameIndex == index :: undefined :: FORM) {
			return index :: undefined :: FORM;
		} else {
			if (identificator->getNext() == NULL) {
				return nameIndex;
			} else {
				const mdl :: Theory* const
					theory = Math :: theories()->get (nameIndex);
				if (theory != NULL) {
					return theory->getIndexDown (identificator->getNext());
				} else {
					return index :: undefined :: FORM;
				}
			}
		}
	}
}
}

#endif /*MDL_FORM_THEORY_CPP_*/
