/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_Link.cpp                                  */
/* Description:     proof link class                                         */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_LINK_CPP_
#define MDL_FORM_PROOF_LINK_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/proof/mdl_form_proof_Link.hpp"

namespace mdl {
namespace form {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Link :: Link
	(
		const Location& location,
		const value :: Kind kind,
		const mdl :: Identificator* const identificator,
		const index :: Arity proposition,
		const mdl :: proof :: References* const references
	) :
	location_ (location),
	kind_ (kind),
	identificator_ (identificator),
	proposition_ (proposition - 1),
	references_ (references),
	assertion_ (NULL) {
	}
	inline
	Link :: Link
	(
		const Location& location,
		const value :: Kind kind
	) :
	location_ (location),
	kind_ (kind),
	identificator_ (NULL),
	proposition_ (index :: undefined :: ARITY),
	references_ (new form :: proof :: References()),
	assertion_ (NULL) {
	}
	inline
	Link :: Link
	(
		const Location& location,
		const mdl :: Proposition* const proposition,
		const mdl :: proof :: References* const references
	) :
	location_ (location),
	kind_ (proposition->getAssertion()->getKind()),
	identificator_ (NULL),
	proposition_ (proposition->getIndex()),
	references_ (references),
	assertion_ (proposition->getAssertion()) {
	}
	Link :: ~ Link()
	{
		deleteAggregateObject (identificator_);
		deleteAggregateObject (references_);
	}


using manipulator :: space;

	// mdl :: proof :: Link implementation
	const mdl :: Assertion*
	Link :: getAssertion() const {
		return assertion_;
	}
	index :: Arity
	Link :: getPropositionIndex() const {
		return proposition_;
	}
	const mdl :: Proposition*
	Link :: getProposition() const {
		return assertion_->getProposition (proposition_);
	}
	void
	Link :: buildScope (const mdl :: proof :: Scope* const scope) {
		const_cast<mdl :: proof :: References* const> (references_)->build (scope);
	}
	const mdl :: proof :: References*
	Link :: getReferences() const {
		return references_;
	}
	mdl :: proof :: References*
	Link :: references() {
		return const_cast<mdl :: proof :: References*>(references_);
	}

	void
	Link :: translateReferences (String& string) const {
		references_->translate (string);
	}
	void
	Link :: translateAssertion (String& string) const
	{
		assertion_->translateName (string, proposition_);
		string << space;
	}

	// object :: Buildable implementation
	void
	Link :: buildTheory (mdl :: Theory* const theory)
	{
		const index :: Assertion
			index = theory->getIndex (identificator_);
		if (index == index :: undefined :: FORM) {
			throw new Error (location_, Error :: SEMANTIC, "reference to undefined assertion");
		}
		assertion_ = Math :: assertions()->get (index);
		if (kind_ != assertion_->getKind()) {
			if ((kind_ != Token :: THEOREM) || (assertion_->getKind() != Token :: PROBLEM)) {
				throw new Error (location_, Error :: SEMANTIC, "wrong assertion kind");
			}
		}
	}

using manipulator :: underline;

	// object :: Writable implementation
	void
	Link :: translate (String& string) const
	{
		references_->translate (string);
		assertion_->translateName (string);
		string << space;
	}
	void
	Link :: replicate (String& string) const
	{
		switch (kind_) {
		case Token :: AXIOM :
			string << Token :: axiom_ << space;
			break;
		case Token :: THEOREM :
		case Token :: PROBLEM :
			string << Token :: theorem_ << space;
			break;
		case Token :: DEFINITION :
			string << Token :: definition_ << space;
			break;
		case Token :: CLAIM :
			string << Token :: claim_ << space;
			break;
		}
		if (identificator_ != NULL) {
			identificator_->replicate (string);
			string << space;
		} else if (kind_ != Token :: CLAIM) {
			assertion_->replicateName (string);
			string << space;
		}
		if ((kind_ != Token :: CLAIM) && (proposition_ != 0)) {
			string << space << Token :: colon_;
			string << space << proposition_ + 1;
			string << space;
		}
		references_->replicate (string);
	}

	// object :: Object implementation
	void
	Link :: commitSuicide() {
		delete this;
	}
	Size_t
	Link :: getVolume() const {
		Size_t volume = 0;
		volume += getAggregateVolume (identificator_);
		volume += getAggregateVolume (references_);
		return volume;
	}
	Size_t
	Link :: getSizeOf() const {
		return sizeof (Link);
	}
}
}
}

#endif /*MDL_FORM_PROOF_LINK_CPP_*/
