/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_Qed.cpp                                   */
/* Description:     proof q.e.d. class                                       */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_QED_CPP_
#define MDL_FORM_PROOF_QED_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/proof/mdl_form_proof_Qed.hpp"

namespace mdl {
namespace form {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Qed :: Qed
	(
		const mdl :: proof :: Line* const prev,
		const Location& location,
		const mdl :: Comments* const comments,
		const value :: Kind kind,
		const index :: Arity indexProp,
		const index :: Step indexStep
	) :
	location_ (location),
	comments_ (comments),
	kind_ (kind),
	indexProp_ (indexProp),
	indexStep_ (indexStep),
	next_ (NULL),
	prev_ (prev),
	prop_ (NULL),
	step_ (NULL),
	scope_ (NULL),
	isVerified_ (false)
	{
		if (prev != NULL) {
			prev->setNext (this);
		}
	}
	inline
	Qed :: Qed
	(
		const mdl :: proof :: Provable* const claim,
		const mdl :: proof :: Step* const step
	) :
	location_ (step->getLocation()),
	comments_ (NULL),
	kind_ (Token :: CLAIM),
	indexProp_ (0),
	indexStep_ (step->getIndex()),
	next_ (NULL),
	prev_ (NULL),
	prop_ (claim),
	step_ (step),
	scope_ (step->getScope()),
	isVerified_ (false) {
	}
	Qed :: ~ Qed() {
		deleteAggregateObject (comments_);
	}

	// proof :: Qed implementation
	const mdl :: proof :: Provable*
	Qed :: getProp() const {
		return prop_;
	}
	const mdl :: proof :: Node*
	Qed :: getStep() const {
		return step_;
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// proof :: Line implementation
	Qed :: Sort
	Qed :: getSort() const {
		return QED;
	}
	const mdl :: proof :: Line*
	Qed :: getNext() const {
		return next_;
	}
	const mdl :: proof :: Line*
	Qed :: getPrev() const {
		return prev_;
	}
	void
	Qed :: incIndex (const index :: Step index) const
	{
		if (indexStep_ >= index) {
			++ indexStep_;
			step_ = NULL;
			const_cast<Qed*>(this)->buildScope (scope_);
		}
	}
	void
	Qed :: decIndex (const index :: Step index) const
	{
		if (indexStep_ > index) {
			-- indexStep_;
			step_ = NULL;
			const_cast<Qed*>(this)->buildScope (scope_);
		}
	}
	void
	Qed :: setNext (const mdl :: proof :: Line* const line) const {
		next_ = line;
	}
	void
	Qed :: setPrev (const mdl :: proof :: Line* const line) const {
		prev_ = line;
	}
	void
	Qed :: completeWithSteps (mdl :: prover :: tree :: Premise* premises) const {
		prev_->completeWithSteps (premises);
	}

	// object :: Scoping implementation
	const mdl :: proof :: Scope*
	Qed :: buildScope (const mdl :: proof :: Scope* const scope)
	{
		if (prop_ != NULL) {
			return scope_;
		}
		try {
			scope_ = scope;
			step_ = scope_->getStep (indexStep_);
			switch (kind_) {
			case PROP : {
				const mdl :: Proposition*
					statement = scope->getAssertion()->getProposition (indexProp_);
				const mdl :: statement :: Provable*
					provable = dynamic_cast<const mdl :: statement :: Provable*>(statement);
				prop_ = provable->getProof();
				break;
			}
			case CLAIM : {
				const mdl :: proof :: Step*
					claim = scope->getClaim();
				prop_ = dynamic_cast<const mdl :: proof :: Provable*>(claim);
				break;
			}
			}
		} catch (Error* error) {
			error->add (location_);
			throw error;
		}
		if (prop_ == NULL) {
			Error* error = new Error (location_, Error :: SEMANTIC);
			error->message() << "undefined reference to proposition ";
			error->message() << indexProp_ + 1 << " in proof." << endLine;
			throw error;
		}
		if (step_ == NULL) {
			Error* error = new Error (location_, Error :: SEMANTIC);
			error->message() << "undefined reference to step ";
			error->message() << indexStep_ + 1 << " in proof." << endLine;
			throw error;
		}
		return scope_;
	}
	const mdl :: proof :: Scope*
	Qed :: getScope() const {
		return scope_;
	}
	void
	Qed :: setScope (const mdl :: proof :: Scope* scope) const {
		scope_ = scope;
	}

	// object :: Verifiable implementation
	void
	Qed :: check() const
	{
		isVerified_ = false;
		const array :: Expression*
			theoremProposition = prop_->getExpression();
		if (step_ == NULL) {
			step_ = scope_->getStep (indexStep_);
		}
		const array :: Expression*
			proofProposition = step_->getExpression();
		if (*theoremProposition != *proofProposition) {
			Error* error = new Error (location_, Error :: CHECKER);
			error->message() << "in proof of ";
			scope_->getAssertion()->replicateName (error->message());
			error->message() << " theorem, "<< endLine;
			error->message() << "proposition does not match proof statement." << endLine;

			error->message() << tab << "proof statement: ";
			proofProposition->write (error->message());
			error->message() << endLine;
			error->message() << tab << "theorem proposition: ";
			theoremProposition->write (error->message());
			error->message() << endLine;
			throw error;
		}
		prop_->setProof (step_);
		isVerified_ = true;
	}
	void
	Qed :: compress() {
	}
	void
	Qed :: prove() {
	}
	bool
	Qed :: isVerified() const {
		return isVerified_;
	}

	// object :: Buildable implementation
	void
	Qed :: buildTheory (mdl :: Theory* const) {
	}

	// object :: Writable implementation
	void
	Qed :: translate (String&) const {
	}
	void
	Qed :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent() << Token :: qed_ << space;
		switch (kind_) {
		case PROP :
			string << Token :: prop_ << space;
			string << indexProp_ + 1 <<space;
			break;
		case CLAIM :
			string << Token :: claim_ << space;
			break;
		}
		string << Token :: equality_ << space;
		string << Token :: step_ << space;
		string << indexStep_ + 1 << space;
		string << Token :: semicolon_ << endLine;
	}

	// object :: Object implementation
	void
	Qed :: commitSuicide() {
		delete this;
	}
	Size_t
	Qed :: getVolume() const {
		return getAggregateVolume (comments_);
	}
	Size_t
	Qed :: getSizeOf() const {
		return sizeof (Qed);
	}
}
}
}

#endif /*MDL_FORM_PROOF_QED_CPP_*/
