/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_Step.cpp                                  */
/* Description:     proof step                                               */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_STEP_CPP_
#define MDL_FORM_PROOF_STEP_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/proof/mdl_form_proof_Step.hpp"

namespace mdl {
namespace form {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Step :: Step
	(
		const mdl :: proof :: Line* const prev,
		const Location& location,
		const mdl :: Comments* const comments,
		const index :: Step index,
		const mdl :: Identificator* const identificator,
		const array :: Expression* const expression,
		const mdl :: proof :: Scope* scope
	) :
	location_ (location),
	comments_ (comments),
	index_ (index),
	identificator_ (identificator),
	type_ (NULL),
	expression_ (expression),
	next_ (NULL),
	prev_ (prev),
	scope_ (scope),
	isVerified_ (false)
	{
		if (prev != NULL) {
			prev->setNext (this);
		}
	}
	inline
	Step :: Step (const mdl :: proof :: Step* const step) :
	location_ (step->getLocation()),
	comments_ (NULL),
	index_ (step->getIndex()),
	identificator_ (NULL),
	type_ (step->getType()),
	expression_ (new array :: Expression (step->getExpression())),
	next_ (step->getNext()),
	prev_ (step->getPrev()),
	scope_ (step->getScope()),
	isVerified_ (false) {
	}
	inline
	Step :: Step
	(
		const mdl :: proof :: Step* const step,
		const index :: Step index,
		const array :: Expression* const expression
	) :
	location_ (step->getLocation()),
	comments_ (NULL),
	index_ (index),
	identificator_ (NULL),
	type_ (expression->getType()),
	expression_ (expression),
	next_ (step->getNext()),
	prev_ (step->getPrev()),
	scope_ (step->getScope()),
	isVerified_ (false) {
	}
	Step :: ~ Step()
	{
		deleteAggregateObject (comments_);
		deleteAggregateObject (identificator_);
		deleteAggregateObject (expression_);
	}

	// proof :: Step implementation
	const mdl :: proof :: Question*
	Step :: questionSelf() const {
		return new proof :: step :: Question (this);
	}
	const Location&
	Step :: getLocation() const {
		return location_;
	}

	// proof :: Node implementation
	Step :: Kind
	Step :: getKind() const {
		return mdl :: proof :: Node :: STEP;
	}
	index :: Step
	Step :: getIndex() const {
		return index_;
	}

	// proof :: Line implementation
	Step :: Sort
	Step :: getSort() const {
		return mdl :: proof :: Line :: STEP;
	}
	const mdl :: proof :: Line*
	Step :: getNext() const {
		return next_;
	}
	const mdl :: proof :: Line*
	Step :: getPrev() const {
		return prev_;
	}
	void
	Step :: incIndex (const index :: Step index) const
	{
		if (index_ >= index) {
			++ index_;
		}
	}
	void
	Step :: decIndex (const index :: Step index) const
	{
		if (index_ > index) {
			-- index_;
		}
	}
	void
	Step :: setNext (const Line* const line) const {
		next_ = line;
	}
	void
	Step :: setPrev (const Line* const line) const {
		prev_ = line;
	}
	void
	Step :: completeWithSteps (mdl :: prover :: tree :: Premise* premises) const
	{
		if (prev_ != NULL) {
			prev_->completeWithSteps (premises);
		} else if (scope_->getClaim() != NULL) {
			scope_->getClaim()->completeWithHypotheses (premises);
		} else {
			scope_->getAssertion()->completeWithHypotheses (premises);
		}
	}

	// object :: Scoping implementation
	const mdl :: proof :: Scope*
	Step :: buildScope (const mdl :: proof :: Scope* const scope)
	{
		try {
			scope_ = scope;
			if (expression_->getTerm() == NULL) {
				const object :: Typing* const
					variables = scope_->getVariables();
				const_cast<array :: Expression* const>(expression_)->build (variables, true);
				const_cast<array :: Expression* const>(expression_)->parse (type_);
			}
			return scope;
		} catch (Error* error) {
			error->add (location_);
			throw error;
		}
	}
	const mdl :: proof :: Scope*
	Step :: getScope() const {
		return scope_;
	}
	void
	Step :: setScope (const mdl :: proof :: Scope* scope) const {
		scope_ = scope;
	}

	// object :: Expressive implementation
	const array :: Expression*
	Step :: getExpression() const {
		return expression_;
	}
	const mdl :: Type*
	Step :: getType() const {
		return type_;
	}

	// object :: Verifiable implementation
	bool
	Step :: isVerified() const {
		return isVerified_;
	}

	// object :: Buildable implementation
	void
	Step :: buildTheory (mdl :: Theory* const theory)
	{
		const index :: Type
			typeIndex = theory->getIndex (identificator_);
		type_ = Math :: types()->get (typeIndex);
	}

using manipulator :: endLine;
using manipulator :: space;

	// object :: Object implementation
	Size_t
	Step :: getVolume() const
	{
		Size_t volume = 0;
		volume += getAggregateVolume (comments_);
		volume += getAggregateVolume (identificator_);
		volume += getAggregateVolume (expression_);
		return volume;
	}
	Size_t
	Step :: getSizeOf() const {
		return sizeof (Step);
	}
}
}
}

#endif /*MDL_FORM_PROOF_STEP_CPP_*/
