/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_Variable.cpp                              */
/* Description:     variable declaration in proof                            */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_VARIABLE_CPP_
#define MDL_FORM_PROOF_VARIABLE_CPP_

#include "mdl/lexer/mdl_lexer.hpp"
#include "mdl/form/proof/scope/mdl_form_proof_scope_Variable.hpp"
#include "mdl/form/proof/mdl_form_proof_Variable.hpp"
#include "mdl/form/proof/mdl_form_proof_Variable_Declaration.cpp"

#include "smm/lexer/smm_lexer_Token.hpp"

namespace mdl {
namespace form {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Variable :: Variable
	(
		const mdl :: proof :: Line* const prev,
		const Location& location,
		const mdl :: Comments* const comments,
		const mdl :: header :: Variable* const variable
	) :
	location_ (location),
	comments_ (comments),
	variable_ (variable),
	scope_ (NULL),
	previous_ (NULL),
	nextLine_ (NULL),
	prevLine_ (prev),
	next_ (NULL)
	{
		if (prev != NULL) {
			prev->setNext (this);
		}
	}
	Variable :: ~ Variable()
	{
		deleteAggregateObject (comments_);
		deleteAggregateObject (variable_);
		deleteAggregateObject (scope_);
	}

	// object :: Typing implementation
	const mdl :: Type*
	Variable :: getType (const value :: Literal literal) const
	{
		if (variable_->getLiteral() == literal) {
			return variable_->getType();
		}
		if (previous_ != NULL) {
			return previous_->getType (literal);
		} else {
			return NULL;
		}
	}
	const mdl :: header :: Variable*
	Variable :: getDeclaration (const value :: Literal literal) const
	{
		if (variable_->getLiteral() == literal) {
			return variable_;
		}
		if (next_ != NULL) {
			return next_->getDeclaration (literal);
		} else {
			return NULL;
		}
	}
	const object :: Typing*
	Variable :: getPrevious() const {
		return previous_;
	}
	void
	Variable :: setNext (const object :: Typing*  next) const {
		next_ = next;
	}
	void
	Variable :: translateVariables (String& string) const
	{
		string << tab << :: smm :: Token :: variable_ << space;
		const value :: Literal literal = variable_->getLiteral();
		string << Table :: literals()->get (literal) << space;
		string << :: smm :: Token :: end_ << endLine;
		if (next_ != NULL) {
			next_->translateVariables (string);
		}
	}
	void
	Variable :: translateFloatings (String& string, const index :: Arity index) const
	{
		string << tab;
		variable_->setIndex (index);
		variable_->translate (string);
		if (next_ != NULL) {
			next_->translateFloatings (string, index + 1);
		}
	}

	// mdl :: proof :: Line implementation
	Variable :: Sort
	Variable :: getSort() const {
		return VARIABLE;
	}
	const mdl :: proof :: Line*
	Variable :: getNext() const {
		return nextLine_;
	}
	const mdl :: proof :: Line*
	Variable :: getPrev() const {
		return prevLine_;
	}
	void
	Variable :: incIndex (const index :: Step) const {
	}
	void
	Variable :: decIndex (const index :: Step) const {
	}
	void
	Variable :: setNext (const mdl :: proof :: Line* const line) const {
		nextLine_ = line;
	}
	void
	Variable :: setPrev (const mdl :: proof :: Line* const line) const {
		prevLine_ = line;
	}
	void
	Variable :: completeWithSteps (mdl :: prover :: tree :: Premise* premises) const
	{
		if (prevLine_ != NULL) {
			prevLine_->completeWithSteps (premises);
		} else if (scope_->getClaim() != NULL) {
			scope_->getClaim()->completeWithHypotheses (premises);
		} else {
			scope_->getAssertion()->completeWithHypotheses (premises);
		}
	}

	// object :: Scoping implementation
	const mdl :: proof :: Scope*
	Variable :: buildScope (const mdl :: proof :: Scope* const scope)
	{
		previous_ = scope->getVariables();
		previous_->setNext (this);
		scope_ = new scope :: Variable (scope, this);
		return scope_;
	}
	const mdl :: proof :: Scope*
	Variable :: getScope() const {
		return scope_;
	}
	void
	Variable :: setScope (const mdl :: proof :: Scope* scope) const {
		scope_ = scope;
	}

	// object :: Verifiable implementation
	void
	Variable :: check() const {
	}
	void
	Variable :: compress() {
	}
	void
	Variable :: prove() {
	}
	bool
	Variable :: isVerified() const {
		return true;
	}

	// object :: Buildable implementation
	void
	Variable :: buildTheory (mdl :: Theory* const theory) {
		const_cast<mdl :: header :: Variable* const> (variable_)->buildTheory (theory);
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// object :: Writable implementation
	void
	Variable :: translate (String& string) const
	{
		string << location_.indent();
		variable_->translate (string);
	}
	void
	Variable :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (comments_ != NULL)) {
			string << location_.indent();
			comments_->replicate (string);
			string << endLine;
		}
		string << location_.indent();
		variable_->replicate (string);
		string << Token :: semicolon_  << endLine;
	}

	// object :: Object implementation
	void
	Variable :: commitSuicide() {
		delete this;
	}
	Size_t
	Variable :: getVolume() const {
		Size_t volume = 0;
		volume += getAggregateVolume (comments_);
		volume += getAggregateVolume (variable_);
		volume += getAggregateVolume (scope_);
		return volume;
	}
	Size_t
	Variable :: getSizeOf() const {
		return sizeof (Variable);
	}
}
}
}

#include "mdl/form/proof/mdl_form_proof_Variable_Declaration.cpp"

#endif /*MDL_FORM_PROOF_VARIABLE_CPP_*/
