/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_step_Assertion.cpp                        */
/* Description:     assertion step inference                                 */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_STEP_ASSERTION_CPP_
#define MDL_FORM_PROOF_STEP_ASSERTION_CPP_

namespace mdl {
namespace form {
namespace proof {
namespace step {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Assertion :: Assertion
	(
		const mdl :: proof :: Line* const line,
		const Location& location,
		const mdl :: Comments* const comments,
		const index :: Step index,
		const mdl :: Identificator* const type,
		const mdl :: proof :: Link* link,
		const array :: Expression* const expression
	) :
	Step (line, location, comments, index, type, expression),
	link_ (link),
	substitution_ (NULL) {
	}
	inline
	Assertion :: Assertion
	(
		const mdl :: proof :: Step* const step,
		const index :: Step index,
		const mdl :: proof :: Link* link,
		const array :: Expression* const expression
	) :
	Step (step, index, expression),
	link_ (link),
	substitution_ (NULL) {
	}
	Assertion :: ~ Assertion()
	{
		deleteAggregateObject (link_);
		delete substitution_;
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;
using manipulator :: mode;

	// proof :: Step implementation
	mdl :: proof :: Step :: Class
	Assertion :: getClass() const {
		return mdl :: proof :: Step :: ASSERTION;
	}
	void
	Assertion :: addReference (const mdl :: proof :: Reference* const reference)
	{
		mdl :: proof :: References* const
			references = const_cast<mdl :: proof :: Link*>(link_)->references();
		references->add (reference);
	}

	// proof :: Node implementation
	void
	Assertion :: replicateReference (String& string) const {
		string << Token :: step_  << space << (index_ + 1);
	}
	void
	Assertion :: translateReference (String& string) const {
		link_->translate (string);
	}

	// proof :: Line implementation
	inline const mdl :: proof :: Line*
	Assertion :: getNext() const {
		return Step :: getNext();
	}
	inline const mdl :: proof :: Line*
	Assertion :: getPrev() const {
		return Step :: getPrev();
	}
	inline void
	Assertion :: incIndex (const index :: Step index) const {
		Step :: incIndex (index);
	}
	inline void
	Assertion :: decIndex (const index :: Step index) const {
		Step :: decIndex (index);
	}
	inline void
	Assertion :: setNext (const Line* const line) const {
		Step :: setNext (line);
	}
	inline void
	Assertion :: setPrev (const Line* const line) const {
		Step :: setPrev (line);
	}
	void
	Assertion :: completeWithSteps (mdl :: prover :: tree :: Premise* premises) const
	{
		premises->addBranch (expression_, this);
		Step :: completeWithSteps (premises);
	}

	// object :: Scoping implementation
	const mdl :: proof :: Scope*
	Assertion :: buildScope (const mdl :: proof :: Scope* const scope)
	{
		try {
			Step :: buildScope (scope);
			const_cast<mdl :: proof :: Link* const> (link_)->buildScope (scope_);
			return scope_;
		} catch (Error* error) {
			error->add (location_);
			error->message() << "inference " << *this << "  build failed." << endLine;
			throw error;
		}
	}
	inline const mdl :: proof :: Scope*
	Assertion :: getScope() const {
		return Step :: getScope();
	}
	inline void
	Assertion :: setScope (const mdl :: proof :: Scope* scope) const {
		Step :: setScope (scope);
	}

	// object :: Expressive implementation
	inline const array :: Expression*
	Assertion :: getExpression() const {
		return Step :: getExpression();
	}
	inline const mdl :: Type*
	Assertion :: getType() const {
		return Step :: getType();
	}

	// object :: Verifiable implementation
	void
	Assertion :: check() const
	{
		Memory :: stack().push();
		checkProposition();
		checkHypotheses();
		checkDisjoineds();
		typedef
			pool :: Unary<const array :: Substitution>
			Pool;
		Pool* const pool = new Pool();
		substitution_ = substitution_->clone<Pool> (pool);
		Memory :: stack().pop();
		isVerified_ = true;
	}
	void
	Assertion :: compress()
	{
		//addPositiveExperience();
		//addNegativeExperience();
	}
	void
	Assertion :: prove()
	{
		addPositiveExperience();
		addNegativeExperience();
	}

	// object :: Buildable implementation
	void
	Assertion :: buildTheory (mdl :: Theory* const theory)
	{
		try {
			const_cast<mdl :: proof :: Link* const> (link_)->buildTheory (theory);
			Step :: buildTheory (theory);
		} catch (Error* error) {
			error->add (location_);
			error->message() << "inference " << *this << "  build failed." << endLine;
			throw error;
		}
	}

	// object :: Writable implementation
	void
	Assertion :: translate (String& string) const
	{
		link_->translateReferences (string);
		translateVariables (string);
		link_->translateAssertion (string);
	}
	void
	Assertion :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (Step :: comments_ != NULL)) {
			string << Step :: location_.indent();
			Step :: comments_->replicate (string);
			string << endLine;
		}
		string << Step :: location_.indent() << Token :: step_ << space;
		string << (index_ + 1) << space << Token :: colon_ << space;

		if (type_ != NULL) {
			Step :: type_->replicateName (string); string << space;
		} else {
			string << "<unknown>"<< space;
		}
		string << Token :: equality_ << space;

		link_->replicate (string);
		string << space << Token :: turnstile_ << space;

		Step :: expression_->write (string); string << space;
		string << Token :: semicolon_ << endLine;
	}

	// object :: Object implementation
	void
	Assertion :: commitSuicide() {
		delete this;
	}
	Size_t
	Assertion :: getVolume() const
	{
		Size_t volume = 0;
		volume += Step :: getVolume();
		volume += getAggregateVolume (link_);
		volume += getAggregateVolume (substitution_);
		return volume;
	}
	Size_t
	Assertion :: getSizeOf() const {
		return sizeof (Assertion);
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Assertion :: translateVariables (String& string) const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();
		const mdl :: Variables* const
			variables = assertion->getVariables();
		for (index :: Arity i = 0; i < variables->getSize(); ++ i) {
			const mdl :: header :: Variable* const
				variable = variables->getVariable (i);
			const array :: Substitution*
				substitution = substitution_->find (variable->getSymbol());
			substitution->getTerm()->translate (string, variable->getType());
		}
	}
	void
	Assertion :: checkProposition() const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();

		if (!assertion->isVerified()) {
			referenceToUnverifiedError();
		}
		const mdl :: Proposition* const
			prop = link_->getProposition();
		const array :: Expression*
			proposition = prop->getExpression();

		const array :: Substitution* substitution = NULL;
		const bool
			success = proposition->unifyDirect (expression_, substitution, substitution_);
		if (!success || substitution != NULL) {
			propositionUnificationError();
		}
	}
	void
	Assertion :: checkHypotheses() const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();
		for (index :: Arity index = 0; index < link_->getReferences()->getArity(); ++ index) {
			const mdl :: Hypothesis* const
				hyp = assertion->getHypothesis (index);
			const array :: Expression*
				hypothesis = hyp->getExpression();
			const mdl :: proof :: Node*
				node = link_->getReferences()->get (index)->getNode();
			const array :: Expression*
				reference = node->getExpression();

			const array :: Substitution* substitution = NULL;
			const bool
				success = hypothesis->unifyDirect (reference, substitution, substitution_);
			if (!success || substitution != NULL) {
				hypothesisUnificationError (index);
			}
		}
	}
	void
	Assertion :: checkDisjoineds() const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();
		const mdl :: Disjointeds* const
			disjointeds = assertion->getDisjointeds();
		const mdl :: Assertion* const
			theorem = scope_->getAssertion();
		disjointeds->check (substitution_, theorem);
	}

	void
	Assertion :: addPositiveExperience()
	{
		const mdl :: Proposition* const
			proposition = link_->getProposition();
		proposition->getExperience().addPositive (this);
	}
	void
	Assertion :: addNegativeExperience()
	{
		typedef
			forest :: Proposition :: Unify_<array :: Expression> :: Result_
			Unified_;
		const forest :: Proposition*
			propositions = Math :: assertions()->getPropositionTree();
		Memory :: stack().push();
		const mdl :: Proposition* const
			firstTheoremProposition = scope_->getAssertion()->getProposition (0);
		Unified_ result =
			propositions->unifyDirect<array :: Expression>
			(
				expression_, NULL, NULL,
				firstTheoremProposition->getSourceIndex() - 1
			);
		const Unified_ :: Node_*
			node = result.getFirst();
		while (node != NULL) {
			typedef
				forest :: Proposition :: Data_
				Data_;
			const Data_&
				data = node->getReference().getSource()->getData();
			const Data_ :: Node_*
				dataNode = data.getFirst();
			while (dataNode != NULL) {
				const mdl :: Proposition*
					proposition = dataNode->getValue();
				if (proposition != link_->getProposition()) {
					proposition->getExperience().addNegative (this);
				}
				dataNode = dataNode->getNext();
			}
			node = node->getNext();
		}
		Memory :: stack().pop();
	}

	void
	Assertion :: referenceToUnverifiedError () const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();

		Error* error = new Error (location_, Error :: CHECKER);
		error->message() << "in ";
		scope_->getAssertion()->replicateName (error->message());
		error->message() << " theorem proof"<< endLine;
		error->message() << "assertion ";
		assertion->replicateName (error->message());
		error->message() << " is not verified." << endLine;
		throw error;
	}
	void
	Assertion :: propositionUnificationError () const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();
		const index :: Arity
			index = link_->getPropositionIndex();
		const mdl :: Statement* const
			prop = link_->getProposition ();
		const array :: Expression*
			proposition = prop->getExpression();

		expression :: Mode mode;
		//mode << expression :: Mode :: SHOW_REPLACEABILITY;
		//mode << expression :: Mode :: SHOW_TERMS;

		Error* error = new Error (location_, Error :: CHECKER);
		error->message() << mode << "in ";
		scope_->getAssertion()->replicateName (error->message());
		error->message() << " theorem proof"<< endLine;
		error->message() << "assertion ";
		assertion->replicateName (error->message());
		if (assertion->getPropArity() > 1) {
			error->message() << " proposition " << (index + 1) << " ";
		} else {
			error->message() << " ";
		}
		error->message() << "unification " << endLine;
		error->message() << "with proof step " << (index_ + 1) << " fails." << endLine;

		error->message() << tab << "proof step expression: ";
		expression_->write (error->message());
		expression_->showTerms (error->message());
		error->message() << endLine;
		error->message() << tab << "assertion proposition: ";
		proposition->write (error->message());
		proposition->showTerms (error->message());
		error->message() << endLine;
		if (substitution_ != NULL) {
			error->message() << tab << "substitution: ";
			substitution_->show (error->message());
		} else {
			error->message() << tab << "substitution: NULL";
			error->message() << endLine;
		}
		throw error;
	}
	void
	Assertion :: hypothesisUnificationError (const index :: Arity index) const
	{
		const mdl :: Assertion* const
			assertion = link_->getAssertion();
		const mdl :: Hypothesis* const
			hyp = assertion->getHypothesis (index);
		const array :: Expression*
			hypothesis = hyp->getExpression();
		const array :: Expression*
			reference = link_->getReferences()->get (index)->getNode()->getExpression();

		Error* error = new Error (location_, Error :: CHECKER);
		error->message() << "in ";
		scope_->getAssertion()->replicateName (error->message());
		error->message() << " theorem proof"<< endLine;
		error->message() << "assertion ";
		assertion->replicateName (error->message());
		error->message() << " hypothesis " << (index + 1) << " unification" << endLine;
		error->message() << "with reference in step " << (index_ + 1) << " fails." << endLine;

		error->message() << tab << "reference expression: ";
		reference->write (error->message());
		error->message() << endLine;
		error->message() << tab << "assertion hypothesis: ";
		hypothesis->write (error->message());
		error->message() << endLine;
		if (substitution_ != NULL) {
			error->message() << tab << "substitution: ";
			substitution_->show (error->message());
		} else {
			error->message() << tab << "substitution: NULL";
			error->message() << endLine;
		}
		throw error;
	}
}
}
}
}

#endif /*MDL_FORM_PROOF_STEP_ASSERTION_CPP_*/
