/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_step_Claim.cpp                            */
/* Description:     step with claim inference                                */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_STEP_CLAIM_CPP_
#define MDL_FORM_PROOF_STEP_CLAIM_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form.hpp"

namespace mdl {
namespace form {
namespace proof {
namespace step {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Claim :: Claim
	(
		const mdl :: proof :: Line* const line,
		const Location& location,
		const mdl :: Comments* const comments,
		const index :: Step index,
		const mdl :: Identificator* const type,
		const mdl :: proof :: Link* const link,
		const array :: Expression* const expression,
		const mdl :: Proof* const proof
	) :
	Ancestor_ (line, location, comments, index, type, expression),
	link_ (link),
	proof_ (proof),
	node_ (NULL),
	scope_ (NULL) {
	}
	inline
	Claim :: Claim
	(
		const mdl :: proof :: Step* question,
		const index :: Step proofLength
	) :
	Ancestor_ (question),
	link_ (new form :: proof :: Link (Ancestor_ :: location_, Token :: CLAIM)),
	proof_ (new form :: Proof (question, proofLength)),
	node_ (NULL),
	scope_ (new scope :: Claim (Ancestor_ :: scope_, this)) {
		const_cast<mdl :: Proof*>(proof_)->buildScope (scope_);
	}
	Claim :: ~ Claim()
	{
		deleteAggregateObject (proof_);
		deleteAggregateObject (link_);
		deleteAggregateObject (scope_);
	}

	// proof :: Claim implementation
	const mdl :: Proof*
	Claim :: getProof() const {
		return proof_;
	}
	void
	Claim :: completeWithHypotheses (mdl :: prover :: tree :: Premise* premises) const
	{
		for (index :: Arity i = 0; i < link_->getReferences()->getArity(); ++ i) {
			const mdl :: proof :: Node* const
				node = link_->getReferences()->get (i)->getNode();
			premises->addBranch (node->getExpression(), node);
		}
	}
	void
	Claim :: addStep (const mdl :: proof :: Step* const step) {
		const_cast<mdl :: Proof*>(proof_)->addStep (step);
	}
	void
	Claim :: addVariable (const Symbol& symbol)
	{
		const object :: Typing* const
			typing =
				(proof_->getLast() == NULL) ?
				proof_->getScope()->getVariables() :
				proof_->getLast()->getScope()->getVariables();

		if (typing->getType (symbol.getLiteral()) == NULL) {
			const mdl :: proof :: Line* const
				last = proof_->getLast();

			const mdl :: proof :: Variable* const
				variable = new form :: proof :: Variable
				(
					last,
					Ancestor_ :: location_,
					NULL,
					new form :: header :: Variable
					(
						Ancestor_ :: location_,
						symbol
					)
				);
			const_cast<mdl :: Proof*>(proof_)->addVariable (variable);
		}
	}
	const mdl :: proof :: Link*
	Claim :: getLink() const {
		return link_;
	}
	mdl :: proof :: Link*
	Claim :: link() {
		return const_cast<mdl :: proof :: Link*>(link_);
	}

	// proof :: Provable implementation
	void
	Claim :: setProof (const mdl :: proof :: Node* const node) const {
		node_ = node;
	}

	// proof :: Step implementation
	mdl :: proof :: Step :: Class
	Claim :: getClass() const {
		return mdl :: proof :: Step :: CLAIM;
	}
	void
	Claim :: addReference (const mdl :: proof :: Reference* const reference)
	{
		mdl :: proof :: References* const
			references = const_cast<mdl :: proof :: Link*>(link_)->references();
		references->add (reference);
	}

using manipulator :: endLine;
using manipulator :: space;

	// proof :: Node implementation
	void
	Claim :: replicateReference (String& string) const {
		string << Token :: step_  << space << index_;
	}
	void
	Claim :: translateReference (String& string) const {
	}

	// proof :: Line implementation
	inline const mdl :: proof :: Line*
	Claim :: getNext() const {
		return Ancestor_ :: getNext();
	}
	inline const mdl :: proof :: Line*
	Claim :: getPrev() const {
		return Ancestor_ :: getPrev();
	}
	inline void
	Claim :: incIndex (const index :: Step index) const {
		Ancestor_ :: incIndex (index);
	}
	inline void
	Claim :: decIndex (const index :: Step index) const {
		Ancestor_ :: decIndex (index);
	}
	inline void
	Claim :: setNext (const mdl :: proof :: Line* const line) const {
		Ancestor_ :: setNext (line);
	}
	inline void
	Claim :: setPrev (const mdl :: proof :: Line* const line) const {
		Ancestor_ :: setPrev (line);
	}
	void
	Claim :: completeWithSteps (mdl :: prover :: tree :: Premise* premises) const
	{
		premises->addBranch (expression_, this);
		Step :: completeWithSteps (premises);
	}

	// object :: Scoping implementation
	const mdl :: proof :: Scope*
	Claim :: buildScope (const mdl :: proof :: Scope* const scope)
	{
		Ancestor_ :: buildScope (scope);
		try {
			const_cast<mdl :: proof :: Link*>(link_)->buildScope (scope);
			scope_ = new scope :: Claim (scope, this);
			const_cast<mdl :: Proof* const>(proof_)->buildScope (scope_);
			return scope;
		} catch (Error* error) {
			error->add (location_);
			throw error;
		}
	}
	inline const mdl :: proof :: Scope*
	Claim :: getScope() const {
		return Ancestor_ :: getScope();
	}
	inline void
	Claim :: setScope (const mdl :: proof :: Scope* scope) const {
		Ancestor_ :: setScope (scope);
	}

	// object :: Expressive implementation
	inline const array :: Expression*
	Claim :: getExpression() const {
		return Ancestor_ :: expression_;
	}
	inline const mdl :: Type*
	Claim :: getType() const {
		return Ancestor_ :: type_;
	}

	// object :: Verifiable implementation
	void
	Claim :: check() const {
		proof_->check();
	}
	void
	Claim :: compress() {
		const_cast<mdl :: Proof*>(proof_)->compress();
	}
	void
	Claim :: prove() {
		try {
			const_cast<mdl :: Proof* const>(proof_)->prove();
		} catch (Error* error) {
			error->add (location_);
			error->message() << "claim proof failed." << endLine;
			throw error;
		}
	}

	// object :: Buildable implementation
	void
	Claim :: buildTheory (mdl :: Theory* const theory) {
		Ancestor_ :: buildTheory (theory);
		const_cast<mdl :: Proof* const>(proof_)->buildTheory (theory);
	}

	// object :: Writable implementation
	void
	Claim :: translate (String& string) const {
		node_->translate (string);
	}
	void
	Claim :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (Ancestor_ :: comments_ != NULL)) {
			string << Ancestor_ :: location_.indent();
			Ancestor_ :: comments_->replicate (string);
			string << endLine;
		}
		string << Ancestor_ :: location_.indent() << Token :: step_ << space;
		string << (index_ + 1) << space << Token :: colon_ << space;

		Ancestor_ :: type_->replicateName (string); string << space;
		string << Token :: equality_ << space;

		link_->replicate (string);
		string << space << Token :: turnstile_ << space;

		Ancestor_ :: expression_->write (string); string << space;
		string << Token :: semicolon_ << endLine;
		proof_->replicate (string);
	}

	// object :: Object implementation
	void
	Claim :: commitSuicide() {
		delete this;
	}
	Size_t
	Claim :: getVolume() const
	{
		Size_t volume = 0;
		volume += Ancestor_ :: getVolume();
		volume += getAggregateVolume (proof_);
		volume += getAggregateVolume (link_);
		volume += getAggregateVolume (scope_);
		return volume;
	}
	Size_t
	Claim :: getSizeOf() const {
		return sizeof (Claim);
	}
}
}
}
}

#endif /*MDL_FORM_PROOF_STEP_CLAIM_CPP_*/
