/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_form_proof_answer_Question.cpp                         */
/* Description:     unknown step inference                                   */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_FORM_PROOF_STEP_QUESTION_CPP_
#define MDL_FORM_PROOF_STEP_QUESTION_CPP_

#include "mdl/math/mdl_math.hpp"
#include "mdl/form/mdl_form.hpp"

namespace mdl {
namespace form {
namespace proof {
namespace step {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Question :: Question
	(
		const mdl :: proof :: Line* const line,
		const Location& location,
		const mdl :: Comments* const comments,
		const index :: Step localIndex,
		const mdl :: Identificator* const type,
		const array :: Expression* const expression,
		const mdl :: proof :: Scope* scope
	) :
	Ancestor_ (line, location, comments, localIndex, type, expression, scope),
	isProved_ (false),
	answer_ (NULL),
	guess_ (NULL) {
	}
	inline
	Question :: Question (const proof :: Step* step) :
	Ancestor_ (step),
	isProved_ (false),
	answer_ (step),
	guess_ (NULL) {
	}
	Question :: ~ Question()
	{
		deleteAggregateObject (answer_);
		deleteAggregateObject (guess_);
	}

	// proof :: step :: Question implementation
	bool
	Question :: isProved() const {
		return isProved_;
	}
	void
	Question :: confirmCompression()
	{
		if (guess_ == NULL) {
			return;
		}
		if (answer_ != NULL) {
			const_cast<mdl :: proof :: Step*>(answer_)->commitSuicide();
		}
		answer_ = guess_;
		guess_ = NULL;
	}
	void
	Question :: rejectCompression()
	{
		if (guess_ != NULL) {
			const_cast<mdl :: proof :: Step*>(guess_)->commitSuicide();
			guess_ = NULL;
		}
	}
	const mdl :: proof :: Step*
	Question :: getAnswer() const {
		return answer_;
	}
	const mdl :: proof :: Step*
	Question :: getGuess() const {
		return guess_;
	}


	// proof :: Step implementation
	mdl :: proof :: Step :: Class
	Question :: getClass() const {
		return mdl :: proof :: Step :: QUESTION;
	}
	void
	Question :: addReference (const mdl :: proof :: Reference* const reference) {
		throw new Error (location_, Error :: SEMANTIC, "proving in a non-proved step.\n");
	}

using manipulator :: endLine;
using manipulator :: space;

	// proof :: Node implementation
	void
	Question :: replicateReference (String& string) const {
		string << Token :: step_  << space << index_;
	}
	void
	Question :: translateReference (String& string) const {
	}

	// proof :: Line implementation
	inline const mdl :: proof :: Line*
	Question :: getNext() const {
		return Ancestor_ :: getNext();
	}
	inline const mdl :: proof :: Line*
	Question :: getPrev() const {
		return Ancestor_ :: getPrev();
	}
	void
	Question :: incIndex (const index :: Step index) const
	{
		Ancestor_ :: incIndex (index);
		if (answer_ != NULL) {
			answer_->incIndex (index);
		}
	}
	void
	Question :: decIndex (const index :: Step index) const
	{
		Ancestor_ :: decIndex (index);
		if (answer_ != NULL) {
			answer_->decIndex (index);
		}
	}
	inline void
	Question :: setNext (const mdl :: proof :: Line* const line) const {
		Ancestor_ :: setNext (line);
	}
	inline void
	Question :: setPrev (const mdl :: proof :: Line* const line) const {
		Ancestor_ :: setPrev (line);
	}
	void
	Question :: completeWithSteps (mdl :: prover :: tree :: Premise* premises) const
	{
		if (isProved_) {
			premises->addBranch (expression_, this);
		}
		Step :: completeWithSteps (premises);
	}

	// object :: Scoping implementation
	inline const mdl :: proof :: Scope*
	Question :: buildScope (const mdl :: proof :: Scope* const scope) {
		return Ancestor_ :: buildScope (scope);
	}
	inline const mdl :: proof :: Scope*
	Question :: getScope() const {
		return Ancestor_ :: getScope();
	}
	inline void
	Question :: setScope (const mdl :: proof :: Scope* scope) const {
		Ancestor_ :: setScope (scope);
	}

	// object :: Expressive implementation
	inline const array :: Expression*
	Question :: getExpression() const {
		return Ancestor_ :: getExpression();
	}
	inline const mdl :: Type*
	Question :: getType() const {
		return Ancestor_ :: type_;
	}

	// object :: Verifiable implementation
	void
	Question :: check() const
	{
		isVerified_ = false;
		if (answer_ == NULL) {
			Error* error = new Error (location_, Error :: PROVER, "question is not proved.");
			error->message() << *this << endLine;
			throw error;
		}
		try {
			answer_->check();
		} catch (Error* error) {
			error->add (location_);
			error->message() << "question proof failed verification." << endLine;
			error->message() << tab << *this << endLine;
			throw error;
		}
		isVerified_ = true;
	}
	void
	Question :: compress()
	{
		const mdl :: Assertion* const
			theorem = scope_->getAssertion();
		isProved_ = false;
		isProved_ = Prover :: prove (theorem, this, Config :: getMaxProverDepth());
		if (isProved_) {
			guess_ = Prover :: getAnswerStep();
		}
	}
	void
	Question :: prove()
	{
		const mdl :: Assertion* const
			theorem = scope_->getAssertion();
		isProved_ = false;
		isProved_ = Prover :: prove (theorem, this, Config :: getMaxProverDepth());
		if (isProved_) {
			if (answer_ != NULL) {
				const_cast<mdl :: proof :: Step*>(answer_)->commitSuicide();
			}
			answer_ = Prover :: getAnswerStep();
		}
	}

	// object :: Buildable implementation
	void
	Question :: buildTheory (mdl :: Theory* const theory) {
		Ancestor_ :: buildTheory (theory);
	}

	// object :: Writable implementation
	void
	Question :: translate (String& string) const {
		answer_->translate (string);
	}
	void
	Question :: replicate (String& string) const
	{
		if (!Config :: removeComments() && (Ancestor_ :: comments_ != NULL)) {
			string << Ancestor_ :: location_.indent();
			Ancestor_ :: comments_->replicate (string);
			string << endLine;
		}
		if ((answer_ == NULL) || Config :: compress()) {
			string << Ancestor_ :: location_.indent() << Token :: step_ << space;
			string << (index_ + 1) << space << Token :: colon_ << space;

			Ancestor_ :: type_->replicateName (string); string << space;
			string << Token :: equality_ << space;

			string << Token :: question_ << space;
			string << Token :: turnstile_ << space;

			Ancestor_ :: expression_->write (string); string << space;
			string << Token :: semicolon_ << endLine;
		} else {
			answer_->replicate (string);
		}
	}

	// object :: Object implementation
	void
	Question :: commitSuicide() {
		delete this;
	}
	Size_t
	Question :: getVolume() const
	{
		Size_t volume = 0;
		volume += Ancestor_ :: getVolume();
		volume += getAggregateVolume (answer_);
		volume += getAggregateVolume (guess_);
		return volume;
	}
	Size_t
	Question :: getSizeOf() const {
		return sizeof (Question);
	}
}
}
}
}

#endif /*MDL_FORM_PROOF_STEP_QUESTION_CPP_*/
