/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_Mdl.cpp                                              */
/* Description:     mdl translator/prover                                    */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_MDL_CPP_
#define MDL_MDL_CPP_

#include <exception>

#include "mdl/auxiliary/mdl_auxiliary.hpp"
#include "mdl/math/mdl_math.hpp"
#include "mdl/parser/mdl_parser.hpp"
#include "mdl/mdl_Mdl.hpp"

namespace mdl {

using namespace exception;

	/****************************
	 *		Public members
	 ****************************/

	Mdl :: Mdl() :
	file_ (Config :: getSource()->getString()),
	path_ (new Path (file_)),
	parser_ (new parser :: Source ()),
	source_ (NULL),
	timers_ (),
	volume_ (this)
	{
		Table :: init();
		Statistics :: init();
		Math :: init();
		Prover :: init();
	}
	Mdl :: ~ Mdl()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "deleting ... " << std :: flush;
		}
		Timers :: release().start(); {
			deleteAggregateObject (source_);
			deleteAggregateObject (parser_);
			deleteAggregateObject (path_);
			mdl_ = NULL;
		}
		Timers :: release().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << Timers :: release() << std :: endl;
		}
	}

using std :: endl;
using std :: cerr;
using std :: cout;

	Mdl :: Result_
	Mdl :: run ()
	{
		if (mdl_ != NULL) {
			cerr << "mdl already is initialized." << endl;
			return GENERAL_FAILURE;
		}
		mdl_ = this;
		try {
			process ();
		} catch (const Error* error) {
			timers_.stopAll();
			cerr << *error << endl;
			Result_ result = static_cast<Result_>(error->getType());
			delete error;
			return result;
		} catch (const OutOfMemory& error) {
			timers_.stopAll();
			cerr << error << endl;
			Memory :: show (cerr);
			return OUT_OF_MEMORY;
		} catch (const std :: exception& exception) {
			timers_.stopAll();
			cerr << exception.what() << endl;
			return GENERAL_FAILURE;
		} catch (...) {
			timers_.stopAll();
			cerr << "exception." << endl;
			return GENERAL_FAILURE;
		}
		return SUCCESS;
	}
	const Timers&
	Mdl :: getTimers() const {
		return timers_;
	}

	const Mdl*
	Mdl :: getMdl() {
		return mdl_;
	}
	Mdl :: Result_
	Mdl :: initMemory()
	{
		Result_ result = SUCCESS;
		if (Config :: verbosity() > 1) {
			std :: cout << "initializing ... " << std :: flush;
		}
		Timers :: init().start();
		try {
			const Size_t
				stackVolume = Config :: getStackVolume();
			Memory :: standard().create();
			Memory :: heap().create();
			Memory :: stack().create (stackVolume);
			if (Config :: useProver()) {
				const Size_t
					proverVolume = Config :: getProverVolume();
				const Size_t
					reserveVolume = Config :: getReserveVolume();
				Memory :: boxed().create (proverVolume, reserveVolume);
			}
		} catch (OutOfMemory&) {
			result = Mdl :: INIT_ERROR;
		}
		Timers :: init().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << ((result == SUCCESS) ? "done in " : "failed in ") << Timers :: init() << std :: endl;
		}
		return result;
	}
	Mdl :: Result_
	Mdl :: releaseMemory()
	{
		bool success = true;
		if (Config :: verbosity() >= Config :: VERBOSE_STANDARD) {
			std :: cout << "releasing ... " << std :: flush;
		}
		Prover :: release();
		Math :: release();
		Statistics :: release();
		Table :: release();
		Config :: release();
		object :: Object :: release();
		form :: block :: Proof :: release();

		success &= releaseHeap();
		success &= releaseStack();
		success &= releaseBoxed();
		success &= releaseStandard();

		if (Config :: verbosity() >= Config :: VERBOSE_STANDARD) {
			std :: cout << (success ? "done." : "failed.") << std :: endl;
		}
		return (success ? SUCCESS : RELEASE_ERROR);
	}
	inline void
	Mdl :: showMemory()
	{
		if (mdl_ != NULL) {
			cerr << mdl_->volume_ << endl;
		}
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// object :: Object implementation
	void
	Mdl :: commitSuicide() {
		delete this;
	}
	Size_t
	Mdl :: getVolume() const
	{
		Size_t volume = 0;
		volume += object :: Object :: getVolume();
		volume += file_.getVolume();
		volume += getAggregateVolume (path_);
		volume += getAggregateVolume (parser_);
		volume += getAggregateVolume (source_);

		volume += timers_.getVolume();
		volume += volume_.getVolume();
		return volume;
	}
	Size_t
	Mdl :: getSizeOf() const {
		return sizeof (Mdl);
	}
	void
	Mdl :: show (String& string) const
	{
		if (path_ == NULL) {
			string << "file <none>" << space;
		} else {
			string << "file" << space;
			string << *path_ << space;
		}
		string << "is done in" << space << timers_.total() << endLine;
		string << endLine;

		try {
			if (Config :: showTimings()) {
				timers_.show (string);
				string << endLine;
			}
			if (Config :: showMath()) {
				Math :: show (string);
				string << endLine;
			}
			if (Config :: showVolume() || Config :: showMemory()) {
				volume_.show (string);
			}
		} catch (const Error* error) {
			cerr << *error << endl;
			delete error;
		} catch (const OutOfMemory& error) {
			cerr << error << endl;
		} catch (const std :: exception& e) {
			cerr << "exception: " << e.what() << endl;
		} catch (...) {
			cerr << "exception." << endl;
		}
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Mdl :: process()
	{
		timers_.total().start();

		read();
		build();

		switch (Config :: getRunMode()) {
		case Config :: PROVE_WORK :
			prove(); break;
		case Config :: CHECK_WORK :
			check(); break;
		case Config :: TRANSLATE_WORK :
			check(); break;
		case Config :: COMPRESS_WORK :
			compress(); break;
		default :
		             break;
		}
		write();

		timers_.total().stop();
	}
	void
	Mdl :: read()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "reading ... " << std :: flush;
		}
		timers_.read().start();
		source_ = const_cast<mdl :: Source*>(parser_->parse (*path_));
		timers_.read().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.read() << std :: endl;
		}
	}
	void
	Mdl :: build()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "building ... " << std :: flush;
		}
		timers_.build().start();
		source_->build();
		timers_.build().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.build() << std :: endl;
		}
	}
	void
	Mdl :: prove()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "proving ... " << std :: endl;
		}
		timers_.prove().start();
		source_->prove();
		timers_.prove().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "proving done in " << timers_.prove() << std :: endl;
		}
	}
	void
	Mdl :: check()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "checking ... " << std :: flush;
		}
		timers_.check().start();
		source_->check();
		timers_.check().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.check() << std :: endl;
		}
	}
	void
	Mdl :: compress()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "compressing ... " << std :: endl;
		}
		timers_.compress().start();
		source_->compress();
		timers_.compress().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "compressing done in " << timers_.compress() << std :: endl;
		}
		form :: block :: Proof :: getLog().write();
	}
	void
	Mdl :: write()
	{
		if (Config :: check()) {
			return;
		}
		if (Config :: verbosity() > 1) {
			std :: cout << "writing ... " << std :: flush;
		}
		timers_.write().start();
		source_->write();
		timers_.write().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.write() << std :: endl;
		}
	}

	bool
	Mdl :: releaseHeap()
	{
		bool success = true;
		try {
			Memory :: heap().destroy();
		} catch (memory :: MemoryLeaks& exception) {
			std :: cerr << std :: endl << chars :: TAB << exception.what() << ": ";
			std :: cerr << exception.getLeak() << " bytes are lost ";
			success = false;
		} catch (std :: exception& exception) {
			std :: cerr << exception.what() << std :: endl;
			success = false;
		} catch (...) {
			success = false;
		}
		return success;
	}
	bool
	Mdl :: releaseStack()
	{
		bool success = true;
		try {
			Memory :: stack().destroy();
		} catch (memory :: MemoryLeaks& exception) {
			std :: cerr  << std :: endl << chars :: TAB << exception.what() << ": ";
			std :: cerr << exception.getLeak() << " bytes are lost ";
			success = false;
		} catch (std :: exception& exception) {
			std :: cerr << exception.what() << std :: endl;
			success = false;
		} catch (...) {
			success = false;
		}
		return success;
	}
	bool
	Mdl :: releaseBoxed()
	{
		bool success = true;
		try {
			if (Config :: useProver()) {
				Memory :: boxed().destroy();
			}
		} catch (memory :: MemoryLeaks& exception) {
			std :: cerr  << std :: endl << chars :: TAB << exception.what() << ": ";
			std :: cerr << exception.getLeak() << " bytes are lost ";
			success = false;
		} catch (std :: exception& exception) {
			std :: cerr << exception.what() << std :: endl;
			success = false;
		} catch (...) {
			success = false;
		}
		return success;
	}
	bool
	Mdl :: releaseStandard()
	{
		bool success = true;
		try {
			Memory :: standard().destroy();
		} catch (memory :: MemoryLeaks& exception) {
			std :: cerr  << std :: endl << chars :: TAB << exception.what() << ": ";
			std :: cerr << exception.getLeak() << " bytes are lost ";
			success = false;
		} catch (std :: exception& exception) {
			std :: cerr << exception.what() << std :: endl;
			success = false;
		} catch (...) {
			success = false;
		}
		return success;
	}

	/*************************************
	 *		Private static attributes
	 *************************************/

	Mdl* Mdl :: mdl_ = NULL;
}

#endif /*MDL_MDL_CPP_*/
