/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_parser_block_Proof.cpp                               */
/* Description:     proof parser                                             */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PARSER_BLOCK_PROOF_CPP_
#define MDL_PARSER_BLOCK_PROOF_CPP_

#include "mdl/lexer/mdl_lexer.hpp"
#include "mdl/form/mdl_form.hpp"
#include "mdl/parser/mdl_parser.hpp"

namespace mdl {
namespace parser {
namespace block {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Proof :: Proof() :
	stepParser_ (new parser :: proof :: Step()),
	qedParser_ (new parser :: proof :: Qed()),
	variableParser_ (new parser :: proof :: Variable()),
	notationParser_ (new parser :: proof :: Notation()),

	location_ (),
	comments_ (NULL),
	stepVector_ (INITIAL_STEP_VECTOR_CAPACITY),
	firstLine_ (NULL),
	line_ (NULL) {
	}
	Proof :: ~ Proof()
	{
		deleteAggregateObject (stepParser_);
		deleteAggregateObject (qedParser_);
		deleteAggregateObject (variableParser_);
		deleteAggregateObject (notationParser_);
	}

	const mdl :: Proof*
	Proof :: parse (Lexer* lexer)
	{
		stepVector_.clear();
		location_ = lexer->getLocation();
		firstLine_ = NULL;
		line_ = NULL;
		lexer->comments().clear();
		lexer->scan <Token :: PROOF>();
		lexer->scan <Token :: OPEN_BRACE>();
		comments_ =
			lexer->comments().isEmpty() ?
			NULL :
			new form :: Comments (lexer->comments(), false);
		lexer->comments().clear();

		while (true) {
			lexer->scan
			<
				Token :: STEP,
				Token :: QED,
				Token :: VARIABLE,
				Token :: NOTATION,
				Token :: CLOSE_BRACE
			>();
			if (lexer->getTokenType() == Token :: CLOSE_BRACE) {
				lexer->comments().clear();
				return new form :: Proof (location_, comments_, firstLine_, stepVector_);
			}
			line_ = parseLine (lexer);
			if (firstLine_ == NULL) {
				firstLine_ = line_;
			}
		}
	}

	// object :: Object implementation
	void
	Proof :: commitSuicide() {
		delete this;
	}
	Size_t
	Proof :: getVolume() const
	{
		Size_t volume = 0;
		volume += getAggregateVolume (stepParser_);
		volume += getAggregateVolume (qedParser_);
		volume += getAggregateVolume (variableParser_);
		volume += getAggregateVolume (notationParser_);
		volume += stepVector_.getVolume();
		return volume;
	}
	Size_t
	Proof :: getSizeOf() const {
		return sizeof (Proof);
	}
	void
	Proof :: show (String&) const {
	}

	/****************************
	 *		Private members
	 ****************************/

	const mdl :: proof :: Line*
	Proof :: parseLine (Lexer* lexer)
	{
		switch (lexer->getTokenType()) {
		case Token :: STEP :
			stepVector_.add (stepParser_->parse (lexer, line_));
			return stepVector_.getLastValue();
		case Token :: QED :
			return qedParser_->parse (lexer, line_);
		case Token :: VARIABLE :
			return variableParser_->parse (lexer, line_);
		case Token :: NOTATION :
			return notationParser_->parse (lexer, line_);
		default :
			return NULL;
		}
	}
}
}
}

#endif /*MDL_PARSER_BLOCK_PROOF_CPP_*/
