/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_parser_Source.cpp                                    */
/* Description:     source parser                                            */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PARSER_SOURCE_CPP_
#define MDL_PARSER_SOURCE_CPP_

#include "mdl/lexer/mdl_lexer.hpp"
#include "mdl/form/mdl_form.hpp"
#include "mdl/parser/mdl_parser.hpp"

namespace mdl {
namespace parser {

	/****************************
	 *		Public members
	 ****************************/

	Source :: Source (const bool isTop) :
	isTop_ (isTop),
	lexer_ (NULL),
	importParser_ (new parser :: Import()),
	theoryParser_ (new parser :: Theory()),
	sourceParser_ (NULL),
	contentsParser_ (new parser :: Contents()) {
	}
	Source :: ~ Source()
	{
		deleteAggregateObject (lexer_);
		deleteAggregateObject (importParser_);
		deleteAggregateObject (theoryParser_);
		deleteAggregateObject (sourceParser_);
		deleteAggregateObject (contentsParser_);
	}

	const mdl :: Source*
	Source :: parse (const mdl :: Path& path)
	{
		Table :: names()->add (path.getName());
		if (isTop_) {
			Table :: files()->add (path.getString());
		}
		open (path);
		lexer_->scanComments();
		const mdl :: Comments* comments =
			lexer_->comments().isEmpty() ?
			NULL :
			new form :: Comments (lexer_->comments());

		lexer_->comments().clear();
		mdl :: Source* const source =
			isTop_?
			new form :: Source (isTop_, lexer_, comments, path, *Config :: getTarget()) :
			new form :: Source (isTop_, lexer_, comments, path, path);

		while (true) {
			lexer_->scan
			<
				Token :: IMPORT,
				Token :: THEORY,
				Token :: CONTENTS,
				Token :: EOF_
			>();
			switch (lexer_->getTokenType()) {
			case Token :: IMPORT : {
				const mdl :: Import* const
					import = importParser_->parse (lexer_);
				const String& importPath = import->getPath().getString();
				source->addComponent (import);

				if (!Table :: files()->contains (importPath)) {
					Table :: files()->add (importPath);
					if (sourceParser_ == NULL) {
						sourceParser_ = new parser :: Source (false);
					}
					const mdl :: Source* const
						importedSource = sourceParser_->parse (import->getPath());
					source->addComponent (importedSource);
				}
				break;
			}
			case Token :: THEORY : {
				const mdl :: Theory* const
					theory = theoryParser_->parse (lexer_);
				source->addComponent (theory);
				break;
			}
			case Token :: CONTENTS : {
				const mdl :: Contents* const
					contents = contentsParser_->parse (lexer_);
				source->addComponent (contents);
				break;
			}
			case Token :: EOF_ : {
				goto end;
			}
			default :
				return NULL;
			}
		}
		end :
		close (path);
		return source;
	}

	// object :: Object implementation
	void
	Source :: commitSuicide() {
		delete this;
	}
	Size_t
	Source :: getVolume() const
	{
		Size_t volume = 0;
		volume += getAggregateVolume (lexer_);
		volume += getAggregateVolume (importParser_);
		volume += getAggregateVolume (theoryParser_);
		volume += getAggregateVolume (sourceParser_);
		volume += getAggregateVolume (contentsParser_);
		return volume;
	}
	Size_t
	Source :: getSizeOf() const {
		return sizeof (Source);
	}
	void
	Source :: show (String&) const {
	}

	/****************************
	 *		Private members
	 ****************************/

using manipulator :: endLine;

	void
	Source :: open (const Path& path)
	{
		const String& file = path.getString();
		if (lexer_ != NULL) {
			lexer_->open (file);
		} else {
			lexer_ = new Lexer (file);
			lexer_->open();
		}
	}
	void
	Source :: close (const Path& path)
	{
		if (lexer_ == NULL) {
			Error* error = new Error (Error :: INPUT);
			error->message() << "attempt to close non opened " << path << endLine;
			throw error;
		}
		if (lexer_->getTokenType() != Token :: EOF_) {
			Error* error = new Error (Error :: INPUT);
			error->message() << "unexpected token: "<< lexer_->getToken() << endLine;
			error->message() << "EOF expected in file " << path << endLine;
			throw error;
		}
		lexer_->close();
	}
}
}

#endif /*MDL_PARSER_SOURCE_CPP_*/
