/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_parser_proof_Step.cpp                                */
/* Description:     proof step parser                                        */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PARSER_PROOF_STEP_CPP_
#define MDL_PARSER_PROOF_STEP_CPP_

#include "mdl/lexer/mdl_lexer.hpp"
#include "mdl/form/mdl_form.hpp"
#include "mdl/parser/mdl_parser.hpp"

namespace mdl {
namespace parser {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Step :: Step() :
	identificatorParser_ (new parser :: Identificator()),
	linkParser_(new parser :: proof :: Link()),
	expressionParser_ (new parser :: Expression (false)),
	proofParser_ (NULL),

	location_(),
	comments_ (NULL),
	index_ (index :: undefined :: STEP),
	type_ (NULL),
	link_ (NULL),
	expression_ (NULL),
	proof_ (NULL) {
	}
	Step :: ~ Step()
	{
		deleteAggregateObject (identificatorParser_);
		deleteAggregateObject (linkParser_);
		deleteAggregateObject (expressionParser_);
		deleteAggregateObject (proofParser_);
	}

	const mdl :: proof :: Step*
	Step :: parse (Lexer* lexer, const mdl :: proof :: Line* const line)
	{
		location_ = lexer->getLocation();
		index_ = lexer->scanNumber<index :: Step>() - 1;

		lexer->scan<Token :: COLON>();
		type_ = identificatorParser_->parse (lexer);
		lexer->scan<Token :: EQUALITY>();

		lexer->scan
		<
			Token :: QUESTION,
			Token :: CLAIM,
			Token :: AXIOM,
			Token :: THEOREM,
			Token :: DEFINITION
		>();
		switch (lexer->getTokenType ()) {
		case Token :: QUESTION :
			return parseQuestion (lexer, line);
		case Token :: CLAIM :
			return parseClaim (lexer, line);
		case Token :: AXIOM :
		case Token :: THEOREM :
		case Token :: DEFINITION :
			return parseAssertion (lexer, line);
		default :
			return NULL;
		}
	}

	// object :: Object implementation
	void
	Step :: commitSuicide() {
		delete this;
	}
	Size_t
	Step :: getVolume() const
	{
		Size_t volume = 0;
		volume += getAggregateVolume (identificatorParser_);
		volume += getAggregateVolume (linkParser_);
		volume += getAggregateVolume (expressionParser_);
		volume += getAggregateVolume (proofParser_);
		return volume;
	}
	Size_t
	Step :: getSizeOf() const {
		return sizeof (Step);
	}
	void
	Step :: show (String&) const {
	}

	/****************************
	 *		Private members
	 ****************************/

	const mdl :: proof :: Step*
	Step :: parseQuestion (Lexer* lexer, const mdl :: proof :: Line* const line)
	{
		lexer->scan<Token :: TURNSTILE>();
		comments_ =
			lexer->comments().isEmpty() ?
			NULL :
			new form :: Comments (lexer->comments());
		lexer->comments().clear();
		expression_ = expressionParser_->parse (lexer);
		return new form :: proof :: step :: Question
		(
			line,
			location_,
			comments_,
			index_,
			type_,
			expression_
		);
	}
	const mdl :: proof :: Step*
	Step :: parseAssertion (Lexer* lexer, const mdl :: proof :: Line* const line)
	{
		link_ = linkParser_->parse (lexer);
		lexer->scan<Token :: TURNSTILE>();
		comments_ =
			lexer->comments().isEmpty() ?
			NULL :
			new form :: Comments (lexer->comments());
		lexer->comments().clear();
		expression_ = expressionParser_->parse (lexer);
		return new form :: proof :: step :: Assertion
		(
			line,
			location_,
			comments_,
			index_,
			type_,
			link_,
			expression_
		);
	}
	const mdl :: proof :: Step*
	Step :: parseClaim (Lexer* lexer, const mdl :: proof :: Line* const line)
	{
		if (proofParser_ == NULL) {
			proofParser_ = new parser :: Proof ();
		}
		link_ = linkParser_->parse (lexer, false);
		lexer->scan<Token :: TURNSTILE>();
		comments_ =
			lexer->comments().isEmpty() ?
			NULL :
			new form :: Comments (lexer->comments());
		lexer->comments().clear();
		expression_ = expressionParser_->parse (lexer);
		proof_ = proofParser_->parse (lexer);
		return new form :: proof :: step :: Claim
		(
			line,
			location_,
			comments_,
			index_,
			type_,
			link_,
			expression_,
			proof_
		);
	}
}
}
}

#endif /*MDL_PARSER_PROOF_STEP_CPP_*/
