/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_evidence_Hyp.cpp                              */
/* Description:     hyp evidence element                                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_EVIDENCE_HYP_CPP_
#define MDL_PROVER_EVIDENCE_HYP_CPP_

namespace mdl {
namespace prover {
namespace evidence {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Hyp :: Hyp (const Hyp& hyp) :
	node_ (NULL),
	substitution_ (new Substitution (*hyp.substitution_)),
	up_ (NULL),
	down_ (hyp.down_),
	depth_ (hyp.depth_),
	cardinality_ (hyp.cardinality_),
	clone_ (NULL) {
	}
	inline
	Hyp :: Hyp
	(
		const Node* node,
		const Substitution* substitution,
		const Evidence* up
	) :
	node_ (node),
	substitution_ (substitution),
	up_ (up),
	down_ (NULL),
	depth_ (up_->getDepth()),
	cardinality_ (up_->getCardinality()),
	clone_ (NULL) {
	}
	inline
	Hyp :: Hyp
	(
		const Node* node,
		const Evidence* down,
		const Evidence* up
	) :
	node_ (node),
	substitution_ (NULL),
	up_ (up),
	down_ (down),
	depth_ (up_->getDepth()),
	cardinality_ (up_->getCardinality()),
	clone_ (NULL) {
	}
	Hyp :: ~ Hyp() {
	}

	inline const Evidence*
	Hyp :: getUp() const {
		return up_;
	}

	// prover :: Evidence implementation
	inline value :: Kind
	Hyp :: getKind() const {
		return HYP;
	}
	inline const Node*
	Hyp :: getNode() const {
		return node_;
	}
	inline void
	Hyp :: verifyIntegrity() const {
		up_->verifyIntegrity();
	}
	const mdl :: prover :: Substitution*
	Hyp :: getSubstitution() const {
		return (down_ == NULL) ? substitution_ : down_->getSubstitution();
	}
	inline index :: Step
	Hyp :: getDepth() const {
		return depth_;
	}
	inline value :: Integer
	Hyp :: getCardinality() const {
		return cardinality_;
	}
	const Proof*
	Hyp :: constructProof (const Substitution* substutition) const
	{
		const Proof* const
			upProof = up_->constructProof (substutition);
		return new proof :: Hyp (this, upProof);
	}
	vector :: Proof
	Hyp :: constructProofs (const Substitution* substutition) const
	{
		const vector :: Proof
			upProofs = up_->constructProofs (substutition);
		const value :: Integer size = upProofs.getSize();
		vector :: Proof proofs (size);
		for (value :: Integer i = 0; i < size; ++ i) {
			const Proof* const
				upProof = upProofs.getValue (i);
			const Proof* const
				proof = new proof :: Hyp (this, upProof);
			proofs.add (proof);
		}
		return proofs;
	}

	// object :: Cloneable implementation
	const Evidence*
	Hyp :: clone() const
	{
		if (clone_ != NULL) {
			return clone_;
		}
		Hyp* clone = new Hyp (*this);
		clone_ = clone;
		clone->node_ = node_->clone();
		clone->up_ = up_->clone();
		return clone_;
	}

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;

	// object :: Object interface
	void
	Hyp :: commitSuicide() {
		delete this;
	}
	Size_t
	Hyp :: getVolume() const {
		return getAggregateVolume (substitution_);
	}
	Size_t
	Hyp :: getSizeOf() const {
		return sizeof (Hyp);
	}
	void
	Hyp :: show (String& string) const
	{
		string << "############################### - Hyp Node - ################################################" << endLine;
		string << "this: " << this << endLine;
		string << "node: " << node_ << endLine;
		string << "expression: " << *(node_->getExpression()) << endLine;
		string << "substitution: " << *getSubstitution() << endLine;
		string << "up: " << up_ << endLine;
		string << "############################### - Hyp Node - ################################################" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
	}
}
}
}

#endif /*MDL_PROVER_EVIDENCE_HYP_CPP_*/
