/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_evidence_Prop.cpp                             */
/* Description:     prop proof evidence element                              */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_EVIDENCE_PROP_CPP_
#define MDL_PROVER_EVIDENCE_PROP_CPP_

namespace mdl {
namespace prover {
namespace evidence {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Prop :: Prop (const Prop& prop) :
	node_ (prop.node_),
	substitution_ (new Substitution (prop.substitution_)),
	up_ (prop.getArity()),
	down_ (prop.down_),
	depth_ (prop.depth_),
	cardinality_ (prop.cardinality_),
	clone_ (NULL) {
	}
	inline
	Prop :: Prop
	(
		const Node* node,
		const Substitution* substitution,
		const index :: Arity arity
	) :
	node_ (node),
	substitution_ (substitution),
	up_ (arity),
	down_ (NULL),
	depth_ (1),
	cardinality_ (1),
	clone_ (NULL) {
	}
	inline
	Prop :: Prop
	(
		const Node* node,
		const Evidence* down,
		const index :: Arity arity
	) :
	node_ (node),
	substitution_ (NULL),
	up_ (arity),
	down_ (down),
	depth_ (1),
	cardinality_ (1),
	clone_ (NULL) {
	}
	Prop :: ~ Prop() {
	}

	inline index :: Arity
	Prop :: getArity() const {
		return up_.getSize();
	}
	inline const Evidence*
	Prop :: getUp (const index :: Arity i) const {
		return up_.getValue (i);
	}
	inline void
	Prop :: addUp (const index :: Arity i, const Evidence* evidence)
	{
		up_.getReference (i) = evidence;
		const index :: Step
			upDepth = evidence->getDepth();
		depth_ =
			(upDepth + 1 > depth_) ?
			upDepth + 1 :
			depth_;
		cardinality_ += evidence->getCardinality();
	}

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;

	// prover :: Evidence implementation
	inline value :: Kind
	Prop :: getKind() const {
		return PROP;
	}
	inline const Node*
	Prop :: getNode() const {
		return node_;
	}
	inline void
	Prop :: verifyIntegrity() const
	{
		Memory :: stack().push();
		verifySubstitutions();

		expression :: Mode mode;
		mode << expression :: Mode :: SHOW_REPLACEABILITY;
		verifyUp();

		const node :: Prop*
			prop = dynamic_cast<const node :: Prop*>(node_);
		const stack :: Substitution* const
			substitutionExt = stack :: Substitution :: create (prop->substitutionExt_);

		stack :: Expression* const
			proposition_1 = new stack :: Expression (node_->getExpression());
		proposition_1->execute (getSubstitution());

		const mdl :: Proposition* const
			proposition = node_->getProposition();
		const mdl :: Assertion* const
			assertion = proposition->getAssertion();
		stack :: Expression* const
			proposition_2 = new stack :: Expression (proposition->getExpression());
		proposition_2->execute (substitutionExt);
		proposition_2->execute (getSubstitution());

		if (*proposition_1 != *proposition_2) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << mode << "debug proposition ";
			assertion->replicateName (error->message());
			error->message() << " verification fail" << endLine;
			error->message() << *proposition_1 << " != " << *proposition_2 << endLine;
			showVerification (error->message());
			throw error;
		}

		if (assertion->getHypArity() != up_.getSize()) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << mode  << "debug proposition ";
			assertion->replicateName (error->message());
			error->message() << " verification fail: arity mismatch" << endLine;
			error->message() << assertion->getHypArity() << " != " << up_.getSize() << endLine;
			showVerification (error->message());
			throw error;
		}

		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis_1 = assertion->getHypothesis (i);
			stack :: Expression* const
				hyp_1 = new stack :: Expression (hypothesis_1->getExpression());
			hyp_1->execute (substitutionExt);
			hyp_1->execute (getSubstitution());

			stack :: Expression* const
				hyp_2 = new stack :: Expression (up_.getValue (i)->getNode()->getExpression());
			hyp_2->execute (getSubstitution());

			if (*hyp_1 != *hyp_2) {
				Error* error = new Error (Error :: SEMANTIC);
				error->message() << mode  << "debug hypothesis ";
				assertion->replicateName (error->message());
				error->message() << " verification fail" << endLine;
				error->message() << *hyp_1 << " != " <<  *hyp_2 << endLine;
				showVerification (error->message());
				//show (error->message());
				throw error;
			}
		}
		Memory :: stack().pop();
	}
	const mdl :: prover :: Substitution*
	Prop :: getSubstitution() const {
		return (down_ == NULL) ? substitution_ : down_->getSubstitution();
	}
	inline index :: Step
	Prop :: getDepth() const {
		return depth_;
	}
	inline value :: Integer
	Prop :: getCardinality() const {
		return cardinality_;
	}
	const Proof*
	Prop :: constructProof (const Substitution* substitution) const
	{
		Substitution*
			global = Substitution :: create (getSubstitution());
		global = global->compose (substitution);

		proof :: Prop* const
			proof = new proof :: Prop (this, global, up_.getSize());

		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Evidence* const
				upEvidence = up_.getValue (i);
			const Proof* const
				upProof = upEvidence->constructProof (global);
			proof->addUp (i, upProof);
		}
		return proof;
	}
	vector :: Proof
	Prop :: constructProofs (const Substitution* substitution) const
	{
		value :: Integer size = 1;
		index :: Arity arity = up_.getSize();
		vector :: Proof upProofs [arity];
		for (index :: Arity i = 0; i < arity; ++ i) {
			const Evidence* evidence = up_.getValue (i);
			upProofs [i] = evidence->constructProofs (substitution);
			size *= upProofs [i].getSize();
		}

		vector :: Proof proofs (size);
		Multyindex multyindex (arity, upProofs);
		while (true) {
			proof :: Prop* const
				proof = new proof :: Prop (this, substitution, up_.getSize());
			for (index :: Arity i = 0; i < arity; ++ i) {
				const Proof* upProof = multyindex.getValue (i);
				proof->addUp (i, upProof);
			}
			proofs.add (proof);
			if (++ multyindex) {
				break;
			}
		}
		return proofs;
	}

	// object :: Cloneable implementation
	const Evidence*
	Prop :: clone() const
	{
		if (clone_ != NULL) {
			return clone_;
		}
		Prop* clone = new Prop (*this);
		clone_ = clone;
		clone->node_ = node_->clone();
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Evidence*
				evidence = up_.getValue (i);
			clone->up_.getReference (i) = evidence->clone();
		}
		return clone_;
	}

	// object :: Object interface
	void
	Prop :: commitSuicide() {
		delete this;
	}
	Size_t
	Prop :: getVolume() const
	{
		Size_t volume = 0;
		volume += getAggregateVolume (substitution_);
		volume += up_.getVolume();
		return volume;
	}
	Size_t
	Prop :: getSizeOf() const {
		return sizeof (Prop);
	}
	void
	Prop :: show (String& string) const
	{
		string << "############################### - Prop Node - ################################################" << endLine;
		string << "prop: "; // << endLine;
		node_->getProposition()->getAssertion()->replicateName (string);
		string << endLine;
		string << "this: " << this << endLine;
		string << "node: " << node_ << endLine;
		string << "substitution: " << *getSubstitution() << endLine;
		string << "up:" << space;
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			string << up_.getValue (i);
			if (i < up_.getSize() - 1) {
				string << "," << space;
			}
		}
		string << endLine;
		string << "up subs:" << endLine;
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			string << tab << i << " : ";
			up_.getValue (i)->getSubstitution()->show (string);
			string << endLine;
		}
		string << endLine;
		string << "****************************************************" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << "********************** NODE ************************" << endLine;
		node_->show (string);
		string << "****************************************************" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		showVerification (string);
		string << "############################### - Prop Node - ################################################" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Prop :: verifyUp() const
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			up_.getValue (i)->verifyIntegrity();
		}
	}
	void
	Prop :: verifySubstitutions() const {
		getSubstitution()->verifyIntegrity();
	}
	void
	Prop :: verifyVariables (const Substitution* substitution) const
	{
		while (substitution != NULL) {
			const Symbol& variable = substitution->getVariable();
			verifyVariable (variable, substitution);
			substitution = substitution->getNext();
		}
	}
	void
	Prop :: verifyVariable
	(
		const Symbol& variable,
		const Substitution* substitution
	) const
	{
		if (node_->getExpression()->getTerm()->contains(variable)) {
			return;
		}
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Evidence* const
				hyp = up_.getValue (i);
			if (hyp->getNode()->getExpression()->getTerm()->contains(variable)) {
				return;
			}
		}
		expression :: Mode mode;
		mode << expression :: Mode :: SHOW_REPLACEABILITY;
		Error* error = new Error (Error :: SEMANTIC);
		error->message() << mode  << "dumb variable in substitution" << endLine;
		error->message() << tab << "substitution = " << *substitution;
		error->message() << tab << "variable = " << variable << endLine;
		error->message() << endLine;
		error->message() << endLine;
		show (error->message());
		std :: cerr << error->message() << std :: endl;
		throw error;
	}
	inline void
	Prop :: showVerification (String& string) const
	{
		Memory :: stack().push();
		expression :: Mode mode;
		//mode << expression :: Mode :: SHOW_REPLACEABILITY;
		string << mode;

		const mdl :: Proposition* const
			proposition = node_->getProposition();
		const mdl :: Assertion* const
			assertion = proposition->getAssertion();
		string << "************* VERIFICATION *************" << endLine;
		string << mode << "assertion ";
		assertion->replicateName (string);
		string << endLine;
		string << "=============================================" << endLine;

		const node :: Prop*
			prop = dynamic_cast<const node :: Prop*>(node_);
		const stack :: Substitution* const
			substitutionInt = stack :: Substitution :: create (prop->substitutionInt_);
		const stack :: Substitution* const
			substitutionExt = stack :: Substitution :: create (prop->substitutionExt_);

		string << "substitution_ = " << *getSubstitution() << endLine;
		string << endLine;
		string << "substitution ext = " << *substitutionExt << endLine;
		string << "substitution int = " << *substitutionInt << endLine;

		stack :: Expression* const
			proposition_1 = new stack :: Expression (node_->getExpression());
		proposition_1->execute (getSubstitution());

		stack :: Expression* const
			proposition_2 = new stack :: Expression (proposition->getExpression());
		proposition_2->execute (substitutionExt);
		proposition_2->execute (getSubstitution());

		string << "proposition_1 = " << *proposition_1 << endLine;
		string << "proposition_2 = " << *proposition_2 << endLine;

		string << endLine;
		string << "node_ = " << node_ << endLine;
		string << "node_->getExpression() = " << *(node_->getExpression()) << endLine;
		string << "proposition->getExpression() = " << *(proposition->getExpression()) << endLine;
		string << endLine;


		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis_1 = assertion->getHypothesis (i);
			stack :: Expression* const
				hyp_1 = new stack :: Expression (hypothesis_1->getExpression());
			hyp_1->execute (substitutionExt);
			hyp_1->execute (getSubstitution());


			stack :: Expression* const
				hyp_2 = new stack :: Expression (up_.getValue (i)->getNode()->getExpression());
			hyp_2->execute (getSubstitution());

			string << "hypothesis " << i << endLine;
			string << "--------------------------------------" << endLine;
			string << "hyp_1 = " << *hyp_1 << endLine;
			string << "hyp_2 = " << *hyp_2 << endLine;

			string << endLine;
			string << "hypothesis_1->getExpression() = " << *(hypothesis_1->getExpression()) << endLine;
			string << "up_.getAll (i)->getExpression() = " << *(up_.getValue (i)->getNode()->getExpression()) << endLine;
			string << endLine;
		}
		string << "=============================================" << endLine;
		string << endLine << endLine;
		Memory :: stack().pop();
	}
}
}
}

#endif /*MDL_PROVER_EVIDENCE_PROP_CPP_*/
