/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_evidence_Ref.cpp                              */
/* Description:     reference proof evidence element                         */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_EVIDENCE_REF_CPP_
#define MDL_PROVER_EVIDENCE_REF_CPP_

namespace mdl {
namespace prover {
namespace evidence {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Ref :: Ref
	(
		const Node* node,
		const Substitution* replacement,
		const Substitution* substitution,
		const Evidence* up
	) :
	node_ (node),
	replacement_ (replacement),
	substitution_ (substitution),
	up_ (up),
	down_ (NULL),
	depth_ (up_->getDepth()),
	cardinality_ (up_->getCardinality()),
	clone_ (NULL) {
	}
	inline
	Ref :: Ref
	(
		const Node* node,
		const Substitution* replacement,
		const Evidence* down,
		const Evidence* up
	) :
	node_ (node),
	replacement_ (replacement),
	substitution_ (NULL),
	up_ (up),
	down_ (down),
	depth_ (up_->getDepth()),
	cardinality_ (up_->getCardinality()),
	clone_ (NULL) {
	}
	Ref :: ~ Ref() {
	}

	inline const Evidence*
	Ref :: getUp() const {
		return up_;
	}

	// prover :: Evidence implementation
	inline value :: Kind
	Ref :: getKind() const {
		return REF;
	}
	inline const Node*
	Ref :: getNode() const {
		return node_;
	}
	inline void
	Ref :: verifyIntegrity() const
	{
		up_->verifyIntegrity();
		Expression* const
			expressionUp = new Expression (up_->getNode()->getExpression());
		Expression* const
			expressionDown = new Expression (node_->getExpression());
		expressionDown->execute (replacement_);

		if (*expressionUp != *expressionDown) {
			Error* error = new Error (Error :: SEMANTIC);
			expression :: Mode mode;
			mode << expression :: Mode :: SHOW_REPLACEABILITY;
			error->message() << mode << "reference unification fail" << endLine;
			error->message() << *expressionUp << " != " << *expressionDown << endLine;
			error->message() << "replacement = " << *replacement_ << endLine;
			showVerification (error->message());
			throw error;
		}
	}
	const mdl :: prover :: Substitution*
	Ref :: getSubstitution() const {
		return (down_ == NULL) ? substitution_ : down_->getSubstitution();
	}
	inline index :: Step
	Ref :: getDepth() const {
		return depth_;
	}
	inline value :: Integer
	Ref :: getCardinality() const {
		return cardinality_;
	}
	const Proof*
	Ref :: constructProof (const Substitution* substitution) const
	{
		Substitution*
			global = Substitution :: create (getSubstitution());
		global = global->compose (substitution);

		const Proof* const
			upProof = up_->constructProof (global);

		return new proof :: Ref (this, global, upProof);
	}
	vector :: Proof
	Ref :: constructProofs (const Substitution* substitution) const
	{
		const vector :: Proof
			upProofs = up_->constructProofs (substitution);
		const value :: Integer size = upProofs.getSize();
		vector :: Proof proofs (size);
		for (value :: Integer i = 0; i < size; ++ i) {
			const Proof* const
				upProof = upProofs.getValue (i);
			const Proof* const
				proof = new proof :: Ref (this, substitution, upProof);
			proofs.add (proof);
		}
		return proofs;
	}

	// object :: Cloneable implementation
	const Evidence*
	Ref :: clone() const
	{
		if (clone_ != NULL) {
			return clone_;
		}
		return clone_;
	}

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;

	// object :: Object interface
	void
	Ref :: commitSuicide() {
		delete this;
	}
	Size_t
	Ref :: getVolume() const
	{
		Size_t volume = 0;
		volume += getAggregateVolume (replacement_);
		volume += getAggregateVolume (substitution_);
		return volume;
	}
	Size_t
	Ref :: getSizeOf() const {
		return sizeof (Ref);
	}
	void
	Ref :: show (String& string) const
	{
		string << "############################### - Ref Node - ################################################" << endLine;
		string << "this: " << this << endLine;
		string << "node: " << node_ << endLine;
		string << "replacement: " << *replacement_ << endLine;
		string << "substitution: " << *getSubstitution() << endLine;
		string << "up: " << up_ << endLine;
		string << "****************************************************" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		showVerification (string);
		string << "############################### - Ref Node - ################################################" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Ref :: showVerification (String& string) const
	{
		Expression* const
			expressionUp = new Expression (up_->getNode()->getExpression());
		Expression* const
			expressionDown = new Expression (node_->getExpression());
		expressionDown->execute (replacement_);

		string << "************* VERIFICATION *************" << endLine;
		string << "node_: " << endLine;
		string << *node_ << endLine;
		string << endLine;
		string << "expressionUp = " << *expressionUp << endLine;
		string << "expressionDown->execute (replacement_) = " << *expressionDown << endLine;
		string << "expressionDown = " << *(node_->getExpression()) << endLine;
		string << endLine;
		//string << "proof_ = " << endLine;
		//string << *this;
	}
}
}
}

#endif /*MDL_PROVER_EVIDENCE_REF_CPP_*/
