/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_evidence_Top.cpp                              */
/* Description:     top (premise) proof evidence element                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_EVIDENCE_TOP_CPP_
#define MDL_PROVER_EVIDENCE_TOP_CPP_

namespace mdl {
namespace prover {
namespace evidence {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Top :: Top (const Top& top) :
	node_ (top.node_),
	premise_ (top.premise_),
	down_ (top.down_),
	substitution_ (new Substitution (*top.substitution_)),
	clone_ (NULL) {
	}
	inline
	Top :: Top
	(
		const Node* node,
		const Premise* premise,
		const Substitution* substitution
	) :
	node_ (node),
	premise_ (premise),
	down_ (NULL),
	substitution_ (substitution),
	clone_ (NULL) {
		#ifdef DEBUG
		verifyIntegrity();
		#endif
	}
	inline
	Top :: Top
	(
		const Node* node,
		const Evidence* down,
		const Premise* premise
	) :
	node_ (node),
	premise_ (premise),
	down_ (down),
	substitution_ (NULL),
	clone_ (NULL) {
		#ifdef DEBUG
		verifyIntegrity();
		#endif
	}
	Top :: ~ Top() {
	}

	// prover :: Evidence implementation
	inline value :: Kind
	Top :: getKind() const {
		return TOP;
	}
	inline const Node*
	Top :: getNode() const {
		return node_;
	}
	inline void
	Top :: verifyIntegrity() const
	{
		Expression* const
			expression = new Expression (node_->getExpression());
		expression->execute (getSubstitution());
		const Expression* const
			premise = new Expression (premise_->getExpression());

		if (*expression != *premise) {
			Error* error = new Error (Error :: SEMANTIC);
			expression :: Mode mode;
			mode << expression :: Mode :: SHOW_REPLACEABILITY;
			error->message() << mode << "premise unification fail" << endLine;
			error->message() << *expression << " != " << *premise << endLine;
			error->message() << "substitution = " << *getSubstitution() << endLine;
			showVerification (error->message());
			throw error;
		}
	}
	inline const mdl :: prover :: Substitution*
	Top :: getSubstitution() const {
		return (down_ == NULL) ? substitution_ : down_->getSubstitution();
	}
	inline index :: Step
	Top :: getDepth() const {
		return 0;
	}
	inline value :: Integer
	Top :: getCardinality() const {
		return 0;
	}
	const Proof*
	Top :: constructProof (const Substitution*) const {
		return new proof :: Top (this);
	}
	vector :: Proof
	Top :: constructProofs (const Substitution* substutition) const
	{
		vector :: Proof proofs (1);
		const Proof* const
			proof = new proof :: Top (this);
		proofs.add (proof);
		return proofs;
	}

	// object :: Cloneable implementation
	const Evidence*
	Top :: clone() const
	{
		if (clone_ != NULL) {
			return clone_;
		}
		clone_ = new Top (*this);
		return clone_;
	}

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;

	// object :: Object implementation
	void
	Top :: commitSuicide() {
		delete this;
	}
	Size_t
	Top :: getVolume() const {
		return getAggregateVolume (substitution_);
	}
	Size_t
	Top :: getSizeOf() const {
		return sizeof (Top);
	}
	void
	Top :: show (String& string) const
	{
		string << "############################### - Top Node - ################################################" << endLine;
		string << "this: " << this << endLine;
		string << "node: " << node_ << endLine;
		string << "premise: " << premise_ << endLine;
		string << "substitution: " << *getSubstitution() << endLine;
		string << "****************************************************" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		showVerification (string);
		string << "############################### - Top Node - ################################################" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Top :: showVerification (String& string) const
	{
		stack :: Expression* const
			expression = new stack :: Expression (node_->getExpression());
		expression->execute (getSubstitution());
		const stack :: Expression* const
			premise = new stack :: Expression (premise_->getExpression());

		string << "************* VERIFICATION *************" << endLine;
		string << "node_ = " << endLine;
		string << *node_ << endLine;
		string << endLine;
		string << "expression_ = " << *(node_->getExpression()) << endLine;
		string << "expression_->execute (substitution_) = " << *expression << endLine;
		string << "premise_ = " << *premise << endLine;
		string << endLine;
		//string << "proof_ = " << endLine;
		//string << *this;
	}
}
}
}

#endif /*MDL_PROVER_EVIDENCE_TOP_CPP_*/
