/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_evidence_Vector.cpp                           */
/* Description:     confirmation: vector of evidences                        */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_EVIDENCE_VECTOR_CPP_
#define MDL_PROVER_EVIDENCE_VECTOR_CPP_

namespace mdl {
namespace prover {
namespace evidence {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Vector :: Vector (const Node* node, const value :: Integer capacity) :
	evidenceVector_ (capacity),
	garbageVector_ (0),
	boundary_ (0),
	node_ (node) {
	}
	inline
	Vector :: Vector (const Vector& vector) :
	evidenceVector_ (vector.evidenceVector_),
	garbageVector_ (vector.garbageVector_),
	boundary_ (vector.boundary_),
	node_ (vector.node_) {
	}
	Vector :: ~ Vector() {
	}

using manipulator :: endLine;
using manipulator :: tab;
using manipulator :: space;

	inline void
	Vector :: clear()
	{
		evidenceVector_.clear();
		boundary_ = 0;
	}
	inline bool
	Vector :: isEqual (const Vector* vector) const
	{
		const bool a = this->isSubset (vector);
		const bool b = vector->isSubset (this);
		const bool c = vector->getSize() == getSize();
		const bool d = vector->getDelta() == getDelta();
		return a && b && c && d;
	}
	inline void
	Vector :: showDifference (String& string, const Vector* vector) const
	{
		string << "direction A, size = " << getSize() << ", ";
		string << "delta = " << getDelta() << ":" << endLine;
		showDiffSubset (string, vector);
		string << endLine;
		string << "direction B, size = " << vector->getSize() << ", ";
		string << "delta = " << vector->getDelta() << ":" << endLine;
		vector->showDiffSubset (string, this);
		string << endLine;
	}
	void
	Vector :: showSubstitutions (String& string, const bool showUp) const
	{
		string << "evidences: " << endLine;
		if (evidenceVector_.getSize() == 0) {
			string << tab << "none." << endLine;
			return;
		}
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			string << tab << evidence << tab;
			string << "depth = " << evidence->getDepth() << ";" << tab;
			if (i < boundary_) {
				string << "old" << ";" << tab;
			} else {
				string << "new" << ";" << tab;
			}
			if (showUp && (evidence->getKind() == Evidence :: PROP)) {
				string << "up: ";
				const evidence :: Prop* const
					prop = dynamic_cast<const evidence :: Prop*>(evidence);
				for (index :: Arity i = 0; i < prop->getArity(); ++ i) {
					const Evidence* const
						upEvidence = prop->getUp (i);
					string << upEvidence << space;
				}
			}
			if (showUp && (evidence->getKind() == Evidence :: SUB)) {
				string << "up: ";
				const evidence :: Sub* const
					sub = dynamic_cast<const evidence :: Sub*>(evidence);
				for (value :: Integer i = 0; i < sub->getSize(); ++ i) {
					const Evidence* const
						upEvidence = sub->getUp (i);
					const evidence :: Prop* const
						prop = dynamic_cast<const evidence :: Prop*>(upEvidence);
					if (i != 0) {
						string << ";" << space;
					}
					for (index :: Arity j = 0; j < prop->getArity(); ++ j) {
						const Evidence* const
							up = prop->getUp (j);
						string << up << space;
					}
				}
			}
			const Substitution* const
				substitution = evidence->getSubstitution();
			if (substitution == NULL) {
				string << "NULL";
			} else {
				string << *substitution;
			}
			string << endLine;
			if (i == evidenceVector_.getSize() - 1) {
				string << endLine;
			}
		}
	}

	// prover :: Confirmation implementation
	inline void
	Vector :: verifyIntegrity() const
	{
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			evidenceVector_.getValue (i)->verifyIntegrity();
		}
	}
	inline bool
	Vector :: boundaryMoved() const {
		return (evidenceVector_.getSize() > boundary_);
	}
	inline void
	Vector :: markBoundary()
	{
		value :: Integer counter = 0;
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			if (Prover :: maxDepth (evidence->getDepth())) {
				evidenceVector_.shiftToEnd (i);
				++ counter;
			}
		}
		boundary_ = evidenceVector_.getSize() - counter;
	}
	inline value :: Integer
	Vector :: getBoundary() const {
		return boundary_;
	}
	inline value :: Integer
	Vector :: getSize() const {
		return evidenceVector_.getSize();
	}
	inline value :: Integer
	Vector :: getDelta() const {
		return evidenceVector_.getSize() - boundary_;
	}
	Evidence*
	Vector :: supply (const stack :: Substitution* sub)
	{
		//static int counterA = 0;
		//static int counterB = 0;
		if (!Prover :: fitsSubstitutionComplexity (sub->getComplexity())) {
			if (Counter :: stop (0) && Counter :: stop (1)) {
				//std :: cerr << "AAA = " << ++ counterA << std :: endl;
			}
			if (Counter :: stop (0) && Counter :: stop (2)) {
				//std :: cerr << "BBB = " << ++ counterB << std :: endl;
			}
			return NULL;
		}
		Evidence* result = NULL;
		const Substitution* substitution_1 = NULL;
		const Substitution* substitution_2 = sub->castAway<allocator :: Prover>();
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const evidence = evidenceVector_.getValue (i);
			substitution_1 = evidence->getSubstitution();
			if (substitution_2->isIdentical (substitution_1)) {
				return result = const_cast<Evidence*>(evidence);
				break;
			}
		}
		if (result == NULL) {
			const Substitution* const
				substitution = Substitution :: create (sub);
			result = new evidence :: Sub (node_, substitution);
			evidenceVector_.add (result);
		}
		return result;
	}
	inline void
	Vector :: add (const Evidence* evidence) {
		evidenceVector_.add (evidence);
	}
	inline void
	Vector :: replace (const Evidence* evidence, const value :: Integer i)
	{
		garbageVector_.add (evidenceVector_.getValue (i));
		evidenceVector_.getReference (i) = evidence;
	}
	inline const Evidence*
	Vector :: get (const value :: Integer index) const {
		return evidenceVector_.getValue (index);
	}
	inline const Evidence*
	Vector :: getNew (const value :: Integer index) const {
		return evidenceVector_.getValue (index + boundary_);
	}

	Size_t
	Vector :: getEvidenceVolume() const
	{
		Size_t volume = 0;
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			volume += getAggregateVolume (evidence);
		}
		return volume;
	}
	Size_t
	Vector :: getSubstitutionVolume() const
	{
		Size_t volume = 0;
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			volume += getAggregateVolume (evidence->getSubstitution());
		}
		return volume;
	}
	Size_t
	Vector :: getEvidenceNodeVolume() const
	{
		Size_t volume = 0;
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			volume += evidence->getSizeOf();
		}
		return volume;
	}
	Size_t
	Vector :: getGarbageVolume() const
	{
		Size_t volume = 0;
		for (value :: Integer i = 0; i < garbageVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = garbageVector_.getValue (i);
			volume += getAggregateVolume (evidence);
		}
		return volume;
	}

	// object :: Object interface
	void
	Vector :: commitSuicide() {
		delete this;
	}
	Size_t
	Vector :: getVolume() const
	{
		Size_t volume = 0;
		volume += evidenceVector_.getVolume();
		volume += garbageVector_.getVolume();
		return volume;
	}
	Size_t
	Vector :: getSizeOf() const {
		return sizeof (Vector);
	}
	void
	Vector :: show (String& string) const
	{
		if (evidenceVector_.getSize() == 0) {
			//string << "evidences: none." << endLine;
			return;
		}
		string << "evidences:" << endLine;
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence*
				evidence = evidenceVector_.getValue (i);
			evidence->show (string);
		}
	}

	/****************************
	 *		Public members
	 ****************************/

	inline bool
	Vector :: isSubset (const Vector* vector) const
	{
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			if (!vector->contains (evidence)) {
				return false;
			}
		}
		return true;
	}
	inline bool
	Vector :: contains (const Evidence* evidence_1) const
	{
		const Substitution* const
			substutution_1 = evidence_1->getSubstitution();
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence_2 = evidenceVector_.getValue (i);
			const Substitution* const
				substutution_2 = evidence_2->getSubstitution();
			if (substutution_1->isIdentical (substutution_2)) {
				return true;
			}
		}
		return false;
	}
	inline void
	Vector :: showDiffSubset (String& string, const Vector* vector) const
	{
		for (value :: Integer i = 0; i < evidenceVector_.getSize(); ++ i) {
			const Evidence* const
				evidence = evidenceVector_.getValue (i);
			if (!vector->contains (evidence)) {
				const Substitution* const
					substitution = evidence->getSubstitution();
				string << tab << evidence << tab;
				if (substitution == NULL) {
					string << "NULL"<< space;
				} else {
					string << *substitution << space;
				}
				string << "is absent" << endLine;
			}
		}
	}

	inline bool
	Vector :: disjointedsAreNew
	(
		const Evidence* upEvidence,
		const Evidence* oldEvidence
	) const
	{
		const mdl :: Proposition* const
			propUp = upEvidence->getNode()->getProposition();
		return disjointedsAreNew (propUp, oldEvidence);
	}
	inline bool
	Vector :: disjointedsAreNew
	(
		const mdl :: Proposition* propUp,
		const Evidence* oldEvidence
	) const
	{
		const mdl :: Proposition* const
			propOld = oldEvidence->getNode()->getProposition();
		if ((propUp == NULL) || (propOld == NULL)) {
			return !(propUp == propOld);
		}
		const mdl :: Disjointeds* const
			disjUp = propUp->getAssertion()->getDisjointeds();
		const mdl :: Disjointeds* const
			disjOld = propOld->getAssertion()->getDisjointeds();

		return !(disjOld->isLessRestrictive (disjUp));
	}
}
}
}

#endif /*MDL_PROVER_EVIDENCE_VECTOR_CPP_*/

