/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_Prover.cpp                                    */
/* Description:     main static prover class                                 */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_PROVER_CPP_
#define MDL_PROVER_PROVER_CPP_

namespace mdl {
namespace prover {

	/****************************
	 *		Public members
	 ****************************/

using manipulator :: endLine;
using manipulator :: underline;

	// prover :: Prover implementation
	bool
	Prover :: prove
	(
		const mdl :: Assertion* problem,
		const mdl :: proof :: Step* question,
		const index :: Step limitDepth
	)
	{
		//Counter :: inc();
		//Counter :: show();
		//Counter :: set (75);
		/*Counter :: set (3432);
		if (Counter :: stop()) {
			//std :: cerr << *problem << std :: endl;
			//std :: cerr << *question << std :: endl;
		}*/

		Memory :: stack().push();
 		Memory :: boxed().clear(); // prepare the battlefield
 		init (problem, question, limitDepth);
		timers_.start(); // death or glory!
		showStart();
		try {
			while (!root_->isProved()) {
				try {
					+ top_; // scaling a tree level up,
					while (!top_.finished()) {
						++ top_; // increment a top tree level,
						//verifyIntegrity();
					}
					/*if (Counter :: stop()) {
						show();
					}*/
				} catch (const OutOfMemory&) { // we are not afraid of that.
					throw Exception ("prover has ran out of memory. "); // Death.
					//-- top_; // we have 'unlimited' memory
				} catch (Error* error) {
					refresh ("error", true);
					throw;
				}
			}
		} catch (const Exception& exception) {
			if (!root_->isProved() && (Config :: verbosity() >= Config :: VERBOSE_STANDARD)) {
				std :: cerr << exception.getMessage() << std :: flush; // Death.
			}
		}
		if (root_->isProved()) { // the proof is found, Glory!
			answer_ = root_->getProofs().compileProof (question_);
		}
		timers_.stop(); // the battle is over ...
		volume_.calculate();
		showResult();
		Memory :: boxed().clear(); // bury all the dead heroes ...
		Memory :: stack().pop();
		Memory :: stack().clear();
		return (root_->isProved()); // write this to the history pages of the endless war for the Truth.
	}
	inline const mdl :: Assertion*
	Prover :: getProblem() {
		return problem_;
	}
	inline const tree :: Premise*
	Prover :: getPremises() {
		return premises_;
	}
	inline index :: Assertion
	Prover :: getPropositionBoundary()
	{
		const mdl :: Proposition*
			firstProposition = problem_->getProposition (0);
		const index :: Assertion
			boundary = firstProposition->getSourceIndex();
		return boundary;
	}
	inline index :: Step
	Prover :: getStepBoundary() {
		return question_->getIndex() + problem_->getHypArity();
	}
	inline const mdl :: proof :: Step*
	Prover :: getQestionStep() {
		return question_;
	}
	inline const mdl :: proof :: Step*
	Prover :: getAnswerStep() {
		return answer_;
	}
	inline index :: Step
	Prover :: getDepth() {
		return top_.getDepth();
	}
	inline Symbol
	Prover :: createVariable (const Symbol& symbol)
	{
		Symbol newSymbol (symbol);
		value :: Integer
			counter = variables_->addFind (symbol.getLiteral(), 0);

		Memory :: stack().push();
		nstd :: String<allocator :: Stack> string (64);
		const String&
			originalString = Table :: literals()->get (symbol.getLiteral());
		while (true) {
			++ counter; string.clear();
			string << originalString << underline << counter;
			const value :: Literal
				literal = Table :: literals()->add (string);
			if (question_->getScope()->getVariables()->getType (literal) == NULL) {
				newSymbol.setLiteral (literal);
				variables_->set (symbol.getLiteral(), counter);
				break;
			}
		}
		Memory :: stack().pop();

		return newSymbol;
	}

	inline bool
	Prover :: maxDepth (const index :: Step depth) {
		return (depth == top_.getDepth());
	}
	inline bool
	Prover :: fitsDepth (const index :: Step depth)
	{
		return
			(top_.getDepth() <= 1) ||
			(depth  < top_.getDepth());
	}
	inline bool
	Prover :: fitsSubstitutionComplexity (const value :: Integer complexity)
	{
		return
			(top_.getDepth() <= 1) ||
			(complexity <= Config :: getMaxSubstitutionComplexity());
	}
	inline bool
	Prover :: fitsExpressionComplexity (const value :: Integer complexity)
	{
		return
			(top_.getDepth() <= 1) ||
			(complexity <= Config :: getMaxExpressionComplexity());
	}

	inline void
	Prover :: refresh (const char* message, const bool force)
	{
		if (force || timers_.refresh()) {
			showStatus (message);
		}
	}
	inline void
	Prover :: show()
	{
		static String string (1024 * 64);
		show (string);
		std :: cerr << string << std :: endl;
	}
	inline const Log&
	Prover :: getLog() {
		return log_;
	}
	const Timers&
	Prover :: getTimers() {
		return timers_;
	}
	const Volume&
	Prover :: getVolumes() {
		return volume_;
	}
	inline Log&
	Prover :: log() {
		return log_;
	}
	inline Timers&
	Prover :: timers() {
		return timers_;
	}

	inline void
	Prover :: verifyIntegrity() {
		volume_.verifyIntegrity();
	}

	inline void
	Prover :: init() {
		log_.init();
	}
	inline void
	Prover :: release() {
		log_.destroy();
	}

using manipulator :: iterate;
using manipulator :: mode;
using manipulator :: tab;
using manipulator :: space;

	// object :: Object implementation
	Size_t
	Prover :: getVolume()
	{
		Size_t volume = 0;
		volume += timers_.getVolume();
		volume += log_.getVolume();
		return volume;
	}
	Size_t
	Prover :: getSizeOf() {
		return sizeof (Prover);
	}
	void
	Prover :: show (String& string)
	{
		string << manipulator :: mode;
		expression :: Mode mode;
		string << (mode << expression :: Mode :: SHOW_REPLACEABILITY);
		string << "variant tree." << endLine;
		string << "=============" << endLine;
		if (root_ != NULL) {
			root_->showTree (string);
		} else {
			string << "NULL" << endLine;
		}
		string << iterate << 16 << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Prover :: init
	(
		const mdl :: Assertion* problem,
		const mdl :: proof :: Step* question,
		const index :: Step limitDepth
	)
	{
		node :: Hyp :: expressions_ = new tree :: Node();
		variables_ = new Variables_ (INITIAL_VARIABLES_CAPACITY);

		premises_ = new tree :: Premise();
		question->completeWithSteps (premises_);

		problem_  = problem;
		question_ = question;
		answer_  =  NULL;
		root_ = NULL;
		top_.init (limitDepth);
		Expression :: clearCounter();
		if (Config :: getLog() != NULL) {
			log_.init (Config :: getLog()->getString());
		} else {
			log_.init ();
		}
		const array :: Expression* const
			expression = question_->getExpression();
		const Expression* const
			target = new Expression (expression);
		root_ = new Root (target);
		top_.add (root_);
	}
	void
	Prover :: showStart()
	{
		if (Config :: verbosity() < Config :: VERBOSE_STANDARD) {
			return;
		}
		log_.message (0) << "proving ";
		log_.message (0) << "step " << question_->getIndex() + 1 << " in ";
		problem_->replicateName (log_.message (0));
		log_.message (0) << " (" << problem_->getIndex() << ") ... ";
		log_.show (0);
	}
	void
	Prover :: showResult()
	{
		if (Config :: verbosity() < Config :: VERBOSE_STANDARD) {
			return;
		}
		if (Config :: verbosity() == Config :: VERBOSE_STANDARD) {
			if (root_->isProved()) {
				log_.message (0) << "done in " << timers_.prove();
				log_.message (0) << "." << endLine;
			} else {
				log_.message (0) << "failed." << endLine;
			}
		}
		if (Config :: verbosity() > Config :: VERBOSE_STANDARD) {
			if (root_->isProved()) {
				log_.message (0) << "done in " << timers_.prove();
				log_.message (0) << "." << endLine;
			} else {
				log_.message (0) << "failed." << endLine;
			}
		}
		log_.show (0);
	}
	void
	Prover :: showStatus (const char* message)
	{
		if (Config :: verbosity() < Config :: VERBOSE_STANDARD) {
			return;
		}
		if (Config :: verbosity() == Config :: VERBOSE_STANDARD) {
			log_.message (1) << endLine << tab;
			log_.message (1) << timers_ << ", ";
			top_.show (log_.message (1));
			log_.message (1) << ", ";
			volume_.calculate();
			volume_.show (log_.message (1));
			log_.message (1) << space;
			//log_.message (1) << message;
		}
		if (Config :: verbosity() > Config :: VERBOSE_STANDARD) {
			log_.message (1) << endLine;
			log_.message (1) << message << endLine;
			log_.message (1) << timers_;
			log_.message (1) << "=================================" << endLine;
			top_.show (log_.message (1));
			volume_.calculate();
			volume_.show (log_.message (1));
			log_.message (1) << "=================================" << endLine;
		}
		log_.show (1);
	}

	/************************************
	 *		Static private attributes
	 ************************************/

	const mdl :: Assertion*
		Prover :: problem_ = NULL;
	const mdl :: proof :: Step*
		Prover :: question_ = NULL;
	const mdl :: proof :: Step*
		Prover :: answer_ = NULL;

	tree :: Premise*
		Prover :: premises_ = NULL;
	Prover :: Variables_*
		Prover :: variables_ = NULL;
	Root*
		Prover :: root_ = NULL;
	Top
		Prover :: top_;
	Log
		Prover :: log_;
	Timers
		Prover :: timers_;
	Volume
		Prover :: volume_;
}
}

#endif /*MDL_PROVER_PROVER_CPP_*/
