/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_Volume.cpp                                    */
/* Description:     volume counter for prover                                */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_VOLUME_CPP_
#define MDL_PROVER_VOLUME_CPP_

namespace mdl {
namespace prover {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Volume :: Volume () :
	treeNodeCount_ (0), evidenceNodeCount_ (0), expressionCount_(0),
	hypNodeCount_ (0), propNodeCount_ (0), refNodeCount_(0), nodeCountSum_ (0),
	tree_(0), evidences_ (0), top_ (0), expression_ (0), garbage_ (0),
	treeNode_ (0), evidenceNode_(0), evidenceContainer_ (0), expressions_ (0),
	substitutionAll_ (0), substitutionInt_ (0), substitutionExt_ (0),
	substitution_ (0), substitutionSum_ (0),
	allocatedBoxed_ (0), usedBoxed_ (0), spaceBoxed_ (0),
	maxStack_ (0), usedStack_ (0), spaceStack_ (0),
	deltaBoxed_ (0), deltaSubstitutions_ (0), deltaNodeCount_ (0) {
	}

using manipulator :: endLine;

	inline void
	Volume :: verifyIntegrity()
	{
		calculate();
		if ((deltaBoxed_ != 0) || (deltaSubstitutions_ != 0) || (deltaNodeCount_ != 0)) {
			Error* error = new Error (Error :: PROVER);
			error->message() << "prover memory leaks" << endLine;
			error->message() << endLine;
			show (error->message());
			//std :: cerr << error->message() << std :: endl;
			throw error;
		}
	}
	inline void
	Volume :: calculate()
	{
		if (Prover :: root_ == NULL) {
			return;
		}
		calculateData();
		calculateBalance();
	}

using manipulator :: tab;
using manipulator :: iterate;
using manipulator :: mode;


	// object :: Object implementation
	inline Size_t
	Volume :: getVolume() const
	{
		Size_t volume = 0;
		volume += tree_;
		volume += evidences_;
		volume += top_;
		volume += garbage_;
		return volume;
	}
	inline Size_t
	Volume :: getSizeOf() const {
		return sizeof (Volume);
	}
	inline void
	Volume :: show (String& string) const
	{
		if (Prover :: root_ == NULL) {
			return;
		}
		if (Config :: verbosity() > Config :: VERBOSE_STANDARD) {
			showVolume (string);
			showBalance (string);
		} else {
			string << "volume: " << (usedBoxed_ / nstd :: MEGABYTE) << " mb, ";
			string << "proofs: " << (evidences_ / nstd :: MEGABYTE) << " mb, ";
			string << "tree: " << (tree_ / nstd :: MEGABYTE) << " mb, ";
			/*string  << endLine << tab;
			string << "hyp %: " << 100 * (double)hypNodeCount_ / (double)treeNodeCount_ << ", ";
			string << "prop %: " << 100 * (double)propNodeCount_ / (double)treeNodeCount_ << ", ";
			string << "ref %: " << 100 * (double)refNodeCount_ / (double)treeNodeCount_ << ", ";
			string << "node: " << treeNodeCount_ << ", ";
			string << "proof: " << evidenceNodeCount_ << ", ";
			//string << "node count: " << treeNodeCount_ << ", ";*/

			//string << "tree: " << treeNodeCount_ << ", ";
			//string << "allocated: " << (allocatedBoxed_ / nstd :: MEGABYTE) << " mb, ";
			//string << "proofs: " << evidenceNodeCount_ << ", ";
			//string << "tree: " << treeNodeCount_ << ", ";
			//string << "balance: " << (deltaBoxed_ / nstd :: MEGABYTE) << " mb ";
			//string << "garbage: " << (garbage_ / nstd :: MEGABYTE) << " mb ";
			//string << "max stack: " << (maxStack_ / nstd :: MEGABYTE) << " mb, ";
			//string << "used stack: " << (usedStack_ / nstd :: MEGABYTE) << " mb";
		}
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Volume :: calculateData()
	{
		treeNodeCount_ = Prover :: root_->getTreeNodeCount();
		evidenceNodeCount_ = Prover :: root_->getEvidenceNodeCount();
		expressionCount_ = Prover :: root_->getExpressionCount();
		hypNodeCount_ = Prover :: root_->getHypCount();
		propNodeCount_ = Prover :: root_->getPropCount();
		refNodeCount_ = Prover :: root_->getRefCount();

		tree_ = Prover :: root_->getTreeVolume();
		evidences_ = Prover :: root_->getEvidenceVolume();
		expression_ = Prover :: root_->getExpressionVolume();

		treeNode_ = Prover :: root_->getTreeNodeVolume();
		evidenceNode_ = Prover :: root_->getEvidenceNodeVolume();
		evidenceContainer_ = Prover :: root_->getEvidenceContainerVolume();
		expressions_ = node :: Hyp :: getExpressionsVolume();
		garbage_ = Prover :: root_->getGarbageVolume();

		substitutionAll_ = Prover :: root_->getSubstitutionAllVolume();
		substitutionInt_ = Prover :: root_->getSubstitutionIntVolume();
		substitutionExt_ = Prover :: root_->getSubstitutionExtVolume();
		substitution_ = Prover :: root_->getSubstitutionVolume();

		top_ = Prover :: top_.getVolume();
	}
	inline void
	Volume :: calculateBalance()
	{
		calculateMemoryBalance();
		calculateSubstitutionBalance();
		calculateNodeCountBalance();
	}
	inline void
	Volume :: calculateMemoryBalance()
	{
		usedBoxed_ = 0;
		usedBoxed_ += Prover :: root_->getSizeOf();
		usedBoxed_ += tree_;
		usedBoxed_ += evidences_;
		usedBoxed_ += top_;
		usedBoxed_ += expressions_;
		usedBoxed_ += garbage_;

		usedStack_ = Memory :: stack().getAllocatedVolume();
		maxStack_ = Memory :: stack().getMaxVolume();
		spaceStack_ = Memory :: stack().getOverallVolume();

		allocatedBoxed_ = Memory :: boxed().getAllocatedVolume();
		spaceBoxed_ = Memory :: boxed().getOverallVolume();
		deltaBoxed_ =
			static_cast<Ptrdiff_t> (allocatedBoxed_) -
			static_cast<Ptrdiff_t> (usedBoxed_);
	}
	inline void
	Volume :: calculateSubstitutionBalance()
	{
		substitutionSum_ = 0;
		substitutionSum_ += substitutionInt_;
		substitutionSum_ += substitutionExt_;
		substitutionSum_ += substitution_;

		deltaSubstitutions_ =
			static_cast<Ptrdiff_t> (substitutionAll_) -
			static_cast<Ptrdiff_t> (substitutionSum_);
	}
	inline void
	Volume :: calculateNodeCountBalance()
	{
		nodeCountSum_ = 0;
		nodeCountSum_ += hypNodeCount_;
		nodeCountSum_ += propNodeCount_;
		nodeCountSum_ += refNodeCount_;

		deltaNodeCount_ =
			static_cast<Ptrdiff_t> (treeNodeCount_) -
			static_cast<Ptrdiff_t> (nodeCountSum_);
	}

	inline void
	Volume :: showVolume (String& string) const
	{
		if (Prover :: root_ == NULL) {
			return;
		}
		string << "prover memory usage: " << endLine;
		string << "---------------------" << endLine;
		string << memory :: UsageMessage (usedBoxed_, usedBoxed_, "total volume");
		string << memory :: UsageMessage (usedBoxed_, tree_, "tree volume");
		string << memory :: UsageMessage (usedBoxed_, evidences_, "evidences volume");
		string << memory :: UsageMessage (usedBoxed_, expression_, "expression volume");
		string << memory :: UsageMessage (usedBoxed_, treeNode_, "tree node volume");
		string << memory :: UsageMessage (usedBoxed_, evidenceNode_, "evidence node volume");
		string << memory :: UsageMessage (usedBoxed_, evidenceContainer_, "evidence container volume");
		string << memory :: UsageMessage (usedBoxed_, top_, "top volume");
		string << memory :: UsageMessage (usedBoxed_, expressions_, "expression tree volume");
		string << memory :: UsageMessage (usedBoxed_, garbage_, "garbage volume");
		string << endLine;
		string << memory :: UsageMessage (usedBoxed_, substitutionAll_, "all substitution volume");
		string << memory :: UsageMessage (usedBoxed_, substitutionInt_, "int substitution volume");
		string << memory :: UsageMessage (usedBoxed_, substitutionExt_, "ext substitution volume");
		string << memory :: UsageMessage (usedBoxed_, substitution_, "substitution volume");
		string << endLine;
		string << "tree node count: " << treeNodeCount_ << endLine;
		string << "evidence node count: " << evidenceNodeCount_ << endLine;
		string << "expression count: " << expressionCount_ << endLine;
		string << "hyp node count: " << hypNodeCount_ << endLine;
		string << "prop node count: " << propNodeCount_ << endLine;
		string << "ref node count: " << refNodeCount_ << endLine;
		string << "average tree node volume: " << tree_ / treeNodeCount_ << endLine;
		string << "average evidence node volume: " << evidences_ / evidenceNodeCount_ << endLine;
		string << "average expression volume: " << expression_ / expressionCount_ << endLine;
		string << endLine;
		string << memory :: UsageMessage (spaceBoxed_, spaceBoxed_, "prover space volume");
		string << memory :: UsageMessage (spaceStack_, spaceStack_, "stack  space volume");
		string << endLine;
	}
	inline void
	Volume :: showBalance (String& string) const
	{
		string << "prover memory balance (must be 0):" << endLine;
		string << "----------------------------------" << endLine;
		string << "delta:" << tab << deltaBoxed_ << " = ";
		string << allocatedBoxed_ << " (allocated) - " << usedBoxed_ << " (used) " << endLine;
		string << "delta substitutions:" << tab << deltaSubstitutions_ << " = ";
		string << substitutionAll_ << " (all) - " << substitutionSum_ << " (sum) " << endLine;
		string << "delta node count:" << tab << deltaNodeCount_ << " = ";
		string << treeNodeCount_ << " (all) - " << nodeCountSum_ << " (sum) " << endLine;
		string << endLine;
	}
}
}

#endif /*MDL_PROVER_VOLUME_CPP_*/
