/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Hyp.cpp                                  */
/* Description:     hypothesis node in proof variant tree                    */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_HYP_CPP_
#define MDL_PROVER_NODE_HYP_CPP_

#include "mdl/prover/node/mdl_prover_node.hpp"

namespace mdl {
namespace prover {
namespace node {

	/****************************
	 *		Public members
	 ****************************/

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;
using manipulator :: iterate;

	Hyp :: Hyp
	(
		const Expression* const expression,
		const Node* const node,
		const bool isRoot
	) :
	expression_ (expression),
	up_ (),
	down_ (),
	clone_ (NULL),
	rate_ (value :: undefined :: RATE),
	weight_ (0),
	confirmation_ (isRoot ? node : this)
	{
		if (isRoot) {
			unifyPremises (node);
		} else {
			const_cast<list :: Node&>(down_).addLast (node);
			Prover :: refresh ("creating hyp node");
			unifyPremises (this);
		}
	}
	Hyp :: ~ Hyp() {
	}

	// prover :: Node implementation
	const Expression*
	Hyp :: getExpression() const {
		return expression_;
	}
	const mdl :: Proposition*
	Hyp :: getProposition() const {
		return NULL;
	}
	const Confirmation&
	Hyp :: getConfirmation() const {
		return confirmation_;
	}
	value :: Weight
	Hyp :: getWeight() const {
		return weight_;
	}
	void
	Hyp :: calculateWeightUp() const
	{
		calculateWeight();
		const list :: Node :: Node_*
			node = up_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			prop->calculateWeightUp();
			node = node->getNext();
		}
	}
	void
	Hyp :: calculateWeightDown() const
	{
		const list :: Node :: Node_*
			node = down_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			prop->calculateWeightDown();
			node = node->getNext();
		}
		calculateWeight();
	}
	bool
	Hyp :: operator < (const Node& node) const
	{
		if (rate_ == value :: undefined :: RATE) {
			return (weight_ > node.getWeight());
		} else {
			return (rate_ < node.getRate());
		}
	}
	value :: Rate
	Hyp :: getRate() const {
		return rate_;
	}
	void
	Hyp :: setRate (const value :: Rate rate) const
	{
		if (rate_ <= rate) {
			return;
		}
		rate_ = rate;
		const list :: Node :: Node_*
			node = down_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			prop->setRate (rate);
			node = node->getNext();
		}
	}
	void
	Hyp :: searchProof()
	{
		const list :: Node :: Node_*
			node = up_.getFirst();
		while (node != NULL) {
			Node* const
				prop = const_cast<Node*>(node->getValue());

			if (Counter :: stop (1)) {
				Counter :: inc (2);
				Counter :: set (23, 2);
				//if (Counter :: stop (2)) {
				//	std :: cerr << *prop << std :: endl;
				//}
			}

			prop->searchProof();
			node = node->getNext();
		}
		buildDown();
	}
	bool
	Hyp :: isProved() const {
		return (confirmation_.getSize() > 0);
	}

	void
	Hyp :: buildUp (Level* const level) {
		buildUp (this, level);
	}
	void
	Hyp :: buildDown (const bool)
	{
		Prover :: refresh ("building down hyp node");
		collectConfirmation (this);
		if (confirmation_.boundaryMoved()) {
			calculateWeightUp();
			followDown();
			confirmation_.markBoundary();
		}
	}
	bool
	Hyp :: isRoot() const {
		return (down_.isEmpty());
	}
	void
	Hyp :: addDown (const Node* node) {
		const_cast<list :: Node&>(down_).addLast (node);
	}
	bool
	Hyp :: liesDown (const Node* node) const
	{
		if (node == this) {
			return true;
		}
		const list :: Node :: Node_*
			listNode = down_.getFirst();
		while (listNode != NULL) {
			const Node* const
				prop = listNode->getValue();
			if (prop->liesDown (node)) {
				return true;
			}
			listNode = listNode->getNext();
		}
		return false;
	}

	void
	Hyp :: showNode
	(
		String& string,
		const value :: Integer depth
	) const
	{
		string << iterate << depth << tab << "hyp: " << *expression_ << (isProved() ? " proved " : "") << endLine;
		string << iterate << depth << tab << "this: " << this << endLine;
		confirmation_.show (string);
		string << iterate << depth << tab << "down: ";
		const list :: Node :: Node_*
			node = down_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			string << prop;
			if (node->getNext() != NULL) {
				string << ", ";
			} else {
				string << endLine;
			}
			node = node->getNext();
		}
		//string << endLine;
	}
	void
	Hyp :: showTree
	(
		String& string,
		const value :: Integer depthStart,
		const value :: Integer depthEnd
	) const
	{
		if (depthEnd > 0) {
			string << iterate << depthStart << tab;
			string << iterate << 64 << '-' << endLine;
		}
		showNode (string, depthStart);

		if (depthEnd == 0) {
			return;
		}
		//confirmation_.show (string);
		const list :: Node :: Node_*
			node = up_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			prop->showTree (string, depthStart + 1, depthEnd - 1);
			node = node->getNext();
		}
		if (depthEnd > 0) {
			string << iterate << depthStart << tab;
			string << iterate << 64 << '-' << endLine;
		}
	}

	// prover :: Statistics implementation
	Size_t
	Hyp :: getEvidenceVolume() const {
		return Statistics :: getEvidenceVolume (this);
	}
	Size_t
	Hyp :: getTreeVolume() const {
		return Statistics :: getTreeVolume (this);
	}
	Size_t
	Hyp :: getExpressionVolume() const {
		return Statistics :: getExpressionVolume (this);
	}
	Size_t
	Hyp :: getTreeNodeVolume() const {
		return Statistics :: getTreeNodeVolume (this);
	}
	Size_t
	Hyp :: getEvidenceNodeVolume() const {
		return Statistics :: getEvidenceNodeVolume (this);
	}
	Size_t
	Hyp :: getEvidenceContainerVolume() const {
		return Statistics :: getEvidenceContainerVolume (this);
	}
	Size_t
	Hyp :: getGarbageVolume() const {
		return Statistics :: getGarbageVolume (this);
	}

	// substitution statistics
	Size_t
	Hyp :: getSubstitutionAllVolume() const {
		return Statistics :: getSubstitutionAllVolume (this);
	}
	Size_t
	Hyp :: getSubstitutionIntVolume() const {
		return Statistics :: getSubstitutionIntVolume (this);
	}
	Size_t
	Hyp :: getSubstitutionExtVolume() const {
		return Statistics :: getSubstitutionExtVolume (this);
	}
	Size_t
	Hyp :: getSubstitutionVolume() const {
		return Statistics :: getSubstitutionVolume (this);
	}

	// node count statistics
	value :: Integer
	Hyp :: getTreeNodeCount() const {
		return Statistics :: getTreeNodeCount (this);
	}
	value :: Integer
	Hyp :: getEvidenceNodeCount() const {
		return Statistics :: getEvidenceNodeCount (this);
	}
	value :: Integer
	Hyp :: getExpressionCount() const {
		return Statistics :: getExpressionCount (this);
	}
	value :: Integer
	Hyp :: getHypCount() const {
		return Statistics :: getHypCount (this);
	}
	value :: Integer
	Hyp :: getPropCount() const {
		return Statistics :: getPropCount (this);
	}
	value :: Integer
	Hyp :: getRefCount() const {
		return Statistics :: getRefCount (this);
	}

	// object :: Cloneable implementation
	const Node*
	Hyp :: clone () const
	{
		if (clone_ != NULL) {
			return clone_;
		}
		Hyp* clone = new Hyp (*this);
		clone_ = clone;
		cloneDown (clone);
		cloneMaxRate (clone);
		cloneProved (clone);
		cloneConfirmation (clone);
		return clone_;
	}

	// object :: Object implementation
	void
	Hyp :: commitSuicide() {
		delete this;
	}
	Size_t
	Hyp :: getVolume() const
	{
		Size_t volume = 0;
		volume += getTreeVolume();
		volume += getEvidenceVolume();
		return volume;
	}
	Size_t
	Hyp :: getSizeOf() const {
		return sizeof (Hyp);
	}
	void
	Hyp :: show (String& string) const
	{
		string << "hyp: " << *expression_ << endLine;
		string << "this: " << this << endLine;
		string << "up: ";
		const list :: Node :: Node_*
			nodeUp = up_.getFirst();
		while (nodeUp != NULL) {
			const Node* const
				prop = nodeUp->getValue();
			string << prop;
			if (nodeUp->getNext() != NULL) {
				string << ", ";
			} else {
				string << endLine;
			}
			nodeUp = nodeUp->getNext();
		}
		string << "down: ";
		const list :: Node :: Node_*
			nodeDown = down_.getFirst();
		while (nodeDown != NULL) {
			const Node* const
				prop = nodeDown->getValue();
			string << prop;
			if (nodeDown->getNext() != NULL) {
				string << ", ";
			} else {
				string << endLine;
			}
			nodeDown = nodeDown->getNext();
		}
		string << endLine;
	}

	Size_t
	Hyp :: getExpressionsVolume()
	{
		Size_t volume = 0;
		volume += expressions_->getVolume();
		volume += expressions_->getSizeOf();
		return volume;
	}

	/****************************
	 *		Private members
	 ****************************/

	// weight routines
	void
	Hyp :: calculateWeight() const
	{
		weight_ = 0;
		const list :: Node :: Node_*
			node = down_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			if (prop->getWeight() > weight_) {
				weight_ = prop->getWeight();
			}
			node = node->getNext();
		}
		weight_ += Config :: getProofFactor() * confirmation_.getSize();
	}

	// build up routines
	void
	Hyp :: buildUp (const Node* self, Level* const level)
	{
		if (!buildUpRefs (self)) {
			buildUpProps (self, level);
		}
		searchProof();
	}
	bool
	Hyp :: buildUpRefs (const Node* self)
	{
		bool result = true;
		Memory :: stack().push();

		Refs_ refs =
			expressions_->equifyInverse<Expression>(expression_);

		const Refs_ :: Node_*
			node = refs.getFirst();
		if (node == NULL) {
			expressions_->addBranch (expression_, self);
			result = false;
		}
		while (node != NULL) {
			buildUpRef (self, node);
			node = node->getNext();
		}
		Memory :: stack().pop();
		return result;
	}
	void
	Hyp :: buildUpProps (const Node* self, Level* const level)
	{
		const forest :: Proposition*
			propositions = Math :: assertions()->getPropositionTree();

		const index :: Assertion
			boundary = Prover :: getPropositionBoundary() - 1;

		Memory :: stack().push();
		Propositions_ result =
			propositions->unifySymmetric<Expression>
			(
				expression_,
				boundary
			);
		const Propositions_ :: Node_*
		node = result.getFirst();
		while (node != NULL) {
			buildUpProp (self, node, level, boundary);
			node = node->getNext();
		}
		Memory :: stack().pop();
	}
	void
	Hyp :: buildUpProp
	(
		const Node* self,
		const Propositions_ :: Node_* node,
		Level* const level,
		const index :: Assertion boundary
	)
	{
		typedef
			mdl :: tree :: Proposition :: Data_
			Data_;
		Memory :: stack().push();

		const stack :: Substitution*
			internal = node->getReference().getTargetSubstitution<allocator :: Stack>();
		const stack :: Substitution*
			external = node->getReference().getSourceSubstitution<allocator :: Stack>();

		const Data_&
			data = node->getReference().getSource()->getData();
		const Data_ :: Node_*
			dataNode = data.getFirst();
		while (dataNode != NULL) {
			const mdl :: Proposition*
				proposition = dataNode->getValue();
			const mdl :: Assertion*
				assertion = proposition->getAssertion();
			const bool
				fitsComplexity = Prop :: fitsComplexity
				(
					assertion,
					internal,
					external
				);
			bool isAccessible = true;
			for (index :: Arity i = 0; i < assertion->getPropArity(); ++ i) {
				const mdl :: Proposition* const
					proposition = assertion->getProposition (i);
				isAccessible &= (proposition->getSourceIndex() <= boundary);
			}

			if (fitsComplexity && assertion->isVerified() && isAccessible) {

				const Substitution* substitutionInt = NULL;
				const Substitution* substitutionExt = NULL;
				buildSubstitutions
				(
					assertion, proposition,
					internal, external,
					substitutionInt, substitutionExt
				);

				const Node*
					prop = new Prop
					(
						level,
						proposition,
						substitutionInt,
						substitutionExt,
						self
					);
				const_cast<list :: Node&>(up_).addLast (prop);
			}
			dataNode = dataNode->getNext();
		}
		Memory :: stack().pop();
	}
	inline void
	Hyp :: buildUpRef (const Node* self, const Refs_ :: Node_* node)
	{
		const Node* up = node->getValue().getSource()->getData().getValue();
		if (liesDown (up)) {
			return;
		}
		const Substitution* const
			replacement = Substitution :: create
				(node->getReference().getTarget().getSubstitution());

		Node* ref = new Ref (replacement, self, up);
		const_cast<list :: Node&>(up_).addLast (ref);
		ref->buildDown (false);
	}
	inline void
	Hyp :: buildSubstitutions
	(
		const mdl :: Assertion* const assertion,
		const mdl :: Proposition* const proposition,
		const stack :: Substitution* inter,
		const stack :: Substitution* exter,
		const Substitution*& substitutionInt,
		const Substitution*& substitutionExt
	)
	{
		Memory :: stack().push();
		stack :: Substitution*
			internal = stack :: Substitution :: create (inter);

		stack :: Substitution*
			external = stack :: Substitution :: create (exter);
		const stack :: Substitution*
			replacement = buildVariableReplacement (assertion, external);

		external = external->compose (replacement);
		internal = internal->compose (replacement, false);

		substitutionInt = Substitution :: create (internal);
		substitutionExt = Substitution :: create (external);
		Memory :: stack().pop();

		#ifdef DEBUG
		verifySubstitutionInt (substitutionInt);
		verifySubstitutionExt (proposition, substitutionExt);
		#endif
	}
	inline const stack :: Substitution*
	Hyp :: buildVariableReplacement
	(
		const mdl :: Assertion* const assertion,
		const stack :: Substitution* external
	)
	{
		const stack :: Substitution* replacement = NULL;
		const mdl :: Variables*
			variables = assertion->getVariables();
		for (index :: Arity i = 0; i < variables->getSize(); ++ i) {
			const mdl :: header :: Variable* const
				variable = variables->getVariable (i);
			const Symbol& oldSymbol = variable->getSymbol();
			if (external->find (oldSymbol) == NULL) {
				const Symbol
					newSymbol = Prover :: createVariable (oldSymbol);
				const stack :: Expression :: Term_* const
					term = stack :: Expression :: createSingleton (newSymbol);
				replacement = new stack :: Substitution (oldSymbol, term, replacement);
			}
		}
		return replacement;
	}

	// build down routines
	void
	Hyp :: collectConfirmation (const Node* self)
	{
		const list :: Node :: Node_*
			node = up_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			const Confirmation&
				confirmation = prop->getConfirmation();

			uniteConfirmation (self, confirmation);
			node = node->getNext();
		}
	}
	void
	Hyp :: followDown()
	{
		const list :: Node :: Node_*
			node = down_.getFirst();
		while (node != NULL) {
			Node* const
				prop = const_cast<Node*>(node->getValue());
			prop->buildDown();
			node = node->getNext();
		}
	}
	void
	Hyp :: unifyPremises (const Node* self)
	{
		const tree :: Premise*
			premises = Prover :: getPremises();

		Memory :: stack().push();
		Premises_
			result = premises->unifyInverse<Expression>(expression_);
		const Premises_ :: Node_*
			node = result.getFirst();
		while (node != NULL) {
			addEvidence (self, node);
			node = node->getNext();
		}
		Memory :: stack().pop();
	}
	void
	Hyp :: uniteConfirmation (const Node* self, const Confirmation& confirmation)
	{
		for (value :: Integer i = 0; i < confirmation.getDelta(); ++ i) {
			const Evidence* const
				upEvidence = confirmation.getNew (i);
			addEvidence (self, upEvidence);
		}
	}
	void
	Hyp :: addEvidence (const Node* self, const Evidence* upEvidence)
	{
		const Substitution* const
			upSubstitution = upEvidence->getSubstitution();
		Evidence* evidenceSub = NULL;
		Memory :: stack().push(); {
			stack :: Substitution*
				sub = stack :: Substitution :: create (upSubstitution);
			sub = sub->restrictTo (expression_->getTerm()->getEnd());
			evidenceSub = confirmation_.supply (sub);
		}
		Memory :: stack().pop();
		if (evidenceSub == NULL) {
			return;
		}
		evidence :: Hyp* evidenceHyp = new evidence :: Hyp
		(
			self,
			evidenceSub,
			upEvidence
		);
		dynamic_cast<evidence :: Sub*>(evidenceSub)->addUp (evidenceHyp);
	}
	void
	Hyp :: addEvidence (const Node* self, const Premises_ :: Node_* node)
	{
		typedef
			tree :: Premise :: Data_
			Data_;
		const Data_&
			data = node->getReference().getSource()->getData();
		const tree :: Premise :: Substitution_* const
			premiseSubstitution = node->getReference().getTarget().getSubstitution();

		Evidence* evidenceSub = NULL;
		Memory :: stack().push(); {
			const stack :: Substitution*
				sub = stack :: Substitution :: create (premiseSubstitution);
			evidenceSub = confirmation_.supply (sub);
		}
		Memory :: stack().pop();
		if (evidenceSub == NULL) {
			return;
		}
		const Premise*
			premise = data.getValue();
		const Evidence* evidenceTop = new evidence :: Top (self, evidenceSub, premise);
		dynamic_cast<evidence :: Sub*>(evidenceSub)->addUp (evidenceTop);
	}

	// cloning routines
	inline void
	Hyp :: cloneDown (Hyp* const clone ) const
	{
		/*clone->down_=
			(down_ == NULL) ?
			NULL :
			down_->clone();*/
	}
	inline void
	Hyp :: cloneMaxRate (Hyp* const clone) const
	{
		const Node*
			maxProp = NULL;
		value :: Rate
			maxRate = value :: undefined :: RATE;

		const list :: Node :: Node_*
			node = up_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			const value :: Rate
				rate = prop->getRate();
			if (rate < maxRate) {
				maxRate = rate;
				maxProp = prop;
			}
			node = node->getNext();
		}
		const Node* const
			upClone = maxProp->clone();
		const_cast<list :: Node&>(clone->up_).addLast (upClone);
	}
	inline void
	Hyp :: cloneProved (Hyp* const clone) const
	{
		const list :: Node :: Node_*
			node = up_.getFirst();
		while (node != NULL) {
			const Node* const
				prop = node->getValue();
			if (prop->isProved()) {
				const Node* const
					upClone = prop->clone();
				const_cast<list :: Node&>(clone->up_).addLast (upClone);
			}
			node = node->getNext();
		}
	}
	inline void
	Hyp :: cloneConfirmation (Hyp* const clone) const
	{
		for (value :: Integer i = 0; i < confirmation_.getSize(); ++ i) {
			const Evidence* const
				proof = confirmation_.get (i);
			clone->confirmation_.add (proof->clone());
		}
	}

	void
	Hyp :: verifySubstitutionInt (const Substitution* substitutionInt) const
	{
		substitutionInt->verifyIntegrity();
		while (substitutionInt != NULL) {
			const Symbol& variable = substitutionInt->getVariable();
			if (!getExpression()->getTerm()->contains(variable)) {
				expression :: Mode mode;
				mode << expression :: Mode :: SHOW_REPLACEABILITY;
				Error* error = new Error (Error :: SEMANTIC);
				error->message() << mode  << "dumb variable in substitutionInt:" << endLine;
				error->message() << tab << "substitutionInt = " << *substitutionInt;
				error->message() << tab << "variable = " << variable << endLine;
				error->message() << endLine;
				error->message() << endLine;
				show (error->message());
				std :: cerr << error->message() << std :: endl;
				throw error;
			}
			substitutionInt = substitutionInt->getNext();
		}
	}
	void
	Hyp :: verifySubstitutionExt
	(
		const mdl :: Proposition* const proposition,
		const Substitution* substitutionExt
	) const
	{
		substitutionExt->verifyIntegrity();
		while (substitutionExt != NULL) {
			const Symbol& variable = substitutionExt->getVariable();
			verifySubstitutionExt (variable, proposition, substitutionExt);
			substitutionExt = substitutionExt->getNext();
		}
	}
	void
	Hyp :: verifySubstitutionExt
	(
		const Symbol& variable,
		const mdl :: Proposition* const proposition,
		const Substitution* substitutionExt
	) const
	{
		if (proposition->getExpression()->getTerm()->contains(variable)) {
			return;
		}
		const mdl :: Assertion* const
			assertion = proposition->getAssertion();
		for (index :: Arity i = 0; i < assertion->getHypArity(); ++ i) {
			const mdl :: Hypothesis* const
				hyp = assertion->getHypothesis (i);
			if (hyp->getExpression()->getTerm()->contains(variable)) {
				return;
			}
		}
		expression :: Mode mode;
		mode << expression :: Mode :: SHOW_REPLACEABILITY;
		Error* error = new Error (Error :: SEMANTIC);
		error->message() << mode  << "dumb variable in substitutionExt" << endLine;
		error->message() << tab << "substitutionExt = " << *substitutionExt;
		error->message() << tab << "variable = " << variable << endLine;
		error->message() << tab << "assertion = " << endLine;
		assertion->show (error->message());
		error->message() << endLine;
		error->message() << endLine;
		show (error->message());
		std :: cerr << error->message() << std :: endl;
		throw error;
	}

	/************************************
	 *		Static private attributes
	 ************************************/

	tree :: Node*
		Hyp :: expressions_ = NULL;
}
}
}

#endif /*MDL_PROVER_NODE_HYP_CPP_*/
