/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Hyp.hpp                                  */
/* Description:     hypothesis node in proof variant tree                    */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_HYP_HPP_
#define MDL_PROVER_NODE_HYP_HPP_

#include "mdl/prover/mdl_prover.dpp"
#include "mdl/prover/evidence/mdl_prover_evidence_Vector.hpp"

namespace mdl {
namespace prover {
namespace node {

class Hyp :
	public Node,
	public Scalar<Hyp, allocator :: Prover> {
public :
	Hyp
	(
		const Expression* const,
		const Node* const,
		const bool = false
	);
	virtual ~ Hyp();

	// prover :: Node interface
	virtual const Expression* getExpression() const;
	virtual const mdl :: Proposition* getProposition() const;
	virtual const Confirmation& getConfirmation() const;
	virtual value :: Weight getWeight() const;
	virtual void calculateWeightUp() const;
	virtual void calculateWeightDown() const;
	virtual bool operator < (const Node&) const;
	virtual value :: Rate getRate() const;
	virtual void setRate (const value :: Rate) const;
	virtual void searchProof();
	virtual bool isProved() const;
	virtual void buildUp (Level* const);
	virtual void buildDown (const bool = true);
	virtual bool isRoot() const;
	virtual void addDown (const Node*);
	virtual bool liesDown (const Node*) const;
	virtual void showNode
	(
		String&,
		const value :: Integer = 0
	) const;
	virtual void showTree
	(
		String&,
		const value :: Integer = 0,
		const value :: Integer = value :: undefined :: INTEGER
	) const;

	// prover :: Statistics interface
	virtual Size_t getEvidenceVolume() const;
	virtual Size_t getTreeVolume() const;
	virtual Size_t getExpressionVolume() const;
	virtual Size_t getTreeNodeVolume() const;
	virtual Size_t getEvidenceNodeVolume() const;
	virtual Size_t getEvidenceContainerVolume() const;
	virtual Size_t getGarbageVolume() const;

	virtual Size_t getSubstitutionAllVolume() const;
	virtual Size_t getSubstitutionIntVolume() const;
	virtual Size_t getSubstitutionExtVolume() const;
	virtual Size_t getSubstitutionVolume() const;

	virtual value :: Integer getTreeNodeCount() const;
	virtual value :: Integer getEvidenceNodeCount() const;
	virtual value :: Integer getExpressionCount() const;
	virtual value :: Integer getHypCount() const;
	virtual value :: Integer getPropCount() const;
	virtual value :: Integer getRefCount() const;

	// object :: Cloneable interface
	virtual const Node* clone() const;

	// object :: Object interface
	virtual void commitSuicide();
	virtual Size_t getVolume() const;
	virtual Size_t getSizeOf() const;
	virtual void show (String&) const;

	static Size_t getExpressionsVolume();

private :
	typedef
		forest :: Proposition :: Unify_
		<
			Expression,
			unify :: policy :: SYMMETRIC
		> :: Result_
		Propositions_;
	typedef
		tree :: Premise :: Unify_
		<
			Expression,
			unify :: policy :: INVERSE
		> :: Result_
		Premises_;
	typedef
		tree :: Node :: Unify_
		<
			Expression,
			unify :: policy :: INVERSE,
			unify :: kind :: PLAIN,
			allocator :: Stack
		> :: Result_
		Refs_;

	// weight routines
	void calculateWeight() const;

	// build up routines
	void buildUp (const Node*, Level* const);
	bool buildUpRefs (const Node*);
	void buildUpProps (const Node*, Level* const);
	void buildUpRef (const Node*, const Refs_ :: Node_*);
	void buildUpProp
	(
		const Node*,
		const Propositions_ :: Node_*,
		Level* const,
		const index :: Assertion
	);
	void buildSubstitutions
	(
		const mdl :: Assertion* const,
		const mdl :: Proposition* const,
		const stack :: Substitution*,
		const stack :: Substitution*,
		const Substitution*&,
		const Substitution*&
	);
	const stack :: Substitution*
	buildVariableReplacement
	(
		const mdl :: Assertion* const,
		const stack :: Substitution*
	);

	// build down routines
	void collectConfirmation (const Node*);
	void followDown();
	void unifyPremises (const Node*);
	void uniteConfirmation (const Node*, const Confirmation&);
	void addEvidence (const Node* self, const Evidence*);
	void addEvidence (const Node*, const Premises_ :: Node_*);

	// cloning routines
	void cloneDown (Hyp* const) const;
	void cloneMaxRate (Hyp* const) const;
	void cloneProved (Hyp* const) const;
	void cloneConfirmation (Hyp* const) const;

	void verifySubstitutionInt (const Substitution*) const;
	void verifySubstitutionExt
	(
		const mdl :: Proposition* const,
		const Substitution*
	) const;
	void verifySubstitutionExt
	(
		const Symbol& variable,
		const mdl :: Proposition* const proposition,
		const Substitution* substitutionExt
	) const;

	friend class Prop;
	friend class Root;
	friend class prover :: Prover;

	class Statistics;
	friend class Statistics;

	// base
	const Expression* const expression_;

	// variant tree
	const list :: Node up_;
	const list :: Node down_;

	// clone
	mutable const Hyp* clone_;

	// weight & rate
	mutable value :: Rate   rate_;
	mutable value :: Weight weight_;

	// all proof evidences (confirmation)
	evidence :: Vector  confirmation_;

	// tree of all expressions in the variant tree
	static tree :: Node* expressions_;
};

}
}
}

#endif /*MDL_PROVER_NODE_HYP_HPP_*/
