/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Hyp_Statistics.cpp                       */
/* Description:     statistics for hypothesis node in proof variant tree     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_HYP_STATISTICS_CPP_
#define MDL_PROVER_NODE_HYP_STATISTICS_CPP_

#include "mdl/prover/node/mdl_prover_node.hpp"

namespace mdl {
namespace prover {
namespace node {

	/****************************
	 *		Public members
	 ****************************/

	// prover :: Statistics implementation
	Size_t
	Hyp :: Statistics :: getEvidenceVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += node->confirmation_.getVolume();
		volume += node->confirmation_.getEvidenceVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getEvidenceVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getTreeVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += getAggregateVolume (node->expression_);
		volume += node->up_.getVolume();
		volume += node->down_.getVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getTreeVolume();
			volume += prop->getSizeOf();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getExpressionVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += getAggregateVolume (node->expression_);
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getExpressionVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getTreeNodeVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += node->up_.getVolume();
		volume += node->down_.getVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getTreeNodeVolume();
			volume += prop->getSizeOf();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getEvidenceNodeVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += node->confirmation_.getEvidenceNodeVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getEvidenceNodeVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getEvidenceContainerVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += node->confirmation_.getVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getEvidenceContainerVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getGarbageVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += node->confirmation_.getGarbageVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getGarbageVolume();
			n = n->getNext();
		}
		return volume;
	}

	// substitution statistics
	Size_t
	Hyp :: Statistics :: getSubstitutionAllVolume (const Hyp* node)
	{
		Size_t volume = 0;
		volume += node->confirmation_.getSubstitutionVolume();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getSubstitutionAllVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getSubstitutionIntVolume (const Hyp* node)
	{
		Size_t volume = 0;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getSubstitutionIntVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getSubstitutionExtVolume (const Hyp* node)
	{
		Size_t volume = 0;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getSubstitutionExtVolume();
			n = n->getNext();
		}
		return volume;
	}
	Size_t
	Hyp :: Statistics :: getSubstitutionVolume (const Hyp* node)
	{
		Size_t volume = 0;
		for (value :: Integer i = 0; i < node->confirmation_.getSize(); ++ i) {
			const Substitution* const
				substitution = node->confirmation_.get (i)->getSubstitution();
			volume += getAggregateVolume (substitution);
		}
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			volume += prop->getSubstitutionVolume();
			n = n->getNext();
		}
		return volume;
	}

	// node count statistics
	value :: Integer
	Hyp :: Statistics :: getTreeNodeCount (const Hyp* node)
	{
		Size_t count = 1;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			count += prop->getTreeNodeCount();
			n = n->getNext();
		}
		return count;
	}
	value :: Integer
	Hyp :: Statistics :: getEvidenceNodeCount (const Hyp* node)
	{
		Size_t count = 0;
		count += node->confirmation_.getSize();
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			count += prop->getEvidenceNodeCount();
			n = n->getNext();
		}
		return count;
	}
	value :: Integer
	Hyp :: Statistics :: getExpressionCount (const Hyp* node)
	{
		Size_t count = 1;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			count += prop->getExpressionCount();
			n = n->getNext();
		}
		return count;
	}
	value :: Integer
	Hyp :: Statistics :: getHypCount (const Hyp* node)
	{
		Size_t count = 1;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			count += prop->getHypCount();
			n = n->getNext();
		}
		return count;
	}
	value :: Integer
	Hyp :: Statistics :: getPropCount (const Hyp* node)
	{
		Size_t count = 0;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			count += prop->getPropCount();
			n = n->getNext();
		}
		return count;
	}
	value :: Integer
	Hyp :: Statistics :: getRefCount (const Hyp* node)
	{
		Size_t count = 0;
		const list :: Node :: Node_*
			n = node->up_.getFirst();
		while (n != NULL) {
			const Node* const
				prop = n->getValue();
			count += prop->getRefCount();
			n = n->getNext();
		}
		return count;
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class T>
	Size_t
	Hyp :: Statistics :: getAggregateVolume (const T* object)
	{
		Size_t volume = 0;
		if (object != NULL) {
			volume += object->getVolume();
			volume += object->getSizeOf();
		}
		return volume;
	}
}
}
}

#endif /*MDL_PROVER_NODE_HYP_STATISTICS_CPP_*/
