/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Prop.cpp                                 */
/* Description:     proposition node in proof variant tree                   */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_PROP_CPP_
#define MDL_PROVER_NODE_PROP_CPP_

namespace mdl {
namespace prover {
namespace node {

	/****************************
	 *		Public members
	 ****************************/

	Prop :: Prop
	(
		Level* const level,
		const mdl :: Proposition* const proposition,
		const Substitution* const substitutionInt,
		const Substitution* const substitutionExt,
		const Node* const down
	) :
	proposition_ (proposition),
	substitutionInt_ (substitutionInt),
	substitutionExt_ (substitutionExt),
	up_ (proposition->getAssertion()->getHypArity()),
	down_ (down),
	clone_ (NULL),
	rate_ (value :: undefined :: RATE),
	weight_ (0),
	fitness_ (0),
	confirmation_ (this)
	#ifdef DEBUG_BACK
	, confirmationOther_ (this)
	#endif
	{
		Prover :: refresh ("building up prop node");
		#ifdef DEBUG
		verifySubstitutions();
		#endif
		calculateWeight();
		buildUp (level);
	}
	Prop :: ~ Prop() {
	}

	inline bool
	Prop :: fitsComplexity
	(
		const mdl :: Assertion* const assertion,
		const stack :: Substitution* const internal,
		const stack :: Substitution* const external
	)
	{
		for (index :: Arity i = 0; i < assertion->getHypArity(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis = assertion->getHypothesis (i);

			Memory :: stack().push();
			stack :: Expression* const
				expression = new stack :: Expression (hypothesis->getExpression());
			expression->execute (external);
			expression->execute (internal);
			const bool
				fits = Prover :: fitsExpressionComplexity (expression->getComplexity());
			Memory :: stack().pop();
			if (!fits) {
				return false;
			}
		}
		return true;
	}

	// prover :: Node implementation
	const Expression*
	Prop :: getExpression() const {
		return down_->getExpression();
	}
	const mdl :: Proposition*
	Prop :: getProposition() const {
		return proposition_;
	}
	const Confirmation&
	Prop :: getConfirmation() const {
		return confirmation_;
	}
	value :: Weight
	Prop :: getWeight() const {
		return weight_;
	}
	void
	Prop :: calculateWeightUp() const
	{
		calculateWeight();
		for (value :: Integer i = 0; i < up_.getSize(); ++ i) {
			up_.getValue (i)->calculateWeightUp();
		}
	}
	void
	Prop :: calculateWeightDown() const
	{
		down_->calculateWeightDown();
		calculateWeight();
	}
	bool
	Prop :: operator < (const Node& node) const
	{
		if (rate_ == value :: undefined :: RATE) {
			return (weight_ > node.getWeight());
		} else {
			return (rate_ < node.getRate());
		}
	}
	value :: Rate
	Prop :: getRate() const {
		return rate_;
	}
	void
	Prop :: setRate (const value :: Rate rate) const
	{
		if (rate_ <= rate) {
			return;
		}
		rate_ = rate;
		down_->setRate (rate);
	}
	void
	Prop :: searchProof()
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			Node* const
				node = const_cast<Node*>(up_.getValue (i));
			node->searchProof();
		}
		buildDown();
	}
	inline bool
	Prop :: isProved() const {
		return (confirmation_.getSize() > 0);
	}
	void
	Prop :: buildUp (Level* const level)
	{
		const mdl :: Assertion*
			assertion = proposition_->getAssertion();
		for (index :: Arity i = 0; i < assertion->getHypArity(); ++ i) {
			const mdl :: Hypothesis* const
				hypothesis = assertion->getHypothesis (i);

			Memory :: stack().push();
			stack :: Expression* const
				exp = new stack :: Expression (hypothesis->getExpression());
			exp->execute (substitutionExt_);
			exp->execute (substitutionInt_);
			const Expression* const
				expression = new Expression (exp);
			Memory :: stack().pop();

			const Node* const
				hyp = new Hyp (expression, this);
			level->add (hyp);
			const_cast<vector :: Node&>(up_).add (hyp);
		}
		#ifdef DEBUG
		verifySubstitutions();
		#endif
	}
	void
	Prop :: buildDown (const bool)
	{
		Prover :: refresh ("building down prop node");
		if (!checkUp()) {
			return;
		}
		unifyMatrix();
		if (confirmation_.boundaryMoved()) {
			calculateWeightUp();
			const_cast<Node*>(down_)->buildDown();
			confirmation_.markBoundary();
			#ifdef DEBUG_BACK
			confirmationOther_.markBoundary();
			#endif
		}
	}
	bool
	Prop :: isRoot() const {
		return false;
	}
	void
	Prop :: addDown (const Node*) {
	}
	bool
	Prop :: liesDown (const Node* node) const
	{
		if (node == this) {
			return true;
		}
		return down_->liesDown (node);
	}

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;
using manipulator :: iterate;

	void
	Prop :: showNode
	(
		String& string,
		const value :: Integer depth
	) const
	{
		string << iterate << depth << tab << "prop: ";
		proposition_->getAssertion()->replicateName(string);
		if (proposition_->getAssertion()->getPropArity() > 1) {
			string << ":" << proposition_->getIndex();
		}
		string << endLine;
		string << iterate << depth << tab << "this: " << this << endLine;
		string << iterate << depth << tab << "substitution int: " << *substitutionInt_ << endLine;
		string << iterate << depth << tab << "substitution ext: " << *substitutionExt_ << endLine;
		if (isProved()) {
			confirmation_.show (string);
		}
		//string << iterate << depth << tab << "down: " << down_ << endLine;
		string << endLine;
	}
	void
	Prop :: showTree
	(
		String& string,
		const value :: Integer depthStart,
		const value :: Integer depthEnd
	) const
	{
		if (depthEnd > 0) {
			string << iterate << depthStart << tab;
			string << iterate << 64 << '-' << endLine;
		}
		showNode (string, depthStart);

		if (depthEnd == 0) {
			return;
		}
		//confirmation_.show (string);
		/*if (up_.isEmpty()) {
			string << iterate << depthStart << tab;
			string << "up: empty" << endLine;
		}*/
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Node* const
				hyp = up_.getValue (i);
			hyp->showTree (string, depthStart + 1, depthEnd - 1);
		}
		if (depthEnd > 0) {
			string << iterate << depthStart << tab;
			string << iterate << 64 << '-' << endLine;
		}
	}

	// prover :: Statistics implementation
	Size_t
	Prop :: getEvidenceVolume() const {
		return Statistics :: getEvidenceVolume (this);
	}
	Size_t
	Prop :: getTreeVolume() const {
		return Statistics :: getTreeVolume (this);
	}
	Size_t
	Prop :: getExpressionVolume() const {
		return Statistics :: getExpressionVolume (this);
	}
	Size_t
	Prop :: getTreeNodeVolume() const {
		return Statistics :: getTreeNodeVolume (this);
	}
	Size_t
	Prop :: getEvidenceNodeVolume() const {
		return Statistics :: getEvidenceNodeVolume (this);
	}
	Size_t
	Prop :: getEvidenceContainerVolume() const {
		return Statistics :: getEvidenceContainerVolume (this);
	}
	Size_t
	Prop :: getGarbageVolume() const {
		return Statistics :: getGarbageVolume (this);
	}

	// substitution statistics
	Size_t
	Prop :: getSubstitutionAllVolume() const {
		return Statistics :: getSubstitutionAllVolume (this);
	}
	Size_t
	Prop :: getSubstitutionIntVolume() const {
		return Statistics :: getSubstitutionIntVolume (this);
	}
	Size_t
	Prop :: getSubstitutionExtVolume() const {
		return Statistics :: getSubstitutionExtVolume (this);
	}
	Size_t
	Prop :: getSubstitutionVolume() const {
		return Statistics :: getSubstitutionVolume (this);
	}

	// node count statistics
	value :: Integer
	Prop :: getTreeNodeCount() const {
		return Statistics :: getTreeNodeCount (this);
	}
	value :: Integer
	Prop :: getEvidenceNodeCount() const {
		return Statistics :: getEvidenceNodeCount (this);
	}
	value :: Integer
	Prop :: getExpressionCount() const {
		return Statistics :: getExpressionCount (this);
	}
	value :: Integer
	Prop :: getHypCount() const {
		return Statistics :: getHypCount (this);
	}
	value :: Integer
	Prop :: getPropCount() const {
		return Statistics :: getPropCount (this);
	}
	value :: Integer
	Prop :: getRefCount() const {
		return Statistics :: getRefCount (this);
	}

	// object :: Cloneable implementation
	const Node*
	Prop :: clone () const
	{
		if (clone_ != NULL) {
			return clone_;
		}
		Prop* clone = new Prop (*this);
		clone_ = clone;
		cloneUp (clone);
		cloneConfirmation (clone);
		return clone_;
	}

	// object :: Object implementation
	void
	Prop :: commitSuicide() {
		delete this;
	}
	Size_t
	Prop :: getVolume() const
	{
		Size_t volume = 0;
		volume += getTreeVolume();
		volume += getEvidenceVolume();
		return volume;
	}
	Size_t
	Prop :: getSizeOf() const {
		return sizeof (Prop);
	}
	void
	Prop :: show (String& string) const
	{
		string << "prop: ";
		proposition_->getAssertion()->replicateName(string);
		if (proposition_->getAssertion()->getPropArity() > 1) {
			string << ":" << proposition_->getIndex();
		}
		string << endLine;
		string << "proposition: " << endLine;
		string << tab << "prop 0 = " << *getExpression() << endLine;
		string << "hypothesis: " << endLine;
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Node* const
				hyp = up_.getValue (i);
			string << tab << "hyp " << i << " = " << *(hyp->getExpression()) << endLine;
		}
		string << endLine;
		string << "substitution int: " << *substitutionInt_ << endLine;
		string << "substitution ext: " << *substitutionExt_ << endLine;
		string << "this: " << this << endLine;
		//string << "down: " << down_ << endLine;
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	// weight routines
	void
	Prop :: calculateWeight() const
	{
		weight_ = down_->getWeight();
		weight_ += Config :: getFitnessFactor() * calculateFitness();
		weight_ += Config :: getProvedFactor() * calculateProved();
	}
	value :: Weight
	Prop :: calculateFitness() const
	{
		if (fitness_ != value :: undefined :: WEIGHT) {
			return fitness_;
		}
		fitness_ = 0;
			/*proposition_->getExperience().evaluateFitness
			(
				Prover :: getPremises(),
				Prover :: getStepBoundary()
			);*/
		return fitness_;
	}
	value :: Weight
	Prop :: calculateProved() const
	{
		const index :: Arity
			arity = proposition_->getAssertion()->getHypArity();
		if (arity == 0) {
			return 1;
		}
		index :: Arity
			proved = 0;
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Node* const
				hyp = up_.getValue (i);
			if (hyp->isProved()) {
				++ proved;
			}
		}
		return proved / arity;
	}

	// build down routines
	inline bool
	Prop :: checkUp() const
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			if (!up_.getValue (i)->isProved()) {
				return false;
			}
		}
		return true;
	}
	inline void
	Prop :: unifyMatrix()
	{
		if (key :: TREE_BACK_UNIFICATION) {
			Tree :: unify (this, confirmation_);
		} else {
			Prod :: unify (this, confirmation_);
		}
		Prover :: refresh ("matrix unification");
		#ifdef DEBUG_BACK
		if (key :: TREE_BACK_UNIFICATION) {
			Prod :: unify (this, confirmationOther_);
		} else {
			Tree :: unify (this, confirmationOther_);
		}
		if (!confirmation_.isEqual (&confirmationOther_)) {
			Error* error = new Error (Error :: SEMANTIC);
			error->message() << "incorrect tree matrix unification." << endLine;
			error->message() << "----------------------------------" << endLine;
			error->message() << "hypotheses proof substitutions:" << endLine;
			showHypConfirmations (error->message());
			error->message() << endLine;
			if (key :: TREE_BACK_UNIFICATION) {
				error->message() << "product matrix unification result:" << endLine;
			} else {
				error->message() << "tree matrix unification result:" << endLine;
			}
			confirmationOther_.showSubstitutions (error->message(), true);
			error->message() << endLine;
			if (key :: TREE_BACK_UNIFICATION) {
				error->message() << "tree matrix unification result:" << endLine;
			} else {
				error->message() << "product matrix unification result:" << endLine;
			}
			confirmation_.showSubstitutions (error->message(), true);
			error->message() << endLine;
			confirmationOther_.showDifference (error->message(), &confirmation_);
			std :: cerr << error->message() << std :: endl;
			confirmation_.isEqual (&confirmationOther_);
			Counter :: set (0, 1);
			confirmationOther_.clear();
			confirmation_.clear();

			if (key :: TREE_BACK_UNIFICATION) {
				Tree :: unify (this, confirmation_);
				Prod :: unify (this, confirmationOther_);
			} else {
				Prod :: unify (this, confirmation_);
				Tree :: unify (this, confirmationOther_);
			}
			throw error;
		}
		#endif
	}
	inline value :: Integer
	Prop :: getMatrixSearchVolume() const {
		return Prod :: getMatrixSearchVolume (this);
	}

	// cloning routines
	inline void
	Prop :: cloneUp (Prop* const clone) const
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Node* const
				upClone = up_.getValue (i)->clone();
			const_cast<vector :: Node&>(clone->up_).add (upClone);
		}

		for (value :: Integer i = 0; i < confirmation_.getSize(); ++ i) {
			const Evidence* const
				proof = confirmation_.get (i);
			clone->confirmation_.add (proof->clone());
		}
	}
	inline void
	Prop :: cloneConfirmation (Prop* const clone) const
	{
		for (value :: Integer i = 0; i < confirmation_.getSize(); ++ i) {
			const Evidence* const
				evidence = confirmation_.get (i);
			clone->confirmation_.add (evidence->clone());
		}
	}

	void
	Prop :: showHypConfirmations (String& string) const
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Node* const
				hyp = up_.getValue (i);
			const Confirmation&
				confirmation = hyp->getConfirmation();
			string << "hyp " << i << " :" << endLine;
			dynamic_cast<const evidence :: Vector&>(confirmation).showSubstitutions (string);
			string << endLine;
		}
	}
	void
	Prop :: verifySubstitutions() const
	{
		substitutionInt_->verifyIntegrity();
		substitutionExt_->verifyIntegrity();
		verifyVariables (substitutionInt_);
	}
	void
	Prop :: verifyVariables (const Substitution* substitution) const
	{
		while (substitution != NULL) {
			const Symbol& variable = substitution->getVariable();
			verifyVariable (variable, substitution);
			substitution = substitution->getNext();
		}
	}
	void
	Prop :: verifyVariable
	(
		const Symbol& variable,
		const Substitution* substitution
	) const
	{
		if (getExpression()->getTerm()->contains(variable)) {
			return;
		}
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Node* const
				hyp = up_.getValue (i);
			if (hyp->getExpression()->getTerm()->contains(variable)) {
				return;
			}
		}
		expression :: Mode mode;
		mode << expression :: Mode :: SHOW_REPLACEABILITY;
		Error* error = new Error (Error :: SEMANTIC);
		error->message() << mode  << "dumb variable in substitution" << endLine;
		error->message() << tab << "substitution = " << *substitution;
		error->message() << tab << "variable = " << variable << endLine;
		error->message() << endLine;
		error->message() << endLine;
		show (error->message());
		std :: cerr << error->message() << std :: endl;
		throw error;
	}
}
}
}

#endif /*MDL_PROVER_NODE_PROP_CPP_*/
