/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Prop_Tree.cpp                            */
/* Description:     matrix tree unification                                  */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_PROP_TREE_CPP_
#define MDL_PROVER_NODE_PROP_TREE_CPP_

namespace mdl {
namespace prover {
namespace node {

	/****************************
	 *		Public members
	 ****************************/

	void
	Prop :: Tree :: unify (const Prop* prop, evidence :: Vector& proofs)
	{
		Memory :: stack().push();
		const index :: Arity
			dimension = prop->up_.getSize();

		vector :: Symbol variables;
		matrix :: Cell   matrix;

		createVariables (prop, variables);
		const value :: Integer
			boundary = calculateBoundary (prop, variables);
		if (!createMatrix (prop, variables, matrix)) {
			Memory :: stack().pop();
			return;
		}

		if (matrix.getSize() == 0) {
			unifyTrivialMatrix (prop, proofs, dimension);
		} else {
			if (Counter :: stop (1)) {
				showMatrix (variables, matrix);
			}
			substitution :: tree :: Unify<Cell, evidence :: Prop>
				unify
				(
					dimension,
					prop->substitutionInt_->castAway<allocator :: Stack>(),
					proofs,
					boundary
				);
			unify (matrix);
		}
		Memory :: stack().pop();
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Prop :: Tree :: unifyTrivialMatrix
	(
		const Prop* prop,
		evidence :: Vector& confirmation,
		const index :: Arity dimension
	)
	{
		for (index :: Arity d = 0; d < dimension; ++ d) {
			const Node* const up = prop->up_.getValue (d);
			const Confirmation& confirmation = up->getConfirmation();
			const Evidence* upEvidence = confirmation.get (0);
			if (!Prover :: fitsDepth (upEvidence->getDepth())) {
				return;
			}
		}
		Evidence* evidenceSub = confirmation.supply
		(
			prop->substitutionInt_->castAway<allocator :: Stack>()
		);
		if (evidenceSub == NULL) {
			return;
		}
		evidence :: Prop* evidenceProp = new evidence :: Prop (prop, evidenceSub, dimension);
		for (index :: Arity i = 0; i < dimension; ++ i) {
			const Node* const up = prop->up_.getValue (i);
			const Confirmation& confirmation = up->getConfirmation();
			const Evidence* upEvidence = confirmation.get (0);
			evidenceProp->addUp (i, upEvidence);
		}
		dynamic_cast<evidence :: Sub*>(evidenceSub)->addUp (evidenceProp);
	}
	inline void
	Prop :: Tree :: createVariables (const Prop* prop, vector :: Symbol& variables)
	{
		for (index :: Arity i = 0; i < prop->up_.getSize(); ++ i) {
			const Node* const
				hyp = prop->up_.getValue (i);
			const Confirmation&
				confirmation = hyp->getConfirmation();
			for (value :: Integer j = 0; j < confirmation.getSize(); ++ j) {
				const Evidence* const
					evidence = confirmation.get (j);
				const Substitution*
					substitution = evidence->getSubstitution();
				while (substitution != NULL) {
					const Symbol& variable = substitution->getVariable();
					if (!variables.contains (variable)) {
						variables.add (variable);
					}
					substitution = substitution->getNext();
				}
			}
		}
	}
	inline bool
	Prop :: Tree :: createMatrix
	(
		const Prop* prop,
		const vector :: Symbol& variables,
		matrix :: Cell& matrix
	)
	{
		for (index :: Arity i = 0; i < variables.getSize(); ++ i) {
			const Symbol& variable = variables.getReference (i);
			vector :: Cell* column = new vector :: Cell (prop->up_.getSize());
			matrix.add (column);
			for (index :: Arity j = 0; j < prop->up_.getSize(); ++ j) {
				Cell* cell = new Cell();
				column->add (cell);
				if (!fillCell (prop, i, j, variable, cell)) {
					return false;
				}
			}
		}
		return true;
	}
	inline value :: Integer
	Prop :: Tree :: calculateBoundary (const Prop* prop, const vector :: Symbol& variables)
	{
		value :: Integer result = 0;
		result += variables.getSize();
		/*const Substitution* substitution = prop->substitutionInt_;
		while (substitution != NULL) {
			const Symbol& variable = substitution->getVariable();
			if (!variables.contains (variable)) {
				++ result;
			}
			substitution = substitution->getNext();
		}*/
		return result;
	}
	inline bool
	Prop :: Tree :: fillCell
	(
		const Prop* prop,
		const index :: Arity i, // variable (column)
		const index :: Arity j, // hyp (row)
		const Symbol& variable, Cell* cell
	)
	{
		const Node* const
			hyp = prop->up_.getValue (j);
		const Confirmation&
			confirmation = hyp->getConfirmation();
		for (value :: Integer j = 0; j < confirmation.getSize(); ++ j) {
			const Evidence* const
				evidence = confirmation.get (j);
			const bool
				isNew = !(j < confirmation.getBoundary());
			const bool
				fitsDepth = Prover :: fitsDepth (evidence->getDepth());
			const bool
				fitsComplexity = Prover :: fitsSubstitutionComplexity (evidence->getSubstitution()->getComplexity());

			if (!fitsDepth || !fitsComplexity) {
				continue;
			}
			const DotEvidence
				dotEvidence (evidence, isNew);

			const Substitution*
				substitution = evidence->getSubstitution();
			while (substitution != NULL) {
				const Symbol& var = substitution->getVariable();
				if (var == variable) {
					const Term* const term = substitution->getTerm();
					cell->addBranch (term, dotEvidence, true);
					break;
				}
				substitution = substitution->getNext();
			}
			if (substitution == NULL) {
				stack :: Term :: Node_*
					node = new stack :: Term :: Node_ (Symbol :: getLambda());
				stack :: Term :: Iterator_
					begin (node);
				const stack :: Term* term_1 = new stack :: Term (begin);
				term_1->addSelf (true);

				Cell :: Iterator_ end = cell->addBranch
				(
					term_1->castAway<allocator :: Prover>(),
					dotEvidence,
					true
				);
				++ end;
				end->setLambda();
				end->data().addValue (dotEvidence);
				const Cell :: Term_* term_2 = new Cell :: Term_ (end);
				term_2->addSelf (true);
			}
		}
		return !cell->isEmpty();
	}
	inline void
	Prop :: Tree :: showMatrix
	(
		const vector :: Symbol& variables,
		const matrix :: Cell& matrix
	)
	{
		for (index :: Arity i = 0; i < matrix.getSize(); ++ i) {
			const Symbol& variable = variables.getReference (i);
			const vector :: Cell* column = matrix.getValue (i);
			std :: cerr << "variable " << variable << " : " << std :: endl;
			std :: cerr << "-----------------------------" << std :: endl;
			for (index :: Arity j = 0; j < column->getSize(); ++ j) {
				std :: cerr << "hypothesis " << j << " : " << std :: endl;
				std :: cerr << "********************" << std :: endl;
				const Cell* cell = column->getValue (j);
				std :: cerr << *cell << std :: endl;
				std :: cerr << "********************" << std :: endl;
				std :: cerr << std :: endl;
				std :: cerr << std :: endl;
			}
			std :: cerr << std :: endl;
			std :: cerr << std :: endl;
		}
	}
}
}
}

#endif /*MDL_PROVER_NODE_PROP_TREE_CPP_*/
