/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Ref.cpp                                  */
/* Description:     reference node in proof variant tree                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_REF_CPP_
#define MDL_PROVER_NODE_REF_CPP_

#include "mdl/prover/node/mdl_prover_node.hpp"

namespace mdl {
namespace prover {
namespace node {

	/****************************
	 *		Public members
	 ****************************/

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;
using manipulator :: iterate;

	Ref :: Ref
	(
		const Substitution* const substitution,
		const Node* const down,
		const Node* const up
	) :
	replacement_ (substitution),
	up_ (up),
	down_ (down),
	clone_ (NULL),
	rate_ (value :: undefined :: RATE),
	weight_ (0),
	confirmation_ (this)
	{
		Prover :: refresh ("building up ref node");
		const_cast<Node*>(up_)->addDown (this);
	}
	Ref :: ~ Ref(){
	}

	// prover :: Node implementation
	const Expression*
	Ref :: getExpression() const {
		return down_->getExpression();
	}
	const mdl :: Proposition*
	Ref :: getProposition() const {
		return NULL;
	}
	const Confirmation&
	Ref :: getConfirmation() const {
		return confirmation_;
	}
	value :: Weight
	Ref :: getWeight() const {
		return weight_;
	}
	void
	Ref :: calculateWeightUp() const
	{
		calculateWeight();
		//up_->calculateWeightUp();
	}
	void
	Ref :: calculateWeightDown() const
	{
		down_->calculateWeightDown();
		calculateWeight();
	}
	bool
	Ref :: operator < (const Node& node) const
	{
		if (rate_ == value :: undefined :: RATE) {
			return (weight_ > node.getWeight());
		} else {
			return (rate_ < node.getRate());
		}
	}
	value :: Rate
	Ref :: getRate() const {
		return rate_;
	}
	void
	Ref :: setRate (const value :: Rate rate) const
	{
		if (rate_ <= rate) {
			return;
		}
		rate_ = rate;
		down_->setRate (rate);
	}
	void
	Ref :: searchProof() {
		//buildDown();
	}
	bool
	Ref :: isProved() const {
		return (confirmation_.getSize() > 0);
	}

	void
	Ref :: buildUp (Level* const) {
	}
	void
	Ref :: buildDown (const bool delta)
	{
		Prover :: refresh ("building down ref node");
		collectConfirmation (up_->getConfirmation(), delta);
		if (confirmation_.boundaryMoved()) {
			calculateWeightUp();
			const_cast<Node*>(down_)->buildDown();
			confirmation_.markBoundary();
		}
	}
	bool
	Ref :: isRoot() const {
		return false;
	}
	void
	Ref :: addDown (const Node*) {
	}
	bool
	Ref :: liesDown (const Node* node) const
	{
		if (node == this) {
			return true;
		}
		return down_->liesDown (node);
	}

	void
	Ref :: showNode
	(
		String& string,
		const value :: Integer depth
	) const
	{
		string << iterate << depth << tab << "ref: " << *replacement_ << endLine;
		string << iterate << depth << tab << "this: " << this << endLine;
		string << iterate << depth << tab << "up: " << up_ << endLine;
		string << iterate << depth << tab << "down: " << down_ << endLine;
		//string << endLine;
	}
	void
	Ref :: showTree
	(
		String& string,
		const value :: Integer depthStart,
		const value :: Integer depthEnd
	) const
	{
		if (depthEnd > 0) {
			string << iterate << depthStart << tab;
			string << iterate << 64 << '-' << endLine;
		}
		showNode (string, depthStart);

		if (depthEnd == 0) {
			return;
		}
		//confirmation_.show (string);
		string << iterate << depthStart << tab;
		string << "up: " << up_ << endLine;
		//up_->showTree (string, depthStart + 1, depthEnd - 1);
		if (depthEnd > 0) {
			string << iterate << depthStart << tab;
			string << iterate << 64 << '-' << endLine;
		}
	}

	// prover :: Statistics implementation
	Size_t
	Ref :: getEvidenceVolume() const {
		return Statistics :: getEvidenceVolume (this);
	}
	Size_t
	Ref :: getTreeVolume() const {
		return Statistics :: getTreeVolume (this);
	}
	Size_t
	Ref :: getExpressionVolume() const {
		return Statistics :: getExpressionVolume (this);
	}
	Size_t
	Ref :: getTreeNodeVolume() const {
		return Statistics :: getTreeNodeVolume (this);
	}
	Size_t
	Ref :: getEvidenceNodeVolume() const {
		return Statistics :: getEvidenceNodeVolume (this);
	}
	Size_t
	Ref :: getEvidenceContainerVolume() const {
		return Statistics :: getEvidenceContainerVolume (this);
	}
	Size_t
	Ref :: getGarbageVolume() const {
		return Statistics :: getGarbageVolume (this);
	}

	// substitution statistics
	Size_t
	Ref :: getSubstitutionAllVolume() const {
		return Statistics :: getSubstitutionAllVolume (this);
	}
	Size_t
	Ref :: getSubstitutionIntVolume() const {
		return Statistics :: getSubstitutionIntVolume (this);
	}
	Size_t
	Ref :: getSubstitutionExtVolume() const {
		return Statistics :: getSubstitutionExtVolume (this);
	}
	Size_t
	Ref :: getSubstitutionVolume() const {
		return Statistics :: getSubstitutionVolume (this);
	}

	// node count statistics
	value :: Integer
	Ref :: getTreeNodeCount() const {
		return Statistics :: getTreeNodeCount (this);
	}
	value :: Integer
	Ref :: getEvidenceNodeCount() const {
		return Statistics :: getEvidenceNodeCount (this);
	}
	value :: Integer
	Ref :: getExpressionCount() const {
		return Statistics :: getExpressionCount (this);
	}
	value :: Integer
	Ref :: getHypCount() const {
		return Statistics :: getHypCount (this);
	}
	value :: Integer
	Ref :: getPropCount() const {
		return Statistics :: getPropCount (this);
	}
	value :: Integer
	Ref :: getRefCount() const {
		return Statistics :: getRefCount (this);
	}

	// object :: Cloneable implementation
	const Node*
	Ref :: clone () const {
		return NULL;
	}

	// object :: Object implementation
	void
	Ref :: commitSuicide() {
		delete this;
	}
	Size_t
	Ref :: getVolume() const
	{
		Size_t volume = 0;
		volume += getTreeVolume();
		volume += getEvidenceVolume();
		return volume;
	}
	Size_t
	Ref :: getSizeOf() const {
		return sizeof (Ref);
	}
	void
	Ref :: show (String& string) const {
		showNode (string);
	}

	/****************************
	 *		Private members
	 ****************************/

	// weight routines
	void
	Ref :: calculateWeight() const
	{
		weight_ = down_->getWeight();
		weight_ += Config :: getProofFactor() * confirmation_.getSize();
	}

	// build down routines
	void
	Ref :: collectConfirmation (const Confirmation& confirmation, const bool delta)
	{
		const value :: Integer
			size = delta ? confirmation.getDelta() : confirmation.getSize();
		for (value :: Integer i = 0; i < size; ++ i) {
			const Evidence* const
				upEvidence = delta ? confirmation.getNew (i) : confirmation.get (i);
			addEvidence (upEvidence);
		}
	}
	void
	Ref :: addEvidence (const Evidence* upEvidence)
	{
		const Substitution* const
			upSubstitution = upEvidence->getSubstitution();
		Evidence* evidenceSub = NULL;
		Memory :: stack().push(); {
			stack :: Substitution*
				sub = stack :: Substitution :: create (replacement_);
			sub = sub->compose (upSubstitution->castAway<allocator :: Stack>(), false);

			evidenceSub = confirmation_.supply (sub);
		}
		Memory :: stack().pop();
		if (evidenceSub == NULL) {
			return;
		}
		const Substitution* replacement = NULL;
		#ifdef DEBUG
		replacement = Substitution :: create (replacement_);
		#endif
		evidence :: Ref* evidenceRef = new evidence :: Ref
		(
			this,
			replacement,
			evidenceSub,
			upEvidence
		);
		dynamic_cast<evidence :: Sub*>(evidenceSub)->addUp (evidenceRef);
	}

	inline void
	Ref :: verify() const {
	}
}
}
}

#endif /*MDL_PROVER_NODE_REF_CPP_*/
