/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_node_Root.cpp                                 */
/* Description:     root node in proof variant tree                          */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_NODE_ROOT_CPP_
#define MDL_PROVER_NODE_ROOT_CPP_

#include "mdl/prover/node/mdl_prover_node.hpp"

namespace mdl {
namespace prover {
namespace node {

	/****************************
	 *		Public members
	 ****************************/

	Root :: Root (const Expression* const expression) :
	hyp_ (expression, this, true),
	proofVector_ () {
	}
	Root :: ~ Root() {
	}

	const proof :: Vector&
	Root :: getProofs() const {
		return proofVector_;
	}

	// prover :: Node implementation
	const Expression*
	Root :: getExpression() const {
		return hyp_.getExpression();
	}
	const mdl :: Proposition*
	Root :: getProposition() const {
		return hyp_.getProposition();
	}
	const Confirmation&
	Root :: getConfirmation() const {
		return hyp_.getConfirmation();
	}
	value :: Weight
	Root :: getWeight() const {
		return hyp_.getWeight();
	}
	void
	Root :: calculateWeightUp() const {
		hyp_.calculateWeightUp();
	}
	void
	Root :: calculateWeightDown() const {
	}
	bool
	Root :: operator < (const Node& node) const {
		return hyp_.operator < (node);
	}
	value :: Rate
	Root :: getRate() const {
		return hyp_.getRate();
	}
	void
	Root :: setRate (const value :: Rate rate) const {
		hyp_.setRate (rate);
	}
	void
	Root :: searchProof() {
		hyp_.searchProof();
	}
	bool
	Root :: isProved() const {
		return proofVector_.hasValidProofs();
	}

	void
	Root :: buildUp (Level* const level) {
		hyp_.buildUp (this, level);
	}
	void
	Root :: buildDown (const bool flag)
	{
		Prover :: refresh ("building down root node");
		hyp_.collectConfirmation (this);
		if (hyp_.confirmation_.boundaryMoved()) {
			hyp_.calculateWeightUp();
			collectProofs();
			hyp_.confirmation_.markBoundary();
			if (isProved()) {
				if (Config :: getSearchMode() == Config :: MINIMAL_SEARCH) {
					throw Exception ("prover had ran out of work.");
				}
				if (Config :: getSearchMode() == Config :: OPTIMAL_SEARCH) {
					Prover :: timers().recount();
				}
			}
		}
	}
	bool
	Root :: isRoot() const {
		return true;
	}
	void
	Root :: addDown (const Node*) {
	}
	bool
	Root :: liesDown (const Node*) const {
		return false;
	}

	void
	Root :: showNode
	(
		String& string,
		const value :: Integer depth
	) const {
		hyp_.showNode (string, depth);
	}
	void
	Root :: showTree
	(
		String& string,
		const value :: Integer depthStart,
		const value :: Integer depthEnd
	) const {
		hyp_.showTree (string, depthStart, depthEnd);
	}

	// prover :: Statistics implementation
	Size_t
	Root :: getEvidenceVolume() const {
		return hyp_.getEvidenceVolume();
	}
	Size_t
	Root :: getTreeVolume() const {
		return hyp_.getTreeVolume();
	}
	Size_t
	Root :: getExpressionVolume() const {
		return hyp_.getExpressionVolume();
	}
	Size_t
	Root :: getTreeNodeVolume() const {
		return hyp_.getTreeNodeVolume();
	}
	Size_t
	Root :: getEvidenceNodeVolume() const {
		return hyp_.getEvidenceNodeVolume();
	}
	Size_t
	Root :: getEvidenceContainerVolume() const {
		return hyp_.getEvidenceContainerVolume();
	}
	Size_t
	Root :: getGarbageVolume() const {
		return hyp_.getGarbageVolume();
	}

	// substitution statistics
	Size_t
	Root :: getSubstitutionAllVolume() const {
		return hyp_.getSubstitutionAllVolume();
	}
	Size_t
	Root :: getSubstitutionIntVolume() const {
		return hyp_.getSubstitutionIntVolume();
	}
	Size_t
	Root :: getSubstitutionExtVolume() const {
		return hyp_.getSubstitutionExtVolume();
	}
	Size_t
	Root :: getSubstitutionVolume() const {
		return hyp_.getSubstitutionVolume();
	}

	// node count statistics
	value :: Integer
	Root :: getTreeNodeCount() const {
		return hyp_.getTreeNodeCount();
	}
	value :: Integer
	Root :: getEvidenceNodeCount() const {
		return hyp_.getEvidenceNodeCount();
	}
	value :: Integer
	Root :: getExpressionCount() const {
		return hyp_.getExpressionCount();
	}
	value :: Integer
	Root :: getHypCount() const {
		return hyp_.getHypCount();
	}
	value :: Integer
	Root :: getPropCount() const {
		return hyp_.getPropCount();
	}
	value :: Integer
	Root :: getRefCount() const {
		return hyp_.getRefCount();
	}

	// object :: Cloneable implementation
	const Node*
	Root :: clone () const {
		return hyp_.clone();
	}

	// object :: Object implementation
	void
	Root :: commitSuicide() {
		delete this;
	}
	Size_t
	Root :: getVolume() const
	{
		Size_t volume = 0;
		volume += hyp_.getVolume();
		volume += proofVector_.getVolume();
		return volume;
	}
	Size_t
	Root :: getSizeOf() const {
		return sizeof (Root);
	}
	void
	Root :: show (String& string) const
	{
		string << "*** ROOT ***" << endLine;
		hyp_.show (string);
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Root :: collectProofs()
	{
		proofVector_.markBoundary();
		for (value :: Integer i = 0; i < hyp_.confirmation_.getDelta(); ++ i) {
			const Evidence* const
				evidence = hyp_.confirmation_.getNew (i);
			Memory :: stack().push();
			const vector :: Proof
				proofs = evidence->constructProofs();
			for (value :: Integer j = 0; j < proofs.getSize(); ++ j) {
				const Proof* const
					proof = proofs.getValue (j);
				proofVector_.add (proof);
			}
			Memory :: stack().pop();
		}
	}
}
}
}

#endif /*MDL_PROVER_NODE_ROOT_CPP_*/
