/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_proof_Prop.cpp                                */
/* Description:     prop proof element                                       */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_PROOF_PROP_CPP_
#define MDL_PROVER_PROOF_PROP_CPP_

namespace mdl {
namespace prover {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Prop :: Prop (const Prop& prop) :
	evidence_ (prop.evidence_),
	substitutionInner_ (prop.substitutionInner_),
	substitutionOuter_ (prop.substitutionOuter_),
	up_ (prop.up_.getSize()),
	depth_ (prop.depth_),
	cardinality_ (prop.cardinality_),
	isValid_ (prop.isValid_),
	step_ (NULL) {
	}
	inline
	Prop :: Prop
	(
		const Evidence* evidence,
		const Substitution* substitution,
		const index :: Arity arity
	) :
	evidence_ (evidence),
	substitutionInner_ (substitution),
	substitutionOuter_ (NULL),
	up_ (arity),
	depth_ (1),
	cardinality_ (1),
	isValid_ (false),
	step_ (NULL) {
		checkValidity();
	}
	Prop :: ~ Prop() {
	}

	inline void
	Prop :: addUp (const index :: Arity i, const Proof* proof)
	{
		up_.getReference (i) = proof;
		const index :: Step
			proofDepth = proof->getDepth();
		depth_ =
			(proofDepth + 1 > depth_) ?
			proofDepth + 1 :
			depth_;
		cardinality_ += proof->getCardinality();
	}

using manipulator :: space;
using manipulator :: tab;
using manipulator :: endLine;

	// prover :: Proof implementation
	bool
	Prop :: isValid() const
	{
		if (!isValid_) {
			return false;
		}
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Proof* const
				proof = up_.getValue (i);
			if (!proof->isValid()) {
				return false;
			}
		}
		return true;
	}
	void
	Prop :: verifyIntegrity() const {
		evidence_->verifyIntegrity();
	}
	inline const Evidence*
	Prop :: getEvidence() const {
		return evidence_;
	}
	index :: Step
	Prop :: getDepth() const {
		return depth_;
	}
	value :: Integer
	Prop :: getCardinality() const {
		return cardinality_;
	}
	const mdl :: proof :: Step*
	Prop :: compileStep (mdl :: proof :: Step* const question) const
	{
		compileStepUp (question);
		createStep
		(
			question,
			question->getIndex()
		);
		return step_;
	}
	void
	Prop :: compileClaim (mdl :: proof :: Claim* const claim) const
	{
		#ifndef DEBUG
		if (step_ != NULL) {
			return;
		}
		#endif

		compileClaimUp (claim);
		createStep
		(
			claim,
			claim->getProof()->getLength()
		);
		const_cast<Location&>(step_->getLocation()).indent().incTabs();
		addVariables (claim);
		claim->addStep (step_);
	}
	const mdl :: proof :: Reference*
	Prop :: getReference() const {
		return NULL;
	}
	const mdl :: proof :: Step*
	Prop :: getStep() const {
		return step_;
	}

	// object :: Object interface
	void
	Prop :: commitSuicide() {
		delete this;
	}
	Size_t
	Prop :: getVolume() const
	{
		Size_t volume = 0;
		volume += up_.getVolume();
		volume += getAggregateVolume (substitutionInner_);
		volume += getAggregateVolume (substitutionOuter_);
		return volume;
	}
	Size_t
	Prop :: getSizeOf() const {
		return sizeof (Prop);
	}
	void
	Prop :: show (String& string) const
	{
		string << "############################### - Prop Node - ################################################" << endLine;
		string << "prop: "; // << endLine;
		evidence_->getNode()->getProposition()->getAssertion()->replicateName (string);
		string << endLine;
		string << "this: " << this << endLine;
		string << "node: " << evidence_ << endLine;
		string << "up:" << space;
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			string << up_.getValue (i);
			if (i < up_.getSize() - 1) {
				string << "," << space;
			}
		}
		string << endLine;
		string << "up subs:" << endLine;
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			string << tab << i << " : " << *up_.getValue (i)->getEvidence()->getSubstitution() << endLine;
		}
		string << endLine;
		string << "****************************************************" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << "********************** EVIDENCE ************************" << endLine;
		evidence_->show (string);
		string << "****************************************************" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << "############################### - Prop Node - ################################################" << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
		string << endLine;
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Prop :: checkValidity()
	{
		const mdl :: prover :: Node* const
			node = evidence_->getNode();
		const node :: Prop*
			prop = dynamic_cast<const node :: Prop*>(node);
		substitutionOuter_ = Substitution :: create (prop->substitutionExt_);
		substitutionOuter_ = const_cast<Substitution*>(substitutionOuter_)->compose (substitutionInner_);

		const mdl :: Proposition* const
			proposition = node->getProposition();
		const mdl :: Assertion* const
			assertion = proposition->getAssertion();
		const mdl :: Disjointeds* const
			disjointeds = assertion->getDisjointeds();
		const mdl :: Variables* const
			variables = assertion->getVariables();

		for (index :: Arity i = 0; i < variables->getSize(); ++ i) {
			const mdl :: header :: Variable* const
				variable = variables->getVariable (i);
			const Symbol& symbol = variable->getSymbol();
			if (substitutionOuter_->find (symbol) == NULL) {
				const Substitution :: Term_*
					term = Expression :: createSingleton (symbol);
				const Substitution*
					delta = new Substitution (symbol, term);
				substitutionOuter_ = const_cast<Substitution*>(substitutionOuter_)->join (delta);
			}
		}
		isValid_ = disjointeds->check (substitutionOuter_, Prover :: getProblem());
	}
	inline void
	Prop :: compileStepUp (mdl :: proof :: Step* const question) const
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Proof* const
				proof = up_.getValue (i);
			proof->compileStep (question);
		}
	}
	inline void
	Prop :: compileClaimUp (mdl :: proof :: Claim* const claim) const
	{
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Proof* const
				proof = up_.getValue (i);
			proof->compileClaim (claim);
		}
	}
	inline void
	Prop :: createStep
	(
		mdl :: proof :: Step* const step,
		const index :: Step index
	) const
	{
		const mdl :: proof :: References*
			references = createReferences();

		const mdl :: proof :: Link* const
			link = new form :: proof :: Link
			(
				step->getLocation(),
				evidence_->getNode()->getProposition(),
				references
			);

		Expression* const
			expression = new Expression (evidence_->getNode()->getExpression());

		expression->execute (substitutionInner_);

		const array :: Expression* const
			arrayExpression = new array :: Expression (expression);

		step_ = new form :: proof :: step :: Assertion
		(
			step,
			index,
			link,
			arrayExpression
		);
	}
	inline const mdl :: proof :: References*
	Prop :: createReferences() const
	{
		mdl :: proof :: References* const
			references = new form :: proof :: References (up_.getSize());
		for (index :: Arity i = 0; i < up_.getSize(); ++ i) {
			const Proof* const
				proof = up_.getValue (i);
			const mdl :: proof :: Reference* const
				reference = proof->getReference();
			references->add (reference);
		}
		return references;
	}
	void
	Prop :: addVariables (mdl :: proof :: Claim* const claim) const
	{
		const array :: Expression* const
			expression = step_->getExpression();
		for (index :: Literal i = 0; i < expression->getSize(); ++ i) {
			const Symbol&
				symbol = expression->getSymbol (i);
			if (symbol.isVariable()) {
				claim->addVariable (symbol);
			}
		}
	}
}
}
}

#endif /*MDL_PROVER_PROOF_PROP_CPP_*/
