/*****************************************************************************/
/* Project name:    mdl - mathematics development language                   */
/* File Name:       mdl_prover_proof_Vector.cpp                              */
/* Description:     proof vector: vector of proofs                           */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MDL_PROVER_PROOF_VECTOR_CPP_
#define MDL_PROVER_PROOF_VECTOR_CPP_

namespace mdl {
namespace prover {
namespace proof {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Vector :: Vector () :
	proofVector_ (INITIAL_PROOFS_CAPACITY),
	boundary_ (0) {
	}
	inline
	Vector :: Vector (const Vector& vector) :
	proofVector_ (vector.proofVector_),
	boundary_ (vector.boundary_) {
	}
	inline
	Vector :: Vector (const value :: Integer capacity) :
	proofVector_ (capacity),
	boundary_ (0) {
	}
	Vector :: ~ Vector() {
	}

using manipulator :: endLine;
using manipulator :: tab;
using manipulator :: space;

	bool
	Vector :: hasValidProofs() const
	{
		for (value :: Integer i = 0; i < proofVector_.getSize(); ++ i) {
			if (proofVector_.getValue (i)->isValid()) {
				return true;
			}
		}
		return false;
	}
	inline void
	Vector :: verifyIntegrity() const
	{
		for (value :: Integer i = 0; i < proofVector_.getSize(); ++ i) {
			proofVector_.getValue (i)->verifyIntegrity();
		}
	}
	inline bool
	Vector :: boundaryMoved() const {
		return (proofVector_.getSize() > boundary_);
	}
	inline void
	Vector :: markBoundary() {
		boundary_ = proofVector_.getSize();
	}
	inline value :: Integer
	Vector :: getBoundary() const {
		return boundary_;
	}
	inline value :: Integer
	Vector :: getSize() const {
		return proofVector_.getSize();
	}
	inline value :: Integer
	Vector :: getDelta() const {
		return proofVector_.getSize() - boundary_;
	}
	inline void
	Vector :: add (const Proof* proof) {
		proofVector_.add (proof);
	}
	inline const Proof*
	Vector :: get (const value :: Integer index) const {
		return proofVector_.getValue (index);
	}
	inline const Proof*
	Vector :: getNew (const value :: Integer index) const {
		return proofVector_.getValue (index + boundary_);
	}
	const mdl :: proof :: Step*
	Vector :: compileProof (const mdl :: proof :: Step* question) const
	{
		#ifdef DEBUG
		verifyCompilation (question);
		#endif
		const Proof* const
			proof = getLeastProof();
		switch (proof->getCardinality()) {
		case 0  : return NULL;
		case 1  : return createStep (proof, question);
		default : return createClaim (proof, question);
		}
	}
	Size_t
	Vector :: getProofVolume() const
	{
		Size_t volume = 0;
		for (value :: Integer i = 0; i < proofVector_.getSize(); ++ i) {
			const Proof* const
				proof = proofVector_.getValue (i);
			volume += getAggregateVolume (proof);
		}
		return volume;
	}

	// object :: Object interface
	void
	Vector :: commitSuicide() {
		delete this;
	}
	Size_t
	Vector :: getVolume() const
	{
		Size_t volume = 0;
		volume += proofVector_.getVolume();
		return volume;
	}
	Size_t
	Vector :: getSizeOf() const {
		return sizeof (Vector);
	}
	void
	Vector :: show (String& string) const
	{
		if (proofVector_.getSize() == 0) {
			//string << "evidences: none." << endLine;
			return;
		}
		string << "proofs:" << endLine;
		for (value :: Integer i = 0; i < proofVector_.getSize(); ++ i) {
			const Proof*
				proof = proofVector_.getValue (i);
			proof->show (string);
		}
	}

	/****************************
	 *		Public members
	 ****************************/

	inline const Proof*
	Vector :: getLeastProof() const
	{
		value :: Integer leastIndex = 0;
		value :: Integer leastCardinality = value :: undefined :: INTEGER;

		for (value :: Integer i = 0; i < proofVector_.getSize(); ++ i) {
			const Proof* const
				proof = proofVector_.getValue (i);
			if (!proof->isValid()) {
				continue;
			}
			const value :: Integer
				cardinality = proof->getCardinality();

			if (cardinality < leastCardinality) {
				leastCardinality = cardinality;
				leastIndex = i;
			}
		}
		return proofVector_.getValue (leastIndex);
	}

	const mdl :: proof :: Step*
	Vector :: createStep
	(
		const Proof* proof,
		const mdl :: proof :: Step* step
	) const
	{
		mdl :: proof :: Step* const
			question = const_cast<mdl :: proof :: Step*>(step);
		return proof->compileStep (question);
	}
	const mdl :: proof :: Step*
	Vector :: createClaim
	(
		const Proof* proof,
		const mdl :: proof :: Step* step
	) const
	{
		const form :: proof :: step :: Question* const
			question = dynamic_cast<const form :: proof :: step :: Question*>(step);
		mdl :: proof :: Claim* const
			claim = new form :: proof :: step :: Claim
			(
				question,
				proof->getCardinality()
			);
		proof->compileClaim (claim);
		return claim;
	}

	void
	Vector :: verifyCompilation (const mdl :: proof :: Step* question) const
	{
		value :: Integer failed = 0;
		value :: Integer succeded = 0;
		std :: cerr << "evidences found: " << proofVector_.getSize() << std :: endl;
		std :: cerr << "===============================" << std :: endl;
		for (value :: Integer i = 0; i < proofVector_.getSize(); ++ i) {
			const Proof* const
				proof = proofVector_.getValue (i);
			const mdl :: proof :: Step*
				claim = NULL;
			//std :: cerr << "evidence " << i << " is verifying... ";
			try {
				claim = verifyClaim (proof, question);
			} catch (Error* error) {
				++ failed;
				Counter :: switchOn();
				claim = createClaim (proof, question);
				std :: cerr << "evidence " << i << " failed" << std :: endl;
				std :: cerr << error->message() << std :: endl;
				std :: cerr << *claim << std :: endl;
				try {
					proof->verifyIntegrity();
					claim->check();
				} catch (Error* error) {
					delete error;
				}
				Counter :: switchOff();
				throw error;
				//break;
			}
			++ succeded;
			//std :: cerr << "ok" << std :: endl;
			//std :: cerr << "depth = " << evidence->getDepth() << std :: endl;
			//std :: cerr << *claim << std :: endl;
		}
		//std :: cerr << "===============================" << std :: endl;
		std :: cerr << "failed = " << failed << std :: endl;
		std :: cerr << "succeded = " << succeded << std :: endl;
		std :: cerr << std :: endl;
	}
	const mdl :: proof :: Step*
	Vector :: verifyClaim
	(
		const Proof* proof,
		const mdl :: proof :: Step* question
	) const
	{
		const mdl :: proof :: Step* const
			claim = createClaim (proof, question);
		proof->verifyIntegrity();
		claim->check();
		return claim;
	}
}
}
}

#endif /*MDL_PROVER_PROOF_VECTOR_CPP_*/

