/*****************************************************************************/
/* Project name:    mm - decompiler from metamath to mdl                     */
/* File Name:       mm_Mm.cpp                                                */
/* Description:     mm decompiler                                            */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef MM_MM_CPP_
#define MM_MM_CPP_

#include <exception>

#include "mm/ast/mm_ast.cpp"
#include "mm/auxiliary/mm_auxiliary.cpp"
#include "mm/expression/mm_expression.cpp"
#include "mm/interface/mm_interface.cpp"
#include "mm/lexer/mm_lexer.cpp"
#include "mm/math/mm_math.cpp"
#include "mm/parser/mm_parser.cpp"

namespace mm {

using namespace exception;

	/****************************
	 *		Public members
	 ****************************/

	Mm :: Mm() :
	file_ (Config :: getSource()->getString()),
	path_ (new Path (file_)),
	parser_ (new Parser_()),
	cutParser_ (new CutParser_()),
	source_ (NULL),
	cutSource_ (NULL),
	target_ (NULL),
	timers_ (),
	volume_ (this)
	{
		Math :: init();
		Table :: init();
		Corrector :: init();
	}
	Mm :: ~ Mm()
	{
		Corrector :: release();
		Table :: release();
		Math :: release();
		deleteAggregateObject (target_);
		deleteAggregateObject (source_);
		deleteAggregateObject (cutSource_);
		deleteAggregateObject (parser_);
		deleteAggregateObject (cutParser_);
		deleteAggregateObject (path_);
		object :: Object :: release();
		Config :: release();
	}

using std :: endl;
using std :: cerr;
using std :: cout;

	Mm :: Result_
	Mm :: run ()
	{
		try {
			if (Config :: cutSource()) {
				cut();
			} else {
				process();
			}
		} catch (const Error* error) {
			timers_.stopAll();
			cerr << *error << endl;
			Result_ result = static_cast<Result_>(error->getType());
			delete error;
			return result;
		} catch (const OutOfMemory& error) {
			timers_.stopAll();
			cerr << error << endl;
			Memory :: show (cerr);
			return OUT_OF_MEMORY;
		} catch (const std :: exception& e) {
			timers_.stopAll();
			cerr << "exception: " << e.what() << endl;
			return GENERAL_FAILURE;
		} catch (...) {
			timers_.stopAll();
			cerr << "exception." << endl;
			return GENERAL_FAILURE;
		}
		return SUCCESS;
	}
	Mm :: Result_
	Mm :: initMemory()
	{
		Result_ result = SUCCESS;
		if (Config :: verbosity() > 1) {
			std :: cout << "initializing ... " << std :: flush;
		}
		try {
			Memory :: standard().create();
			Memory :: heap().create();
		} catch (OutOfMemory&) {
			result = Mm :: INIT_ERROR;
		}
		if (Config :: verbosity() > 1) {
			std :: cout << ((result == SUCCESS) ? "done." : "failed.") << std :: endl;
		}
		return result;
	}
	Mm :: Result_
	Mm :: releaseMemory()
	{
		bool success = true;
		if (Config :: verbosity() >= Config :: VERBOSE_STANDARD) {
			std :: cout << "releasing ... " << std :: flush;
		}
		try {
			Memory :: heap().destroy();
		} catch (memory :: MemoryLeaks& exception) {
			std :: cerr << std :: endl << chars :: TAB << exception.what() << ": ";
			std :: cerr << exception.getLeak() << " bytes are lost ";
			success = false;
		} catch (std :: exception& exception) {
			std :: cerr << exception.what() << std :: endl;
			success = false;
		} catch (...) {
			success = false;
		}
		if (Config :: verbosity() >= Config :: VERBOSE_STANDARD) {
			std :: cout << (success ? "done." : "failed.") << std :: endl;
		}
		return (success ? SUCCESS : RELEASE_ERROR);
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// object :: Object implementation
	void
	Mm :: commitSuicide() {
		delete this;
	}
	Size_t
	Mm :: getVolume() const
	{
		Size_t volume = 0;
		volume += object :: Object :: getVolume();
		volume += file_.getVolume();
		volume += getAggregateVolume (path_);
		volume += getAggregateVolume (parser_);
		volume += getAggregateVolume (cutParser_);
		volume += getAggregateVolume (source_);
		volume += getAggregateVolume (cutSource_);
		volume += getAggregateVolume (target_);
		volume += timers_.getVolume();
		volume += volume_.getVolume();
		return volume;
	}
	Size_t
	Mm :: getSizeOf() const {
		return sizeof (Mm);
	}
	void
	Mm :: show (String& string) const
	{
		string << "file" << space << *path_ << space;
		string << "is done in" << space << timers_.total() << endLine;

		if (Config :: showTimings()) {
			timers_.show (string);
			string << endLine;
		}
		if (Config :: showMath()) {
			Math :: show (string);
			string << endLine;
		}
		if (Config :: showVolume() || Config :: showMemory()) {
			volume_.show (string);
			string << endLine;
		}
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Mm :: process ()
	{
		timers_.total().start();

		read();
		check();
		translate();
		write();

		timers_.total().stop();
	}
	void
	Mm :: read ()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "reading ... " << std :: flush;
		}
		timers_.read().start ();
		source_ = parser_->parse (path_);
		timers_.read().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.read() << std :: endl;
		}
	}
	void
	Mm :: check ()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "checking ... " << std :: flush;
		}
		timers_.check().start();
		source_->check();
		timers_.check().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.check() << std :: endl;
		}
	}
	void
	Mm :: translate ()
	{
		if (Config :: translate()) {
			if (Config :: verbosity() > 1) {
				std :: cout << "translating ... " << std :: flush;
			}
			timers_.translate().start();
			target_ = dynamic_cast<const Target*>(source_->translate());
			timers_.translate().stop();
			if (Config :: verbosity() > 1) {
				std :: cout << "done in " << timers_.translate() << std :: endl;
			}
		}
	}
	void
	Mm :: write ()
	{
		if (Config :: translate()) {
			if (Config :: verbosity() > 1) {
				std :: cout << "writing ... " << std :: flush;
			}
			timers_.write().start();
			target_->write();
			timers_.write().stop();
			if (Config :: verbosity() > 1) {
				std :: cout << "done in " << timers_.write() << std :: endl;
			}
		}
	}
	void
	Mm :: cut()
	{
		timers_.total().start();
		cutRead();
		cutWrite();
		timers_.total().stop();
	}
	void
	Mm :: cutRead()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "reading ... " << std :: flush;
		}
		timers_.read().start ();
		cutSource_ = cutParser_->parse (path_);
		timers_.read().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.read() << std :: endl;
		}
	}
	void
	Mm :: cutWrite()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "writing ... " << std :: flush;
		}
		timers_.write().start();
		cutSource_->write();
		timers_.write().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.write() << std :: endl;
		}
	}
}

#endif /*MM_MM_CPP_*/
