/*****************************************************************************/
/* Project name:    nsgl - non-standard general library                      */
/* File name:       nsgl_config_Paths.cpp                                    */
/* Description:     paths config flags                                       */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSGL_CONFIG_PATHS_CPP_
#define NSGL_CONFIG_PATHS_CPP_

#include "system/nsgl_system.hpp"
#include "config/nsgl_config_Paths.hpp"

namespace nstd {
namespace config {

	/****************************
	 *		Public members
	 ****************************/

	inline Paths :: Path_*&
	Paths :: source() {
		return source_;
	}
	inline Paths :: Path_*&
	Paths :: target() {
		return target_;
	}
	inline Paths :: Path_*&
	Paths :: log() {
		return log_;
	}
	inline const Paths :: Path_*
	Paths :: getSource() {
		return source_;
	}
	inline const Paths :: Path_*
	Paths :: getTarget() {
		return target_;
	}
	inline const Paths :: Path_*
	Paths :: getLog() {
		return log_;
	}
	inline bool
	Paths :: overwrite() {
		return overwrite_;
	}

	/****************************
	 *		Protected members
	 ****************************/

using std :: endl;
using std :: cout;
using std :: cerr;

	inline void
	Paths :: showOptions()
	{
		cout << "  -o,  --out <file>             translate to the specified file" << endl;
		cout << "       --log <file>             write log to the specified file" << endl;
		cout << "  -w,  --overwrite              overwrite output file" << endl;
		cout << endl;
	}
	inline void
	Paths :: init () {
	}
	void
	Paths :: release ()
	{
		if (source_ != NULL) {
			delete source_;
			source_ = NULL;
		}
		if (target_ != NULL) {
			delete target_;
			target_ = NULL;
		}
		if (log_ != NULL) {
			delete log_;
			log_ = NULL;
		}
	}
	inline Size_t
	Paths :: getVolume()
	{
		Size_t volume = 0;
		if (source_ != NULL) {
			volume += source_->getVolume();
			volume += source_->getSizeOf();
		}
		if (target_ != NULL) {
			volume += target_->getVolume();
			volume += target_->getSizeOf();
		}
		if (log_ != NULL) {
			volume += log_->getVolume();
			volume += log_->getSizeOf();
		}
		return volume;
	}

	bool
	Paths :: proceedStringOption (const char* opt, const int argCount, const char* argValues[], int& i)
	{
		if (i == argCount - 1) {
			const char* path = argValues [i];
			source_ = new Path_ (path);
			return true;
		} else if (equalStrings (opt, "out")) {
			if (i < argCount - 1) {
				const char* path = argValues [++ i];
				target_ = new Path_ (path);
			} else {
				cerr << "\"--out\" option must be followed by a output file name, not source file." << endl;
				abortParsing_ = true;
			}
			return true;
		} else if (equalStrings (opt, "log")) {
			if (i < argCount - 1) {
				const char* path = argValues [++ i];
				log_ = new Path_ (path);
			} else {
				cerr << "\"--log\" option must be followed by a output log file name." << endl;
				abortParsing_ = true;
			}
			return true;
		} else if (equalStrings (opt, "overwrite")) {
			overwrite_ = true;
			return true;
		}
		return false;
	}
	bool
	Paths :: proceedOneCharOption (const char* opt, const int argCount, const char* argValues[], int& i)
	{
		if (i == argCount - 1) {
			const char* path = argValues [i];
			source_ = new Path_ (path);
			return true;
		} else if (*opt == 'o') {
			if (i < argCount - 1) {
				const char* path = argValues [++ i];
				target_ = new Path_ (path);
			} else {
				cerr << "\"-o\" option must be followed by a output file name, not source file." << endl;
				abortParsing_ = true;
			}
			return true;
		} else if (*opt == 'w') {
			overwrite_= true;
			return true;
		}
		return false;
	}
	bool
	Paths :: check ()
	{
		if (source_ != NULL) {
			if (!checkSource()) {
				return false;
			}
		}
		if (target_ != NULL) {
			if (!checkTarget()) {
				return false;
			}
		}
		if (log_ != NULL) {
			if (!checkLog()) {
				return false;
			}
		}
		if (target_ != NULL  && source_ != NULL) {
			if (!checkBoth()) {
				return false;
			}
		}
		return true;
	}

	/****************************
	 *		Private members
	 ****************************/

	inline bool
	Paths :: checkSource()
	{
		const Size_t
			sourceSize = system :: getFileSize (source_->getString().c_str());
		if (sourceSize == 0) {
			cerr << "source file: " << *source_ << " is empty." << endl;
			return false;
		}
		if (sourceSize == Undefined<Size_t> :: getValue()) {
			cerr << "source file: " << *source_ << " is not found." << endl;
			return false;
		}
		return true;
	}
	inline bool
	Paths :: checkTarget()
	{
		const Size_t
			targetSize = system :: getFileSize (target_->getString().c_str());
		if ((targetSize != Undefined<Size_t> :: getValue()) && !overwrite_) {
			cerr << "target file: " << *target_ << " already exists" << endl;
			return false;
		}
		return true;
	}
	inline bool
	Paths :: checkLog() {
		return true;
	}
	inline bool
	Paths :: checkBoth()
	{
		if (*target_ == *source_) {
			cerr << "source: " << *source_;
			cerr <<" and target " << *target_;
			cerr << " files coincide" << endl;
			return false;
		}
		return true;
	}

	/************************************
	 *		Static private attributes
	 ************************************/

	Paths :: Path_*
		Paths :: source_ = NULL;
	Paths :: Path_*
		Paths :: target_ = NULL;
	Paths :: Path_*
		Paths :: log_ = NULL;
	bool
		Paths :: overwrite_= false;
}
}

#endif /*NSGL_CONFIG_PATHS_CPP_*/
