/*****************************************************************************/
/* Project name:    nsgl - non-standard general library                      */
/* File name:       nsgl_Path.ipp                                            */
/* Description:     path in the filesystem                                   */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSGL_PATH_IPP_
#define NSGL_PATH_IPP_

namespace nstd {

	/****************************
	 *		Public members
	 ****************************/

	template<class A>
	inline
	Path<A> :: Path () :
	changed_ (true),
	path_(),
	prefix_ (NO),
	extension_(),
	name_(),
	directories_() {
	}
	template<class A>
	inline
	Path<A> :: Path (const Path& path) :
	changed_ (path.changed_),
	path_(path.getString()),
	prefix_ (path.prefix_),
	extension_(path.getExtension()),
	name_(path.getName()),
	directories_(path.getDepth())
	{
		for (Depth_ depth = 0; depth < path.getDepth(); ++ depth) {
			String_* dir_1 = directories_.push();
			const String_& dir_2 = path.getDirectory (depth);
			*dir_1 = dir_2;
		}
	}
	template<class A>
	template<class A1>
	inline
	Path<A> :: Path (const Path<A1>& path) :
	changed_ (path.changed_),
	path_(path.getString()),
	prefix_ (path.getPrefix()),
	extension_(path.getExtension()),
	name_(path.getName()),
	directories_(path.getDepth())
	{
		for (Depth_ depth = 0; depth < path.getDepth(); ++ depth) {
			String_* dir_1 = directories_.push();
			const String<A1>& dir_2 = path.getDirectory (depth);
			*dir_1 = dir_2;
		}
	}
	template<class A>
	template<class A1>
	inline
	Path<A> :: Path (const String<A1>& path) :
	changed_ (false),
	path_(path),
	prefix_ (NO),
	extension_(),
	name_(),
	directories_ ()
	{
		parsePath();
		buildPath();
		checkPath (path.c_str());
	}
	template<class A>
	inline
	Path<A> :: Path (const char* path) :
	changed_ (false),
	path_(path),
	prefix_ (NO),
	extension_(),
	name_(),
	directories_ ()
	{
		parsePath ();
		buildPath();
		checkPath (path);
	}
	template<class A>
	Path<A> :: ~ Path() {
	}

	template<class A>
	inline char&
	Path<A> :: prefix() {
		changed_ = true;
		return prefix_;
	}
	template<class A>
	inline typename Path<A> :: String_&
	Path<A> :: name() {
		changed_ = true;
		return name_;
	}
	template<class A>
	inline typename Path<A> :: String_&
	Path<A> :: directory (const Depth_ depth) {
		changed_ = true;
		return *(directories_.getValue (depth));
	}
	template<class A>
	inline typename Path<A> :: String_&
	Path<A> :: extension() {
		changed_ = true;
		return extension_;
	}
	template<class A>
	inline typename Path<A> :: String_&
	Path<A> :: pushDirectory () {
		return *directories_.push();
	}
	template<class A>
	template<class A1>
	inline void
	Path<A> :: addDirectoryFirst (const String<A1>& directory) {
		*directories_.pushFirst() = directory;
	}
	template<class A>
	template<class A1>
	inline void
	Path<A> :: addDirectoryLast (const String<A1>& directory) {
		*directories_.pushLast() = directory;
	}

	template<class A>
	inline char
	Path<A> :: getPrefix() const {
		return prefix_;
	}
	template<class A>
	inline Size_t
	Path<A> :: getDepth() const {
		return directories_.getSize();
	}
	template<class A>
	inline const typename Path<A> :: String_&
	Path<A> :: getString() const
	{
		if (changed_) {
			buildPath();
		}
		return path_;
	}
	template<class A>
	inline const typename Path<A> :: String_&
	Path<A> :: getName() const {
		return name_;
	}
	template<class A>
	inline const typename Path<A> :: String_&
	Path<A> :: getExtension() const {
		return extension_;
	}
	template<class A>
	inline const typename Path<A> :: String_&
	Path<A> :: getDirectory (const Depth_ depth) const {
		return *(directories_.getValue (depth));
	}

	template<class A>
	inline bool
	Path<A> :: isValid() const {
		return true;
	}
	template<class A>
	void
	Path<A> :: provideDirectory() const
	{
		if (changed_) {
			buildPath ();
		}
		const char* const dir = path_.c_str();
		char* ch = path_.c_str();
		switch (prefix_) {
		case DOUBLE_DOT : ++ ch;
		case DOT : ++ ch;
		case DIR : ++ ch;
		default :  break;
		}
		while (*ch != chars :: NULL_CHAR) {
			if (*ch == chars :: DIR_DELIMITER) {
				*ch = chars :: NULL_CHAR;
				system :: createDirectory (dir);
				*ch = chars :: DIR_DELIMITER;
			}
			++ ch;
		}
	}

	template<class A>
	template<class A1>
	inline bool
	Path<A> :: operator == (const Path<A1>& path)
	{
		if (changed_) {
			buildPath ();
		}
		return (path_ == path.path_);
	}
	template<class A>
	template<class A1>
	inline bool
	Path<A> :: operator != (const Path<A1>& path) {
		return !(operator == (path));
	}
	template<class A>
	template<class A1>
	inline void
	Path<A> :: operator = (const Path<A1>& path)
	{
		changed_ = path.changed_;
		path_ = path.path_;
		name_ = path.name_;
		extension_ = path.extension_;

		directories_.clear();
		for (Depth_ depth = 0; depth < path_.getDepth(); ++ depth) {
			String_* dir_1 = directories_.push();
			const String_& dir_2 = path_.getDirectory (depth);
			*dir_1 = dir_2;
		}
	}

	// nstd :: Object implementation
	template<class A>
	void
	Path<A> :: commitSuicide() {
		delete this;
	}
	template<class A>
	Size_t
	Path<A> :: getVolume() const
	{
		Size_t result = 0;
		result += path_.getVolume();
		result += name_.getVolume();
		result += extension_.getVolume();
		result += directories_.getVolume();
		return result;
	}
	template<class A>
	Size_t
	Path<A> :: getSizeOf() const {
		return sizeof (Path);
	}
	template<class A>
	void
	Path<A> :: show (String<>& string) const
	{
		if (changed_) {
			buildPath();
		}
		string << path_;
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class A>
	const char*
	Path<A> :: parsePrefix()
	{
		const char* const initialChar = path_.getFirstChar();
		const char ch0 = initialChar [0];
		if (ch0 == chars :: DIR_DELIMITER) {
			prefix_ = DIR;
			return initialChar + 1;
		}
		if (ch0 != chars :: DOT) {
			prefix_ = NO;
			return initialChar;
		}
		const char ch1 = initialChar [1];
		if (ch1 == chars :: DIR_DELIMITER) {
			prefix_ = DOT;
			return initialChar + 2;
		}
		if (ch1 != chars :: DOT) {
			prefix_ = NO;
			return initialChar;
		}
		const char ch2 = initialChar [2];
		if (ch2 == chars :: DIR_DELIMITER) {
			prefix_ = DOUBLE_DOT;
			return initialChar + 3;
		}
		prefix_ = NO;
		return initialChar;
	}
	template<class A>
	const char*
	Path<A> :: parseDirectories ()
	{
		directories_.clear();

		const char* const initialChar = parsePrefix(); //path_.getFirstChar();
		const char* const terminalChar = path_.getTerminalChar();

		const char* firstChar = initialChar;
		for (const char* ch = initialChar; ch != terminalChar; ++ ch)
		{
			if (*ch == chars :: DIR_DELIMITER) {
				String_* dir = directories_.push();
				dir->copyFrom (firstChar, ch);
				firstChar = ch + 1;
			}
		}
		return firstChar;
	}
	template<class A>
	const char*
	Path<A> :: parseExtension ()
	{
		const char* const initial = path_.getFirstChar();
		const char* const terminal = path_.getTerminalChar();

		const char* delimiter = terminal - 1;
		while (delimiter != initial) {
			if (*delimiter == chars :: EXTENSION_DELIMITER) {
				break;
			}
			-- delimiter;
		}
		delimiter = (delimiter == initial) ? terminal : delimiter;
		extension_.copyFrom (delimiter + 1, terminal);
		return delimiter;
	}
	template<class A>
	void
	Path<A> :: parsePath ()
	{
		const char* const initialNameChar = parseDirectories();
		const char* const terminalNameChar = parseExtension();
		name_.copyFrom (initialNameChar, terminalNameChar);
	}
	template<class A>
	void
	Path<A> :: buildPath() const
	{
		path_.clear();
		switch (prefix_) {
		case DIR :
			path_ << chars :: DIR_DELIMITER;
			break;
		case DOT :
			path_ << chars :: DOT << chars :: DIR_DELIMITER;
			break;
		case DOUBLE_DOT :
			path_ << chars :: DOT << chars :: DOT << chars :: DIR_DELIMITER;
			break;
		case NO :
		default :
			break;
		}
		for (Depth_ depth = 0; depth < getDepth(); ++ depth) {
			path_ << *(directories_.getValue (depth)) << chars :: DIR_DELIMITER;
		}
		path_ << name_;
		path_ << chars :: EXTENSION_DELIMITER;
		path_ << extension_;
		changed_ = false;
	}
	template<class A>
	void
	Path<A> :: checkPath (const char* path) const
	{
		const typename String_ :: Index_
			length = String_ :: getStringLength (path);
		if (length != path_.getSize()) {
			goto bad;
		}
		for (typename String_ :: Index_ i = 0; i < length; ++ i) {
			if (path_ [i] != path [i]) {
				goto bad;
			}
		}
		return;
		bad :
			std :: cerr << "path " << *path << "was parsed incorrectly" << std :: endl;
			std :: cerr << "parse result: " << path_ << std :: endl;
			throw std :: exception();
	}
}

#endif /*NSGL_PATH_IPP_*/
