/*****************************************************************************/
/* Project name:    nsgl - non-standard general library                      */
/* File name:       nsgl_Timer.ipp                                           */
/* Description:     simple timer                                             */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSGL_TIMER_IPP_
#define NSGL_TIMER_IPP_

#include "system/nsgl_system.hpp"

namespace nstd {

	/****************************
	 *		Public members
	 ****************************/

	template<class A>
	inline
	Timer<A> :: Timer (const Mode mode) :
	isOn_ (mode == START),
	startTime_ (Undefined<Time_t> :: getValue()),
	startClock_ (Undefined<Clock_t> :: getValue()),
	stopTime_ (Undefined<Time_t> :: getValue()),
	stopClock_ (Undefined<Clock_t> :: getValue())
	{
		if (isOn_) {
			start();
		}
	}
	template<class A>
	Timer<A> :: ~ Timer() {
	}

	template<class A>
	inline void
	Timer<A> :: start()
	{
		isOn_ = true;
		startTime_  = system :: getTime();
		startClock_ = system :: getClock();
	}
	template<class A>
	inline void
	Timer<A> :: mark()
	{
		stopTime_  = system :: getTime();
		stopClock_ = system :: getClock();
	}
	template<class A>
	inline void
	Timer<A> :: stop()
	{
		isOn_ = false;
		stopTime_  = system :: getTime();
		stopClock_ = system :: getClock();
	}
	template<class A>
	inline Time_t
	Timer<A> :: getTime() const
	{
		if (isOn_) {
			stopTime_  = system :: getTime();
			stopClock_ = system :: getClock();
		}
		return (stopTime_ - startTime_);
	}
	template<class A>
	inline Clock_t
	Timer<A> :: getClock() const
	{
		if (isOn_) {
			stopTime_  = system :: getTime();
			stopClock_ = system :: getClock();
		}
		return (stopClock_ - startClock_);
	}
	template<class A>
	inline double
	Timer<A> :: getSeconds() const {
		return
			static_cast<double>(getClock()) /
			static_cast<double>(system :: getClocksPerSecond());
	}
	template<class A>
	inline double
	Timer<A> :: getMinutes() const {
		return
			static_cast<double>(getTime()) /
			static_cast<double>(60);
	}
	template<class A>
	inline double
	Timer<A> :: getHours() const {
		return
			static_cast<double>(getMinutes()) /
			static_cast<double>(60);
	}

	template<class A>
	inline bool
	Timer<A> :: isUsed() const
	{
		return
			(stopTime_ != Undefined<Time_t> :: getValue()) &&
			(startTime_ != Undefined<Time_t> :: getValue());
	}

	// nstd :: Object implementation
	template<class A>
	void
	Timer<A> :: commitSuicide() {
		delete this;
	}
	template<class A>
	Size_t
	Timer<A> :: getVolume() const {
		return 0;
	}
	template<class A>
	Size_t
	Timer<A> :: getSizeOf() const {
		return sizeof (Timer);
	}
}

#include <sstream>
#include <ctime>

namespace nstd {

	template<class A>
	void
	Timer<A> :: show (String<>& string) const
	{
		if (!isUsed()) {
			string << "unknown period";
			return;
		}
		if (getTime() < 60) {
			const double seconds = getSeconds();
			std :: ostringstream oss;
			oss << seconds;
			string << oss.str() << " seconds";
		} else {
			const Time_t delta = getTime();
			const std :: tm*
				t = std :: gmtime (&delta);
			enum {
				BUFFER_SIZE = 256
			};
			char str [BUFFER_SIZE];
			if (getTime() < 60 * 60) {
				std :: strftime (str, BUFFER_SIZE, "%M minutes, %S seconds", t);
			} else {
				std :: strftime (str, BUFFER_SIZE, "%H hours, %M minutes, %S seconds", t);
			}
			string << str;
		}
	}
}

#endif /*NSGL_TIMER_IPP_*/
