/*****************************************************************************/
/* Project name:    nstl - non-standard template library                     */
/* File name:       nstl_memory_Interval.ipp                                 */
/* Description:     template memory interval                                 */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSTL_MEMORY_INTERVAL_IPP_
#define NSTL_MEMORY_INTERVAL_IPP_

namespace nstd {
namespace memory {

	/****************************
	 *		Public members
	 ****************************/

	template<class T>
	inline
	Interval<T> :: Interval () :
	beg_ (NULL),
	end_ (NULL) {
	}
	template<class T>
	inline
	Interval<T> :: Interval (const Interval& interval) :
	beg_ (interval.beg_),
	end_ (interval.end_) {
	}
	template<class T>
	inline Interval<T> :: Interval (T* const beg, T* const end) :
	beg_ (beg),
	end_ (end) {
	}
	template<class T>
	inline Interval<T> :: Interval (T* const beg, const Size_t length)  :
	beg_ (beg),
	end_ (beg + length) {
	}

	template<class T>
	inline bool
	Interval<T> :: isClear() const
	{
		if ((beg_ == NULL) || (beg_ == end_)) {
			return true;
		}
		// Duff's device
		Byte_t* d = beg<Byte_t>();
		Ptrdiff_t c = getVolume();

		switch (c % 8)  /* c > 0 assumed */
		{
			case 0 : do { if (*d++ != '\0') return false;
			case 7 :      if (*d++ != '\0') return false;
			case 6 :      if (*d++ != '\0') return false;
			case 5 :      if (*d++ != '\0') return false;
			case 4 :      if (*d++ != '\0') return false;
			case 3 :      if (*d++ != '\0') return false;
			case 2 :      if (*d++ != '\0') return false;
			case 1 :      if (*d++ != '\0') return false;
			            } while ((c -= 8) > 0);
		}
		return true;
	}
	template<class T>
	inline bool
	Interval<T> :: isEmpty() const {
		return (beg_ == end_);
	}
	template<class T>
	inline bool
	Interval<T> :: isNull() const {
		return (beg_ == NULL) || (end_ == NULL);
	}
	template<class T>
	inline bool
	Interval<T> :: isRegular() const {
		return !isNull() && !isEmpty();
	}
	template<class T>
	inline bool
	Interval<T> :: isValid() const {
		return (beg_ <= end_);
	}

	template<class T>
	inline bool
	Interval<T> :: operator == (const Interval& interval) const {
		return isEqual (interval);
	}
	template<class T>
	inline bool
	Interval<T> :: operator != (const Interval& interval) const {
		return !isEqual (interval);
	}
	template<class T>
	inline bool
	Interval<T> :: isEqual (const Interval& interval) const
	{
		return
			(beg_ == interval.beg()) &&
			(end_ == interval.end());
	}
	template<class T>
	inline bool
	Interval<T> :: isDisjointed (const Interval& interval) const
	{
		return
			(end_ < interval.beg()) ||
			(interval.end() < beg_);
	}

	template<class T>
	inline T* const
	Interval<T> :: beg() const {
		return beg_;
	}
	template<class T>
	inline T* const
	Interval<T> :: end() const {
		return end_;
	}

	template<class T>
	inline void
	Interval<T> :: operator = (const Interval& interval) {
		copy (interval);
	}
	template<class T>
	inline void
	Interval<T> :: copy (const Interval& interval)
	{
		beg_ = interval.beg();
		end_ = interval.end();
	}

	template<class T>
	inline void
	Interval<T> :: shiftBeg (const Size_t index) {
		beg_ += index;
	}
	template<class T>
	inline void
	Interval<T> :: shiftEnd (const Size_t index) {
		end_ += index;
	}

	template<class T>
	inline void
	Interval<T> :: setBeg (T* const beg) {
		beg_ = beg;
	}
	template<class T>
	inline void
	Interval<T> :: setEnd (T* const end) {
		end_ = end;
	}
	template<class T>
	inline void
	Interval<T> :: setNull ()
	{
		beg_ = NULL;
		end_ = NULL;
	}

	// interval integer parameters
	template<class T>
	inline Size_t
	Interval<T> :: getLength() const {
		return end_ - beg_;
	}
	template<class T>
	inline Size_t
	Interval<T> :: getVolume() const {
		return sizeof (T) * getLength();
	}
	template<class T>
	inline Size_t
	Interval<T> :: getSizeOf() const {
		return sizeof (Interval);
	}

	template<class T>
	template<class S>
	inline void
	Interval<T> :: show (S& string) const
	{
		if (isNull()) {
			string << "NULL";
			return;
		}
		if (isEmpty()) {
			string << "empty ";
		}
		string << "begin = " << beg<void>() << ", ";
		string << "end = " << end<void>() << ", ";
		if (getVolume() == getLength()) {
			string << "byte volume = " << getVolume() << " ";
		} else {
			string << "byte volume = " << getVolume() << ", ";
			string << "array length = " << getLength()<< " ";
		}
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class T>
	inline
	Interval<T> :: Interval (const Size_t length) throw (exception :: OutOfMemory) :
	beg_ ((length == 0) ? NULL : new T [length]),
	end_ ((length == 0) ? NULL : beg_ + length)
	{
		if ((length > 0) && (beg_ == NULL)) {
			throw exception :: OutOfMemory();
		}
		clear();
	}
	template<class T>
	inline void
	Interval<T> :: allocate (const Size_t length) throw (exception :: OutOfMemory)
	{
		if (beg_ != NULL) {
			throw exception :: Exception ("double interval allocation.");
		}
		beg_  = (length == 0) ? NULL : new T [length];
		end_  = (length == 0) ? NULL : (beg_ + length);
		if ((length > 0) && (beg_ == NULL)) {
			throw exception :: OutOfMemory();
		}
		clear();
	}

	template<class T>
	template<class S>
	inline S* const
	Interval<T> :: beg() const {
		return reinterpret_cast<S*> (beg_);
	}
	template<class T>
	template<class S>
	inline S* const
	Interval<T> :: end() const {
		return reinterpret_cast<S*> (end_);
	}

	template<class T>
	template<class S>
	inline void
	Interval<T> :: operator = (const Interval<S>& interval) {
		copy (interval);
	}
	template<class T>
	template<class S>
	inline void
	Interval<T> :: copy (const Interval<S>& interval)
	{
		beg_ = reinterpret_cast<T*> (interval.beg());
		end_ = reinterpret_cast<T*> (interval.end());
	}

	template<class T>
	inline void
	Interval<T> :: empty() {
		end_ = beg_;
	}
	template<class T>
	void
	Interval<T> :: clear()
	{
		if ((beg_ == NULL) || (beg_ == end_)) {
			return;
		}
		// Duff's device
		Byte_t* d = beg<Byte_t>();
		Ptrdiff_t c = getVolume();

		switch (c % 8)  /* c > 0 assumed */
		{
			case 0 : do { *d++ = '\0';
			case 7 :      *d++ = '\0';
			case 6 :      *d++ = '\0';
			case 5 :      *d++ = '\0';
			case 4 :      *d++ = '\0';
			case 3 :      *d++ = '\0';
			case 2 :      *d++ = '\0';
			case 1 :      *d++ = '\0';
			            } while ((c -= 8) > 0);
		}
	}
	template<class T>
	inline void
	Interval<T> :: destroy ()
	{
		if (beg_ != NULL) {
			delete[] beg_;
			setNull();
		}
	}

	template<class T>
	inline Interval<>
	Interval<T> :: byteRepresentation() const
	{
		return Interval<>
		(
			reinterpret_cast<Byte_t*> (beg_),
			reinterpret_cast<Byte_t*> (end_)
		);
	}

	/****************************
	 *		Operators
	 ****************************/

	template<class T, class S>
	inline S&
	operator << (S& s, Interval<T> interval)
	{
		interval.show (s);
		return s;
	}
}
}

#endif /*NSTL_MEMORY_INTERVAL_IPP_*/
