/*****************************************************************************/
/* Project name:    nstl - non-standard template library                     */
/* File name:       nstl_memory_unit_Disposed.cpp                            */
/* Description:     container for disposed memory chunks                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSTL_MEMORY_UNIT_DISPOSED_CPP_
#define NSTL_MEMORY_UNIT_DISPOSED_CPP_

#include "memory/nstl_memory_UsageMessage.hpp"
#include "memory/unit/nstl_memory_unit_Disposed.hpp"

namespace nstd {
namespace memory {
namespace unit {

	/****************************
	 *		Public members
	 ****************************/

	inline
	Disposed :: Disposed() :
	maxFragmentation_ (0),
	maxLengthSize_ (0),
	wasted_ (0),
	vector_ (NULL),
	begins_ (NULL),
	ends_ (NULL),
	lengths_(NULL),
	pool_ (NULL) {
	}

	inline void
	Disposed :: create()
	{
		vector_  = new Vector_ (INITIAL_VECTOR_CAPACITY);
		begins_  = new PointerMap_ (INITIAL_VECTOR_CAPACITY);
		ends_    = new PointerMap_ (INITIAL_VECTOR_CAPACITY);
		lengths_ = new LengthMap_ (INITIAL_VECTOR_CAPACITY);
		pool_    = new IndexesPool_ (INITIAL_POOL_CAPACITY);
	}
	inline void
	Disposed :: destroy()
	{
		for (Index_ i = 0; i < lengths_->getSize(); ++ i) {
			Indexes_* indexes = lengths_->getValue (i);
			indexes->commitSuicide();
		}
		for (Index_ i = 0; i < pool_->getSize(); ++ i) {
			Indexes_* indexes = pool_->getValue (i);
			indexes->commitSuicide();
		}
		vector_->commitSuicide();  vector_ = NULL;
		begins_->commitSuicide();  begins_ = NULL;
		ends_->commitSuicide();    ends_ = NULL;
		lengths_->commitSuicide(); lengths_ = NULL;
		pool_->commitSuicide();    pool_ = NULL;
	}

	inline void
	Disposed :: clear ()
	{
		for (Index_ i = 0; i < lengths_->getSize(); ++ i) {
			Indexes_* indexes = lengths_->getValue (i);
			indexes->clear();
			storeIndexes (indexes);
		}
		maxFragmentation_ = 0;
		maxLengthSize_ = 0;
		wasted_ = 0;
		vector_->clear();
		begins_->clear();
		ends_->clear();
		lengths_->clear();
	}

	inline Size_t
	Disposed :: getUsedVolume() const
	{
		Size_t result = 0;
		if (vector_ != NULL) {
			result += vector_->getVolume();
			result += vector_->getSizeOf();
			result += begins_->getVolume();
			result += begins_->getSizeOf();
			result += ends_->getVolume();
			result += ends_->getSizeOf();
			result += lengths_->getVolume();
			result += lengths_->getSizeOf();
			for (Size_t i = 0; i < lengths_->getSize(); ++ i) {
				const Indexes_* indexes = lengths_->getValue (i);
				result += indexes->getVolume();
				result += indexes->getSizeOf();
			}
			result += pool_->getVolume();
			result += pool_->getSizeOf();
			for (Size_t i = 0; i < pool_->getSize(); ++ i) {
				const Indexes_* indexes = pool_->getValue (i);
				result += indexes->getVolume();
				result += indexes->getSizeOf();
			}
		}
		return result;
	}
	inline Size_t
	Disposed :: getFreeVolume() const
	{
		Size_t result = 0;
		if (vector_ != NULL) {
			for (Index_ i = 0; i < vector_->getSize(); ++ i) {
				result += vector_->getReference (i).getVolume();
			}
		}
		return result;
	}
	inline Size_t
	Disposed :: getWastedVolume() const {
		return wasted_;
	}

	/****************************
	 *		Private members
	 ****************************/

	inline void
	Disposed :: verifyIntegrity() const
	{
		bool verified = true;
		verified &= verifyVector();
		verified &= verifyBegins();
		verified &= verifyEnds();
		verified &= verifyLengths();
		verified &= verifyPool();
		if (!verified) {
			std :: cerr << std :: endl << std :: endl;
			showInside();
			throw std :: exception();
		}
	}
	inline bool
	Disposed :: verifyVector() const
	{
		bool verified = true;
		const Index_ size = vector_->getSize();
		for (Index_ i = 0; i < size; ++ i) {
			const Interval<> interval_1 = vector_->getValue (i);
			if (!interval_1.isValid()) {
				std :: cerr << "disposed interval is not valid:" << std :: endl;
				std :: cerr << chars :: TAB << "interval = " << interval_1 << std :: endl;
				verified = false;
			}
			if (!interval_1.isClear()) {
				std :: cerr << "disposed interval is not clear:" << std :: endl;
				std :: cerr << chars :: TAB << "interval = " << interval_1 << std :: endl;
				verified = false;
			}
			for (Index_ j = i + 1; j < size; ++ j) {
				const Interval<> interval_2 = vector_->getValue (j);
				if (!interval_1.isDisjointed (interval_2)) {
					std :: cerr << "disposed intervals are not disjointed:" << std :: endl;
					std :: cerr << chars :: TAB << "interval_1 = " << interval_1 << std :: endl;
					std :: cerr << chars :: TAB << "interval_2 = " << interval_2 << std :: endl;
					verified = false;
				}
			}
		}
		return verified;
	}
	inline bool
	Disposed :: verifyBegins() const
	{
		bool verified = true;
		const Index_ size = vector_->getSize();
		if (begins_->getSize() != size) {
			std :: cerr << "begins size is wrong:" << std :: endl;
			std :: cerr << chars :: TAB << "vector_->getSize() = " << size << std :: endl;
			std :: cerr << chars :: TAB << "begins_->getSize() = " << begins_->getSize() << std :: endl;
			verified = false;
		}
		for (Index_ i = 0; i < size; ++ i) {
			const Interval<> interval = vector_->getValue (i);
			const Index_ indexBeg = begins_->find (interval.beg());
			if (indexBeg != i) {
				std :: cerr << "begins index is wrong:" << std :: endl;
				std :: cerr << chars :: TAB << "vector_->getSize() = " << size << std :: endl;
				std :: cerr << chars :: TAB << "i = " << i << std :: endl;
				std :: cerr << chars :: TAB << "vector_->getValue (i) = " << interval << std :: endl;
				std :: cerr << chars :: TAB << "begins_->find (interval.beg()) = " << indexBeg << std :: endl;
				if (0 <= indexBeg && indexBeg < size) {
					std :: cerr << chars :: TAB << "begins_->getValue (...) = " << vector_->getValue (indexBeg) << std :: endl;
				}
				verified = false;
			}
		}
		return verified;
	}
	inline bool
	Disposed :: verifyEnds() const
	{
		bool verified = true;
		const Index_ size = vector_->getSize();
		if (ends_->getSize() != size) {
			std :: cerr << "ends size is wrong:" << std :: endl;
			std :: cerr << chars :: TAB << "vector_->getSize() = " << size << std :: endl;
			std :: cerr << chars :: TAB << "ends_->getSize() = " << ends_->getSize() << std :: endl;
			verified = false;
		}
		for (Index_ i = 0; i < size; ++ i) {
			const Interval<> interval = vector_->getValue (i);
			const Index_ indexEnd = ends_->find (interval.end());
			if (indexEnd != i) {
				std :: cerr << chars :: TAB << "vector_->getSize() = " << size << std :: endl;
				std :: cerr << "ends index is wrong:" << std :: endl;
				std :: cerr << chars :: TAB << "i = " << i << std :: endl;
				std :: cerr << chars :: TAB << "vector_->getValue (i) = " << interval << std :: endl;
				std :: cerr << chars :: TAB << "ends_->find (interval.beg()) = " << indexEnd << std :: endl;
				if (0 <= indexEnd && indexEnd < size) {
					std :: cerr << chars :: TAB << "ends_->getValue (...) = " << vector_->getValue (indexEnd) << std :: endl;
				}
				verified = false;
			}
		}
		return verified;
	}
	inline bool
	Disposed :: verifyLengths() const
	{
		bool verified = true;
		const Index_ size = vector_->getSize();
		Index_ lengthsCount = 0;
		for (Size_t i = 0; i < lengths_->getSize(); ++ i) {
			const Indexes_* indexes = lengths_->getValue (i);
				lengthsCount += indexes->getSize();
		}
		if (lengthsCount != size) {
			std :: cerr << "lengthsCount size is wrong:" << std :: endl;
			std :: cerr << chars :: TAB << "vector_->getSize() = " << size << std :: endl;
			std :: cerr << chars :: TAB << "lengthsCount = " << lengthsCount << std :: endl;
			verified = false;
		}
		for (Index_ i = 0; i < vector_->getSize(); ++ i) {
			const Interval<> interval = vector_->getValue (i);
			const Size_t length = interval.getLength();
			const Indexes_* indexes = lengths_->find (length);
			const Size_t index = indexes->find (i);
			if ((index == Undefined<Size_t> :: getValue()) ||
				(index >= indexes->getSize())) {
				std :: cerr << "length is not found:" << std :: endl;
				std :: cerr << chars :: TAB << "vector_->getSize() = " << size << std :: endl;
				std :: cerr << chars :: TAB << "i = " << i << std :: endl;
				std :: cerr << chars :: TAB << "vector_->getValue (i) = " << interval << std :: endl;
				std :: cerr << chars :: TAB << "lengths_->find (interval.lengths()) = " << *indexes << std :: endl;
				verified = false;
			}
		}
		return verified;
	}
	inline bool
	Disposed :: verifyPool() const
	{
		bool verified = true;
		const Index_ size = pool_->getSize();
		for (Size_t i = 0; i < size; ++ i) {
			const Indexes_* indexes = pool_->getValue (i);
			if (!indexes->isEmpty()) {
				std :: cerr << "non-empty indexes in the pool:" << std :: endl;
				std :: cerr << chars :: TAB << "pool_->getSize() = " << size << std :: endl;
				std :: cerr << chars :: TAB << "i = " << i << std :: endl;
				std :: cerr << chars :: TAB << "pool_->getValue (i) = " << *indexes << std :: endl;
				verified = false;
			}
		}
		return verified;
	}

	inline void
	Disposed :: showInside() const
	{
		showVector();
		showBegins();
		showEnds();
		showLengths();
	}
	inline void
	Disposed :: showVector() const
	{
		const Index_ size = vector_->getSize();
		std :: cerr << "vector:" << std :: endl;
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << chars :: TAB << "size = " << size << std :: endl;
		for (Index_ i = 0; i < size; ++ i) {
			std :: cerr << chars :: TAB << i << " = " << vector_->getValue (i) << std :: endl;
		}
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << std :: endl;
	}
	inline void
	Disposed :: showBegins() const
	{
		const Index_ size = begins_->getSize();
		std :: cerr << "begins:" << std :: endl;
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << chars :: TAB << "size = " << size << std :: endl;
		for (Index_ i = 0; i < size; ++ i) {
			const void* beg = begins_->getKey (i);
			std :: cerr << chars :: TAB << i << " key = " << beg << ", ";
			std :: cerr << "value = " << begins_->getValue (i) << std :: endl;
		}
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << std :: endl;
	}
	inline void
	Disposed :: showEnds() const
	{
		const Index_ size = ends_->getSize();
		std :: cerr << "ends:" << std :: endl;
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << chars :: TAB << "size = " << size << std :: endl;
		for (Index_ i = 0; i < size; ++ i) {
			const void* end = ends_->getKey (i);
			std :: cerr << chars :: TAB << i << " key = " << end << ", ";
			std :: cerr << "value = " << ends_->getValue (i) << std :: endl;
		}
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << std :: endl;
	}
	inline void
	Disposed :: showLengths() const
	{
		const Index_ size = lengths_->getSize();
		std :: cerr << "lengths:" << std :: endl;
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << chars :: TAB << "size = " << size << std :: endl;
		for (Index_ i = 0; i < size; ++ i) {
			std :: cerr << chars :: TAB << i << " key = " << lengths_->getKey (i) << ", ";
			std :: cerr << "value = " << *lengths_->getValue (i) << std :: endl;
		}
		std :: cerr << "------------------------" << std :: endl;
		std :: cerr << std :: endl;
	}
}
}
}

#endif /*NSTL_MEMORY_UNIT_DISPOSED_HPP_*/
