/*****************************************************************************/
/* Project name:    nstl - non-standard template library                     */
/* File name:       nstl_memory_unit_Disposed.hpp                            */
/* Description:     container for disposed memory chunks                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSTL_MEMORY_UNIT_DISPOSED_HPP_
#define NSTL_MEMORY_UNIT_DISPOSED_HPP_

#include "map/nstl_map.dpp"
#include "container/nstl_container.dpp"
#include "memory/nstl_memory_Interval.hpp"
#include "memory/allocator/nstl_memory_allocator_Method.hpp"

namespace nstd {
namespace memory {
namespace unit {

class Disposed {
public :
	Disposed();

	void create();
	void destroy();

	template<class T>
	const Interval<T> allocate (const Size_t, const allocator :: Method = allocator :: DEFAULT);

	template<class T>
	void dispose (const Interval<T>&);

	void clear ();

	Size_t getUsedVolume() const;
	Size_t getFreeVolume() const;
	Size_t getWastedVolume() const;

	template<class S>
	void show (S&, const Size_t  = 0, const bool = false) const;

private :
	enum {
		INITIAL_VECTOR_CAPACITY = KILOBYTE,
		INITIAL_POOL_CAPACITY   = KILOBYTE,
		INITIAL_LENGTH_CAPACITY = 1,
		FRAGMENTATION_THRESHOLD = KILOBYTE * 2,
	};
	typedef
		map :: Scalar
		<
			const Byte_t*,
			Size_t,
			allocator :: Standard
		>
		PointerMap_;
	typedef
		container :: Set
		<
			Size_t,
			container :: storage :: ByValue,
			allocator :: Standard
		>
		Indexes_;
	typedef
		map :: Scalar
		<
			Size_t,
			Indexes_*,
			allocator :: Standard
		>
		LengthMap_;
	typedef
		container :: Vector
		<
			Interval<>,
			container :: storage :: ByValue,
			allocator :: Standard
		>
		Vector_;
	typedef
		container :: Vector
		<
			Indexes_*,
			container :: storage :: ByValue,
			allocator :: Standard
		>
		IndexesPool_;

	typedef Size_t Index_;

	template<class T>
	const Interval<T> allocateGreedy (const Size_t);
	template<class T>
	const Interval<T> allocateExactly (const Size_t, bool&);
	template<class T>
	const Interval<T> allocateAnyway (const Size_t);

	template<class T>
	bool disposeSmart (const Interval<T>&);
	void disposeEndBorder (Byte_t*, Byte_t*, const Index_);
	void disposeBegBorder (Byte_t*, Byte_t*, const Index_);
	void disposeBothBorders (Byte_t*, Byte_t*, const Index_, const Index_);

	void addLength (const Size_t length, const Index_ index);
	void removeLength (const Size_t length, const Index_ index);
	void decLengths (const Index_);
	void replaceMinimal (Interval<>);
	void storeIndexes (Indexes_*);
	Indexes_* restoreIndexes();

	void verifyIntegrity() const;
	bool verifyVector() const;
	bool verifyBegins() const;
	bool verifyEnds() const;
	bool verifyLengths() const;
	bool verifyPool() const;

	void showInside() const;
	void showVector() const;
	void showBegins() const;
	void showEnds() const;
	void showLengths() const;

	// Maximum fragmentation.
	Index_ maxFragmentation_;
	Index_ maxLengthSize_;
	Size_t wasted_;

	Vector_*      vector_;
	PointerMap_*  begins_;
	PointerMap_*  ends_;
	LengthMap_*   lengths_;
	IndexesPool_* pool_;
};

}
}
}

#endif /*NSTL_MEMORY_UNIT_DISPOSED_HPP_*/
