/*****************************************************************************/
/* Project name:    nstl - non-standard template library                     */
/* File name:       nstl_memory_unit_Disposed.ipp                            */
/* Description:     container for disposed memory chunks                     */
/* Copyright:       (c) 2006-2009 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef NSTL_MEMORY_UNIT_DISPOSED_IPP_
#define NSTL_MEMORY_UNIT_DISPOSED_IPP_

#include "memory/nstl_memory_UsageMessage.hpp"
#include "memory/nstl_memory_ignorableDifference.hpp"

namespace nstd {
namespace memory {
namespace unit {

	/****************************
	 *		Public members
	 ****************************/

	template<class T>
	const Interval<T>
	Disposed :: allocate (const Size_t length, const allocator :: Method method)
	{
		typedef T Type;
		Interval<Type> result;
		bool lengthFound = false;
		if (result.isEmpty()) {
			result = allocateExactly<Type> (length, lengthFound);
		}
		if (method == allocator :: GREEDY && !lengthFound) {
			result = allocateGreedy<Type> (length);
		}
		if (result.isEmpty() && !lengthFound) {
			result = allocateAnyway<Type> (length);
		}
		if (!result.isClear()) {
			throw exception :: Exception ("allocated space is not clear.");
		}
		#ifdef DEBUG_DISPOSED
		verifyIntegrity();
		#endif
		return result;
	}

	template<class T>
	void
	Disposed :: dispose (const Interval<T>& in)
	{
		if (in.isNull()) {
			return;
		}
		Interval<> interval (in.template beg<Byte_t>(), in.template end<Byte_t>());
		interval.clear();
		if (vector_->getSize() > FRAGMENTATION_THRESHOLD) {
			if (interval.getVolume() < lengths_->getKey (0)) {
				wasted_ += interval.getVolume();
			} else {
				wasted_ += lengths_->getKey (0);
				replaceMinimal (interval);
			}
			return;
		}
		if (disposeSmart (interval)) {
			#ifdef DEBUG_DISPOSED
			verifyIntegrity();
			#endif
			return;
		}
		const Index_ i = vector_->getSize();
		vector_->add (interval);
		if (vector_->getSize() > maxFragmentation_) {
			maxFragmentation_ = vector_->getSize();
		}
		begins_->add (interval.beg(), i);
		ends_->add (interval.end(), i);
		addLength (interval.getLength(), i);
		#ifdef DEBUG_DISPOSED
		verifyIntegrity();
		#endif
	}

	template<class S>
	void
	Disposed :: show (S& string, const Size_t overallVolume, const bool showFragmentation) const
	{
		string << "disposed memory information:" << chars :: NEW_LINE;
		string << "----------------------------" << chars :: NEW_LINE;
		string << UsageMessage (overallVolume, getUsedVolume(), "used memory");
		string << UsageMessage (overallVolume, getFreeVolume(), "free memory");
		string << UsageMessage (overallVolume, getWastedVolume(), "wasted memory");

		const Index_ fragmentation = vector_->getSize();
		string << "fragmentation:" << chars :: TAB << fragmentation << chars :: TAB;
		string << "of: " << maxFragmentation_ << chars :: TAB;
		const double percentFragm = double (fragmentation * 100) / maxFragmentation_;
		string << "= " << percentFragm << " %";
		string << chars :: NEW_LINE;

		//string << "max lengths size:" << chars :: TAB << maxLengthSize_ << chars :: NEW_LINE;

		if (showFragmentation) {
			string << chars :: NEW_LINE;
			string << "disposed intervals:" << chars :: NEW_LINE;
			string << "-------------------" << chars :: NEW_LINE;
			for (Index_ i = 0; i < fragmentation; ++ i) {
				string << chars :: TAB << vector_->getReference (i) << chars :: NEW_LINE;
			}
		}
	}

	/****************************
	 *		Private members
	 ****************************/

	template<class T>
	const Interval<T>
	Disposed :: allocateGreedy (const Size_t length)
	{
		typedef T Type;
		const Size_t size = length * sizeof (Type);
		const Index_ j = lengths_->findPlace (size);
		const Size_t sizeFound = lengths_->getKey (j);
		if ((j == lengths_->getSize()) ||
			!ignorableDifference (size, sizeFound)) {
			return Interval<Type>();
		} else {
			Indexes_* indexes = lengths_->value (j);
			const Index_ i = indexes->getLastValue();
			Interval<>& interval = vector_->getReference (i);
			Type* const memory = interval.beg<Type>();

			indexes->removeLast();
			if (indexes->isEmpty()) {
				storeIndexes (indexes);
				lengths_->remove (interval.getLength());
			}
			const Size_t newLength = interval.getLength() / sizeof (Type);
			const Size_t allocated = newLength * sizeof (Type);

			if (allocated < sizeFound) {
				begins_->remove (interval.beg());
				interval.shiftBeg (allocated);
				begins_->add (interval.beg(), i);
				addLength (interval.getLength(), i);
			} else {
				begins_->removeDec (interval.beg(), i);
				ends_->removeDec (interval.end(), i);
				decLengths (i);
				vector_->remove (i);
			}
			return Interval<Type> (memory, newLength);
		}
	}
	template<class T>
	const Interval<T>
	Disposed :: allocateExactly (const Size_t length, bool& lengthFound)
	{
		typedef T Type;
		const Size_t size = length * sizeof (Type);
		Indexes_* indexes = lengths_->find (size);
		lengthFound = (indexes != NULL);
		if (!lengthFound) {
			return Interval<Type>();
		} else {
			const Index_ i = indexes->getLastValue();
			const Interval<> interval = vector_->getValue (i);
			Type* const memory = interval.beg<Type>();
			begins_->removeDec (interval.beg(), i);
			ends_->removeDec (interval.end(), i);
			indexes->removeLast();
			if (indexes->isEmpty()) {
				storeIndexes (indexes);
				lengths_->remove (size);
			}
			decLengths (i);
			vector_->remove (i);
			return Interval<Type> (memory, length);
		}
	}
	template<class T>
	const Interval<T>
	Disposed :: allocateAnyway (const Size_t length)
	{
		typedef T Type;
		const Size_t size = length * sizeof (Type);
		const Index_ j = lengths_->findPlace (size);
		if (j == lengths_->getSize()) {
			return Interval<Type>();
		} else {
			Indexes_* indexes = lengths_->value (j);
			const Index_ i = indexes->getLastValue();

			Interval<>& interval = vector_->getReference (i);
			Type* const memory = interval.beg<Type>();
			begins_->remove (interval.beg());
			indexes->removeLast();
			if (indexes->isEmpty()) {
				storeIndexes (indexes);
				lengths_->remove (interval.getLength());
			}
			interval.shiftBeg (size);
			begins_->add (interval.beg(), i);
			addLength (interval.getLength(), i);

			return Interval<Type> (memory, length);
		}
	}
	template<class T>
	bool
	Disposed :: disposeSmart (const Interval<T>& interval)
	{
		Byte_t* const
			beg = interval.template beg<Byte_t>();
		Byte_t* const
			end = interval.template end<Byte_t>();

		const Index_ index_1 = ends_->find (beg);
		const Index_ index_2 = begins_->find (end);
		const bool endBorder = (index_1 != Undefined<Index_> :: getValue());
		const bool begBorder = (index_2 != Undefined<Index_> :: getValue());

		if (endBorder && !begBorder) {
			disposeEndBorder (beg, end, index_1);
		}
		if (begBorder && !endBorder) {
			disposeBegBorder (beg, end, index_2);
		}
		if (begBorder && endBorder) {
			disposeBothBorders (beg, end, index_1, index_2);
		}
		return (endBorder || begBorder);
	}
	inline void
	Disposed :: disposeEndBorder (Byte_t* beg, Byte_t* end, const Index_ index)
	{
		Interval<>& interval = vector_->getReference (index);
		ends_->remove (beg);
		removeLength (interval.getLength(), index);
		interval.setEnd (end);
		ends_->add (interval.end(), index);
		addLength (interval.getLength(), index);
	}
	inline void
	Disposed :: disposeBegBorder (Byte_t* beg, Byte_t* end, const Index_ index)
	{
		Interval<>& interval = vector_->getReference (index);
		begins_->remove (end);
		removeLength (interval.getLength(), index);
		interval.setBeg (beg);
		begins_->add (interval.beg(), index);
		addLength (interval.getLength(), index);
	}
	void
	Disposed :: disposeBothBorders
	(
		Byte_t* beg, Byte_t* end,
		const Index_ index_1, const Index_ index_2
	)
	{
		Interval<>& interval_1 = vector_->getReference (index_1);
		Interval<>& interval_2 = vector_->getReference (index_2);

		ends_->remove (beg);
		begins_->removeDec (end, index_2);
		ends_->removeDec (interval_2.end(), index_2);

		removeLength (interval_1.getLength(), index_1);
		removeLength (interval_2.getLength(), index_2);
		decLengths (index_2);

		interval_1.setEnd (interval_2.end());
		const Index_ newIndex = (index_1 < index_2) ? index_1 : index_1 - 1;
		ends_->add (interval_1.end(), newIndex);
		addLength (interval_1.getLength(), newIndex);
		vector_->remove (index_2);
	}

	inline void
	Disposed :: addLength (const Size_t length, const Index_ index)
	{
		Indexes_* indexes = lengths_->find (length);
		if (indexes == NULL) {
			indexes = restoreIndexes();
			lengths_->add (length, indexes);
			if (lengths_->getSize() > maxLengthSize_) {
				maxLengthSize_ = lengths_->getSize();
			}
		}
		indexes->add (index);
	}
	inline void
	Disposed :: removeLength (const Size_t length, const Index_ index)
	{
		Indexes_* indexes = lengths_->find (length);
		if (indexes == NULL) {
			return;
		}
		const Index_ i = indexes->find (index);
		indexes->remove (i);
		if (indexes->isEmpty()) {
			storeIndexes (indexes);
			lengths_->remove (length);
		}
	}
	inline void
	Disposed :: decLengths (const Index_ index)
	{
		for (Index_ i = 0; i < lengths_->getSize(); ++ i) {
			Indexes_* const indexes = lengths_->getValue (i);
			Index_ j = indexes->findPlace (index);
			for (Index_ k = j; k < indexes->getSize(); ++ k) {
				-- indexes->getReference (k);
			}
		}
	}
	void
	Disposed :: replaceMinimal (const Interval<> newInterval)
	{
		Indexes_* indexes = lengths_->getValue (0);
		const Index_ index = indexes->getLastValue();
		const Interval<> oldInterval = vector_->getValue (index);
		vector_->getReference (index) = newInterval;

		begins_->replaceKey (oldInterval.beg(), newInterval.beg());
		ends_->replaceKey (oldInterval.end(), newInterval.end());
		indexes->removeLast();
		if (indexes->isEmpty()) {
			storeIndexes (indexes);
			lengths_->remove (oldInterval.getLength());
		}
		addLength (newInterval.getLength(), index);
	}
	inline void
	Disposed :: storeIndexes (Indexes_* indexes) {
		pool_->add (indexes);
	}
	inline Disposed :: Indexes_*
	Disposed :: restoreIndexes()
	{
		if (pool_->isEmpty()) {
			return new Indexes_ (INITIAL_LENGTH_CAPACITY);
		}
		Indexes_* result = pool_->getLastValue();
		pool_->removeLast();
		return result;
	}
}
}
}

#endif /*NSTL_MEMORY_UNIT_DISPOSED_IPP_*/
