/*****************************************************************************/
/* Project name:    smm - verifier for the Simplified MetaMath language      */
/* File name:       smm_auxiliary_Config.cpp                                 */
/* Description:     static class for command line arguments                  */
/* Copyright:       (c) 2006-2010 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef SMM_AUXILIARY_CONFIG_CPP_
#define SMM_AUXILIARY_CONFIG_CPP_

#include "smm/auxiliary/smm_auxiliary_Config.hpp"

namespace smm {
namespace auxiliary {

	/****************************
	 *		Public members
	 ****************************/

	bool
	Config :: keepChecking () {
		return keepChecking_;
	}
	bool
	Config :: showStack() {
		return showStack_;
	}

	// info output options
	bool
	Config :: showTimings() {
		return showTimings_;
	}
	bool
	Config :: showMath() {
		return showMath_;
	}
	bool
	Config :: showVolume() {
		return showVolume_;
	}

using std :: cout;
using std :: cerr;
using std :: endl;

	bool
	Config :: proceed (const int argCount, const char* argValues[])
	{
		init();
		if (argCount == 1) {
			cerr << getProgramName() << ": no input files" << endl;
			return false;
		}

		for (int i = 1; i < argCount; i ++) {
			 const char* opt = argValues [i];
			if (*opt == '-') {
				opt ++;
			} else {
				if (i != argCount - 1) {
					cerr << "option \"" << opt;
					cerr << "\" does not start with \"-\" or \"--\"" << endl;
					return false;
				}
				if (!proceedStringOption (opt, argCount, argValues, i)) {
					cerr << "option \"" << opt;
					cerr << "\" does not start with \"-\" or \"--\"" << endl;
					return false;
				}
				continue;
			}
			if (*opt == '-') {
				opt ++;
				if (!proceedStringOption (opt, argCount, argValues, i)) {
					cerr << "unknown option: \"" << opt << "\"" << endl;
					return false;
				}
			} else if (!proceedOneCharOption (opt, argCount, argValues, i)) {
				cerr << "unknown option: \"" << opt << "\"" << endl;
				return false;
			}
		}
		return analyseArguments();
	}
	void
	Config :: release() {
		nstd :: Config :: release();
	}
	Size_t
	Config :: getVolume() {
		return nstd :: Config :: getVolume();
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Config :: showUsageMessage()
	{
		Ancestor :: showProgramUsage();
		cout << "Options:" << endl;
		Ancestor :: showOptions();
		cout << "  -k,  --keep-checking          keep checking at check errors" << endl;
		cout << "  -s,  --show-stack             extra debugging facility - " << endl;
		cout << "                                shows proof stack when proof error occurs." << endl;
		cout << endl;
		cout << "       --info-timing            show thorough timing info" << endl;
		cout << "       --info-math              show thorough math info" << endl;
		cout << "       --info-volume            show thorough volume info" << endl;
		cout << "       --info-all               show all thorough info" << endl;
		cout << endl;
	}

	void
	Config :: init ()
	{
		Ancestor :: init ();
		programName() << "smm";
		programDescription() << "smm (Simplified MetaMath) verifier.";
	}

	bool
	Config :: proceedStringOption (const char* opt, const int argCount, const char* argValues[], int& i)
	{
		if (Ancestor :: proceedStringOption (opt, argCount, argValues, i)) {
			return true;
		}
		if (equalStrings (opt, "show-stack")) {
			showStack_ = true;
			return true;
		}
		if (equalStrings (opt, "keep-checking")) {
			keepChecking_ = true;
			return true;
		}
		if (equalStrings (opt, "info-timing")) {
			showTimings_ = true;
			return true;
		}
		if (equalStrings (opt, "info-math")) {
			showMath_ = true;
			return true;
		}
		if (equalStrings (opt, "info-volume")) {
			showVolume_ = true;
			return true;
		}
		if (equalStrings (opt, "info-all")) {
			showTimings_ = true;
			showMath_ = true;
			showVolume_ = true;
			return true;
		}
		return false;
	}

	bool
	Config :: proceedOneCharOption (const char* opt, const int argCount, const char* argValues[], int& i)
	{
		if (Ancestor :: proceedOneCharOption (opt, argCount, argValues, i)) {
			return true;
		}
		switch (*opt) {
		case 'k' :
			keepChecking_ = true;
			return true;
		case 's' :
			showStack_ = true;
			return true;
		default :
			return false;
		}
	}

	bool
	Config :: checkPaths()
	{
		if (source()->extension() != "smm") {
			cerr << "wrong source extension. Should be *.smm" << endl;
			return false;
		}
		return Ancestor :: Paths :: check();
	}
	bool
	Config :: analyseArguments()
	{
		if (showHelp()) {
			showUsageMessage();
			return false;
		}
		if (abortParsing()) {
			return false;
		}
		if (!checkPaths()) {
			return false;
		}
		return true;
	}

	/*************************************
	 *		Static private attributes
	 *************************************/

	bool Config :: keepChecking_ = false;
	bool Config :: showStack_ = false;

	// info output options
	bool Config :: showTimings_ = false;
	bool Config :: showMath_ = false;
	bool Config :: showVolume_ = false;
}
}

#endif /*SMM_AUXILIARY_CONFIG_CPP_*/
