/*****************************************************************************/
/* Project name:    smm - verifier for the Simplified MetaMath language      */
/* File Name:       smm_Smm.cpp                                              */
/* Description:     smm verifier                                             */
/* Copyright:       (c) 2006-2010 Dmitri Vlasov                              */
/* Author:          Dmitri Yurievich Vlasov, Novosibirsk, Russia             */
/* Email:           vlasov at academ.org                                     */
/* URL:             http://mathdevlanguage.sourceforge.net                   */
/* Modified by:                                                              */
/* License:         GNU General Public License Version 3                     */
/*****************************************************************************/

#ifndef SMM_SMM_CPP_
#define SMM_SMM_CPP_

#include <exception>

#include "smm/auxiliary/smm_auxiliary.hpp"
#include "smm/math/smm_math.hpp"
#include "smm/parser/smm_parser.hpp"
#include "smm/smm_Smm.hpp"

namespace smm {

using namespace exception;

	/****************************
	 *		Public members
	 ****************************/

	Smm :: Smm() :
	file_ (Config :: getSource()->getString()),
	path_ (new Path (file_)),
	parser_ (new parser :: Source ()),
	source_ (NULL),
	timers_ (),
	volume_ (this)
	{
		Math :: init();
		Table :: init();
	}
	Smm :: ~ Smm()
	{
		Table :: release();
		Math :: release();
		deleteAggregateObject (source_);
		deleteAggregateObject (parser_);
		deleteAggregateObject (path_);
		Config :: release();
		object :: Object :: release();
	}

using std :: endl;
using std :: cerr;
using std :: cout;

	Smm :: Result_
	Smm :: run ()
	{
		try {
			process();
		} catch (const Error* error) {
			timers_.stopAll();
			cerr << *error << endl;
			Result_ result = static_cast<Result_>(error->getType());
			delete error;
			return result;
		} catch (const OutOfMemory& error) {
			timers_.stopAll();
			cerr << error << endl;
			Memory :: show (cerr);
			return OUT_OF_MEMORY;
		} catch (const std :: exception& e) {
			timers_.stopAll();
			cerr << "exception: " << e.what() << endl;
			return GENERAL_FAILURE;
		} catch (...) {
			timers_.stopAll();
			cerr << "exception." << endl;
			return GENERAL_FAILURE;
		}
		return SUCCESS;
	}
	Smm :: Result_
	Smm :: initMemory()
	{
		Result_ result = SUCCESS;
		if (Config :: verbosity() > 1) {
			std :: cout << "initializing ... " << std :: flush;
		}
		try {
			Memory :: standard().create();
			Memory :: heap().create();
		} catch (OutOfMemory&) {
			result = Smm :: INIT_ERROR;
		}
		if (Config :: verbosity() > 1) {
			std :: cout << ((result == SUCCESS) ? "done." : "failed.") << std :: endl;
		}
		return result;
	}
	Smm :: Result_
	Smm :: releaseMemory()
	{
		bool success = true;
		if (Config :: verbosity() >= Config :: VERBOSE_STANDARD) {
			std :: cout << "releasing ... " << std :: flush;
		}
		try {
			Memory :: heap().destroy();
		} catch (memory :: MemoryLeaks& exception) {
			std :: cerr << std :: endl << chars :: TAB <<exception.what() << ": ";
			std :: cerr << exception.getLeak() << " bytes are lost ";
			success = false;
		} catch (std :: exception& exception) {
			std :: cerr << exception.what() << std :: endl;
			success = false;
		} catch (...) {
			success = false;
		}
		if (Config :: verbosity() >= Config :: VERBOSE_STANDARD) {
			std :: cout << (success ? "done." : "failed.") << std :: endl;
		}
		return (success ? SUCCESS : RELEASE_ERROR);
	}

using manipulator :: endLine;
using manipulator :: space;
using manipulator :: tab;

	// object :: Object implementation
	void
	Smm :: commitSuicide() {
		delete this;
	}
	Size_t
	Smm :: getVolume() const
	{
		Size_t volume = 0;
		volume += object :: Object :: getVolume();
		volume += file_.getVolume();
		volume += getAggregateVolume (path_);
		volume += getAggregateVolume (parser_);
		volume += getAggregateVolume (source_);
		volume += timers_.getVolume();
		volume += volume_.getVolume();
		return volume;
	}
	Size_t
	Smm :: getSizeOf() const {
		return sizeof (Smm);
	}
	void
	Smm :: show (String& string) const
	{
		if (path_ != NULL) {
			string << "file" << space;
			string << *path_ << space;
			string << "is done in" << space << timers_.total() << endLine;
		}
		if (Config :: showTimings()) {
			timers_.show (string);
			string << endLine;
		}
		if (Config :: showMath()) {
			Math :: show (string);
			string << endLine;
		}
		if (Config :: showVolume() || Config :: showMemory()) {
			volume_.show (string);
			string << endLine;
		}
	}

	/****************************
	 *		Private members
	 ****************************/

	void
	Smm :: process ()
	{
		timers_.total().start();
		read();
		//verify();
		timers_.total().stop();
	}
	void
	Smm :: read ()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "reading ... " << std :: flush;
		}
		timers_.read().start ();
		source_ = parser_->parse (path_->getString());
		timers_.read().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.read() << std :: endl;
		}
	}
	void
	Smm :: verify ()
	{
		if (Config :: verbosity() > 1) {
			std :: cout << "verifying ... " << std :: flush;
		}
		timers_.verify().start();
		source_->verify();
		timers_.verify().stop();
		if (Config :: verbosity() > 1) {
			std :: cout << "done in " << timers_.verify() << std :: endl;
		}
	}
}

#endif /*SMM_SMM_CPP_*/
